{*************************************************************************
                Copyright (c) PilotLogic Software House

  Package bs_EnginesPlus
  This file is part of CodeTyphon Studio (https://www.pilotlogic.com/)

 ***** BEGIN LICENSE BLOCK *****
 * Version: LGPL-2 with link exception (Modified LGPL)
 *
 * The contents of this file are subject to the
 * GNU LIBRARY GENERAL PUBLIC LICENSE Version 2 (the "License");
 * with the following modification:
 *
 * As a special exception, the copyright holders of this library give you
 * permission to link this library with independent modules to produce an
 * executable, regardless of the license terms of these independent modules,
 * and to copy and distribute the resulting executable under terms of your choice,
 * provided that you also meet, for each linked independent module, the terms
 * and conditions of the license of that module. An independent module is a
 * module which is not derived from or based on this library. If you modify this
 * library, you may extend this exception to your version of the library, but
 * you are not obligated to do so. If you do not wish to do so, delete this
 * exception statement from your version.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * ***** END LICENSE BLOCK *****

***************************************************************************}

Unit gaBitmapPlus;

{$MODE objfpc}{$H+}

Interface

Uses
   Classes, SysUtils,
   LCLType, Graphics, Math,

   gaResStrings,
   gaEngine,
   gaBitMap,
   gaClipper,
   gaFileFormats,
   gaLayers,
   gaPalette,
   gaSmoothPath,
   gaSVG,
   gaFonts;

Type

   TDrawElements = (deNone, deBoth, deOnlyFill, deOnlyLine, deErasePolygon, deClearType);
   FDrawLineType = (dltNormal, dltInverted, dltDashed, dltInvertedDashed);

   TgaBitmapPlus = Class;

   TgapPalette = Class
   Protected
      FPal: TArrayOfColor32;
   Public
      Constructor Create;
      Destructor Destroy; Override;
      Procedure BlackWhitePal;
      Procedure DefaultMacPal16;
      Procedure DefaultWinPal16;
      Procedure TrimPalette(Const colorFrequency: TArrayOfInteger; newSize: integer); Overload;
      Procedure TrimPalette(Const colorFrequency: TArrayOfInteger; fraction: double); Overload;
      Function CreateLogPalette: TMaxLogPalette;

      Property Palette: TArrayOfColor32 Read FPal Write FPal;
   End;

   TgapPolygonD = Class
   Protected
      FImagePlus: TgaBitmapPlus;
      FPathD: TPathD;
      FDrawElements: TDrawElements;
      FFillColor: TColor32;
      FFillRule: TFillRule;
      FFillRenderer: TgaRendererCustom;
      FFillRendererUsed: Boolean;
      FLineType: FDrawLineType;
      FLineColor: TColor32;
      FLineWidth: double;
      FLineEndStyle: TEndStyle;
      FLineJoinStyle: TJoinStyle;
      FLineMiterLimit: double;
      FLineDashPattern: TArrayOfInteger;
      FLinePatternOffset: Double;
      FLineRenderer: TgaRendererCustom;
      FLineRendererUsed: Boolean;
      Function GetCount: Int64;
   Public
      Constructor Create(aImagePlus: TgaBitmapPlus);
      Destructor Destroy; Override;
      Procedure Clear;
      Procedure SetDefaultValues;

      Procedure Rectangle(Const rec: TRect); Overload;
      Procedure Rectangle(Const rec: TRectD); Overload;
      Procedure Rectangle(l, t, r, b: double); Overload;

      Procedure RoundRect(Const rec: TRect; radius: integer) Overload;
      Procedure RoundRect(Const rec: TRectD; radius: double) Overload;
      Procedure RoundRect(Const rec: TRect; radius: TPoint) Overload;
      Procedure RoundRect(Const rec: TRectD; radius: TPointD) Overload;

      Procedure Ellipse(Const rec: TRect; steps: integer = 0); Overload;
      Procedure Ellipse(Const rec: TRectD; steps: integer = 0); Overload;
      Procedure Ellipse(Const rec: TRectD; scale: double); Overload;

      Procedure Circle(Const pt: TPoint; radius: double); Overload;
      Procedure Circle(Const pt: TPointD; radius: double); Overload;

      Procedure Star(Const focalPt: TPointD; innerRadius, outerRadius: double; points: integer);
      Procedure Arc(Const rec: TRectD; startAngle, endAngle: double; scale: double = 0);
      Procedure Pie(Const rec: TRectD; StartAngle, EndAngle: double; scale: double = 0);

      Procedure FlattenQBezier(Const pt1, pt2, pt3: TPointD; tolerance: double = 0.0); Overload;
      Procedure FlattenQBezier(Const tolerance: double = 0.0); Overload;
      Procedure FlattenQBezier(Const firstPt: TPointD; Const pts: TPathD; tolerance: double = 0.0); Overload;

      Procedure FlattenCBezier(Const pt1, pt2, pt3, pt4: TPointD; tolerance: double = 0.0); Overload;
      Procedure FlattenCBezier(Const tolerance: double = 0.0); Overload;

      Procedure DrawFill;
      Procedure DrawLine;
      Procedure Draw;

      Property PathD: TPathD Read FPathD Write FPathD;
      Property Count: Int64 Read GetCount;
      Property DrawElements: TDrawElements Read FDrawElements Write FDrawElements;

      Property FillColor: TColor32 Read FFillColor Write FFillColor;
      Property FillRule: TFillRule Read FFillRule Write FFillRule;
      Property FillRenderer: TgaRendererCustom Read FFillRenderer Write FFillRenderer;
      Property FillRendererUsed: Boolean Read FFillRendererUsed Write FFillRendererUsed;

      Property LineType: FDrawLineType Read FLineType Write FLineType;
      Property LineColor: TColor32 Read FLineColor Write FLineColor;
      Property LineWidth: double Read FLineWidth Write FLineWidth;
      Property LineEndStyle: TEndStyle Read FLineEndStyle Write FLineEndStyle;
      Property LineJoinStyle: TJoinStyle Read FLineJoinStyle Write FLineJoinStyle;
      Property LineMiterLimit: double Read FLineMiterLimit Write FLineMiterLimit;
      Property LineDashPattern: TArrayOfInteger Read FLineDashPattern Write FLineDashPattern;
      Property LinePatternOffset: Double Read FLinePatternOffset Write FLinePatternOffset;
      Property LineRenderer: TgaRendererCustom Read FLineRenderer Write FLineRenderer;
      Property LineRendererUsed: Boolean Read FLineRendererUsed Write FLineRendererUsed;

      Property ImagePlus: TgaBitmapPlus Read FImagePlus Write FImagePlus;
   End;

   TgapPolygonsD = Class
   Protected
      FImagePlus: TgaBitmapPlus;
      FPathsD: TPathsD;
      FDrawElements: TDrawElements;

      FFillColor: TColor32;
      FFillBackColor: TColor32;
      FFillRule: TFillRule;
      FFillRenderer: TgaRendererCustom;
      FFillRendererUsed: Boolean;

      FLineType: FDrawLineType;
      FLineColor: TColor32;
      FLineWidth: double;
      FLineEndStyle: TEndStyle;
      FLineJoinStyle: TJoinStyle;
      FLineMiterLimit: double;
      FLineDashPattern: TArrayOfInteger;
      FLinePatternOffset: Double;
      FLineRenderer: TgaRendererCustom;
      FLineRendererUsed: Boolean;
      Function GetCount: Int64;
   Public
      Constructor Create(aImagePlus: TgaBitmapPlus);
      Destructor Destroy; Override;
      Procedure Clear;
      Procedure SetDefaultValues;

      Procedure DrawFill;
      Procedure DrawLine;
      Procedure Draw;

      Procedure FlattenQBezier(Const tolerance: double = 0.0); Overload;
      Procedure FlattenCBezier(Const tolerance: double = 0.0); Overload;

      Property PathsD: TPathsD Read FPathsD Write FPathsD;
      Property Count: Int64 Read GetCount;
      Property DrawElements: TDrawElements Read FDrawElements Write FDrawElements;

      Property FillColor: TColor32 Read FFillColor Write FFillColor;
      Property FillBackColor: TColor32 Read FFillBackColor Write FFillBackColor;
      Property FillRule: TFillRule Read FFillRule Write FFillRule;
      Property FillRenderer: TgaRendererCustom Read FFillRenderer Write FFillRenderer;
      Property FillRendererUsed: Boolean Read FFillRendererUsed Write FFillRendererUsed;

      Property LineType: FDrawLineType Read FLineType Write FLineType;
      Property LineColor: TColor32 Read FLineColor Write FLineColor;
      Property LineWidth: double Read FLineWidth Write FLineWidth;
      Property LineEndStyle: TEndStyle Read FLineEndStyle Write FLineEndStyle;
      Property LineJoinStyle: TJoinStyle Read FLineJoinStyle Write FLineJoinStyle;
      Property LineMiterLimit: double Read FLineMiterLimit Write FLineMiterLimit;
      Property LineDashPattern: TArrayOfInteger Read FLineDashPattern Write FLineDashPattern;
      Property LinePatternOffset: Double Read FLinePatternOffset Write FLinePatternOffset;
      Property LineRenderer: TgaRendererCustom Read FLineRenderer Write FLineRenderer;
      Property LineRendererUsed: Boolean Read FLineRendererUsed Write FLineRendererUsed;

      Property ImagePlus: TgaBitmapPlus Read FImagePlus Write FImagePlus;
   End;

   TgapText = Class
   Protected
      FImagePlus: TgaBitmapPlus;
      FGlyphCache: TgaFontCache;
      FFontReader: TgaFontReader;
      FRenderer: TgaRendererCustom;
      FColor: TColor32;
      Function GeTgaFontInfo: TgaFontInfo;
      Function GetWeight: integer;
      Function GetFontFamily: TgaTTFFontFamily;
      Function GetAscent: double;
      Function GetDescent: double;
      Function GetLineHeight: double;
      Function GetScale: double;
      Function GetFontHeight: double;
      Procedure SetFontHeight(Const Val: double);
      Function GetInvertY: boolean;
      Procedure SetInvertY(Const Val: boolean);
      Function GetKerning: boolean;
      Procedure SetKerning(Const Val: boolean);
      Function GetUnderlined: Boolean;
      Procedure SetUnderlined(Const Val: Boolean);
      Function GetStrikeOut: Boolean;
      Procedure SetStrikeOut(Const Val: Boolean);

   Public
      Constructor Create(aImagePlus: TgaBitmapPlus);
      Destructor Destroy; Override;
      Procedure Clear(Const LoadDefaultFont: Boolean = True);
      Procedure SetDefaultValues;

      Function IsValidFontFormat: Boolean;
      Function LoadFromStream(stream: TStream): Boolean;
      Function LoadFromResource(Const resName: String; resType: PChar): Boolean;
      Function LoadFromFile(Const filename: String): Boolean;
      Function GetGlyphInfo(unicode: Word; Out paths: TPathsD; Out nextX: integer; Out glyphMetrics: TgaGlyphMetrics): Boolean;
      Function GetCharInfo(charOrdinal: WORD): PgaGlyphInfo;
      Function GetTextGlyphs(x, y: double; Const text: String): TPathsD; Overload;
      Function GetTextGlyphs(x, y: double; Const text: String; Out nextX: double): TPathsD; Overload;
      Function GetTextGlyphs(Const rec: TRect; Const text: String;
                             textAlign: TgaTextAlign; textAlignV: TgaTextVAlign;
                             Out nextIdx: integer; Out nextPt: TPointD): TPathsD; Overload;
      Function GetAngledTextGlyphs(x, y: double; Const text: String; angleRadians: double; Const rotatePt: TPointD;
                                   Out nextPt: TPointD): TPathsD;
      Function GetCharOffsets(Const text: String; interCharSpace: double = 0): TArrayOfDouble;
      Function GetTextWidth(Const text: String): double;

      Function DrawText(x, y: double; Const text: String): double; Overload;
      Function DrawText(x, y: double; Const text: String; TextColor: TColor32): double; Overload;
      Function DrawTextRenderer(x, y: double; Const text: String): double;
      Function DrawText(Const rec: TRect; Const text: String; textAlign: TgaTextAlign; textAlignV: TgaTextVAlign): TPointD; Overload;
      Function DrawText(Const rec: TRect; Const text: String; textAlign: TgaTextAlign; textAlignV: TgaTextVAlign; TextColor: TColor32): TPointD; Overload;
      Function DrawAngledDegText(x, y: double; angleDeg: double; Const text: String): TPointD; Overload;
      Function DrawAngledDegText(x, y: double; angleDeg: double; Const text: String; TextColor: TColor32): TPointD; Overload;
      Function DrawAngledRadText(x, y: double; angleRad: double; Const text: String): TPointD; Overload;
      Function DrawAngledRadText(x, y: double; angleRad: double; Const text: String; TextColor: TColor32): TPointD; Overload;
      Function DrawVerticalText(x, y, interCharSpace: double; Const text: String): double; Overload;
      Function DrawVerticalText(x, y, interCharSpace: double; Const text: String; TextColor: TColor32): double; Overload;

      Property FontFamily: TgaTTFFontFamily Read GetFontFamily;
      Property FontInfo: TgaFontInfo Read GeTgaFontInfo;
      Property Weight: integer Read GetWeight;
      Property Ascent: double Read GetAscent;
      Property Descent: double Read GetDescent;
      Property LineHeight: double Read GetLineHeight;
      Property Scale: double Read GetScale;

      Property Color: TColor32 Read FColor Write FColor;
      Property FontHeight: double Read GetFontHeight Write SetFontHeight;
      Property InvertY: boolean Read GetInvertY Write SetInvertY;
      Property Kerning: boolean Read GetKerning Write SetKerning;
      Property Underlined: Boolean Read GetUnderlined Write SetUnderlined;
      Property StrikeOut: Boolean Read GetStrikeOut Write SetStrikeOut;
      Property Renderer: TgaRendererCustom Read FRenderer Write FRenderer;

      Property ImagePlus: TgaBitmapPlus Read FImagePlus Write FImagePlus;
   End;

   TgaBitmapPlus = Class(TgaBitMap)
   Protected
      FOutBitMap: TBitMap;
      FPolygonD: TgapPolygonD;
      FPolygonsD: TgapPolygonsD;
      FText: TgapText;
      Procedure SetOutBitMap(Val: TBitMap);
   Public
      Constructor Create(aWidth: Integer = 0; aHeight: Integer = 0);
      Destructor Destroy; Override;
      Procedure Invalidate;

      Function MakePalette(Const MaxColors: integer): TgapPalette; Overload;
      Function MakePalette(Const MaxColors: integer; Out frequencies: TArrayOfInteger): TgapPalette; Overload;
      Function MakeAndApplyPalette(Const MaxColors: integer; UseDithering: Boolean): TgapPalette; Overload;
      Function MakeAndApplyPalette(Const MaxColors: integer; UseDithering: Boolean; Out frequencies: TArrayOfInteger): TgapPalette; Overload;
      Procedure ApplyPalette(Const palette: TgapPalette; UseDithering: Boolean = true);
      Procedure DrawPalette(Const palette: TgapPalette);

      Property OutBitMap: TBitMap Read FOutBitMap Write SetOutBitMap;

      Property PolygonD: TgapPolygonD Read FPolygonD;
      Property PolygonsD: TgapPolygonsD Read FPolygonsD;
      Property Text: TgapText Read FText;
   End;

Implementation

{$R gabitmapplus.res}

Constructor TgapText.Create(aImagePlus: TgaBitmapPlus);
Begin
   FRenderer := Nil;
   FImagePlus := aImagePlus;
   FFontReader := TgaFontReader.Create;
   FGlyphCache := TgaFontCache.Create(FFontReader, DpiAware(36));
   Clear;
   SetDefaultValues;
End;

Destructor TgapText.Destroy;
Begin
   FRenderer := Nil;
   FGlyphCache.Free;
   FFontReader.Free;
   Inherited;
End;

Procedure TgapText.Clear(Const LoadDefaultFont: Boolean = True);
Begin
   FGlyphCache.Clear;
   FFontReader.Clear;
   If LoadDefaultFont Then
     {$IFDEF WinCE}
      FFontReader.LoadFromResource('ADV_GABITMAP_GEORGIA', PChar(RT_RCDATA));
     {$ELSE}
      FFontReader.LoadFromResource('ADV_GABITMAP_GEORGIA', RT_RCDATA);
     {$ENDIF}
End;

Procedure TgapText.SetDefaultValues;
Begin
   FColor := clBlack32;
End;

Function TgapText.GeTgaFontInfo: TgaFontInfo;
Begin
   Result := FFontReader.FontInfo;
End;

Function TgapText.GetWeight: integer;
Begin
   Result := FFontReader.Weight;
End;

Function TgapText.GetFontFamily: TgaTTFFontFamily;
Begin
   Result := FFontReader.FontFamily;
End;

Function TgapText.IsValidFontFormat: Boolean;
Begin
   Result := FFontReader.IsValidFontFormat;
End;

Function TgapText.LoadFromStream(stream: TStream): Boolean;
Begin
   Result := FFontReader.LoadFromStream(stream);
End;

Function TgapText.LoadFromResource(Const resName: String; resType: PChar): Boolean;
Begin
   Result := FFontReader.LoadFromResource(resName, resType);
End;

Function TgapText.LoadFromFile(Const filename: String): Boolean;
Begin
   Result := FFontReader.LoadFromFile(filename);
End;

Function TgapText.GetGlyphInfo(unicode: Word; Out paths: TPathsD; Out nextX: integer; Out glyphMetrics: TgaGlyphMetrics): Boolean;
Begin
   Result := FFontReader.GetGlyphInfo(unicode, paths, nextX, glyphMetrics);
End;

Function TgapText.GetAscent: double;
Begin
   Result := FGlyphCache.Ascent;
End;

Function TgapText.GetDescent: double;
Begin
   Result := FGlyphCache.Descent;
End;

Function TgapText.GetLineHeight: double;
Begin
   Result := FGlyphCache.LineHeight;
End;

Function TgapText.GetScale: double;
Begin
   Result := FGlyphCache.Scale;
End;

Function TgapText.GetFontHeight: double;
Begin
   Result := FGlyphCache.FontHeight;
End;

Function TgapText.GetInvertY: boolean;
Begin
   Result := FGlyphCache.InvertY;
End;

Function TgapText.GetKerning: boolean;
Begin
   Result := FGlyphCache.Kerning;
End;

Function TgapText.GetUnderlined: Boolean;
Begin
   Result := FGlyphCache.Underlined;
End;

Function TgapText.GetStrikeOut: Boolean;
Begin
   Result := FGlyphCache.StrikeOut;
End;

Procedure TgapText.SetFontHeight(Const Val: double);
Begin
   FGlyphCache.FontHeight := Val;
End;

Procedure TgapText.SetInvertY(Const Val: boolean);
Begin
   FGlyphCache.InvertY := Val;
End;

Procedure TgapText.SetKerning(Const Val: boolean);
Begin
   FGlyphCache.Kerning := Val;
End;

Procedure TgapText.SetUnderlined(Const Val: Boolean);
Begin
   FGlyphCache.Underlined := Val;
End;

Procedure TgapText.SetStrikeOut(Const Val: Boolean);
Begin
   FGlyphCache.StrikeOut := Val;
End;

Function TgapText.GetCharInfo(charOrdinal: WORD): PgaGlyphInfo;
Begin
   Result := FGlyphCache.GetCharInfo(charOrdinal);
End;

Function TgapText.GetTextGlyphs(x, y: double; Const text: String): TPathsD;
Begin
   Result := FGlyphCache.GetTextGlyphs(x, y, text);
End;

Function TgapText.GetTextGlyphs(x, y: double; Const text: String; Out nextX: double): TPathsD;
Begin
   Result := FGlyphCache.GetTextGlyphs(x, y, text, nextX);
End;

Function TgapText.GetTextGlyphs(Const rec: TRect; Const text: String;
   textAlign: TgaTextAlign; textAlignV: TgaTextVAlign;
   Out nextIdx: integer; Out nextPt: TPointD): TPathsD;
Begin
   Result := FGlyphCache.GetTextGlyphs(rec, text, textAlign, textAlignV, nextIdx, nextPt);
End;

Function TgapText.GetAngledTextGlyphs(x, y: double; Const text: String; angleRadians: double; Const rotatePt: TPointD;
   Out nextPt: TPointD): TPathsD;
Begin
   Result := FGlyphCache.GetAngledTextGlyphs(x, y, text, angleRadians, rotatePt, nextPt);
End;

Function TgapText.GetCharOffsets(Const text: String; interCharSpace: double = 0): TArrayOfDouble;
Begin
   Result := FGlyphCache.GetCharOffsets(text, interCharSpace);
End;

Function TgapText.GetTextWidth(Const text: String): double;
Begin
   Result := FGlyphCache.GetTextWidth(text);
End;

Function TgapText.DrawText(x, y: double; Const text: String): double;
Begin
   Result := gaFonts.DrawText(FImagePlus, x, y, text, FGlyphCache, FColor);
End;

Function TgapText.DrawText(x, y: double; Const text: String; TextColor: TColor32): double;
Begin
   Result := gaFonts.DrawText(FImagePlus, x, y, text, FGlyphCache, TextColor);
End;

Function TgapText.DrawTextRenderer(x, y: double; Const text: String): double;
Begin
   Result := gaFonts.DrawText(FImagePlus, x, y, text, FGlyphCache, FRenderer);
End;

Function TgapText.DrawText(Const rec: TRect; Const text: String; textAlign: TgaTextAlign; textAlignV: TgaTextVAlign): TPointD;
Begin
   Result := gaFonts.DrawText(FImagePlus, rec, text, textAlign, textAlignV, FGlyphCache, FColor);
End;

Function TgapText.DrawText(Const rec: TRect; Const text: String; textAlign: TgaTextAlign; textAlignV: TgaTextVAlign; TextColor: TColor32): TPointD;
Begin
   Result := gaFonts.DrawText(FImagePlus, rec, text, textAlign, textAlignV, FGlyphCache, TextColor);
End;

Function TgapText.DrawAngledDegText(x, y: double; angleDeg: double; Const text: String): TPointD;
Begin
   Result := gaFonts.DrawAngledText(FImagePlus, x, y, DegToRad(angleDeg), text, FGlyphCache, FColor);
End;

Function TgapText.DrawAngledDegText(x, y: double; angleDeg: double; Const text: String; TextColor: TColor32): TPointD;
Begin
   Result := gaFonts.DrawAngledText(FImagePlus, x, y, DegToRad(angleDeg), text, FGlyphCache, TextColor);
End;

Function TgapText.DrawAngledRadText(x, y: double; angleRad: double; Const text: String): TPointD;
Begin
   Result := gaFonts.DrawAngledText(FImagePlus, x, y, angleRad, text, FGlyphCache, FColor);
End;

Function TgapText.DrawAngledRadText(x, y: double; angleRad: double; Const text: String; TextColor: TColor32): TPointD;
Begin
   Result := gaFonts.DrawAngledText(FImagePlus, x, y, angleRad, text, FGlyphCache, TextColor);
End;

Function TgapText.DrawVerticalText(x, y, interCharSpace: double; Const text: String): double;
Begin
   Result := gaFonts.DrawVerticalText(FImagePlus, x, y, interCharSpace, text, FGlyphCache, FColor);
End;

Function TgapText.DrawVerticalText(x, y, interCharSpace: double; Const text: String; TextColor: TColor32): double;
Begin
   Result := gaFonts.DrawVerticalText(FImagePlus, x, y, interCharSpace, text, FGlyphCache, TextColor);
End;

Constructor TgapPolygonD.Create(aImagePlus: TgaBitmapPlus);
Begin
   FImagePlus := aImagePlus;
   Clear;
   SetDefaultValues;
End;

Destructor TgapPolygonD.Destroy;
Begin
   SetLength(FPathD, 0);
   Inherited;
End;

Function TgapPolygonD.GetCount: Int64;
Begin
   Result := length(FPathD) + 1;
End;

Procedure TgapPolygonD.Clear;
Begin
   SetLength(FPathD, 0);
End;

Procedure TgapPolygonD.SetDefaultValues;
Begin
   FDrawElements := deBoth;

   FFillColor := clBlack32;
   FFillRule := frEvenOdd;
   FFillRenderer := Nil;
   FFillRendererUsed := False;

   FLineType := dltNormal;
   FLineColor := clBlack32;
   FLineWidth := 1;
   FLineEndStyle := esPolygon;
   FLineJoinStyle := jsAuto;
   FLineMiterLimit := 2;
   FLineDashPattern := Nil;
   FLinePatternOffset := 0.0;

   FLineRenderer := Nil;
   FLineRendererUsed := False;
End;

Procedure TgapPolygonD.DrawFill;
Begin
   If (FFillRenderer <> Nil) And FFillRendererUsed Then
      FImagePlus.DrawPolygon(Self.PathD, FFillRule, FFillRenderer)
   Else
      FImagePlus.DrawPolygon(Self.PathD, FFillRule, FFillColor);
End;

Procedure TgapPolygonD.DrawLine;
Var
   xUseRdr: Boolean;
Begin
   xUseRdr := ((FLineRenderer <> Nil) And FLineRendererUsed);

   Case FLineType Of
      dltNormal:
         Begin
            If xUseRdr Then
               FImagePlus.DrawLine(Self.PathD, FlineWidth, FLineRenderer, FLineEndStyle, FLineJoinStyle, FLineMiterLimit)
            Else
               FImagePlus.DrawLine(Self.PathD, FlineWidth, FLineColor, FLineEndStyle, FLineJoinStyle, FLineMiterLimit);
         End;
      dltInverted:
         Begin
            FImagePlus.DrawInvertedLine(Self.PathD, FlineWidth, FLineEndStyle, FLineJoinStyle);
         End;
      dltDashed:
         Begin
            If xUseRdr Then
               FImagePlus.DrawDashedLine(Self.PathD, FLineDashPattern, PDouble(@FlinePatternOffset), FlineWidth, FLineRenderer, FLineEndStyle, FLineJoinStyle)
            Else
               FImagePlus.DrawDashedLine(Self.PathD, FLineDashPattern, PDouble(@FlinePatternOffset), FlineWidth, FLineColor, FLineEndStyle, FLineJoinStyle);
         End;
      dltInvertedDashed:
         Begin
            FImagePlus.DrawInvertedDashedLine(Self.PathD, FLineDashPattern, PDouble(@FlinePatternOffset), FlineWidth, FLineEndStyle, FLineJoinStyle);
         End;

   End;
End;

Procedure TgapPolygonD.Draw;
Begin
   Case FDrawElements Of
      deBoth:
         Begin
            DrawFill;
            DrawLine;
         End;
      deOnlyFill: DrawFill;
      deOnlyLine: DrawLine;
      deErasePolygon: FImagePlus.ErasePolygon(Self.PathD, FFillRule);
   End;
End;

Procedure TgapPolygonD.Rectangle(Const rec: TRect);
Begin
   SetLength(FPathD, 0);
   FPathD := gaEngine.Rectangle(rec);
   Draw;
End;

Procedure TgapPolygonD.Rectangle(Const rec: TRectD);
Begin
   SetLength(FPathD, 0);
   FPathD := gaEngine.Rectangle(rec);
   Draw;
End;

Procedure TgapPolygonD.Rectangle(l, t, r, b: double);
Begin
   SetLength(FPathD, 0);
   FPathD := gaEngine.Rectangle(l, t, r, b);
   Draw;
End;

Procedure TgapPolygonD.RoundRect(Const rec: TRect; radius: integer);
Begin
   SetLength(FPathD, 0);
   FPathD := gaEngine.RoundRect(rec, radius);
   Draw;
End;

Procedure TgapPolygonD.RoundRect(Const rec: TRectD; radius: double);
Begin
   SetLength(FPathD, 0);
   FPathD := gaEngine.RoundRect(rec, radius);
   Draw;
End;

Procedure TgapPolygonD.RoundRect(Const rec: TRect; radius: TPoint);
Begin
   SetLength(FPathD, 0);
   FPathD := gaEngine.RoundRect(rec, radius);
   Draw;
End;

Procedure TgapPolygonD.RoundRect(Const rec: TRectD; radius: TPointD);
Begin
   SetLength(FPathD, 0);
   FPathD := gaEngine.RoundRect(rec, radius);
   Draw;
End;

Procedure TgapPolygonD.Ellipse(Const rec: TRect; steps: integer = 0);
Begin
   SetLength(FPathD, 0);
   FPathD := gaEngine.Ellipse(rec, steps);
   Draw;
End;

Procedure TgapPolygonD.Ellipse(Const rec: TRectD; steps: integer = 0);
Begin
   SetLength(FPathD, 0);
   FPathD := gaEngine.Ellipse(rec, steps);
   Draw;
End;

Procedure TgapPolygonD.Ellipse(Const rec: TRectD; scale: double);
Begin
   SetLength(FPathD, 0);
   FPathD := gaEngine.Ellipse(rec, scale);
   Draw;
End;

Procedure TgapPolygonD.Circle(Const pt: TPoint; radius: double);
Begin
   SetLength(FPathD, 0);
   FPathD := gaEngine.Circle(pt, radius);
   Draw;
End;

Procedure TgapPolygonD.Circle(Const pt: TPointD; radius: double);
Begin
   SetLength(FPathD, 0);
   FPathD := gaEngine.Circle(pt, radius);
   Draw;
End;

Procedure TgapPolygonD.Star(Const focalPt: TPointD; innerRadius, outerRadius: double; points: integer);
Begin
   SetLength(FPathD, 0);
   FPathD := gaEngine.Star(focalPt, innerRadius, outerRadius, points);
   Draw;
End;

Procedure TgapPolygonD.Arc(Const rec: TRectD; startAngle, endAngle: double; scale: double = 0);
Begin
   SetLength(FPathD, 0);
   FPathD := gaEngine.Arc(rec, startAngle, endAngle, scale);
   Draw;
End;

Procedure TgapPolygonD.Pie(Const rec: TRectD; StartAngle, EndAngle: double; scale: double = 0);
Begin
   SetLength(FPathD, 0);
   FPathD := gaEngine.Pie(rec, StartAngle, EndAngle, scale);
   Draw;
End;

Procedure TgapPolygonD.FlattenQBezier(Const pt1, pt2, pt3: TPointD; tolerance: double = 0.0);
Begin
   SetLength(FPathD, 0);
   FPathD := gaEngine.FlattenQBezier(pt1, pt2, pt3, tolerance);
   Draw;
End;

Procedure TgapPolygonD.FlattenQBezier(Const tolerance: double = 0.0);
Begin
   SetLength(FPathD, 0);
   FPathD := gaEngine.FlattenQBezier(FPathD, tolerance);
   Draw;
End;

Procedure TgapPolygonD.FlattenQBezier(Const firstPt: TPointD; Const pts: TPathD; tolerance: double = 0.0);
Begin
   SetLength(FPathD, 0);
   FPathD := gaEngine.FlattenQBezier(pts, tolerance);
   Draw;
End;

Procedure TgapPolygonD.FlattenCBezier(Const pt1, pt2, pt3, pt4: TPointD; tolerance: double = 0.0);
Begin
   SetLength(FPathD, 0);
   FPathD := gaEngine.FlattenCBezier(pt1, pt2, pt3, pt4, tolerance);
   Draw;
End;

Procedure TgapPolygonD.FlattenCBezier(Const tolerance: double = 0.0);
Begin
   SetLength(FPathD, 0);
   FPathD := gaEngine.FlattenCBezier(FPathD, tolerance);
   Draw;
End;

Constructor TgapPolygonsD.Create(aImagePlus: TgaBitmapPlus);
Begin
   FImagePlus := aImagePlus;
   Clear;
   SetDefaultValues;
End;

Destructor TgapPolygonsD.Destroy;
Begin
   SetLength(FPathsD, 0);
   Inherited;
End;

Function TgapPolygonsD.GetCount: Int64;
Begin
   Result := length(FPathsD) + 1;
End;

Procedure TgapPolygonsD.Clear;
Begin
   SetLength(FPathsD, 0);
End;

Procedure TgapPolygonsD.SetDefaultValues;
Begin
   FDrawElements := deBoth;

   FFillColor := clBlack32;
   FFillBackColor := clWhite32;
   FFillRule := frEvenOdd;
   FFillRenderer := Nil;
   FFillRendererUsed := False;

   FLineColor := clBlack32;
   FLineWidth := 1;
   FLineEndStyle := esPolygon;
   FLineJoinStyle := jsAuto;
   FLineMiterLimit := 2;
   FLineRenderer := Nil;
   FLineRendererUsed := False;
End;

Procedure TgapPolygonsD.DrawFill;
Begin
   If (FFillRenderer <> Nil) And FFillRendererUsed Then
      FImagePlus.DrawPolygon(Self.PathsD, FFillRule, FFillRenderer)
   Else
      FImagePlus.DrawPolygon(Self.PathsD, FFillRule, FFillColor);
End;

Procedure TgapPolygonsD.DrawLine;
Var
   xUseRdr: Boolean;
Begin
   xUseRdr := ((FLineRenderer <> Nil) And FLineRendererUsed);

   Case FLineType Of
      dltNormal:
         Begin
            If xUseRdr Then
               FImagePlus.DrawLine(Self.PathsD, FlineWidth, FLineRenderer, FLineEndStyle, FLineJoinStyle, FLineMiterLimit)
            Else
               FImagePlus.DrawLine(Self.PathsD, FlineWidth, FLineColor, FLineEndStyle, FLineJoinStyle, FLineMiterLimit);
         End;
      dltInverted:
         Begin
            FImagePlus.DrawInvertedLine(Self.PathsD, FlineWidth, FLineEndStyle, FLineJoinStyle);
         End;
      dltDashed:
         Begin
            If xUseRdr Then
               FImagePlus.DrawDashedLine(Self.PathsD, FLineDashPattern, PDouble(@FlinePatternOffset), FlineWidth, FLineRenderer, FLineEndStyle, FLineJoinStyle)
            Else
               FImagePlus.DrawDashedLine(Self.PathsD, FLineDashPattern, PDouble(@FlinePatternOffset), FlineWidth, FLineColor, FLineEndStyle, FLineJoinStyle);
         End;
      dltInvertedDashed:
         Begin
            FImagePlus.DrawInvertedDashedLine(Self.PathsD, FLineDashPattern, PDouble(@FlinePatternOffset), FlineWidth, FLineEndStyle, FLineJoinStyle);
         End;

   End;
End;

Procedure TgapPolygonsD.Draw;
Begin
   Case FDrawElements Of
      deBoth:
         Begin
            DrawFill;
            DrawLine;
         End;
      deOnlyFill: DrawFill;
      deOnlyLine: DrawLine;
      deErasePolygon: FImagePlus.ErasePolygon(Self.PathsD, FFillRule);
      deClearType: FImagePlus.DrawPolygon_ClearType(Self.PathsD, FFillRule, FFillColor, FFillBackColor);
   End;
End;

Procedure TgapPolygonsD.FlattenQBezier(Const tolerance: double = 0.0);
Begin
   SetLength(FPathsD, 0);
   FPathsD := gaEngine.FlattenQBezier(FPathsD, tolerance);
End;

Procedure TgapPolygonsD.FlattenCBezier(Const tolerance: double = 0.0);
Begin
   SetLength(FPathsD, 0);
   FPathsD := gaEngine.FlattenQBezier(FPathsD, tolerance);
End;

Constructor TgapPalette.Create;
Begin
   Inherited;
   SetLength(FPal, 0);
End;

Destructor TgapPalette.Destroy;
Begin
   SetLength(FPal, 0);
   Inherited;
End;

Procedure TgapPalette.BlackWhitePal;
Begin
   SetLength(FPal, 0);
   FPal := gaPalette.BlackWhitePal;
End;

Procedure TgapPalette.DefaultMacPal16;
Begin
   SetLength(FPal, 0);
   FPal := gaPalette.DefaultMacPal16;
End;

Procedure TgapPalette.DefaultWinPal16;
Begin
   SetLength(FPal, 0);
   FPal := gaPalette.DefaultWinPal16;
End;

Procedure TgapPalette.TrimPalette(Const colorFrequency: TArrayOfInteger; newSize: integer);
Begin
   FPal := gaPalette.TrimPalette(FPal, colorFrequency, newSize)
End;

Procedure TgapPalette.TrimPalette(Const colorFrequency: TArrayOfInteger; fraction: double);
Begin
   FPal := gaPalette.TrimPalette(FPal, colorFrequency, fraction)
End;

Function TgapPalette.CreateLogPalette: TMaxLogPalette;
Begin
   Result := gaBitMap.CreateLogPalette(FPal);
End;

Constructor TgaBitmapPlus.Create(aWidth: Integer = 0; aHeight: Integer = 0);
Begin
   Inherited Create(aWidth, aHeight);
   FText := TgapText.Create(self);
   FPolygonD := TgapPolygonD.Create(self);
   FPolygonsD := TgapPolygonsD.Create(self);
   FOutBitMap := Nil;
End;

Destructor TgaBitmapPlus.Destroy;
Begin
   FPolygonD.Free;
   FPolygonsD.Free;
   FText.Free;
   FOutBitMap := Nil;
   Inherited;
End;

Procedure TgaBitmapPlus.Invalidate;
Begin
   If FOutBitMap = Nil Then
      Exit;
   Self.CopyToTBitMap(FOutBitMap);
End;

Procedure TgaBitmapPlus.SetOutBitMap(Val: TBitMap);
Begin
   FOutBitMap := Val;
End;

Function TgaBitmapPlus.MakePalette(Const MaxColors: integer): TgapPalette;
Begin
   Result := TgapPalette.Create;
   Result.FPal := gaPalette.MakePalette(Self, MaxColors);
End;

Function TgaBitmapPlus.MakePalette(Const MaxColors: integer; Out frequencies: TArrayOfInteger): TgapPalette;
Begin
   Result := TgapPalette.Create;
   Result.FPal := gaPalette.MakePalette(Self, MaxColors, frequencies);
End;

Function TgaBitmapPlus.MakeAndApplyPalette(Const MaxColors: integer; UseDithering: Boolean): TgapPalette;
Begin
   Result := TgapPalette.Create;
   Result.FPal := gaPalette.MakeAndApplyPalette(Self, MaxColors, UseDithering);
End;

Function TgaBitmapPlus.MakeAndApplyPalette(Const MaxColors: integer; UseDithering: Boolean;
   Out frequencies: TArrayOfInteger): TgapPalette;
Begin
   Result := TgapPalette.Create;
   Result.FPal := gaPalette.MakeAndApplyPalette(Self, MaxColors, UseDithering, frequencies);
End;

Procedure TgaBitmapPlus.ApplyPalette(Const palette: TgapPalette; UseDithering: Boolean = true);
Begin
   gaPalette.ApplyPalette(Self, palette.FPal, UseDithering);
End;

Procedure TgaBitmapPlus.DrawPalette(Const palette: TgapPalette);
Begin
   gaPalette.DrawPalette(Self, palette.FPal);
End;

End.
