unit sysloader;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils,
  ctUtils;

Type

  TLoadType = (ltCTRuntimes, ltOS, ltUser);

  TbaseLoader = Class(TComponent)
  private
  protected        
    FActive:Boolean;
    FUserDirectory:TDirectoryName;
    FLoadType:TLoadType;
    FCTSubDir:String;
    Procedure SetActive(Const Value:Boolean);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    Procedure Libraries_Load; virtual;
    Procedure Libraries_Free; virtual;
    Function  GetLibrariesDir:String;
  published
    Property Active:Boolean read FActive write SetActive;
    Property UserDirectory:String read FUserDirectory write FUserDirectory;
    Property LoadType:TLoadType read FLoadType write FLoadType;
    Property LibrariesDir:String read GetLibrariesDir stored False;
  end;

implementation
{$IF DEFINED(WINDOWS))}
  Uses Windows;
{$ELSEIF DEFINED(UNIX)}
  Uses Unix;
{$ENDIF}

constructor TbaseLoader.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FLoadType:=ltCTRuntimes;
  FCTSubDir:='';
end;

destructor TbaseLoader.Destroy;
begin
  inherited Destroy;
end;

Procedure TbaseLoader.SetActive(Const Value:Boolean);
begin
  If Value=FActive then Exit;

  FActive:=Value;

  if FActive then
   Libraries_Load else
   Libraries_Free;
end;

Function TbaseLoader.GetLibrariesDir:String;
begin
  Result:='';

  case FLoadType of
     ltCTRuntimes: if FCTSubDir='' then
                      Result:=ctUtils.ctGetRuntimesCPUOSDir(False) else 
                      Result:=ctUtils.ctGetRuntimesCPUOSDir(True)+FCTSubDir;
     ltOS        : Result:='';
     ltUser      : Result := FUserDirectory;
  end;
end;

Procedure TbaseLoader.Libraries_Load;
 var
    xORIGINAL_ENV_PATH:String;
    xLoadDir:String;
begin
  Libraries_Free;

  xLoadDir:=GetLibrariesDir;

  xORIGINAL_ENV_PATH:=SysUtils.GetEnvironmentVariable('PATH');

  if xLoadDir<>'' then
   begin
    {$IF DEFINED(MSWINDOWS))}
       if xORIGINAL_ENV_PATH='' then
         Windows.SetEnvironmentVariable('PATH',PCHAR(xLoadDir)) else
         Windows.SetEnvironmentVariable('PATH',PCHAR(xLoadDir+';'+xORIGINAL_ENV_PATH));
    {$elseif defined(WinCE)}
      //...Nothing ...
     {$ELSE}
      fpSystem('export PATH=$PATH:'+xLoadDir);
     {$ENDIF}
   end;

  //.............................


  FActive:=True;
end;

Procedure TbaseLoader.Libraries_Free;
begin
  FActive:=False;
end;


end.

