{**************************************************************************
                Copyright (c) PilotLogic Software House

 Package bs_Engines
 This file is part Of CodeTyphon Studio (https://www.pilotlogic.com)

 ***** BEGIN LICENSE BLOCK *****
 * Version: LGPL-2 with link exception (Modified LGPL)
 *
 * The contents Of this file are subject to the
 * GNU LIBRARY GENERAL PUBLIC LICENSE Version 2 (the "License");
 * with the following modification:
 *
 * As a special exception, the copyright holders Of this library give you
 * permission to link this library with independent modules to produce an
 * executable, regardless Of the license terms Of these independent modules,
 * and to copy and distribute the resulting executable under terms Of your choice,
 * provided that you also meet, For each linked independent module, the terms
 * and conditions Of the license Of that module. An independent module is a
 * module which is NOT derived from or based on this library. If you modify this
 * library, you may extend this exception to your version Of the library, but
 * you are NOT obligated to do so. If you do NOT wish to do so, delete this
 * exception statement from your version.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * For the specific language governing rights and limitations under the
 * License.
 *
 * ***** END LICENSE BLOCK *****

 ***************************************************************************}

Unit gaEngine;

{$MODE objfpc}{$H+}

Interface

Uses
   Types, SysUtils, Classes, 
   Math, gaResStrings;

Const    
   RT_BITMAP = 'RT_BITMAP';
   cnAngle180 = Pi;
   cnAngle360 = Pi * 2;
   cnAngle0 = 0;
   cnAngle1 = Pi / 180;
   cnAngle15 = Pi / 12;
   cnAngle30 = cnAngle15 * 2;
   cnAngle45 = cnAngle15 * 3;
   cnAngle60 = cnAngle15 * 4;
   cnAngle75 = cnAngle15 * 5;
   cnAngle90 = Pi / 2;
   cnAngle105 = cnAngle180 - cnAngle75;
   cnAngle120 = cnAngle180 - cnAngle60;
   cnAngle135 = cnAngle180 - cnAngle45;
   cnAngle150 = cnAngle180 - cnAngle30;
   cnAngle165 = cnAngle180 - cnAngle15;
   cnAngle195 = cnAngle180 + cnAngle15;
   cnAngle210 = cnAngle180 + cnAngle30;
   cnAngle225 = cnAngle180 + cnAngle45;
   cnAngle240 = cnAngle180 + cnAngle60;
   cnAngle255 = cnAngle180 + cnAngle75;
   cnAngle270 = cnAngle360 - cnAngle90;
   cnAngle285 = cnAngle360 - cnAngle75;
   cnAngle300 = cnAngle360 - cnAngle60;
   cnAngle315 = cnAngle360 - cnAngle45;
   cnAngle330 = cnAngle360 - cnAngle30;
   cnAngle345 = cnAngle360 - cnAngle15;
Type
   TRect = Types.TRect;
   TColor32 = Type cardinal;

   TPointD = Record
      X, Y: double;
   End;

Const
   clNone32 = TColor32($00000000);
   clBlack32 = TColor32($FF000000);
   clDimGray32 = TColor32($FF3F3F3F);
   clGray32 = TColor32($FF7F7F7F);
   clLightGray32 = TColor32($FFBFBFBF);
   clWhite32 = TColor32($FFFFFFFF);
   clMaroon32 = TColor32($FF7F0000);
   clGreen32 = TColor32($FF007F00);
   clOlive32 = TColor32($FF7F7F00);
   clNavy32 = TColor32($FF00007F);
   clPurple32 = TColor32($FF7F007F);
   clTeal32 = TColor32($FF007F7F);
   clRed32 = TColor32($FFFF0000);
   clLime32 = TColor32($FF00FF00);
   clYellow32 = TColor32($FFFFFF00);
   clBlue32 = TColor32($FF0000FF);
   clFuchsia32 = TColor32($FFFF00FF);
   clAqua32 = TColor32($FF00FFFF);

   clAliceBlue32 = TColor32($FFF0F8FF);
   clAntiqueWhite32 = TColor32($FFFAEBD7);
   clAquamarine32 = TColor32($FF7FFFD4);
   clAzure32 = TColor32($FFF0FFFF);
   clBeige32 = TColor32($FFF5F5DC);
   clBisque32 = TColor32($FFFFE4C4);
   clBlancheDalmond32 = TColor32($FFFFEBCD);
   clBlueViolet32 = TColor32($FF8A2BE2);
   clBrown32 = TColor32($FFA52A2A);
   clBurlyWood32 = TColor32($FFDEB887);
   clCadetblue32 = TColor32($FF5F9EA0);
   clChartReuse32 = TColor32($FF7FFF00);
   clChocolate32 = TColor32($FFD2691E);
   clCoral32 = TColor32($FFFF7F50);
   clCornFlowerBlue32 = TColor32($FF6495ED);
   clCornSilk32 = TColor32($FFFFF8DC);
   clCrimson32 = TColor32($FFDC143C);
   clDarkBlue32 = TColor32($FF00008B);
   clDarkCyan32 = TColor32($FF008B8B);
   clDarkGoldenRod32 = TColor32($FFB8860B);
   clDarkGray32 = TColor32($FFA9A9A9);
   clDarkGreen32 = TColor32($FF006400);
   clDarkGrey32 = TColor32($FFA9A9A9);
   clDarkKhaki32 = TColor32($FFBDB76B);
   clDarkMagenta32 = TColor32($FF8B008B);
   clDarkOliveGreen32 = TColor32($FF556B2F);
   clDarkOrange32 = TColor32($FFFF8C00);
   clDarkOrchid32 = TColor32($FF9932CC);
   clDarkRed32 = TColor32($FF8B0000);
   clDarkSalmon32 = TColor32($FFE9967A);
   clDarkSeaGreen32 = TColor32($FF8FBC8F);
   clDarkSlateBlue32 = TColor32($FF483D8B);
   clDarkSlateGray32 = TColor32($FF2F4F4F);
   clDarkSlateGrey32 = TColor32($FF2F4F4F);
   clDarkTurquoise32 = TColor32($FF00CED1);
   clDarkViolet32 = TColor32($FF9400D3);
   clDeepPink32 = TColor32($FFFF1493);
   clDeepSkyBlue32 = TColor32($FF00BFFF);
   clDodgerBlue32 = TColor32($FF1E90FF);
   clFireBrick32 = TColor32($FFB22222);
   clFloralWhite32 = TColor32($FFFFFAF0);
   clGainsBoro32 = TColor32($FFDCDCDC);
   clGhostWhite32 = TColor32($FFF8F8FF);
   clGold32 = TColor32($FFFFD700);
   clGoldenRod32 = TColor32($FFDAA520);
   clGreenYellow32 = TColor32($FFADFF2F);
   clGrey32 = TColor32($FF808080);
   clHoneyDew32 = TColor32($FFF0FFF0);
   clHotPink32 = TColor32($FFFF69B4);
   clIndianRed32 = TColor32($FFCD5C5C);
   clIndigo32 = TColor32($FF4B0082);
   clIvory32 = TColor32($FFFFFFF0);
   clKhaki32 = TColor32($FFF0E68C);
   clLavender32 = TColor32($FFE6E6FA);
   clLavenderBlush32 = TColor32($FFFFF0F5);
   clLawnGreen32 = TColor32($FF7CFC00);
   clLemonChiffon32 = TColor32($FFFFFACD);
   clLightBlue32 = TColor32($FFADD8E6);
   clLightCoral32 = TColor32($FFF08080);
   clLightCyan32 = TColor32($FFE0FFFF);
   clLightGoldenRodYellow32 = TColor32($FFFAFAD2);
   clLightGreen32 = TColor32($FF90EE90);
   clLightGrey32 = TColor32($FFD3D3D3);
   clLightPink32 = TColor32($FFFFB6C1);
   clLightSalmon32 = TColor32($FFFFA07A);
   clLightSeagreen32 = TColor32($FF20B2AA);
   clLightSkyblue32 = TColor32($FF87CEFA);
   clLightSlategray32 = TColor32($FF778899);
   clLightSlategrey32 = TColor32($FF778899);
   clLightSteelblue32 = TColor32($FFB0C4DE);
   clLightYellow32 = TColor32($FFFFFFE0);
   clLtGray32 = TColor32($FFC0C0C0);
   clMedGray32 = TColor32($FFA0A0A4);
   clDkGray32 = TColor32($FF808080);
   clMoneyGreen32 = TColor32($FFC0DCC0);
   clLegacySkyBlue32 = TColor32($FFA6CAF0);
   clCream32 = TColor32($FFFFFBF0);
   clLimeGreen32 = TColor32($FF32CD32);
   clLinen32 = TColor32($FFFAF0E6);
   clMediumAquamarine32 = TColor32($FF66CDAA);
   clMediumBlue32 = TColor32($FF0000CD);
   clMediumOrchid32 = TColor32($FFBA55D3);
   clMediumPurple32 = TColor32($FF9370DB);
   clMediumSeaGreen32 = TColor32($FF3CB371);
   clMediumSlateBlue32 = TColor32($FF7B68EE);
   clMediumSpringGreen32 = TColor32($FF00FA9A);
   clMediumTurquoise32 = TColor32($FF48D1CC);
   clMediumVioletRed32 = TColor32($FFC71585);
   clMidnightBlue32 = TColor32($FF191970);
   clMintCream32 = TColor32($FFF5FFFA);
   clMistyRose32 = TColor32($FFFFE4E1);
   clMoccasin32 = TColor32($FFFFE4B5);
   clNavajoWhite32 = TColor32($FFFFDEAD);
   clOldLace32 = TColor32($FFFDF5E6);
   clOliveDrab32 = TColor32($FF6B8E23);
   clOrange32 = TColor32($FFFFA500);
   clOrangeRed32 = TColor32($FFFF4500);
   clOrchid32 = TColor32($FFDA70D6);
   clPaleGoldenRod32 = TColor32($FFEEE8AA);
   clPaleGreen32 = TColor32($FF98FB98);
   clPaleTurquoise32 = TColor32($FFAFEEEE);
   clPaleVioletred32 = TColor32($FFDB7093);
   clPapayaWhip32 = TColor32($FFFFEFD5);
   clPeachPuff32 = TColor32($FFFFDAB9);
   clPeru32 = TColor32($FFCD853F);
   clPlum32 = TColor32($FFDDA0DD);
   clPowderBlue32 = TColor32($FFB0E0E6);
   clRosyBrown32 = TColor32($FFBC8F8F);
   clRoyalBlue32 = TColor32($FF4169E1);
   clSaddleBrown32 = TColor32($FF8B4513);
   clSalmon32 = TColor32($FFFA8072);
   clSandyBrown32 = TColor32($FFF4A460);
   clSeaGreen32 = TColor32($FF2E8B57);
   clSeaShell32 = TColor32($FFFFF5EE);
   clSienna32 = TColor32($FFA0522D);
   clSilver32 = TColor32($FFC0C0C0);
   clSkyblue32 = TColor32($FF87CEEB);
   clSlateBlue32 = TColor32($FF6A5ACD);
   clSlateGray32 = TColor32($FF708090);
   clSlateGrey32 = TColor32($FF708090);
   clSnow32 = TColor32($FFFFFAFA);
   clSpringgreen32 = TColor32($FF00FF7F);
   clSteelblue32 = TColor32($FF4682B4);
   clTan32 = TColor32($FFD2B48C);
   clThistle32 = TColor32($FFD8BFD8);
   clTomato32 = TColor32($FFFF6347);
   clTurquoise32 = TColor32($FF40E0D0);
   clViolet32 = TColor32($FFEE82EE);
   clWheat32 = TColor32($FFF5DEB3);
   clWhitesmoke32 = TColor32($FFF5F5F5);
   clYellowgreen32 = TColor32($FF9ACD32);

   clTrWhite32 = TColor32($7FFFFFFF);
   clTrGray32 = TColor32($7F7F7F7F);
   clTrBlack32 = TColor32($7F000000);
   clTrRed32 = TColor32($7FFF0000);
   clTrGreen32 = TColor32($7F00FF00);
   clTrBlue32 = TColor32($7F0000FF);

Type
   PColor32 = ^TColor32;
   TArrayOfColor32 = Array Of TColor32;
   TArrayOfArrayOfColor32 = Array Of TArrayOfColor32;
   TArrayOfInteger = Array Of integer;
   TArrayOfWord = Array Of word;
   TArrayOfByte = Array Of byte;    

   PGRColor32Array = ^TGRColor32Array;
   TGRColor32Array = Array[0..0] Of TColor32;

   TClipboardPriority = (cpLow, cpMedium, cpHigh);
   TArrowStyle = (asNone, asSimple, asFancy, asDiamond, asCircle, asTail);
   TJoinStyle = (jsAuto, jsSquare, jsMiter, jsRound);
   TEndStyle = (esPolygon, esButt, esSquare, esRound);
   TPathEnd = (peStart, peEnd, peBothEnds);
   TSplineType = (stQuadratic, stCubic);
   TFillRule = (frEvenOdd, frNonZero, frPositive, frNegative);
   TgaButtonShape = (bsRound, bsSquare, bsDiamond);
   TTileFillStyle = (tfsRepeat, tfsMirrorHorz, tfsMirrorVert, tfsRotate180);
   TGradientFillStyle = (gfsClamp, gfsMirror, gfsRepeat);
   TgaButtonAttribute = (baShadow, ba3D);
   TgaButtonAttributes = Set Of TgaButtonAttribute;

   PPointD = ^TPointD;
   TPathD  = Array Of TPointD;
   TPathsD = Array Of TPathD;
   TArrayOfPathsD = Array Of TPathsD;
   TArrayOfDouble = Array Of double;
   TArrayOfString = Array Of String;

   PMatrixD = ^TMatrixD;
   TMatrixD = Array[0..2, 0..2] Of double;

   PARGB = ^TARGB;
   TARGB = Packed Record
      Case boolean Of
         False: (
            B: byte;
            G: byte;
            R: byte;
            A: byte);
         True: (Color: TColor32);
   End;
   TArrayOfARGB = Array Of TARGB;
   PArgbArray = ^TArrayOfARGB;

   THsl = Packed Record
      hue: byte;
      sat: byte;
      lum: byte;
      alpha: byte;
   End;
   PHsl = ^THsl;
   TArrayofHSL = Array Of THsl;

   TGradientColor = Record
      offset: double;
      color: TColor32;
   End;
   TArrayOfGradientColor = Array Of TGradientColor;

   PPt = ^TPt;
   TPt = Record
      pt: TPointD;
      vec: TPointD;
      len: double;
      Next: PPt;
      prev: PPt;
   End;

Type
   TgaGraphMem = Class;

   TBoundsProc = Function(dist, colorsCnt: integer): integer;
   TBoundsProcD = Function(dist: double; colorsCnt: integer): integer;
   TgaGraphMemChangeProc = Procedure Of Object;
   TBlendFunction = Function(bgColor, fgColor: TColor32): TColor32;
   TCompareFunction = Function(master, current: TColor32; Data: integer): boolean;
   TCompareFunctionEx = Function(master, current: TColor32): byte;
   TResamplerFunction = Function(img: TgaGraphMem; x256, y256: integer): TColor32;
   TgaImageFormatClass = Class Of TgaImageFormat;

   TgaImageFormat = Class
   Public
      Class Function IsValidImageStream(stream: TStream): boolean; Virtual; Abstract;
      Procedure SaveToStream(stream: TStream; gaBitMap: TgaGraphMem); Virtual; Abstract;
      Function SaveToFile(Const filename: String; gaBitMap: TgaGraphMem): boolean; Virtual;
      Function LoadFromStream(stream: TStream; gaBitMap: TgaGraphMem): boolean; Virtual; Abstract;
      Function LoadFromFile(Const filename: String; gaBitMap: TgaGraphMem): boolean; Virtual;
      Class Function CanCopyToClipboard: boolean; Virtual;
      Class Function CopyToClipboard(gaBitMap: TgaGraphMem): boolean; Virtual; Abstract;
      Class Function CanPasteFromClipboard: boolean; Virtual; Abstract;
      Class Function PasteFromClipboard(gaBitMap: TgaGraphMem): boolean; Virtual; Abstract;
   End;

   TgaRendererCustom = Class
   Private
      fImgWidth: integer;
      fImgHeight: integer;
      fImgBase: Pointer;
      fCurrY: integer;
      fCurrLinePtr: Pointer;
      fPixelSize: integer;
      fChangeProc: TgaGraphMemChangeProc;
   Protected
      Procedure NotifyChange;
      Function Initialize(imgBase: Pointer; imgWidth, imgHeight, pixelSize: integer): boolean; Overload; Virtual;
      Function Initialize(targetImage: TgaGraphMem): boolean; Overload; Virtual;
      Function GetDstPixel(x, y: integer): Pointer;
      Procedure RenderProc(x1, x2, y: integer; alpha: pbyte); Virtual; Abstract;
      Property ImgWidth: integer Read fImgWidth;
      Property ImgHeight: integer Read fImgHeight;
      Property ImgBase: Pointer Read fImgBase;
      Property PixelSize: integer Read fPixelSize;
   End;

   TgaGraphMem = Class
   Private
   Protected    
      fWidth: integer;
      fHeight: integer;
      fResampler: integer;
      fIsPremultiplied: boolean;
      fColorCount: integer;
      fPixels: TArrayOfColor32;
      fOnChange: TNotifyEvent;
      fOnResize: TNotifyEvent;
      fUpdateCnt: integer;
      Function GetPixel(x, y: integer): TColor32;
      Procedure SetPixel(x, y: integer; color: TColor32);
      Function GetIsBlank: boolean;
      Function GetIsEmpty: boolean;
      Function GetPixelBase: PColor32;
      Function GetPixelRow(row: integer): PColor32;
      Procedure NearestNeighborResize(newWidth, newHeight: integer);
      Procedure ResamplerResize(newWidth, newHeight: integer);
      Procedure RotateLeft90;
      Procedure RotateRight90;
      Procedure Rotate180;
      Function GetColorCount: integer;
      Function GetHasTransparency: boolean;
      Function GetBounds: TRect;
      Function GetMidPoint: TPointD;
      Procedure CopyInternal(src: TgaGraphMem; Const srcRec, dstRec: TRect; blendFunc: TBlendFunction);
      Procedure Changed; Virtual;
      Procedure Resized; Virtual;
      Property UpdateCount: integer Read fUpdateCnt;
   Public
      Constructor Create(Const aWidth: integer = 0; aHeight: integer = 0); Virtual; Overload;
      Constructor Create(aSrc: TgaGraphMem); Virtual; Overload;
      Constructor Create(aSrc: TgaGraphMem; Const srcRec: TRect); Virtual; Overload;
      Destructor Destroy; Override;
      Procedure BeginUpdate;
      Procedure EndUpdate;
      Procedure BlockUpdate;
      Procedure UnblockUpdate;

      Procedure SetSize(newWidth, newHeight: integer; color: TColor32 = 0);
      Procedure Resize(newWidth, newHeight: integer; stretchImage: boolean = True);
      Procedure ScaleToFit(Width, Height: integer);
      Procedure ScaleToFitCentered(Width, Height: integer);
      Procedure Scale(s: double); Overload;
      Procedure Scale(sx, sy: double); Overload;

      Class Function CanPasteFromClipBoard: boolean;
      Function CopyToClipBoard: boolean;
      Function PasteFromClipBoard: boolean;

      Procedure Assign(src: TgaGraphMem);
      Procedure AssignTo(dst: TgaGraphMem);
      Function  RectHasTransparency(rec: TRect): Boolean;

      Function  CopyPixels(rec: TRect): TArrayOfColor32;
      Function  Copy(src: TgaGraphMem; srcRec, dstRec: TRect): boolean;
      Function  CopyBlend(src: TgaGraphMem; srcRec, dstRec: TRect; blendFunc: TBlendFunction = Nil): boolean;
      
      Function  CopyToGR32(aGR32Bits: PGRColor32Array): Boolean;    
      Function  CopyFromGR32(aGR32Bits: PGRColor32Array; Const ScrWidth, ScrHeight: Integer): Boolean;
      
      Procedure Crop(Const rec: TRect);
      Procedure SetBackgroundColor(bgColor: TColor32);
      Procedure Clear(color: TColor32 = 0); Overload;
      Procedure Clear(Const rec: TRect; color: TColor32 = 0); Overload;
      Procedure FillRect(rec: TRect; color: TColor32);
      Procedure ConvertToBoolMask(reference: TColor32; tolerance: integer; colorFunc: TCompareFunction;
                                  maskBg: TColor32 = clWhite32; maskFg: TColor32 = clBlack32);
      Procedure ConvertToAlphaMask(reference: TColor32; colorFunc: TCompareFunctionEx);
      Procedure FlipVertical;
      Procedure FlipHorizontal;
      Procedure PreMultiply;
      Procedure SetAlpha(alpha: byte);
      Procedure ReduceOpacity(opacity: byte); Overload;
      Procedure ReduceOpacity(opacity: byte; rec: TRect); Overload;
      Procedure SetRGB(rgbColor: TColor32); Overload;
      Procedure SetRGB(rgbColor: TColor32; rec: TRect); Overload;
      Procedure Grayscale;
      Procedure InvertColors;
      Procedure InvertAlphas;
      Procedure Tile(newWidth, newHeight: integer; style: TTileFillStyle = tfsRepeat);
      Procedure AdjustHue(percent: integer);
      Procedure AdjustLuminance(percent: integer);
      Procedure AdjustSaturation(percent: integer);
      Function  CropTransparentPixels: TRect;
      Procedure Rotate(angleRads: double);
      Procedure RotateRect(Const rec: TRect; angleRads: double; eraseColor: TColor32 = 0);
      Procedure Skew(dx, dy: double);
      Procedure ScaleAlpha(aScale: double);

      Class Procedure RegisterImageFormatClass(ext: String; bm32ExClass: TgaImageFormatClass; clipPriority: TClipboardPriority);
      Class Function GeTgaImageFormatClass(Const ext: String): TgaImageFormatClass;
      Class Function IsRegisteredFormat(Const ext: String): boolean;
      Function SaveToFile(filename: String): boolean;
      Function SaveToStream(stream: TStream; Const FmtExt: String): boolean;
      Function LoadFromFile(Const filename: String): boolean;
      Function LoadFromStream(stream: TStream): boolean;
      Function LoadFromResource(Const resName: String; resType: PChar): boolean;

      Procedure DrawPoint(Const pt: TPointD; radius: double; color: TColor32); Overload;
      Procedure DrawPoint(Const pt: TPointD; radius: double; renderer: TgaRendererCustom); Overload;
      Procedure DrawPoint(Const points: TPathD; radius: double; color: TColor32); Overload;
      Procedure DrawPoint(Const paths: TPathsD; radius: double; color: TColor32); Overload;

      Procedure DrawLine(Const pt1, pt2: TPointD; lineWidth: double; color: TColor32); Overload;
      Procedure DrawLine(Const line: TPathD; lineWidth: double; color: TColor32; endStyle: TEndStyle;
                         joinStyle: TJoinStyle = jsAuto; miterLimit: double = 2); Overload;
      Procedure DrawLine(Const line: TPathD; lineWidth: double; renderer: TgaRendererCustom; endStyle: TEndStyle;
                         joinStyle: TJoinStyle = jsAuto; miterLimit: double = 2); Overload;
      Procedure DrawLine(Const Lines: TPathsD; lineWidth: double; color: TColor32; endStyle: TEndStyle;
                         joinStyle: TJoinStyle = jsAuto; miterLimit: double = 2); Overload;
      Procedure DrawLine(Const Lines: TPathsD; lineWidth: double; renderer: TgaRendererCustom;
                         endStyle: TEndStyle; joinStyle: TJoinStyle = jsAuto; miterLimit: double = 2); Overload;

      Procedure DrawInvertedLine(Const line: TPathD; lineWidth: double; endStyle: TEndStyle;
                                 joinStyle: TJoinStyle = jsAuto); Overload;
      Procedure DrawInvertedLine(Const Lines: TPathsD; lineWidth: double; endStyle: TEndStyle;
                                 joinStyle: TJoinStyle = jsAuto); Overload;

      Procedure DrawDashedLine(Const line: TPathD; dashPattern: TArrayOfInteger; patternOffset: PDouble;
                               lineWidth: double; color: TColor32; endStyle: TEndStyle; joinStyle: TJoinStyle = jsAuto); Overload;
      Procedure DrawDashedLine(Const Lines: TPathsD; dashPattern: TArrayOfInteger; patternOffset: PDouble;
                               lineWidth: double; color: TColor32; endStyle: TEndStyle; joinStyle: TJoinStyle = jsAuto); Overload;
      Procedure DrawDashedLine(Const line: TPathD; dashPattern: TArrayOfInteger; patternOffset: PDouble;
                               lineWidth: double; renderer: TgaRendererCustom; endStyle: TEndStyle; joinStyle: TJoinStyle = jsAuto); Overload;
      Procedure DrawDashedLine(Const Lines: TPathsD; dashPattern: TArrayOfInteger; patternOffset: PDouble;
                               lineWidth: double; renderer: TgaRendererCustom; endStyle: TEndStyle; joinStyle: TJoinStyle = jsAuto); Overload;

      Procedure DrawInvertedDashedLine(Const line: TPathD; dashPattern: TArrayOfInteger;
                                       patternOffset: PDouble; lineWidth: double; endStyle: TEndStyle; joinStyle: TJoinStyle = jsAuto); Overload;
      Procedure DrawInvertedDashedLine(Const Lines: TPathsD; dashPattern: TArrayOfInteger;
                patternOffset: PDouble; lineWidth: double; endStyle: TEndStyle; joinStyle: TJoinStyle = jsAuto); Overload;
      Procedure DrawPolygon(Const polygon: TPathD; fillRule: TFillRule; color: TColor32); Overload;
      Procedure DrawPolygon(Const polygon: TPathD; fillRule: TFillRule; renderer: TgaRendererCustom); Overload;
      Procedure DrawPolygon(Const polygons: TPathsD; fillRule: TFillRule; color: TColor32); Overload;
      Procedure DrawPolygon(Const polygons: TPathsD; fillRule: TFillRule; renderer: TgaRendererCustom); Overload;
      Procedure DrawPolygon_ClearType(Const polygons: TPathsD; fillRule: TFillRule; color: TColor32; backColor: TColor32 = clWhite32);
      Procedure DrawBoolMask(Const mask: TArrayOfByte; color: TColor32 = clBlack32);
      Procedure DrawAlphaMask(Const mask: TArrayOfByte; color: TColor32 = clBlack32);
      Procedure DrawShadow(Const polygon: TPathD; fillRule: TFillRule; depth: double; angleRads: double = cnAngle45;
                           color: TColor32 = $80000000; cutoutInsideShadow: boolean = False); Overload;
      Procedure DrawShadow(Const polygons: TPathsD; fillRule: TFillRule; depth: double; angleRads: double = cnAngle45;
                           color: TColor32 = $80000000; cutoutInsideShadow: boolean = False); Overload;
      Procedure DrawGlow(Const polygon: TPathD; fillRule: TFillRule; color: TColor32; blurRadius: integer); Overload;
      Procedure DrawGlow(Const polygons: TPathsD; fillRule: TFillRule; color: TColor32; blurRadius: integer); Overload;
      Procedure Draw3D(Const polygon: TPathD; fillRule: TFillRule; Height, blurRadius: double; colorLt: TColor32 = $DDFFFFFF;
                       colorDk: TColor32 = $80000000; angleRads: double = cnAngle45); Overload;
      Procedure Draw3D(Const polygons: TPathsD; fillRule: TFillRule; Height, blurRadius: double; colorLt: TColor32 = $DDFFFFFF;
                       colorDk: TColor32 = $80000000; angleRads: double = cnAngle45); Overload;
      Function DrawButton(Const pt: TPointD; size: double; color: TColor32 = clNone32; buttonShape: TgaButtonShape = bsRound;
                          buttonAttributes: TgaButtonAttributes = [baShadow, ba3D]): TPathD;

      Procedure EraseColor(color: TColor32);
      Procedure ErasePolygon(Const polygon: TPathD; fillRule: TFillRule); Overload;
      Procedure ErasePolygon(Const polygons: TPathsD; fillRule: TFillRule); Overload;
      Procedure EraseInsidePath(Const path: TPathD; fillRule: TFillRule);
      Procedure EraseInsidePaths(Const paths: TPathsD; fillRule: TFillRule);
      Procedure EraseOutsidePath(Const path: TPathD; fillRule: TFillRule; Const outsideBounds: TRect);
      Procedure EraseOutsidePaths(Const paths: TPathsD; fillRule: TFillRule; Const outsideBounds: TRect);

      Procedure FloodFill(x, y: integer; newColor: TColor32; compareFunc: TCompareFunction = Nil; tolerance: integer = 0);
      Procedure FastGaussianBlur(Const rec: TRect; stdDev: integer; repeats: integer = 2);
      Procedure GaussianBlur(rec: TRect; radius: integer);
      Procedure Emboss(radius: integer = 1; depth: integer = 10; luminance: integer = 75; preserveColor: boolean = False);
      Procedure Sharpen(radius: integer = 2; amount: integer = 10);
      Procedure HatchBackground(color1: TColor32 = clWhite32; color2: TColor32 = $FFE8E8E8; hatchSize: integer = 10);
      Procedure GridBackground(majorInterval, minorInterval: integer; fillColor: TColor32 = clWhite32;
                               majColor: TColor32 = $30000000; minColor: TColor32 = $20000000);
      Procedure ReplaceColor(oldColor, newColor: TColor32);
      Procedure RedEyeRemove(Const rect: TRect);
      Procedure PencilEffect(intensity: integer = 0);
      Procedure TraceContours(intensity: integer);
      Function  Vectorize(compareColor: TColor32; compareFunc: TCompareFunction; colorTolerance: integer;
                          roundingTolerance: integer = 2): TPathsD;
      Function GetFloodFillMask(x, y: integer; compareFunc: TCompareFunction; tolerance: integer): TArrayOfByte;
      Procedure SymmetricCropTransparent;
      Procedure BoxDownSampling(newWidth, newHeight: integer);
      Procedure ApplyClearType(textColor: TColor32 = clBlack32; bkColor: TColor32 = clWhite32);
      Function AffineTransformImage(matrix: TMatrixD): TPoint;
      Function ProjectiveTransform(Const srcPts, dstPts: TPathD; Const margins: TRect): boolean;
      Function SplineVertTransform(Const topSpline: TPathD; splineType: TSplineType; backColor: TColor32;
                                   reverseFill: boolean; Out offset: TPoint): boolean;
      Function SplineHorzTransform(Const leftSpline: TPathD; splineType: TSplineType; backColor: TColor32;
                                   reverseFill: boolean; Out offset: TPoint): boolean;

      Function GetTransformBounds(Const matrix: TMatrixD): TRect;
      Function GetBoolMask(reference: TColor32; compareFunc: TCompareFunction; tolerance: integer): TArrayOfByte;
      Function GetByteMask(reference: TColor32; compareFunc: TCompareFunctionEx): TArrayOfByte;
      Function GetSymmetricCropTransparentRect: TRect;
      Function GetFloodFillBounds(x, y: integer; Const byteArray: TArrayOfByte): TRect;

      Property Width: integer Read fWidth;
      Property Height: integer Read fHeight;
      Property Bounds: TRect Read GetBounds;
      Property IsBlank: boolean Read GetIsBlank;
      Property IsEmpty: boolean Read GetIsEmpty;
      Property MidPoint: TPointD Read GetMidPoint;
      Property Pixel[x, y: integer]: TColor32 Read GetPixel Write SetPixel;
      Property Pixels: TArrayOfColor32 Read fPixels;
      Property PixelBase: PColor32 Read GetPixelBase;
      Property PixelRow[row: integer]: PColor32 Read GetPixelRow;
      Property ColorCount: integer Read GetColorCount;
      Property HasTransparency: boolean Read GetHasTransparency;
      Property Resampler: integer Read fResampler Write fResampler;
      Property OnChange: TNotifyEvent Read fOnChange Write fOnChange;
      Property OnResize: TNotifyEvent Read fOnResize Write fOnResize;
   End;

   TgaGraphMemList = Class
   Private
      fList: TList;
      fIsImageOwner: boolean;
      Function GetImage(index: integer): TgaGraphMem;
      Procedure SetImage(index: integer; img: TgaGraphMem);
      Function GetLast: TgaGraphMem;
   Public
      Constructor Create;
      Destructor Destroy; Override;
      Procedure Clear;
      Function  Count: integer;
      Procedure Add(image: TgaGraphMem); Overload;
      Function  Add(Width, Height: integer): TgaGraphMem; Overload;
      Procedure Insert(index: integer; image: TgaGraphMem);
      Procedure Move(currentIndex, newIndex: integer);
      Procedure Delete(index: integer);
                                                                                    
      Property IsImageOwner: boolean Read fIsImageOwner Write fIsImageOwner;
      Property Image[index: integer]: TgaGraphMem Read GetImage Write SetImage; Default;
      Property Last: TgaGraphMem Read GetLast;
   End;

   TRectD = Object
      Left, Top, Right, Bottom: double;
      Function IsEmpty: boolean;
      Function Width: double;
      Function Height: double;
      Function Normalize: boolean;
      Function Contains(Const Pt: TPoint): boolean; Overload;
      Function Contains(Const Pt: TPointD): boolean; Overload;
      Function TopLeft: TPointD;
      Function BottomRight: TPointD;
      Function MidPoint: TPointD;
   End;

Type
   TSizeD = Object
      sx: double;
      sy: double;
      Function average: double;
   End;

   TRectWH = Object
   Public
      Left, Top, Width, Height: double;
      Function IsEmpty: boolean;
      Function Right: double;
      Function Bottom: double;
      Function Contains(Const Pt: TPoint): boolean; Overload;
      Function Contains(Const Pt: TPointD): boolean; Overload;
      Function MidPoint: TPointD;
      Function RectD: TRectD;
   End;

   TgaFitCurveContainer = Class
   Private
      ppts: PPt;
      solution: TPathD;
      tolSqrd: double;
      Function Count(First, last: PPt): integer;
      Function AddPt(Const pt: TPointD): PPt;
      Procedure Clear;
      Function ComputeLeftTangent(p: PPt): TPointD;
      Function ComputeRightTangent(p: PPt): TPointD;
      Function ComputeCenterTangent(p: PPt): TPointD;
      Function ChordLengthParameterize(First: PPt; cnt: integer): TArrayOfDouble;
      Function GenerateBezier(First, last: PPt; cnt: integer; Const u: TArrayOfDouble; Const firstTan, lastTan: TPointD): TPathD;
      Function Reparameterize(First: PPt; cnt: integer; Const u: TArrayOfDouble; Const bezier: TPathD): TArrayOfDouble;
      Function NewtonRaphsonRootFind(Const q: TPathD; Const pt: TPointD; u: double): double;
      Function ComputeMaxErrorSqrd(First, last: PPt; Const bezier: TPathD; Const u: TArrayOfDouble; Out SplitPoint: PPt): double;
      Function FitCubic(First, last: PPt; firstTan, lastTan: TPointD): boolean;
      Procedure AppendSolution(Const bezier: TPathD);
   Public
      Function FitCurve(Const path: TPathD; closed: boolean; tolerance: double; minSegLength: double): TPathD;
   End;

   PgaWeightedColor = ^TgaWeightedColor;
   TgaWeightedColor = Object
   Private
      fAddCount: integer;
      fAlphaTot: int64;
      fColorTotR: int64;
      fColorTotG: int64;
      fColorTotB: int64;
      Function GetColor: TColor32;
   Public
      Procedure Reset;
      Procedure Add(c: TColor32; w: integer = 1); Overload;
      Procedure Add(Const other: TgaWeightedColor); Overload;
      Procedure Subtract(c: TColor32; w: integer = 1); Overload;
      Procedure Subtract(Const other: TgaWeightedColor); Overload;
      Procedure AddWeight(w: integer);

      Property AddCount: integer Read fAddCount;
      Property Color: TColor32 Read GetColor;
      Property Weight: integer Read fAddCount;
   End;
   TArrayOfWeightedColor = Array Of TgaWeightedColor;

   TgaRendererColor = Class(TgaRendererCustom)
   Private
      fAlpha: byte;
      fColor: TColor32;
   Protected
      Procedure RenderProc(x1, x2, y: integer; alpha: pbyte); Override;
      Function  Initialize(targetImage: TgaGraphMem): boolean; Override;
   Public
      Constructor Create(color: TColor32 = clNone32);
      Procedure SetColor(Value: TColor32);
   End;

   TgaRendererErase = Class(TgaRendererCustom)
   Protected
      Procedure RenderProc(x1, x2, y: integer; alpha: pbyte); Override;
   End;

   TgaRendererInverse = Class(TgaRendererCustom)
   Protected
      Procedure RenderProc(x1, x2, y: integer; alpha: pbyte); Override;
   End;

   TgaRendererImage = Class(TgaRendererCustom)
   Private
      fImage: TgaGraphMem;
      fOffset: TPoint;
      fBrushPixel: PARGB;
      fLastYY: integer;
      fMirrorY: boolean;
      fBoundsProc: TBoundsProc;
      Function GetFirstBrushPixel(x, y: integer): PARGB;
   Protected
      Procedure RenderProc(x1, x2, y: integer; alpha: pbyte); Override;
      Function  Initialize(targetImage: TgaGraphMem): boolean; Override;
   Public
      Constructor Create(tileFillStyle: TTileFillStyle = tfsRepeat; brushImage: TgaGraphMem = Nil);
      Destructor Destroy; Override;
      Procedure SetTileFillStyle(Value: TTileFillStyle);
      Property Image: TgaGraphMem Read fImage;
      Property Offset: TPoint Read fOffset Write fOffset;
   End;

   TgaRendererGradientCustom = Class(TgaRendererCustom)
   Private
      fBoundsProc: TBoundsProc;
      fGradientColors: TArrayOfGradientColor;
   Protected
      fColors: TArrayOfColor32;
      fColorsCnt: integer;
      Procedure SetGradientFillStyle(Value: TGradientFillStyle); Virtual;
   Public
      Constructor Create;
      Procedure SetParameters(startColor, endColor: TColor32; gradFillStyle: TGradientFillStyle = gfsClamp); Virtual;
      Procedure InsertColorStop(offsetFrac: double; color: TColor32);
      Procedure Clear;
   End;

   TgaRendererGradientLinear = Class(TgaRendererGradientCustom)
   Private
      fStartPt: TPointD;
      fEndPt: TPointD;
      fPerpendicOffsets: TArrayOfInteger;
      fIsVert: boolean;
   Protected
      Procedure RenderProc(x1, x2, y: integer; alpha: pbyte); Override;
      Function Initialize(targetImage: TgaGraphMem): boolean; Override;
   Public
      Procedure SetParameters(Const startPt, endPt: TPointD; startColor, endColor: TColor32;
                              gradFillStyle: TGradientFillStyle = gfsClamp); Reintroduce;
   End;

   TgaRendererGradientRadial = Class(TgaRendererGradientCustom)
   Private
      fCenterPt: TPointD;
      fScaleX: double;
      fScaleY: double;
   Protected
      Procedure RenderProc(x1, x2, y: integer; alpha: pbyte); Override;
      Function Initialize(targetImage: TgaGraphMem): boolean; Override;
   Public
      Procedure SetParameters(Const focalRect: TRect; innerColor, outerColor: TColor32;
                              gradientFillStyle: TGradientFillStyle = gfsClamp); Reintroduce;
   End;

   TgaRendererGradientSvgRadial = Class(TgaRendererGradientCustom)
   Private
      fA, fB: double;
      fAA, fBB: double;
      fCenterPt: TPointD;
      fFocusPt: TPointD;
      fBoundsProcD: TBoundsProcD;
   Protected
      Procedure RenderProc(x1, x2, y: integer; alpha: pbyte); Override;
      Function Initialize(targetImage: TgaGraphMem): boolean; Override;
   Public
      Procedure SetParameters(Const ellipseRect: TRect; Const focus: TPoint; innerColor, outerColor: TColor32;
                              gradientFillStyle: TGradientFillStyle = gfsClamp); Reintroduce;
   End;

   TgaRendererBarycentric = Class(TgaRendererCustom)
   Private
      a: TPointD;
      c1, c2, c3: TARGB;
      v0, v1: TPointD;
      d00, d01, d11, invDenom: double;
      Function GetColor(Const pt: TPointD): TColor32;
   Protected
      Procedure RenderProc(x1, x2, y: integer; alpha: pbyte); Override;
   Public
      Procedure SetParameters(Const aa, bb, cc: TPointD; cc1, cc2, cc3: TColor32);
   End;
                                              
Procedure Monochrome(Var color: TColor32);
Procedure SwapRedBlue(color: PColor32; Count: integer); overload;
Function Color32(a, r, g, b: byte): TColor32; overload;
Function RGBColor(color: TColor32): cardinal;
Function InvertColor(color: TColor32): TColor32;
Function CompareHue(master, current: TColor32; tolerance: integer): boolean;
Function CompareAlpha(master, current: TColor32; tolerance: integer): boolean;
Function CompareRgbEx(master, current: TColor32): byte;
Function CompareAlphaEx(master, current: TColor32): byte;
Function RgbToHsl(color: TColor32): THsl;
Function HslToRgb(hslColor: THsl): TColor32;
Function AdjustHue(color: TColor32; percent: integer): TColor32;
Function ArrayOfColor32ToArrayHSL(Const clr32Arr: TArrayOfColor32): TArrayofHSL;
Function ArrayOfHSLToArrayColor32(Const hslArr: TArrayofHSL): TArrayOfColor32;
Function Alpha(color: TColor32): TColor32;
Function NoAlpha(color: TColor32): TColor32;
Function BlendAverage(bgColor, fgColor: TColor32): TColor32;
Function BlendLinearBurn(bgColor, fgColor: TColor32): TColor32;
Function BlendColorDodge(bgColor, fgColor: TColor32): TColor32;
Function RainbowColor(fraction: double): TColor32;
Function GradientColor(color1, color2: TColor32; frac: single): TColor32;
Function MakeDarker(color: TColor32; percent: cardinal): TColor32;
Function MakeLighter(color: TColor32; percent: cardinal): TColor32;
Function BlendToOpaque(bgColor, fgColor: TColor32): TColor32;
Function BlendToAlpha(bgColor, fgColor: TColor32): TColor32;
Function BlendMask(bgColor, alphaMask: TColor32): TColor32;
Function BlendInvertedMask(bgColor, alphaMask: TColor32): TColor32;
Function BlendBlueChannel(bgColor, blueMask: TColor32): TColor32;
Function CompareRGB(master, current: TColor32; tolerance: integer): boolean;
Function SwapRedBlue(color: TColor32): TColor32; overload;
Function IncPColor32(pc: Pointer; cnt: integer): PColor32;
Function GrayScale(color: TColor32): TColor32;
                                           
Procedure NormalizeAngle(Var angle: double; tolerance: double = Pi / 360); 
Procedure GetResamplerList(stringList: TStringList);                       
Procedure GetSinCos(angle: double; Out sinA, cosA: double);
Function IsValid(Value: integer): boolean; overload;
Function IsValid(Value: double): boolean; overload;
Function ClampByte(val: integer): byte; overload;
Function ClampByte(val: double): byte; overload;
Function ClampRange(val, min, max: integer): integer; overload;
Function ClampRange(val, min, max: double): double; overload;
Function DPIAware(val: integer): integer; overload;
Function DPIAware(val: double): double; overload;
Function CreateResourceStream(Const resName: String; resType: PChar): TResourceStream;
Function GetResampler(id: integer): TResamplerFunction;
Function RegisterResampler(func: TResamplerFunction; Const Name: String): integer;
Function MulBytes(b1, b2: byte): byte;
Function Size(sx, sy: integer): TSize;
Function SizeD(sx, sy: double): TSizeD;
Function MakeArrayOfInteger(Const ints: Array Of integer): TArrayOfInteger;
Function MakeArrayOfDouble(Const doubles: Array Of double): TArrayOfDouble;
Function Average(val1, val2: integer): integer; overload;
Function Average(val1, val2: double): double; overload;
Function ValueAlmostZero(val: double; epsilon: double = 0.001): boolean;
Function ValueAlmostOne(val: double; epsilon: double = 0.001): boolean;

Function IsValid(Const pt: TPoint): boolean; overload;
Function IsValid(Const pt: TPointD): boolean; overload;
Function Point(X, Y: integer): TPoint; overload;
Function Point(Const pt: TPointD): TPoint; overload;
Function PointD(Const X, Y: double): TPointD; overload;
Function PointD(Const pt: TPoint): TPointD; overload;
Function ScalePoint(Const pt: TPointD; scale: double): TPointD; overload;
Function ScalePoint(Const pt: TPointD; sx, sy: double): TPointD; overload;
Function OffsetPoint(Const pt: TPoint; dx, dy: integer): TPoint; overload;
Function OffsetPoint(Const pt: TPointD; dx, dy: double): TPointD; overload;
Function DistanceSqrd(Const pt1, pt2: TPoint): double; overload;
Function DistanceSqrd(Const pt1, pt2: TPointD): double; overload;
Function Distance(Const pt1, pt2: TPoint): double; overload;
Function Distance(Const pt1, pt2: TPointD): double; overload;
Function MidPoint(Const pt1, pt2: TPoint): TPoint; overload;
Function MidPoint(Const pt1, pt2: TPointD): TPointD; overload;
Function GetAngle(Const origin, pt: TPoint): double; overload;
Function GetAngle(Const origin, pt: TPointD): double; overload;
Function GetAngle(Const a, b, c: TPoint): double; overload;
Function GetAngle(Const a, b, c: TPointD): double; overload;
Function PointsEqual(Const pt1, pt2: TPointD): boolean; overload;
Function PointsNearEqual(Const pt1, pt2: TPointD; distSqrd: double): boolean;
Function PointInPolygon(Const pt: TPointD; Const polygon: TPathD; fillRule: TFillRule): boolean;
Function PointInPolygons(Const pt: TPointD; Const polygons: TPathsD; fillRule: TFillRule): boolean;
Function PerpendicularDist(Const pt, line1, line2: TPointD): double;
Function ClosestPointOnLine(Const pt, linePt1, linePt2: TPointD): TPointD;
Function ClosestPointOnSegment(Const pt, segPt1, segPt2: TPointD): TPointD;
Function GetPointAtAngleAndDist(Const origin: TPointD; angle, distance: double): TPointD;
Function CrossProduct(Const pt1, pt2, pt3: TPointD): double;
Function DotProduct(Const pt1, pt2, pt3: TPointD): double;
Function TurnsLeft(Const pt1, pt2, pt3: TPointD): boolean;
Function TurnsRight(Const pt1, pt2, pt3: TPointD): boolean;
Function GetUnitVector(Const pt1, pt2: TPointD): TPointD;
Function GetUnitNormal(Const pt1, pt2: TPointD): TPointD;
Function PerpendicularDistSqrd(Const pt, line1, line2: TPointD): double;
Function IntersectPoint(Const ln1a, ln1b, ln2a, ln2b: TPointD; Out ip: TPointD): boolean;
Function ReflectPoint(Const pt, pivot: TPointD): TPointD;  
Procedure RotatePoint(Var pt: TPointD; Const focalPoint: TPointD; sinA, cosA: double); overload;
Procedure RotatePoint(Var pt: TPointD; Const focalPoint: TPointD; angleRad: double); overload;

Function Rect(Const recD: TRectD): TRect; overload;
Function Rect(Const left, top, right, bottom: integer): TRect; overload;
Function RectD(left, top, right, bottom: double): TRectD; overload;
Function RectD(Const rec: TRect): TRectD; overload;
Function MidPoint(Const rec: TRect): TPoint; overload;
Function MidPoint(Const rec: TRectD): TPointD; overload;
Function IsEmptyRect(Const rec: TRect): boolean; overload;
Function IsEmptyRect(Const rec: TRectD): boolean; overload;
Function ScaleRect(Const rec: TRect; scale: double): TRect; overload;
Function ScaleRect(Const rec: TRectD; scale: double): TRectD; overload;
Function ScaleRect(Const rec: TRect; sx, sy: double): TRect; overload;
Function ScaleRect(Const rec: TRectD; sx, sy: double): TRectD; overload;
Function RectWidth(Const rec: TRect): integer;
Function RectHeight(Const rec: TRect): integer;
Function GetRotatedRectBounds(Const rec: TRect; angle: double): TRect; overload;
Function GetRotatedRectBounds(Const rec: TRectD; angle: double): TRectD; overload;
Function RectsEqual(Const rec1, rec2: TRect): boolean;
Function IntersectRect(Const rec1, rec2: TRectD): TRectD; overload;
Function UnionRect(Const rec1, rec2: TRectD): TRectD; overload;
Function IsPointInEllipse(Const ellipseRec: TRect; Const pt: TPoint): boolean;
Function GetLineEllipseIntersects(Const ellipseRect: TRect; Var linePt1, linePt2: TPointD): boolean;
Function RectWH(left, top, Width, Height: integer): TRectWH; overload;
Function RectWH(left, top, Width, Height: double): TRectWH; overload;
Function RectWH(Const rec: TRectD): TRectWH; overload;         
Procedure InflateRect(Var rec: TRect; dx, dy: integer); overload;
Procedure InflateRect(Var rec: TRectD; dx, dy: double); overload;     
Procedure OffsetRect(Var rec: TRectD; dx, dy: double); overload;    
Procedure RectWidthHeight(Const rec: TRect; Out width, height: Integer);

Function MakePathI(Const pts: Array Of integer): TPathD; overload;
Function MakePathD(Const pts: Array Of double): TPathD; overload;
Function IsPathConvex(Const path: TPathD): boolean;
Function Paths(Const path: TPathD): TPathsD;
Function CopyPath(Const path: TPathD): TPathD;
Function CopyPaths(Const paths: TPathsD): TPathsD;
Function gaGetBounds(Const path: TPathD): TRect; overload;
Function gaGetBounds(Const paths: TPathsD): TRect; overload;
Function gaGetBoundsD(Const path: TPathD): TRectD; overload;
Function gaGetBoundsD(Const paths: TPathsD): TRectD; overload;
Function ScalePath(Const path: TPathD; sx, sy: double): TPathD; overload;
Function ScalePath(Const path: TPathD; scale: double): TPathD; overload;
Function ScalePath(Const paths: TPathsD; sx, sy: double): TPathsD; overload;
Function ScalePath(Const paths: TPathsD; scale: double): TPathsD; overload;
Function ReversePath(Const path: TPathD): TPathD; overload;
Function ReversePath(Const paths: TPathsD): TPathsD; overload;
Function OpenPathToFlatPolygon(Const path: TPathD): TPathD;
Procedure AppendPoint(Var path: TPathD; Const extra: TPointD);
                                                                    
Procedure AppendPath(Var path: TPathD; Const pt: TPointD); overload;
Procedure AppendPath(Var path1: TPathD; Const path2: TPathD); overload;
Procedure AppendPath(Var paths: TPathsD; Const extra: TPathD); overload;
Procedure AppendPath(Var paths: TPathsD; Const extra: TPathsD); overload;
Procedure AppendPath(Var ppp: TArrayOfPathsD; Const extra: TPathsD); overload;

Function OffsetPath(Const path: TPathD; dx, dy: double): TPathD; overload;
Function OffsetPath(Const paths: TPathsD; dx, dy: double): TPathsD; overload;
Function OffsetPath(Const ppp: TArrayOfPathsD; dx, dy: double): TArrayOfPathsD; overload;
Function RotatePath(Const path: TPathD; Const focalPoint: TPointD; angleRads: double): TPathD; overload;
Function RotatePath(Const paths: TPathsD; Const focalPoint: TPointD; angleRads: double): TPathsD; overload;
Function Grow(Const path, normals: TPathD; delta: double; joinStyle: TJoinStyle; miterLimOrRndScale: double): TPathD;
Function GetVectors(Const path: TPathD): TPathD;
Function GetNormals(Const path: TPathD): TPathD;
Function Distance(Const path: TPathD; stopAt: integer = 0): double; overload;
Function GetDistances(Const path: TPathD): TArrayOfDouble;
Function GetCumulativeDistances(Const path: TPathD): TArrayOfDouble;
Function Area(Const path: TPathD): double;
Function RoundRect(Const rec: TRect; radius: integer): TPathD; overload;
Function RoundRect(Const rec: TRectD; radius: double): TPathD; overload;
Function RoundRect(Const rec: TRect; radius: TPoint): TPathD; overload;
Function RoundRect(Const rec: TRectD; radius: TPointD): TPathD; overload;
Function Rectangle(Const rec: TRect): TPathD; overload;
Function Rectangle(Const rec: TRectD): TPathD; overload;
Function Rectangle(l, t, r, b: double): TPathD; overload;
Function Ellipse(Const rec: TRect; steps: integer = 0): TPathD; overload;
Function Ellipse(Const rec: TRectD; steps: integer = 0): TPathD; overload;
Function Ellipse(Const rec: TRectD; scale: double): TPathD; overload;
Function Circle(Const pt: TPoint; radius: double): TPathD; overload;
Function Circle(Const pt: TPointD; radius: double): TPathD; overload;
Function Star(Const focalPt: TPointD; innerRadius, outerRadius: double; points: integer): TPathD;
Function Arc(Const rec: TRectD; startAngle, endAngle: double; scale: double = 0): TPathD;
Function Pie(Const rec: TRectD; StartAngle, EndAngle: double; scale: double = 0): TPathD;
Function FlattenQBezier(Const pt1, pt2, pt3: TPointD; tolerance: double = 0.0): TPathD; overload;
Function FlattenQBezier(Const pts: TPathD; tolerance: double = 0.0): TPathD; overload;
Function FlattenQBezier(Const pts: TPathsD; tolerance: double = 0.0): TPathsD; overload;
Function FlattenQBezier(Const firstPt: TPointD; Const pts: TPathD;tolerance: double = 0.0): TPathD; Overload;
Function GetPointInQuadBezier(Const a, b, c: TPointD; t: double): TPointD;
Function FlattenCBezier(Const pt1, pt2, pt3, pt4: TPointD; tolerance: double = 0.0): TPathD; overload;
Function FlattenCBezier(Const pts: TPathD; tolerance: double = 0.0): TPathD; overload;
Function FlattenCBezier(Const pts: TPathsD; tolerance: double = 0.0): TPathsD; overload;      
Function FlattenCBezier(Const firstPt: TPointD; Const pts: TPathD; tolerance: double = 0.0): TPathD; overload;
Function GetPointInCubicBezier(Const a, b, c, d: TPointD; t: double): TPointD;
Function FlattenCSpline(Const pts: TPathD; tolerance: double = 0.0): TPathD; overload;
Function FlattenCSpline(Const priorCtrlPt, startPt: TPointD; Const pts: TPathD; tolerance: double = 0.0): TPathD; overload;
Function FlattenQSpline(Const pts: TPathD; tolerance: double = 0.0): TPathD; overload;
Function FlattenQSpline(Const priorCtrlPt, startPt: TPointD; Const pts: TPathD; tolerance: double = 0.0): TPathD; overload;
Function ArrowHead(Const arrowTip, ctrlPt: TPointD; size: double; arrowStyle: TArrowStyle): TPathD;
Function GetDefaultArrowHeadSize(lineWidth: double): double;
Function ShortenPath(Const path: TPathD; pathEnd: TPathEnd; amount: double): TPathD;
Function GetDashedPath(Const path: TPathD; closed: boolean; Const pattern: TArrayOfInteger; patternOffset: PDouble): TPathsD;
Function GetDashedOutLine(Const path: TPathD; closed: boolean; Const pattern: TArrayOfInteger; patternOffset: PDouble;
                          lineWidth: double; joinStyle: TJoinStyle; endStyle: TEndStyle): TPathsD;
Function Outline(Const line: TPathD; lineWidth: double; joinStyle: TJoinStyle; endStyle: TEndStyle;
                 miterLimOrRndScale: double = 0): TPathsD; overload;
Function Outline(Const Lines: TPathsD; lineWidth: double; joinStyle: TJoinStyle; endStyle: TEndStyle;
                 miterLimOrRndScale: double = 0): TPathsD; overload;
Function VectorizeMask(Const mask: TArrayOfByte; maskWidth: integer): TPathsD;
Function RamerDouglasPeucker(Const path: TPathD; epsilon: double): TPathD; overload;
Function RamerDouglasPeucker(Const paths: TPathsD; epsilon: double): TPathsD; overload;
Function SmoothToBezier(Const path: TPathD; closed: boolean; tolerance: double; minSegLength: double = 2): TPathD; overload;
Function SmoothToBezier(Const paths: TPathsD; closed: boolean; tolerance: double; minSegLength: double = 2): TPathsD; overload;
Function StripNearDuplicates(Const path: TPathD; minDist: double; isClosedPath: boolean): TPathD; overload;
Function StripNearDuplicates(Const paths: TPathsD; minLength: double; isClosedPaths: boolean): TPathsD; overload;
Procedure Rasterize(Const paths: TPathsD; Const clipRec: TRect; fillRule: TFillRule; renderer: TgaRendererCustom);

Function IsIdentityMatrix(Const matrix: TMatrixD): boolean;
Function IsValidMatrix(Const matrix: TMatrixD): boolean;
Function Matrix(Const m00, m01, m02, m10, m11, m12, m20, m21, m22: double): TMatrixD;
Function MatrixDeterminant(Const matrix: TMatrixD): double;
Function MatrixAdjugate(Const matrix: TMatrixD): TMatrixD;
Function MatrixMultiply(Const modifier, matrix: TMatrixD): TMatrixD; 
Function MatrixInvert(Var matrix: TMatrixD): boolean;         
Function ExtractAngleFromMatrix(Const mat: TMatrixD): double;
Function ExtractScaleFromMatrix(Const mat: TMatrixD): TSizeD;
Function ExtractAvgScaleFromMatrix(Const mat: TMatrixD): double;
Procedure MatrixApply(Const matrix: TMatrixD; Var x, y: double); overload;
Procedure MatrixApply(Const matrix: TMatrixD; Var pt: TPointD); overload;
Procedure MatrixApply(Const matrix: TMatrixD; Var rec: TRect); overload;
Procedure MatrixApply(Const matrix: TMatrixD; Var rec: TRectD); overload;
Procedure MatrixApply(Const matrix: TMatrixD; Var rec: TRectWH); overload;
Procedure MatrixApply(Const matrix: TMatrixD; Var path: TPathD); overload;
Procedure MatrixApply(Const matrix: TMatrixD; Var paths: TPathsD); overload;
Procedure MatrixSkew(Var matrix: TMatrixD; angleX, angleY: double);
Procedure MatrixScale(Var matrix: TMatrixD; scale: double); overload;
Procedure MatrixScale(Var matrix: TMatrixD; scaleX, scaleY: double); overload;
Procedure MatrixRotate(Var matrix: TMatrixD; Const center: TPointD; angRad: double);
Procedure MatrixTranslate(Var matrix: TMatrixD; dx, dy: double);
Procedure ExtractAllFromMatrix(Const mat: TMatrixD; Out angle: double; Out scale, skew, trans: TPointD);

Function PrePendPoint(Const pt: TPointD; Const p: TPathD): TPathD;
Function PrePendPoints(Const pt1, pt2: TPointD; Const p: TPathD): TPathD;

Function GetPtOnEllipseFromAngle(Const ellipseRect: TRectD; angle: double): TPointD;
Function GetPtOnRotatedEllipseFromAngle(Const ellipseRect: TRectD; ellipseRotAngle, angle: double): TPointD;
Function GetEllipticalAngleFromPoint(Const ellipseRect: TRectD; Const pt: TPointD): double;
Function GetRotatedEllipticalAngleFromPoint(Const ellipseRect: TRectD; ellipseRotAngle: double; pt: TPointD): double;
Function GetClosestPtOnRotatedEllipse(Const ellipseRect: TRectD; ellipseRotation: double; Const pt: TPointD): TPointD;

Function MakePath(Const pts: Array Of integer): TPathD; overload;
Function MakePath(Const pts: Array Of double): TPathD; overload;

Const
   cnInvalid = -MaxInt;
   cnInvalidD = -Infinity;
   cnNullPoint: TPoint = (X: 0; Y: 0);
   cnNullPointD: TPointD = (X: 0; Y: 0);
   cnInvalidPoint: TPoint = (X: - MaxInt; Y: - MaxInt);
   cnInvalidPointD: TPointD = (X: - Infinity; Y: - Infinity);
   cnNullRect: TRect = (left: 0; top: 0; right: 0; Bottom: 0);
   cnNullRectD: TRectD = (left: 0; top: 0; right: 0; Bottom: 0);
   cnBezierTolerance: double = 0.25;
   cnIdentityMatrix: TMatrixD = ((1, 0, 0), (0, 1, 0), (0, 0, 1));

Var
   VarAutoWidthThreshold: double = 5.0;
   VarMinStrokeWidth: double = 0.5;
   VarDefaultMiterLimit: double = 4.0;
   VarClockwiseRotationIsAnglePositive: boolean = True;
   VarNearestResampler: integer;
   VarBilinearResampler: integer;
   VarBicubicResampler: integer;
   VarDefaultResampler: integer = 0;
   VarMulTable: Array[byte, byte] Of byte;
   VarDivTable: Array[byte, byte] Of byte;
   VarScreenScale: double = 1.0;
   VarDpiAwareI: integer;
   VarDpiAwareD: double;

Implementation

Const
   cnDiv255: double = 1 / 255;
   cnBuffSize = 64;

Type
   TByteArray = Array[0..MaxInt - 1] Of byte;
   PByteArray = ^TByteArray;

   TImgFmtRec = Record
      Fmt: String;
      SortOrder: TClipboardPriority;
      Obj: TgaImageFormatClass;
   End;
   PImgFmtRec = ^TImgFmtRec;

   TResamplerObj = Class
      id: integer;
         Name: String;
      func: TResamplerFunction;
   End;

Var
   ImageFormatClassList: TList;
   ResamplerList: TList;

Procedure CreateImageFormatList;
Begin
   If Assigned(ImageFormatClassList) Then
      Exit;
   ImageFormatClassList := TList.Create;
End;

Procedure NormalizeAngle(Var angle: double; tolerance: double = Pi / 360);
Var
   aa: double;
Begin
   While angle < -cnAngle180 Do
      angle := angle + cnAngle360;
   While angle > cnAngle180 Do
      angle := angle - cnAngle360;

   aa := Abs(angle);
   If aa < tolerance Then
      angle := 0
   Else If aa > cnAngle180 - tolerance Then
      angle := cnAngle180
   Else If (aa < cnAngle90 - tolerance) Or (aa > cnAngle90 + tolerance) Then
      Exit
   Else If angle < 0 Then
      angle := -cnAngle90
   Else
      angle := cnAngle90;
End;

Function SwapRedBlue(color: TColor32): TColor32;
Var
   c: Array[0..3] Of byte Absolute color;
   r: Array[0..3] Of byte Absolute Result;
Begin
   Result := color;
   r[0] := c[2];
   r[2] := c[0];
End;

Procedure SwapRedBlue(color: PColor32; Count: integer);
Var
   i: integer;
Begin
   For i := 1 To Count Do
   Begin
      color^ := SwapRedBlue(color^);
      Inc(color);
   End;
End;

Function MulBytes(b1, b2: byte): byte;
Begin
   Result := VarMulTable[b1, b2];
End;

Function ImageFormatClassListSort(item1, item2: Pointer): integer;
Var
   imgFmtRec1: PImgFmtRec Absolute item1;
   imgFmtRec2: PImgFmtRec Absolute item2;
Begin
   Result := integer(imgFmtRec1^.SortOrder) - integer(imgFmtRec2^.SortOrder);
End;

Function BlendToOpaque(bgColor, fgColor: TColor32): TColor32;
Var
   res: TARGB Absolute Result;
   bg: TARGB Absolute bgColor;
   fg: TARGB Absolute fgColor;
   fw, bw: PByteArray;
Begin
   If fg.A = 0 Then
      Result := bgColor
   Else If fg.A = 255 Then
      Result := fgColor
   Else
   Begin
      res.A := 255;
      fw := PByteArray(@VarMulTable[fg.A]);
      bw := PByteArray(@VarMulTable[Not fg.A]);
      res.R := fw^[fg.R] + bw^[bg.R];
      res.G := fw^[fg.G] + bw^[bg.G];
      res.B := fw^[fg.B] + bw^[bg.B];
   End;
End;

Function BlendToAlpha(bgColor, fgColor: TColor32): TColor32;
Var
   res: TARGB Absolute Result;
   bg: TARGB Absolute bgColor;
   fg: TARGB Absolute fgColor;
   fgWeight: byte;
   R, InvR: PByteArray;
Begin
   If fg.A = 0 Then
      Result := bgColor
   Else If fg.A = 255 Then
      Result := fgColor
   Else
   Begin
      res.A := (((fg.A Xor 255) * (bg.A Xor 255)) Shr 8) Xor 255;

      fgWeight := VarDivTable[fg.A, res.A];

      R := PByteArray(@VarMulTable[fgWeight]);
      InvR := PByteArray(@VarMulTable[Not fgWeight]);
      res.R := R^[fg.R] + InvR^[bg.R];
      res.G := R^[fg.G] + InvR^[bg.G];
      res.B := R^[fg.B] + InvR^[bg.B];
   End;
End;

Function BlendMask(bgColor, alphaMask: TColor32): TColor32;
Var
   res: TARGB Absolute Result;
   bg: TARGB Absolute bgColor;
   fg: TARGB Absolute alphaMask;
Begin
   Result := bgColor;
   res.A := VarMulTable[bg.A, fg.A];
   If res.A = 0 Then
      Result := 0;
End;

Function BlendBlueChannel(bgColor, blueMask: TColor32): TColor32;
Var
   res: TARGB Absolute Result;
   bg: TARGB Absolute bgColor;
   fg: TARGB Absolute blueMask;
Begin
   Result := bgColor;
   res.A := VarMulTable[bg.A, VarMulTable[fg.B, fg.A]];
End;

Function BlendInvertedMask(bgColor, alphaMask: TColor32): TColor32;
Var
   res: TARGB Absolute Result;
   bg: TARGB Absolute bgColor;
   fg: TARGB Absolute alphaMask;
Begin
   Result := bgColor;
   res.A := VarMulTable[bg.A, 255 - fg.A];
   If res.A < 2 Then
      Result := 0;
End;

Function CompareRGB(master, current: TColor32; tolerance: integer): boolean;
Var
   mast: TARGB Absolute master;
   curr: TARGB Absolute current;
Begin
   If curr.A < $80 Then
      Result := False
   Else If (master And $FFFFFF) = (current And $FFFFFF) Then
      Result := True
   Else If tolerance = 0 Then
      Result := False
   Else
      Result :=
         (Abs(curr.R - mast.R) <= tolerance) And (Abs(curr.G - mast.G) <= tolerance) And (Abs(curr.B - mast.B) <= tolerance);
End;

Function CompareAlpha(master, current: TColor32; tolerance: integer): boolean;
Var
   mast: TARGB Absolute master;
   curr: TARGB Absolute current;
Begin
   If mast.A = curr.A Then
      Result := True
   Else If tolerance = 0 Then
      Result := False
   Else
      Result := Abs(curr.A - mast.A) <= tolerance;
End;

Function CompareHue(master, current: TColor32; tolerance: integer): boolean;
Var
   curr, mast: THsl;
   val: integer;
Begin
   If TARGB(current).A < $80 Then
   Begin
      Result := False;
      Exit;
   End;
   curr := RgbToHsl(current);
   mast := RgbToHsl(master);
   If curr.hue > mast.hue Then
   Begin
      val := curr.hue - mast.hue;
      If val > 127 Then
         val := mast.hue - curr.hue + 255;
   End
   Else
   Begin
      val := mast.hue - curr.hue;
      If val > 127 Then
         val := curr.hue - mast.hue + 255;
   End;
   Result := val <= tolerance;
End;

Function CompareRgbEx(master, current: TColor32): byte;
Var
   mast: TARGB Absolute master;
   curr: TARGB Absolute current;
   res: cardinal;
Begin
   If curr.A = 0 Then
   Begin
      Result := 0;
   End
   Else
   Begin
      res := Sqr(mast.R - curr.R) + Sqr(mast.G - curr.G) + Sqr(mast.B - curr.B);
      If res >= 65025 Then
         Result := 0
      Else
         Result := 255 - Round(Sqrt(res));
      If curr.A < 255 Then
         Result := VarMulTable[Result, curr.A];
   End;
End;

Function CompareAlphaEx(master, current: TColor32): byte;
Var
   mast: TARGB Absolute master;
   curr: TARGB Absolute current;
Begin
   Result := 255 - abs(mast.A - curr.A);
End;

Function IsAlphaChar(c: char): boolean;
Begin
   Result := ((c >= 'A') And (c <= 'Z')) Or ((c >= 'a') And (c <= 'z'));
End;

Procedure RectWidthHeight(Const rec: TRect; Out width, height: Integer);
Begin
   width := rec.Right - rec.Left;
   height := rec.Bottom - rec.Top;
End;

Function RectWidth(Const rec: TRect): integer;
Begin
   Result := rec.Right - rec.Left;
End;

Function RectHeight(Const rec: TRect): integer;
Begin
   Result := rec.Bottom - rec.Top;
End;

Function IsEmptyRect(Const rec: TRect): boolean;
Begin
   Result := (rec.Right <= rec.Left) Or (rec.Bottom <= rec.Top);
End;

Function IsEmptyRect(Const rec: TRectD): boolean;
Begin
   Result := (rec.Right <= rec.Left) Or (rec.Bottom <= rec.Top);
End;

Function InvertColor(color: TColor32): TColor32;
Var
   c: TARGB Absolute color;
   r: TARGB Absolute Result;
Begin
   r.A := c.A;
   r.R := 255 - c.R;
   r.G := 255 - c.G;
   r.B := 255 - c.B;
End;

Procedure Monochrome(Var color: TColor32);
Var
   c: TARGB Absolute color;
Begin
   c.R := (c.R * 61 + c.G * 174 + c.B * 21) Shr 8;
   If c.R > 127 Then
      c.R := 255
   Else
      c.R := 0;
   c.G := c.R;
   c.B := c.R;
   c.A := 255;
End;

Function Alpha(color: TColor32): TColor32;
Begin
   Result := (color And $FF000000);
End;

Function NoAlpha(color: TColor32): TColor32;
Begin
   Result := (color And $FFFFFF);
End;

Function RGBColor(color: TColor32): cardinal;
Var
   c: TARGB Absolute color;
   res: TARGB Absolute Result;
Begin
   res.R := c.B;
   res.G := c.G;
   res.B := c.R;
   res.A := 0;
End;

Function Color32(a, r, g, b: byte): TColor32;
Var
   res: TARGB Absolute Result;
Begin
   res.A := a;
   res.R := r;
   res.G := g;
   res.B := b;
End;

Function DPIAware(val: integer): integer;
Begin
   Result := Round(val * VarScreenScale);
End;

Function DPIAware(val: double): double;
Begin
   Result := val * VarScreenScale;
End;

Function ClampByte(val: integer): byte;
Begin
   If val < 0 Then
      Result := 0
   Else If val > 255 Then
      Result := 255
   Else
      Result := val;
End;

Function ClampByte(val: double): byte;
Begin
   If val <= 0 Then
      Result := 0
   Else If val >= 255 Then
      Result := 255
   Else
      Result := Round(val);
End;

Function GrayScale(color: TColor32): TColor32;
Var
   c: TARGB Absolute color;
   r: TARGB Absolute Result;
   g: byte;
Begin
   g := ClampByte(0.299 * c.R + 0.587 * c.G + 0.114 * c.B);
   r.A := c.A;
   r.R := g;
   r.G := g;
   r.B := g;
End;

Function ClampRange(val, min, max: integer): integer;
Begin
   If val < min Then
      Result := min
   Else If val > max Then
      Result := max
   Else
      Result := val;
End;

Function ClampRange(val, min, max: double): double;
Begin
   If val < min Then
      Result := min
   Else If val > max Then
      Result := max
   Else
      Result := val;
End;

Procedure ScaleRect(Var rec: TRect; scale: TPointD);
Begin
   rec.Right := rec.Left + Round((rec.Right - rec.Left) * scale.X);
   rec.Bottom := rec.Top + Round((rec.Bottom - rec.Top) * scale.Y);
End;

Function IncPColor32(pc: Pointer; cnt: integer): PColor32;
Begin
   Result := PColor32(pbyte(pc) + cnt * SizeOf(TColor32));
End;

Function PointD(Const X, Y: double): TPointD;
Begin
   Result.X := X;
   Result.Y := Y;
End;

Function PointD(Const pt: TPoint): TPointD;
Begin
   Result.X := pt.X;
   Result.Y := pt.Y;
End;

Function GetAlpha(master, current: TColor32): byte;
Var
   curr: TARGB Absolute current;
Begin
   Result := curr.A;
End;

Function RgbToHsl(color: TColor32): THsl;
Var
   rgba: TARGB Absolute color;
   hsl: THsl Absolute Result;
   r, g, b: byte;
   maxRGB, minRGB, mAdd, mSub: integer;
Begin

   r := rgba.R;
   g := rgba.G;
   b := rgba.B;
   maxRGB := Max(r, Max(g, b));
   minRGB := Min(r, Min(g, b));
   mAdd := maxRGB + minRGB;
   hsl.lum := mAdd Shr 1;
   hsl.alpha := rgba.A;

   If maxRGB = minRGB Then
   Begin
      hsl.hue := 0;
      hsl.sat := 0;
      Exit;
   End;

   mSub := maxRGB - minRGB;
   If mAdd <= 255 Then
      hsl.sat := VarDivTable[mSub, mAdd]
   Else
      hsl.sat := VarDivTable[mSub, 511 - mAdd];

   mSub := mSub * 6;
   If r = maxRGB Then
   Begin
      If g >= b Then
         hsl.hue := (g - b) * 255 Div mSub
      Else
         hsl.hue := 255 - ((b - g) * 255 Div mSub);
   End
   Else If G = maxRGB Then
   Begin
      If b > r Then
         hsl.hue := 85 + (b - r) * 255 Div mSub
      Else
         hsl.hue := 85 - (r - b) * 255 Div mSub;
   End
   Else
   Begin
      If r > g Then
         hsl.hue := 170 + (r - g) * 255 Div mSub
      Else
         hsl.hue := 170 - (g - r) * 255 Div mSub;
   End;
End;

Function HslToRgb(hslColor: THsl): TColor32;
Var
   rgba: TARGB Absolute Result;
   hsl: THsl Absolute hslColor;
   c, x, m: integer;
Begin
   c := (255 - abs(2 * hsl.lum - 255)) * hsl.sat Div 255;
   x := c * (255 - abs(integer((hsl.hue Mod 85) * 6 - 255)) Div 255);
   m := hsl.lum - c Div 2;
   rgba.A := hsl.alpha;
   Case (hsl.hue * 6) Shr 8 Of
      0:
         Begin
            rgba.R := c + m;
            rgba.G := x + m;
            rgba.B := 0 + m;
         End;
      1:
         Begin
            rgba.R := x + m;
            rgba.G := c + m;
            rgba.B := 0 + m;
         End;
      2:
         Begin
            rgba.R := 0 + m;
            rgba.G := c + m;
            rgba.B := x + m;
         End;
      3:
         Begin
            rgba.R := 0 + m;
            rgba.G := x + m;
            rgba.B := c + m;
         End;
      4:
         Begin
            rgba.R := x + m;
            rgba.G := 0 + m;
            rgba.B := c + m;
         End;
      5:
         Begin
            rgba.R := c + m;
            rgba.G := 0 + m;
            rgba.B := x + m;
         End;
   End;
End;

Function AdjustHue(color: TColor32; percent: integer): TColor32;
Var
   hsl: THsl;
Begin
   percent := percent Mod 100;
   If percent < 0 Then
      Inc(percent, 100);
   hsl := RgbToHsl(color);
   hsl.hue := (hsl.hue + Round(percent * 255 / 100)) Mod 256;
   Result := HslToRgb(hsl);
End;

Function ArrayOfColor32ToArrayHSL(Const clr32Arr: TArrayOfColor32): TArrayofHSL;
Var
   i, len: integer;
Begin
   len := length(clr32Arr);
   setLength(Result, len);
   For i := 0 To len - 1 Do
      Result[i] := RgbToHsl(clr32Arr[i]);
End;

Function ArrayOfHSLToArrayColor32(Const hslArr: TArrayofHSL): TArrayOfColor32;
Var
   i, len: integer;
Begin
   len := length(hslArr);
   setLength(Result, len);
   For i := 0 To len - 1 Do
      Result[i] := HslToRgb(hslArr[i]);
End;

Function NameToId(Name: PChar): longint;
Begin
   If cardinal(PWord(Name)) < 30 Then
   Begin
      Result := cardinal(PWord(Name));
   End
   Else
   Begin
      If Name^ = '#' Then
         Inc(Name);
      Result := StrToIntDef(Name, 0);
      If Result > 65535 Then
         Result := 0;
   End;
End;


Function CreateResourceStream(Const resName: String; resType: PChar): TResourceStream;
Var
   nameId, typeId: cardinal;
Begin
   Result := Nil;
   typeId := NameToId(resType);
   If (typeId > 0) Then
      resType := PChar(typeId)
   Else If (resType = 'BMP') Then
      resType := RT_BITMAP;

   nameId := NameToId(PChar(resName));
   If nameId > 0 Then
   Begin
      {$IFDEF WinCE}
      If FindResource(hInstance, PWideChar(nameId), resType) <> 0 Then
         Result := TResourceStream.CreateFromID(hInstance, nameId, PWideChar(resType));
      {$ELSE}    
      If FindResource(hInstance, PChar(nameId), resType) <> 0 Then
         Result := TResourceStream.CreateFromID(hInstance, nameId, resType);
     {$ENDIF}
   End
   Else
   Begin     
      {$IFDEF WinCE}
      If FindResource(hInstance, PWideChar(resName), resType) <> 0 Then
         Result := TResourceStream.Create(hInstance, PChar(resName), PWideChar(resType));
      {$ELSE}
      If FindResource(hInstance, PChar(resName), resType) <> 0 Then
         Result := TResourceStream.Create(hInstance, PChar(resName), resType);
     {$ENDIF}
   End;
End;

Function RectD(left, top, right, bottom: double): TRectD;
Begin
   Result.Left := left;
   Result.Top := top;
   Result.Right := right;
   Result.Bottom := bottom;
End;

Function RectD(Const rec: TRect): TRectD;
Begin
   With rec Do
   Begin
      Result.Left := left;
      Result.Top := top;
      Result.Right := right;
      Result.Bottom := bottom;
   End;
End;

Procedure MakeBlendTables;
Var
   i, j: integer;
Begin
   For j := 0 To 255 Do
      VarMulTable[0, j] := 0;
   For i := 0 To 255 Do
      VarMulTable[i, 0] := 0;
   For j := 0 To 255 Do
      VarDivTable[0, j] := 0;
   For i := 0 To 255 Do
      VarDivTable[i, 0] := 0;
   For i := 1 To 255 Do
      For j := 1 To 255 Do
      Begin
         VarMulTable[i, j] := Round(i * j * cnDiv255);
         If i >= j Then
            VarDivTable[i, j] := 255
         Else
            VarDivTable[i, j] := Round(i * $FF / j);
      End;
End;

Procedure CleanUpImageFormatClassList;
Var
   i: integer;
Begin
   For i := ImageFormatClassList.Count - 1 Downto 0 Do
      Dispose(PImgFmtRec(ImageFormatClassList[i]));
   ImageFormatClassList.Free;
End;

Procedure CreateResamplerList;
Begin
   ResamplerList := TList.Create;
End;

Function GetResampler(id: integer): TResamplerFunction;
Var
   i: integer;
Begin
   Result := Nil;
   If Not Assigned(ResamplerList) Then
      Exit;

   For i := ResamplerList.Count - 1 Downto 0 Do
      If TResamplerObj(ResamplerList[i]).id = id Then
      Begin
         Result := TResamplerObj(ResamplerList[i]).func;
         Break;
      End;
End;

Function RegisterResampler(func: TResamplerFunction; Const Name: String): integer;
Var
   resampleObj: TResamplerObj;
Begin
   If Not Assigned(ResamplerList) Then
      CreateResamplerList;

   resampleObj := TResamplerObj.Create;
   Result := ResamplerList.Add(resampleObj) + 1;
   resampleObj.id := Result;
   resampleObj.Name := Name;
   resampleObj.func := func;
End;

Procedure GetResamplerList(stringList: TStringList);
Var
   i: integer;
   resampleObj: TResamplerObj;
Begin
   stringList.Clear;
   stringList.Capacity := ResamplerList.Count;
   For i := 0 To ResamplerList.Count - 1 Do
   Begin
      resampleObj := TResamplerObj(ResamplerList[i]);
      stringList.AddObject(resampleObj.Name, resampleObj);
   End;
End;

Procedure CleanUpResamplerClassList;
Var
   i: integer;
Begin
   If Not Assigned(ResamplerList) Then
      Exit;
   For i := ResamplerList.Count - 1 Downto 0 Do
      TResamplerObj(ResamplerList[i]).Free;
   ResamplerList.Free;
End;

Function RectWH(left, top, Width, Height: integer): TRectWH;
Begin
   Result.Left := left;
   Result.Top := top;
   Result.Width := Width;
   Result.Height := Height;
End;

Function RectWH(left, top, Width, Height: double): TRectWH;
Begin
   Result.Left := left;
   Result.Top := top;
   Result.Width := Width;
   Result.Height := Height;
End;

Function RectWH(Const rec: TRectD): TRectWH;
Begin
   Result.Left := rec.left;
   Result.Top := rec.top;
   Result.Width := rec.Width;
   Result.Height := rec.Height;
End;

Function RectsEqual(Const rec1, rec2: TRect): boolean;
Begin
   Result := (rec1.Left = rec2.Left) And (rec1.Top = rec2.Top) And (rec1.Right = rec2.Right) And (rec1.Bottom = rec2.Bottom);
End;

Function Rect(Const left, top, right, bottom: integer): TRect;
Begin
   Result.Left := left;
   Result.Top := top;
   Result.Right := right;
   Result.Bottom := bottom;
End;

Function IsValid(Value: integer): boolean;
Begin
   Result := Value <> -MaxInt;
End;

Function IsValid(Value: double): boolean;
Begin
   Result := Value <> cnInvalidD;
End;

Function IsValid(Const pt: TPoint): boolean;
Begin
   Result := (pt.X <> cnInvalid) And (pt.Y <> cnInvalid);
End;

Function IsValid(Const pt: TPointD): boolean;
Begin
   Result := (pt.X <> -Infinity) And (pt.Y <> -Infinity);
End;

Function Point(X, Y: integer): TPoint;
Begin
   Result.X := X;
   Result.Y := Y;
End;

Function Point(Const pt: TPointD): TPoint;
Begin
   Result.X := Round(pt.x);
   Result.Y := Round(pt.y);
End;

Function PointsEqual(Const pt1, pt2: TPointD): boolean;
Begin
   Result := (pt1.X = pt2.X) And (pt1.Y = pt2.Y);
End;

Function PointsNearEqual(Const pt1, pt2: TPointD; distSqrd: double): boolean;
Begin
   Result := Sqr(pt1.X - pt2.X) + Sqr(pt1.Y - pt2.Y) < distSqrd;
End;

Function StripNearDuplicates(Const path: TPathD; minDist: double; isClosedPath: boolean): TPathD;
Var
   i, j, len: integer;
Begin
   len := length(path);
   SetLength(Result, len);
   If len = 0 Then
      Exit;
   Result[0] := path[0];
   j := 0;
   minDist := minDist * minDist;
   For i := 1 To len - 1 Do
      If Not PointsNearEqual(Result[j], path[i], minDist) Then
      Begin
         Inc(j);
         Result[j] := path[i];
      End;
   If isClosedPath And PointsNearEqual(Result[j], Result[0], minDist) Then
      Dec(j);
   SetLength(Result, j + 1);
End;

Function StripNearDuplicates(Const paths: TPathsD; minLength: double; isClosedPaths: boolean): TPathsD;
Var
   i, len: integer;
Begin
   len := Length(paths);
   SetLength(Result, len);
   For i := 0 To len - 1 Do
      Result[i] := StripNearDuplicates(paths[i], minLength, isClosedPaths);
End;

Function ValueAlmostZero(val: double; epsilon: double = 0.001): boolean;
Begin
   Result := Abs(val) < epsilon;
End;

Function ValueAlmostOne(val: double; epsilon: double = 0.001): boolean;
Begin
   Result := Abs(val - 1) < epsilon;
End;

Procedure GetSinCos(angle: double; Out sinA, cosA: double);
Begin
   Math.SinCos(angle, sinA, cosA);
End;

Function GetRotatedRectBounds(Const rec: TRect; angle: double): TRect;
Var
   sinA, cosA: double;
   w, h, recW, recH: integer;
   mp: TPoint;
Begin
   NormalizeAngle(angle);
   If angle <> 0 Then
   Begin
      GetSinCos(angle, sinA, cosA);
      sinA := Abs(sinA);
      cosA := Abs(cosA);
      recW := RectWidth(rec);
      recH := RectHeight(rec);
      w := Ceil((recW * cosA + recH * sinA) / 2);
      h := Ceil((recW * sinA + recH * cosA) / 2);
      mp := MidPoint(rec);
      Result := Rect(mp.X - w, mp.Y - h, mp.X + w, mp.Y + h);
   End
   Else
      Result := rec;
End;

Function GetRotatedRectBounds(Const rec: TRectD; angle: double): TRectD;
Var
   sinA, cosA: double;
   w, h: double;
   mp: TPointD;
Begin
   NormalizeAngle(angle);
   If angle <> 0 Then
   Begin
      GetSinCos(angle, sinA, cosA);
      sinA := Abs(sinA);
      cosA := Abs(cosA);
      w := (rec.Width * cosA + rec.Height * sinA) / 2;
      h := (rec.Width * sinA + rec.Height * cosA) / 2;
      mp := rec.MidPoint;
      Result := RectD(mp.X - w, mp.Y - h, mp.X + w, mp.Y + h);
   End
   Else
      Result := rec;
End;

Function Rect(Const recD: TRectD): TRect;
Begin
   Result.Left := Floor(recD.Left);
   Result.Top := Floor(recD.Top);
   Result.Right := Ceil(recD.Right);
   Result.Bottom := Ceil(recD.Bottom);
End;

Function Size(sx, sy: integer): TSize;
Begin
   Result.cx := sx;
   Result.cy := sy;
End;

Function SizeD(sx, sy: double): TSizeD;
Begin
   Result.sx := sx;
   Result.sy := sy;
End;

Function Area(Const path: TPathD): double;
Var
   i, j, highI: integer;
   d: double;
Begin
   Result := 0.0;
   highI := High(path);
   If (highI < 2) Then
      Exit;
   j := highI;
   For i := 0 To highI Do
   Begin
      d := (path[j].X + path[i].X);
      Result := Result + d * (path[j].Y - path[i].Y);
      j := i;
   End;
   Result := -Result * 0.5;
End;

Procedure OffsetRect(Var rec: TRectD; dx, dy: double);
Begin
   rec.Left := rec.Left + dx;
   rec.Top := rec.Top + dy;
   rec.Right := rec.Right + dx;
   rec.Bottom := rec.Bottom + dy;
End;

Function MidPoint(Const rec: TRect): TPoint;
Begin
   Result.X := (rec.Left + rec.Right) Div 2;
   Result.Y := (rec.Top + rec.Bottom) Div 2;
End;

Function MidPoint(Const rec: TRectD): TPointD;
Begin
   Result.X := (rec.Left + rec.Right) * 0.5;
   Result.Y := (rec.Top + rec.Bottom) * 0.5;
End;

Function MidPoint(Const pt1, pt2: TPoint): TPoint;
Begin
   Result.X := (pt1.X + pt2.X) Div 2;
   Result.Y := (pt1.Y + pt2.Y) Div 2;
End;

Function MidPoint(Const pt1, pt2: TPointD): TPointD;
Begin
   Result.X := (pt1.X + pt2.X) * 0.5;
   Result.Y := (pt1.Y + pt2.Y) * 0.5;
End;

Function Average(val1, val2: integer): integer;
Begin
   Result := (val1 + val2) Div 2;
End;

Function Average(val1, val2: double): double;
Begin
   Result := (val1 + val2) * 0.5;
End;

Function IntersectRect(Const rec1, rec2: TRectD): TRectD;
Begin
   Result.Left := Max(rec1.Left, rec2.Left);
   Result.Top := Max(rec1.Top, rec2.Top);
   Result.Right := Min(rec1.Right, rec2.Right);
   Result.Bottom := Min(rec1.Bottom, rec2.Bottom);
End;

Function UnionRect(Const rec1, rec2: TRectD): TRectD;
Begin
   If IsEmptyRect(rec1) Then
      Result := rec2
   Else If IsEmptyRect(rec2) Then
      Result := rec1
   Else
   Begin
      Result.Left := Min(rec1.Left, rec2.Left);
      Result.Top := Min(rec1.Top, rec2.Top);
      Result.Right := Max(rec1.Right, rec2.Right);
      Result.Bottom := Max(rec1.Bottom, rec2.Bottom);
   End;
End;

Function MakeArrayOfInteger(Const ints: Array Of integer): TArrayOfInteger;
Var
   i, len: integer;
Begin
   len := Length(ints);
   SetLength(Result, len);
   For i := 0 To len - 1 Do
      Result[i] := ints[i];
End;

Function MakeArrayOfDouble(Const doubles: Array Of double): TArrayOfDouble;
Var
   i, len: integer;
Begin
   len := Length(doubles);
   SetLength(Result, len);
   For i := 0 To len - 1 Do
      Result[i] := doubles[i];
End;

Function CrossProduct(Const pt1, pt2, pt3: TPointD): double;
Var
   x1, x2, y1, y2: double;
Begin
   x1 := pt2.X - pt1.X;
   y1 := pt2.Y - pt1.Y;
   x2 := pt3.X - pt2.X;
   y2 := pt3.Y - pt2.Y;
   Result := (x1 * y2 - y1 * x2);
End;

Function DotProduct(Const pt1, pt2, pt3: TPointD): double;
Var
   x1, x2, y1, y2: double;
Begin
   x1 := pt2.X - pt1.X;
   y1 := pt2.Y - pt1.Y;
   x2 := pt2.X - pt3.X;
   y2 := pt2.Y - pt3.Y;
   Result := (x1 * x2 + y1 * y2);
End;

Function TurnsLeft(Const pt1, pt2, pt3: TPointD): boolean;
Begin
   Result := CrossProduct(pt1, pt2, pt3) < 0;
End;

Function TurnsRight(Const pt1, pt2, pt3: TPointD): boolean;
Begin
   Result := CrossProduct(pt1, pt2, pt3) > 0;
End;

Function IsPathConvex(Const path: TPathD): boolean;
Var
   i, pathLen: integer;
   dir: boolean;
Begin
   Result := False;
   pathLen := length(path);
   If pathLen < 3 Then
      Exit;
   dir := TurnsRight(path[0], path[1], path[2]);
   For i := 1 To pathLen - 1 Do
      If TurnsRight(path[i], path[(i + 1) Mod pathLen], path[(i + 2) Mod pathLen]) <> dir Then
         Exit;
   Result := True;
End;

Function GetUnitVector(Const pt1, pt2: TPointD): TPointD;
Var
   dx, dy, inverseHypot: double;
Begin
   If (pt1.x = pt2.x) And (pt1.y = pt2.y) Then
   Begin
      Result.X := 0;
      Result.Y := 0;
      Exit;
   End;
   dx := (pt2.X - pt1.X);
   dy := (pt2.Y - pt1.Y);
   inverseHypot := 1 / Hypot(dx, dy);
   dx := dx * inverseHypot;
   dy := dy * inverseHypot;
   Result.X := dx;
   Result.Y := dy;
End;

Function GetUnitNormal(Const pt1, pt2: TPointD): TPointD;
Var
   dx, dy, inverseHypot: double;
Begin
   If PointsNearEqual(pt1, pt2, 0.001) Then
   Begin
      Result.X := 0;
      Result.Y := 0;
      Exit;
   End;
   dx := (pt2.X - pt1.X);
   dy := (pt2.Y - pt1.Y);
   inverseHypot := 1 / Hypot(dx, dy);
   dx := dx * inverseHypot;
   dy := dy * inverseHypot;
   Result.X := dy;
   Result.Y := -dx;
End;

Function Paths(Const path: TPathD): TPathsD;
Begin
   SetLength(Result, 1);
   Result[0] := Copy(path, 0, length(path));
End;

Function CopyPath(Const path: TPathD): TPathD;
Begin
   Result := Copy(path, 0, Length(path));
End;

Function CopyPaths(Const paths: TPathsD): TPathsD;
Var
   i, len1: integer;
Begin
   len1 := length(paths);
   setLength(Result, len1);
   For i := 0 To len1 - 1 Do
      Result[i] := Copy(paths[i], 0, length(paths[i]));
End;

Function OffsetPoint(Const pt: TPoint; dx, dy: integer): TPoint;
Begin
   Result.x := pt.x + dx;
   Result.y := pt.y + dy;
End;

Function OffsetPoint(Const pt: TPointD; dx, dy: double): TPointD;
Begin
   Result.x := pt.x + dx;
   Result.y := pt.y + dy;
End;

Function OffsetPath(Const path: TPathD; dx, dy: double): TPathD;
Var
   i, len: integer;
Begin
   len := length(path);
   setLength(Result, len);
   For i := 0 To len - 1 Do
   Begin
      Result[i].x := path[i].x + dx;
      Result[i].y := path[i].y + dy;
   End;
End;

Function OffsetPath(Const paths: TPathsD; dx, dy: double): TPathsD;
Var
   i, len: integer;
Begin
   len := length(paths);
   setLength(Result, len);
   For i := 0 To len - 1 Do
      Result[i] := OffsetPath(paths[i], dx, dy);
End;

Function OffsetPath(Const ppp: TArrayOfPathsD; dx, dy: double): TArrayOfPathsD;
Var
   i, len: integer;
Begin
   len := length(ppp);
   setLength(Result, len);
   For i := 0 To len - 1 Do
      Result[i] := OffsetPath(ppp[i], dx, dy);
End;

Function ScalePoint(Const pt: TPointD; scale: double): TPointD;
Begin
   Result.X := pt.X * scale;
   Result.Y := pt.Y * scale;
End;

Function ScalePoint(Const pt: TPointD; sx, sy: double): TPointD;
Begin
   Result.X := pt.X * sx;
   Result.Y := pt.Y * sy;
End;

Function ScalePath(Const path: TPathD; sx, sy: double): TPathD;
Var
   i, len: integer;
Begin
   If (sx = 0) Or (sy = 0) Then
      Result := Nil
   Else If ((sx = 1) And (sy = 1)) Then
   Begin
      Result := Copy(path, 0, Length(path));
   End
   Else
   Begin
      len := length(path);
      setLength(Result, len);
      For i := 0 To len - 1 Do
      Begin
         Result[i].x := path[i].x * sx;
         Result[i].y := path[i].y * sy;
      End;
   End;
End;

Function ScalePath(Const path: TPathD; scale: double): TPathD;
Begin
   Result := ScalePath(path, scale, scale);
End;

Function ScalePath(Const paths: TPathsD; sx, sy: double): TPathsD;
Var
   i, len: integer;
Begin
   len := length(paths);
   setLength(Result, len);
   For i := 0 To len - 1 Do
      Result[i] := ScalePath(paths[i], sx, sy);
End;

Function ScalePath(Const paths: TPathsD; scale: double): TPathsD;
Begin
   Result := ScalePath(paths, scale, scale);
End;

Function ScaleRect(Const rec: TRect; scale: double): TRect;
Begin
   Result := rec;
   Result.Left := Round(Result.Left * scale);
   Result.Top := Round(Result.Top * scale);
   Result.Right := Round(Result.Right * scale);
   Result.Bottom := Round(Result.Bottom * scale);
End;

Function ScaleRect(Const rec: TRect; sx, sy: double): TRect;
Begin
   Result := rec;
   Result.Left := Round(Result.Left * sx);
   Result.Top := Round(Result.Top * sy);
   Result.Right := Round(Result.Right * sx);
   Result.Bottom := Round(Result.Bottom * sy);
End;

Function ScaleRect(Const rec: TRectD; scale: double): TRectD;
Begin
   Result := rec;
   Result.Left := Result.Left * scale;
   Result.Top := Result.Top * scale;
   Result.Right := Result.Right * scale;
   Result.Bottom := Result.Bottom * scale;
End;

Function ScaleRect(Const rec: TRectD; sx, sy: double): TRectD;
Begin
   Result := rec;
   Result.Left := Result.Left * sx;
   Result.Top := Result.Top * sy;
   Result.Right := Result.Right * sx;
   Result.Bottom := Result.Bottom * sy;
End;

Function ReversePath(Const path: TPathD): TPathD;
Var
   i, highI: integer;
Begin
   highI := High(path);
   SetLength(Result, highI + 1);
   For i := 0 To highI Do
      Result[i] := path[highI - i];
End;

Function ReversePath(Const paths: TPathsD): TPathsD;
Var
   i, len: integer;
Begin
   len := Length(paths);
   SetLength(Result, len);
   For i := 0 To len - 1 Do
      Result[i] := ReversePath(paths[i]);
End;

Function OpenPathToFlatPolygon(Const path: TPathD): TPathD;
Var
   i, len, len2: integer;
Begin
   len := Length(path);
   len2 := Max(0, len - 2);
   setLength(Result, len + len2);
   If len = 0 Then
      Exit;
   Move(path[0], Result[0], len * SizeOf(TPointD));
   If len2 = 0 Then
      Exit;
   For i := 0 To len - 3 Do
      Result[len + i] := path[len - 2 - i];
End;

Function GetVectors(Const path: TPathD): TPathD;
Var
   i, j, len: cardinal;
   pt: TPointD;
Begin
   len := length(path);
   setLength(Result, len);
   If len = 0 Then
      Exit;
   pt := path[0];

   i := len - 1;
   While (i > 0) And (path[i].X = pt.X) And (path[i].Y = pt.Y) Do
      Dec(i);
   If (i = 0) Then
   Begin
      For i := 0 To len - 1 Do
         Result[i] := PointD(0, 0);
      Exit;
   End;
   Result[i] := GetUnitVector(path[i], pt);

   For j := i + 1 To len - 1 Do
      Result[j] := Result[j - 1];

   pt := path[i];
   For i := i - 1 Downto 0 Do
   Begin
      If (path[i].X <> pt.X) Or (path[i].Y <> pt.Y) Then
      Begin
         Result[i] := GetUnitVector(path[i], pt);
         pt := path[i];
      End
      Else
         Result[i] := Result[i + 1];
   End;
End;

Function GetNormals(Const path: TPathD): TPathD;
Var
   i, j, len: cardinal;
   pt: TPointD;
Begin
   len := length(path);
   setLength(Result, len);
   If len = 0 Then
      Exit;
   pt := path[0];

   i := len - 1;
   While (i > 0) And PointsNearEqual(path[i], pt, 0.001) Do
      Dec(i);
   If (i = 0) Then
   Begin
      For i := 0 To len - 1 Do
         Result[i] := PointD(0, 0);
      Exit;
   End;
   Result[i] := GetUnitNormal(path[i], pt);

   For j := i + 1 To len - 1 Do
      Result[j] := Result[j - 1];

   pt := path[i];
   For i := i - 1 Downto 0 Do
   Begin
      If (path[i].X <> pt.X) Or (path[i].Y <> pt.Y) Then
      Begin
         Result[i] := GetUnitNormal(path[i], pt);
         pt := path[i];
      End
      Else
         Result[i] := Result[i + 1];
   End;
End;

Function DistanceSqrd(Const pt1, pt2: TPoint): double;
Begin
   Result := Sqr(pt1.X - pt2.X) + Sqr(pt1.Y - pt2.Y);
End;

Function DistanceSqrd(Const pt1, pt2: TPointD): double;
Begin
   Result := Sqr(pt1.X - pt2.X) + Sqr(pt1.Y - pt2.Y);
End;

Function Distance(Const pt1, pt2: TPoint): double;
Begin
   Result := Sqrt(DistanceSqrd(pt1, pt2));
End;

Function Distance(Const pt1, pt2: TPointD): double;
Begin
   Result := Sqrt(DistanceSqrd(pt1, pt2));
End;

Function Distance(Const path: TPathD; stopAt: integer): double;
Var
   i, highI: integer;
Begin
   Result := 0;
   highI := High(path);
   If (stopAt > 0) And (stopAt < HighI) Then
      highI := stopAt;
   For i := 1 To highI Do
      Result := Result + Distance(path[i - 1], path[i]);
End;

Function GetDistances(Const path: TPathD): TArrayOfDouble;
Var
   i, len: integer;
Begin
   len := Length(path);
   SetLength(Result, len);
   If len = 0 Then
      Exit;
   Result[0] := 0;
   For i := 1 To len - 1 Do
      Result[i] := Distance(path[i - 1], path[i]);
End;

Function GetCumulativeDistances(Const path: TPathD): TArrayOfDouble;
Var
   i, len: integer;
Begin
   len := Length(path);
   SetLength(Result, len);
   If len = 0 Then
      Exit;
   Result[0] := 0;
   For i := 1 To len - 1 Do
      Result[i] := Result[i - 1] + Distance(path[i - 1], path[i]);
End;

Function PerpendicularDistSqrd(Const pt, line1, line2: TPointD): double;
Var
   a, b, c, d: double;
Begin
   a := pt.X - line1.X;
   b := pt.Y - line1.Y;
   c := line2.X - line1.X;
   d := line2.Y - line1.Y;
   If (c = 0) And (d = 0) Then
      Result := 0
   Else
      Result := Sqr(a * d - c * b) / (c * c + d * d);
End;

Function PointInPolyWindingCount(Const pt: TPointD; Const path: TPathD; Out PointOnEdgeDir: integer): integer;
Var
   i, len: integer;
   prevPt: TPointD;
   isAbove: boolean;
   crossProd: double;
Begin
   Result := 0;
   PointOnEdgeDir := 0;
   i := 0;
   len := Length(path);
   If len = 0 Then
      Exit;
   prevPt := path[len - 1];
   While (i < len) And (path[i].Y = prevPt.Y) Do
      Inc(i);
   If i = len Then
      Exit;
   isAbove := (prevPt.Y < pt.Y);
   While (i < len) Do
   Begin
      If isAbove Then
      Begin
         While (i < len) And (path[i].Y < pt.Y) Do
            Inc(i);
         If i = len Then
            break
         Else If i > 0 Then
            prevPt := path[i - 1];
         crossProd := CrossProduct(prevPt, path[i], pt);
         If crossProd = 0 Then
         Begin
            PointOnEdgeDir := -1;
         End
         Else If crossProd < 0 Then
            Dec(Result);
      End
      Else
      Begin
         While (i < len) And (path[i].Y > pt.Y) Do
            Inc(i);
         If i = len Then
            break
         Else If i > 0 Then
            prevPt := path[i - 1];
         crossProd := CrossProduct(prevPt, path[i], pt);
         If crossProd = 0 Then
         Begin
            PointOnEdgeDir := 1;
         End
         Else If crossProd > 0 Then
            Inc(Result);
      End;
      Inc(i);
      isAbove := Not isAbove;
   End;
End;

Function PointInPolygon(Const pt: TPointD; Const polygon: TPathD; fillRule: TFillRule): boolean;
Var
   wc: integer;
   PointOnEdgeDir: integer;
Begin
   wc := PointInPolyWindingCount(pt, polygon, PointOnEdgeDir);
   Case fillRule Of
      frEvenOdd: Result := (PointOnEdgeDir <> 0) Or Odd(wc);
      frNonZero: Result := (PointOnEdgeDir <> 0) Or (wc <> 0);
      frPositive: Result := (PointOnEdgeDir + wc > 0);
      Else
         Result := (PointOnEdgeDir + wc < 0);
   End;
End;

Function PointInPolysWindingCount(Const pt: TPointD; Const paths: TPathsD; Out PointOnEdgeDir: integer): integer;
Var
   i, j, len: integer;
   p: TPathD;
   prevPt: TPointD;
   isAbove: boolean;
   crossProd: double;
Begin
   Result := 0;
   PointOnEdgeDir := 0;
   For i := 0 To High(paths) Do
   Begin
      j := 0;
      p := paths[i];
      len := Length(p);
      If len < 3 Then
         Continue;
      prevPt := p[len - 1];
      While (j < len) And (p[j].Y = prevPt.Y) Do
         Inc(j);
      If j = len Then
         continue;
      isAbove := (prevPt.Y < pt.Y);
      While (j < len) Do
      Begin
         If isAbove Then
         Begin
            While (j < len) And (p[j].Y < pt.Y) Do
               Inc(j);
            If j = len Then
               break
            Else If j > 0 Then
               prevPt := p[j - 1];
            crossProd := CrossProduct(prevPt, p[j], pt);
            If crossProd = 0 Then
               PointOnEdgeDir := -1
            Else If crossProd < 0 Then
               Dec(Result);
         End
         Else
         Begin
            While (j < len) And (p[j].Y > pt.Y) Do
               Inc(j);
            If j = len Then
               break
            Else If j > 0 Then
               prevPt := p[j - 1];
            crossProd := CrossProduct(prevPt, p[j], pt);
            If crossProd = 0 Then
               PointOnEdgeDir := 1
            Else If crossProd > 0 Then
               Inc(Result);
         End;
         Inc(j);
         isAbove := Not isAbove;
      End;
   End;
End;

Function PointInPolygons(Const pt: TPointD; Const polygons: TPathsD; fillRule: TFillRule): boolean;
Var
   wc: integer;
   PointOnEdgeDir: integer;
Begin
   wc := PointInPolysWindingCount(pt, polygons, PointOnEdgeDir);
   Case fillRule Of
      frEvenOdd: Result := (PointOnEdgeDir <> 0) Or Odd(wc);
      frNonZero: Result := (PointOnEdgeDir <> 0) Or (wc <> 0);
      frPositive: Result := (PointOnEdgeDir + wc > 0);
      Else
         Result := (PointOnEdgeDir + wc < 0);
   End;
End;

Function PerpendicularDist(Const pt, line1, line2: TPointD): double;
Var
   a, b, c, d: double;
Begin
   a := pt.X - line1.X;
   b := pt.Y - line1.Y;
   c := line2.X - line1.X;
   d := line2.Y - line1.Y;
   Result := abs(a * d - c * b) / Sqrt(c * c + d * d);
End;

Function ClosestPoint(Const pt, linePt1, linePt2: TPointD; constrainToSegment: boolean): TPointD;
Var
   q: double;
Begin
   If (linePt1.X = linePt2.X) And (linePt1.Y = linePt2.Y) Then
   Begin
      Result := linePt1;
   End
   Else
   Begin
      q := ((pt.X - linePt1.X) * (linePt2.X - linePt1.X) + (pt.Y - linePt1.Y) * (linePt2.Y - linePt1.Y)) /
         (sqr(linePt2.X - linePt1.X) + sqr(linePt2.Y - linePt1.Y));
      If constrainToSegment Then
      Begin
         If q < 0 Then
            q := 0
         Else If q > 1 Then
            q := 1;
      End;
      Result.X := round((1 - q) * linePt1.X + q * linePt2.X);
      Result.Y := round((1 - q) * linePt1.Y + q * linePt2.Y);
   End;
End;

Function ClosestPointOnLine(Const pt, linePt1, linePt2: TPointD): TPointD;
Begin
   Result := ClosestPoint(pt, linePt1, linePt2, False);
End;

Function ClosestPointOnSegment(Const pt, segPt1, segPt2: TPointD): TPointD;
Begin
   Result := ClosestPoint(pt, segPt1, segPt2, True);
End;

Function IsPointInEllipse(Const ellipseRec: TRect; Const pt: TPoint): boolean;
Var
   rec: TRectD;
   x, y, y2, a, b, dx, dy: double;
Begin
   a := RectWidth(ellipseRec) * 0.5;
   b := RectHeight(ellipseRec) * 0.5;
   dx := ellipseRec.Left + a;
   dy := ellipseRec.Top + b;
   rec := RectD(ellipseRec);
   OffsetRect(rec, -dx, -dy);
   x := pt.X - dx;
   y := pt.Y - dy;
   Result := (abs(x) <= a) And (abs(y) <= b);
   If Not Result Then
      Exit;
   y2 := Sqrt((b * b * (1 - (x * x) / (a * a))));
   Result := (y >= -y2) And (y <= y2);
End;

Function GetLineEllipseIntersects(Const ellipseRect: TRect; Var linePt1, linePt2: TPointD): boolean;
Var
   dx, dy, m, a, b, c, q: double;
   qa, qb, qc, qs: double;
   rec: TRectD;
   pt1, pt2: TPointD;
Begin
   a := RectWidth(ellipseRect) * 0.5;
   b := RectHeight(ellipseRect) * 0.5;
   dx := ellipseRect.Left + a;
   dy := ellipseRect.Top + b;
   rec := RectD(ellipseRect);
   offsetRect(rec, -dx, -dy);
   pt1 := OffsetPoint(linePt1, -dx, -dy);
   pt2 := OffsetPoint(linePt2, -dx, -dy);

   If (pt1.X = pt2.X) Then
   Begin

      q := (b * b) * (1 - Sqr(pt1.X) / (a * a));
      Result := q >= 0;
      If Not Result Then
         Exit;
      q := Sqrt(q);
      pt1.Y := q;
      pt2.Y := -q;
   End
   Else
   Begin

      m := (pt1.Y - pt2.Y) / (pt1.X - pt2.X);
      c := pt1.Y - m * pt1.X;
      qa := ((b * b) / (a * a) + (m * m));
      qb := 2 * m * c;
      qc := (c * c) - (b * b);
      qs := (qb * qb) - 4 * qa * qc;
      Result := qs >= 0;
      If Not Result Then
         Exit;
      qs := Sqrt(qs);
      pt1.X := (-qb + qs) / (2 * qa);
      pt1.Y := m * pt1.X + c;
      pt2.X := (-qb - qs) / (2 * qa);
      pt2.Y := m * pt2.X + c;
   End;

   linePt1 := OffsetPoint(pt1, dx, dy);
   linePt2 := OffsetPoint(pt2, dx, dy);
End;

Function Sign(Const Value: double): integer;
Begin
   If Value < 0 Then
      Result := -1
   Else If Value > 0 Then
      Result := 1
   Else
      Result := 0;
End;

Function Grow(Const path, normals: TPathD; delta: double; joinStyle: TJoinStyle; miterLimOrRndScale: double): TPathD;
Var
   i, prevI, highI: cardinal;
   buffLen, resultLen: cardinal;
   steps360, stepsPerRad, pathArea, absPathArea, resultArea: double;
   pathBounds: TRectD;
   stepSin, stepCos, cosA: double;
   isShrinking: boolean;
   P: TPointD;
   norms: TPathD;
   normalA, normalB: TPointD;

   Procedure AddPoint(Const pt: TPointD);
   Begin
      If resultLen >= buffLen Then
      Begin
         Inc(buffLen, 128);
         setLength(Result, buffLen);
      End;
      Result[resultLen] := pt;
      Inc(resultLen);
   End;

   Function GetNormal(Const pt: TPointD; Const norm: TPointD): TPointD;
   Begin
      Result := PointD(pt.x + norm.x * delta, pt.y + norm.y * delta);
   End;

   Procedure AddMiter(Const N1, N2: TPointD; cosAplus1: double);
   Var
      q: double;
   Begin
      q := delta / cosAplus1;
      AddPoint(PointD(P.X + (N2.X + N1.X) * q, P.Y + (N1.Y + N2.Y) * q));
   End;

   Procedure AddBevel(Const N1, N2: TPointD);
   Begin
      AddPoint(GetNormal(P, N1));
      AddPoint(GetNormal(P, N2));
   End;

   Procedure AddRound(Const N1, N2: TPointD; cosA: double);
   Var
      i, steps: integer;
      a, sinA: double;
      P2: TPointD;
   Begin
      sinA := normalA.x * normalB.y - normalA.y * normalB.x;

      If (delta * sinA < -0.01) Then
      Begin
         AddBevel(N1, N2);
      End
      Else
      Begin
         a := ArcTan2(sinA, cosA);
         steps := Round(stepsPerRad * Abs(a));
         P2 := PointD(N1.x * delta, N1.y * delta);
         AddPoint(PointD(P.x + P2.x, P.y + P2.y));
         For i := 1 To steps Do
         Begin
            P2 := PointD(P2.x * stepCos - stepSin * P2.y, P2.x * stepSin + P2.y * stepCos);
            AddPoint(PointD(P.x + P2.x, P.y + P2.y));
         End;
      End;
   End;

Begin
   highI := high(path);
   If (highI < 1) Then
      exit;
   If delta < VarMinStrokeWidth / 2 Then
      delta := VarMinStrokeWidth / 2;

   pathArea := Area(path);
   absPathArea := Abs(pathArea);
   isShrinking := (absPathArea > 0.001) And (Sign(pathArea) <> Sign(delta));
   If isShrinking Then
   Begin
      pathBounds := gaGetBoundsD(path);
      If (pathBounds.Width < abs(delta) * 2) Or (pathBounds.Height < abs(delta) * 2) Then
         Exit;
   End;

   If delta < 1 Then
      joinStyle := jsSquare
   Else If joinStyle = jsAuto Then
   Begin
      If delta < VarAutoWidthThreshold / 2 Then
         joinStyle := jsSquare
      Else
         joinStyle := jsRound;
   End;

   If joinStyle = jsRound Then
   Begin
      If abs(delta) < 0.25 Then
         Exit;
      If miterLimOrRndScale <= 0 Then
         miterLimOrRndScale := 1;
      steps360 := 4 * Max(1, Sqrt(delta * miterLimOrRndScale));
      stepsPerRad := steps360 / (Pi * 2);
      GetSinCos(Pi * 2 / steps360, stepSin, stepCos);
   End
   Else
   Begin
      If miterLimOrRndScale < 1 Then
         miterLimOrRndScale := VarDefaultMiterLimit;
      miterLimOrRndScale := 2 / (sqr(miterLimOrRndScale));
   End;

   Result := Nil;
   resultLen := 0;
   buffLen := 0;
   norms := normals;
   If Not assigned(norms) Then
      norms := GetNormals(path);

   PrevI := HighI;
   For i := 0 To HighI Do
   Begin
      P := path[i];
      normalA := norms[PrevI];
      normalB := norms[i];
      cosA := normalB.x * normalA.x + normalB.y * normalA.y;

      Case
         joinStyle Of
         jsRound: AddRound(normalA, normalB, cosA);
         jsSquare:
            If cosA > 0 Then
               AddMiter(normalA, normalB, cosA + 1)
            Else
               AddBevel(normalA, normalB);
         jsMiter:
            If (1 + cosA < miterLimOrRndScale) Then
               AddBevel(normalA, normalB)
            Else
               AddMiter(normalA, normalB, cosA + 1);
      End;
      PrevI := i;
   End;

   setLength(Result, resultLen);

   If Not isShrinking Then
      Exit;
   resultArea := area(Result);
   If (Sign(resultArea) <> Sign(pathArea)) Or (abs(resultArea) > absPathArea) Then
      Result := Nil;
End;

Procedure AppendPath(Var path: TPathD; Const pt: TPointD);
Var
   len: integer;
Begin
   len := length(path);
   If (len > 0) And PointsEqual(pt, path[len - 1]) Then
      Exit;
   setLength(path, len + 1);
   path[len] := pt;
End;

Procedure AppendPath(Var path1: TPathD; Const path2: TPathD);
Var
   len1, len2: integer;
Begin
   len1 := length(path1);
   len2 := length(path2);
   If len2 = 0 Then
      Exit;
   If (len1 > 0) And PointsEqual(path2[0], path1[len1 - 1]) Then
      Dec(len1);
   setLength(path1, len1 + len2);
   Move(path2[0], path1[len1], len2 * SizeOf(TPointD));
End;

Procedure AppendPoint(Var path: TPathD; Const extra: TPointD);
Var
   len: integer;
Begin
   len := length(path);
   SetLength(path, len + 1);
   path[len] := extra;
End;

Procedure AppendPath(Var paths: TPathsD; Const extra: TPathD);
Var
   len1, len2: integer;
Begin
   len2 := length(extra);
   If len2 = 0 Then
      Exit;
   len1 := length(paths);
   setLength(paths, len1 + 1);
   paths[len1] := Copy(extra, 0, len2);
End;

Procedure AppendPath(Var paths: TPathsD; Const extra: TPathsD);
Var
   i, len1, len2: integer;
Begin
   len2 := length(extra);
   If len2 = 0 Then
      Exit;
   len1 := length(paths);
   setLength(paths, len1 + len2);
   For i := 0 To len2 - 1 Do
      paths[len1 + i] := Copy(extra[i], 0, length(extra[i]));
End;

Procedure AppendPath(Var ppp: TArrayOfPathsD; Const extra: TPathsD);
Var
   len: integer;
Begin
   len := length(ppp);
   setLength(ppp, len + 1);
   If Assigned(extra) Then
      AppendPath(ppp[len], extra)
   Else
      ppp[len] := Nil;
End;

Procedure RotatePoint(Var pt: TPointD; Const focalPoint: TPointD; sinA, cosA: double);
Var
   tmpX, tmpY: double;
Begin
   tmpX := pt.X - focalPoint.X;
   tmpY := pt.Y - focalPoint.Y;
   pt.X := tmpX * cosA - tmpY * sinA + focalPoint.X;
   pt.Y := tmpX * sinA + tmpY * cosA + focalPoint.Y;
End;

Procedure RotatePoint(Var pt: TPointD; Const focalPoint: TPointD; angleRad: double);
Var
   sinA, cosA: double;
Begin
   If Not VarClockwiseRotationIsAnglePositive Then
      angleRad := -angleRad;
   GetSinCos(angleRad, sinA, cosA);
   RotatePoint(pt, focalPoint, sinA, cosA);
End;

Function RotatePathInternal(Const path: TPathD; Const focalPoint: TPointD; sinA, cosA: double): TPathD;
Var
   i: integer;
   x, y: double;
Begin
   SetLength(Result, length(path));
   For i := 0 To high(path) Do
   Begin
      x := path[i].X - focalPoint.X;
      y := path[i].Y - focalPoint.Y;
      Result[i].X := x * cosA - y * sinA + focalPoint.X;
      Result[i].Y := x * sinA + y * cosA + focalPoint.Y;
   End;
End;

Function RotatePath(Const path: TPathD; Const focalPoint: TPointD; angleRads: double): TPathD;
Var
   sinA, cosA: double;
Begin
   If Not VarClockwiseRotationIsAnglePositive Then
      angleRads := -angleRads;
   GetSinCos(angleRads, sinA, cosA);
   Result := RotatePathInternal(path, focalPoint, sinA, cosA);
End;

Function RotatePath(Const paths: TPathsD; Const focalPoint: TPointD; angleRads: double): TPathsD;
Var
   i: integer;
   sinA, cosA: double;
Begin
   If Not VarClockwiseRotationIsAnglePositive Then
      angleRads := -angleRads;
   GetSinCos(angleRads, sinA, cosA);
   SetLength(Result, length(paths));
   For i := 0 To high(paths) Do
      Result[i] := RotatePathInternal(paths[i], focalPoint, sinA, cosA);
End;

Function GetAngle(Const origin, pt: TPoint): double;
Var
   x, y: double;
Begin
   x := pt.X - origin.X;
   y := pt.Y - origin.Y;
   If x = 0 Then
   Begin
      If y > 0 Then
         Result := cnAngle90
      Else
         Result := -cnAngle90;
   End
   Else If y = 0 Then
   Begin
      If x > 0 Then
         Result := 0
      Else
         Result := cnAngle180;
   End
   Else
      Result := arctan2(y, x);
   If Not VarClockwiseRotationIsAnglePositive Then
      Result := -Result;
End;

Function GetAngle(Const origin, pt: TPointD): double;
Var
   x, y: double;
Begin
   x := pt.X - origin.X;
   y := pt.Y - origin.Y;
   If x = 0 Then
   Begin
      If y > 0 Then
         Result := cnAngle90
      Else
         Result := -cnAngle90;
   End
   Else If y = 0 Then
   Begin
      If x > 0 Then
         Result := 0
      Else
         Result := cnAngle180;
   End
   Else
      Result := arctan2(y, x);
   If Not VarClockwiseRotationIsAnglePositive Then
      Result := -Result;
End;

Function GetAngle(Const a, b, c: TPoint): double;
Var
   ab, bc: TPointD;
   dp, cp: double;
Begin
   ab := PointD(b.x - a.x, b.y - a.y);
   bc := PointD(b.x - c.x, b.y - c.y);
   dp := (ab.x * bc.x + ab.y * bc.y);
   cp := (ab.x * bc.y - ab.y * bc.x);
   Result := arctan2(cp, dp);
   If Not VarClockwiseRotationIsAnglePositive Then
      Result := -Result;
End;

Function GetAngle(Const a, b, c: TPointD): double;
Var
   ab, bc: TPointD;
   dp, cp: double;
Begin
   ab := PointD(b.x - a.x, b.y - a.y);
   bc := PointD(b.x - c.x, b.y - c.y);
   dp := (ab.x * bc.x + ab.y * bc.y);
   cp := (ab.x * bc.y - ab.y * bc.x);
   Result := arctan2(cp, dp);
   If Not VarClockwiseRotationIsAnglePositive Then
      Result := -Result;
End;

Function GetPointAtAngleAndDist(Const origin: TPointD; angle, distance: double): TPointD;
Begin
   Result := origin;
   Result.X := Result.X + distance;
   RotatePoint(Result, origin, angle);
End;

Function IntersectPoint(Const ln1a, ln1b, ln2a, ln2b: TPointD; Out ip: TPointD): boolean;
Var
   m1, b1, m2, b2: double;
Begin
   Result := False;

   If (ln1B.X = ln1A.X) Then
   Begin
      If (ln2B.X = ln2A.X) Then
         exit;
      m2 := (ln2B.Y - ln2A.Y) / (ln2B.X - ln2A.X);
      b2 := ln2A.Y - m2 * ln2A.X;
      ip.X := ln1A.X;
      ip.Y := m2 * ln1A.X + b2;
   End
   Else If (ln2B.X = ln2A.X) Then
   Begin
      m1 := (ln1B.Y - ln1A.Y) / (ln1B.X - ln1A.X);
      b1 := ln1A.Y - m1 * ln1A.X;
      ip.X := ln2A.X;
      ip.Y := m1 * ln2A.X + b1;
   End
   Else
   Begin
      m1 := (ln1B.Y - ln1A.Y) / (ln1B.X - ln1A.X);
      b1 := ln1A.Y - m1 * ln1A.X;
      m2 := (ln2B.Y - ln2A.Y) / (ln2B.X - ln2A.X);
      b2 := ln2A.Y - m2 * ln2A.X;
      If m1 = m2 Then
         exit;
      ip.X := (b2 - b1) / (m1 - m2);
      ip.Y := m1 * ip.X + b1;
   End;
   Result := True;
End;

Function GrowOpenLine(Const line: TPathD; Width: double; joinStyle: TJoinStyle; endStyle: TEndStyle; miterLimOrRndScale: double): TPathsD;
Var
   len, x, y, wDiv2: integer;
   wd2: double;
   line1, line2: TPathD;
   vec1, vec2: TPointD;
Begin
   Result := Nil;
   len := length(line);
   If len = 0 Then
      Exit;

   If Width < VarMinStrokeWidth Then
      Width := VarMinStrokeWidth;

   If len = 1 Then
   Begin
      wDiv2 := Round(Width / 2);
      x := Round(line[0].X);
      y := Round(line[0].Y);
      SetLength(Result, 1);
      Result[0] := Ellipse(RectD(x - wDiv2, y - wDiv2, x + wDiv2, y + wDiv2));
      Exit;
   End;

   If joinStyle = jsAuto Then
   Begin
      If (endStyle = esRound) And (Width >= VarAutoWidthThreshold) Then
         joinStyle := jsRound
      Else
         joinStyle := jsSquare;
   End;

   If endStyle = esSquare Then
   Begin
      line1 := Copy(line, 0, len);

      wd2 := Width / 2;
      vec1 := GetUnitVector(line1[1], line1[0]);
      vec2 := GetUnitVector(line1[len - 2], line1[len - 1]);
      With line1[0] Do
         line1[0] := PointD(X + vec1.X * wd2, Y + vec1.y * wd2);
      With line1[len - 1] Do
         line1[len - 1] := PointD(X + vec2.X * wd2, Y + vec2.y * wd2);
      line2 := OpenPathToFlatPolygon(line1);
   End
   Else
      line2 := OpenPathToFlatPolygon(line);

   SetLength(Result, 1);
   Result[0] := Grow(line2, Nil, Width / 2, joinStyle, miterLimOrRndScale);
End;

Function ReverseNormals(Const norms: TPathD): TPathD;
Var
   i, highI: integer;
Begin
   highI := high(norms);
   setLength(Result, highI + 1);
   For i := 1 To highI Do
   Begin
      Result[i - 1].X := -norms[highI - i].X;
      Result[i - 1].Y := -norms[highI - i].Y;
   End;
   Result[highI].X := -norms[highI].X;
   Result[highI].Y := -norms[highI].Y;
End;

Function GrowClosedLine(Const line: TPathD; Width: double; joinStyle: TJoinStyle; miterLimOrRndScale: double): TPathsD;
Var
   len: integer;
   line2, norms: TPathD;
   rec: TRectD;
   skipHole: boolean;
Begin
   len := length(line);
   If len < 3 Then
   Begin
      Result := GrowOpenLine(line, Width, joinStyle, esPolygon, miterLimOrRndScale);
      Exit;
   End;

   If Width < VarMinStrokeWidth Then
      Width := VarMinStrokeWidth;
   If joinStyle = jsAuto Then
   Begin
      If Width < VarAutoWidthThreshold Then
         joinStyle := jsSquare
      Else
         joinStyle := jsRound;
   End;

   rec := gaGetBoundsD(line);
   skipHole := (rec.Width <= Width) Or (rec.Height <= Width);
   If skipHole Then
      SetLength(Result, 1)
   Else
      SetLength(Result, 2);
   norms := GetNormals(line);
   Result[0] := Grow(line, norms, Width / 2, joinStyle, miterLimOrRndScale);

   If skipHole Then
      Exit;
   line2 := ReversePath(line);
   norms := ReverseNormals(norms);
   Result[1] := Grow(line2, norms, Width / 2, joinStyle, miterLimOrRndScale);
End;

Function Outline(Const line: TPathD; lineWidth: double; joinStyle: TJoinStyle; endStyle: TEndStyle; miterLimOrRndScale: double): TPathsD;
Begin
   If Not assigned(line) Then
      Result := Nil
   Else If endStyle = esPolygon Then
      Result := GrowClosedLine(line, lineWidth, joinStyle, miterLimOrRndScale)
   Else
      Result := GrowOpenLine(line, lineWidth, joinStyle, endStyle, miterLimOrRndScale);
End;

Function Outline(Const Lines: TPathsD; lineWidth: double; joinStyle: TJoinStyle; endStyle: TEndStyle; miterLimOrRndScale: double): TPathsD;
Var
   i: integer;
   Procedure AddPaths(Const paths: TPathsD);
   Var
      i, len, len2: integer;
   Begin
      len := length(Result);
      len2 := length(paths);
      setLength(Result, len + len2);
      For i := 0 To len2 - 1 Do
         Result[len + i] := paths[i];
   End;

Begin
   Result := Nil;
   If Not assigned(Lines) Then
      exit;
   If joinStyle = jsAuto Then
   Begin
      If endStyle In [esPolygon, esRound] Then
         joinStyle := jsRound
      Else
         joinStyle := jsSquare;
   End;
   If endStyle = esPolygon Then
      For i := 0 To high(Lines) Do
         AddPaths(GrowClosedLine(Lines[i], lineWidth, joinStyle, miterLimOrRndScale))
   Else
      For i := 0 To high(Lines) Do
         AddPaths(GrowOpenLine(Lines[i], lineWidth, joinStyle, endStyle, miterLimOrRndScale));
End;

Function Rectangle(Const rec: TRect): TPathD;
Begin
   setLength(Result, 4);
   With rec Do
   Begin
      Result[0] := PointD(left, top);
      Result[1] := PointD(right, top);
      Result[2] := PointD(right, bottom);
      Result[3] := PointD(left, bottom);
   End;
End;

Function Rectangle(Const rec: TRectD): TPathD;
Begin
   setLength(Result, 4);
   With rec Do
   Begin
      Result[0] := PointD(left, top);
      Result[1] := PointD(right, top);
      Result[2] := PointD(right, bottom);
      Result[3] := PointD(left, bottom);
   End;
End;

Function Rectangle(l, t, r, b: double): TPathD;
Begin
   setLength(Result, 4);
   Result[0] := PointD(l, t);
   Result[1] := PointD(r, t);
   Result[2] := PointD(r, b);
   Result[3] := PointD(l, b);
End;

Procedure InflateRect(Var rec: TRect; dx, dy: integer);
Begin
   rec.Left := rec.Left - dx;
   rec.Top := rec.Top - dy;
   rec.Right := rec.Right + dx;
   rec.Bottom := rec.Bottom + dy;
End;

Procedure InflateRect(Var rec: TRectD; dx, dy: double);
Begin
   rec.Left := rec.Left - dx;
   rec.Top := rec.Top - dy;
   rec.Right := rec.Right + dx;
   rec.Bottom := rec.Bottom + dy;
End;

Function RoundRect(Const rec: TRect; radius: integer): TPathD;
Begin
   Result := RoundRect(RectD(rec), PointD(radius, radius));
End;

Function RoundRect(Const rec: TRect; radius: TPoint): TPathD;
Begin
   Result := RoundRect(RectD(rec), PointD(radius));
End;

Function RoundRect(Const rec: TRectD; radius: double): TPathD;
Begin
   Result := RoundRect(rec, PointD(radius, radius));
End;

Function RoundRect(Const rec: TRectD; radius: TPointD): TPathD;
Var
   i, j: integer;
   corners: TPathD;
   bezPts: TPathD;
   magic: TPointD;
Const
   magicC: double = 0.55228475;
Begin
   Result := Nil;
   If rec.IsEmpty Then
      Exit;
   radius.X := Min(radius.X, rec.Width / 2);
   radius.Y := Min(radius.Y, rec.Height / 2);
   If (radius.X < 1) And (radius.Y < 1) Then
   Begin
      Result := Rectangle(rec);
      Exit;
   End;

   magic.X := radius.X * magicC;
   magic.Y := radius.Y * magicC;

   SetLength(Corners, 4);
   With rec Do
   Begin
      corners[0] := PointD(Right, Top);
      corners[1] := BottomRight;
      corners[2] := PointD(Left, Bottom);
      corners[3] := TopLeft;
   End;
   SetLength(Result, 1);
   Result[0].X := corners[3].X + radius.X;
   Result[0].Y := corners[3].Y;

   SetLength(bezPts, 4);
   For i := 0 To High(corners) Do
   Begin
      For j := 0 To 3 Do
         bezPts[j] := corners[i];
      Case i Of
         3:
            Begin
               bezPts[0].Y := bezPts[0].Y + radius.Y;
               bezPts[1].Y := bezPts[0].Y - magic.Y;
               bezPts[3].X := bezPts[3].X + radius.X;
               bezPts[2].X := bezPts[3].X - magic.X;
            End;
         0:
            Begin
               bezPts[0].X := bezPts[0].X - radius.X;
               bezPts[1].X := bezPts[0].X + magic.X;
               bezPts[3].Y := bezPts[3].Y + radius.Y;
               bezPts[2].Y := bezPts[3].Y - magic.Y;
            End;
         1:
            Begin
               bezPts[0].Y := bezPts[0].Y - radius.Y;
               bezPts[1].Y := bezPts[0].Y + magic.Y;
               bezPts[3].X := bezPts[3].X - radius.X;
               bezPts[2].X := bezPts[3].X + magic.X;
            End;
         2:
            Begin
               bezPts[0].X := bezPts[0].X + radius.X;
               bezPts[1].X := bezPts[0].X - magic.X;
               bezPts[3].Y := bezPts[3].Y - radius.Y;
               bezPts[2].Y := bezPts[3].Y + magic.Y;
            End;
      End;
      AppendPath(Result, FlattenCBezier(bezPts));
   End;
End;

Function Circle(Const pt: TPoint; radius: double): TPathD;
Var
   rec: TRectD;
Begin
   rec.Left := pt.X - radius;
   rec.Right := pt.X + radius;
   rec.Top := pt.Y - radius;
   rec.Bottom := pt.Y + radius;
   Result := Ellipse(rec);
End;

Function Circle(Const pt: TPointD; radius: double): TPathD;
Var
   rec: TRectD;
Begin
   rec.Left := pt.X - radius;
   rec.Right := pt.X + radius;
   rec.Top := pt.Y - radius;
   rec.Bottom := pt.Y + radius;
   Result := Ellipse(rec);
End;

Function Ellipse(Const rec: TRectD; scale: double): TPathD;
Var
   steps: integer;
Begin
   If scale <= 0 Then
      scale := 1;
   steps := 4 * Max(1, Trunc(Sqrt((rec.Width + rec.Height) * scale)));
   Result := Ellipse(rec, steps);
End;

Function Ellipse(Const rec: TRect; steps: integer): TPathD;
Begin
   Result := Ellipse(RectD(rec), steps);
End;

Function Ellipse(Const rec: TRectD; steps: integer): TPathD;
Var
   i: integer;
   sinA, cosA: double;
   centre, radius, delta: TPointD;
Begin
   Result := Nil;
   If rec.IsEmpty Then
      Exit;
   With rec Do
   Begin
      centre := rec.MidPoint;
      radius := PointD(Width * 0.5, Height * 0.5);
   End;

   If steps < 4 Then
   Begin
      steps := 4 * Max(1, Trunc(Sqrt(rec.Width + rec.Height)));
   End;
   GetSinCos(2 * Pi / Steps, sinA, cosA);
   delta.x := cosA;
   delta.y := sinA;

   SetLength(Result, Steps);
   Result[0] := PointD(centre.X + radius.X, centre.Y);
   For i := 1 To steps - 1 Do
   Begin
      Result[i] := PointD(centre.X + radius.X * delta.x, centre.Y + radius.y * delta.y);
      delta := PointD(delta.X * cosA - delta.Y * sinA, delta.Y * cosA + delta.X * sinA);
   End;
End;

Function Star(Const focalPt: TPointD; innerRadius, outerRadius: double; points: integer): TPathD;
Var
   i: integer;
   sinA, cosA: double;
   delta: TPointD;
Begin
   Result := Nil;
   If (innerRadius <= 0) Or (outerRadius <= 0) Then
      Exit;
   If points <= 5 Then
      points := 10
   Else
      points := points * 2;
   GetSinCos(2 * Pi / points, sinA, cosA);
   delta.x := cosA;
   delta.y := sinA;
   SetLength(Result, points);
   Result[0] := PointD(focalPt.X + innerRadius, focalPt.Y);
   For i := 1 To points - 1 Do
   Begin
      If Odd(i) Then
         Result[i] := PointD(focalPt.X + outerRadius * delta.x, focalPt.Y + outerRadius * delta.y)
      Else
         Result[i] := PointD(focalPt.X + innerRadius * delta.x, focalPt.Y + innerRadius * delta.y);
      delta := PointD(delta.X * cosA - delta.Y * sinA, delta.Y * cosA + delta.X * sinA);
   End;
End;

Function Arc(Const rec: TRectD; startAngle, endAngle: double; scale: double): TPathD;
Var
   i, steps: integer;
   angle: double;
   sinA, cosA: double;
   centre, radius: TPointD;
   deltaX, deltaX2, deltaY: double;
Begin
   Result := Nil;
   If (endAngle = startAngle) Or IsEmptyRect(rec) Then
      Exit;
   If scale <= 0 Then
      scale := 4.0;

   If Not VarClockwiseRotationIsAnglePositive Then
   Begin
      startAngle := -startAngle;
      endAngle := -endAngle;
   End;

   NormalizeAngle(startAngle);
   NormalizeAngle(endAngle);

   With rec Do
   Begin
      centre := rec.MidPoint;
      radius := PointD(Width * 0.5, Height * 0.5);
   End;

   If endAngle < startAngle Then
      angle := endAngle - startAngle + cnAngle360
   Else
      angle := endAngle - startAngle;

   steps := 4 * Trunc(angle * Sqrt((rec.Width + rec.Height) * scale));
   If steps < 2 Then
      steps := 2;
   SetLength(Result, Steps + 1);

   GetSinCos(startAngle, deltaY, deltaX);
   Result[0].X := centre.X + radius.X * deltaX;
   Result[0].Y := centre.Y + radius.y * deltaY;

   GetSinCos(angle / Steps, sinA, cosA);
   For i := 1 To steps Do
   Begin
      deltaX2 := deltaX * cosA - deltaY * sinA;
      deltaY := deltaY * cosA + deltaX * sinA;
      deltaX := deltaX2;
      Result[i].X := centre.X + radius.X * deltaX;
      Result[i].Y := centre.Y + radius.y * deltaY;
   End;
End;

Function Pie(Const rec: TRectD; StartAngle, EndAngle: double; scale: double): TPathD;
Var
   len: integer;
Begin
   Result := Arc(rec, StartAngle, EndAngle, scale);
   len := length(Result);
   setLength(Result, len + 1);
   Result[len] := PointD((rec.Left + rec.Right) / 2, (rec.Top + rec.Bottom) / 2);
End;

Function ArrowHead(Const arrowTip, ctrlPt: TPointD; size: double; arrowStyle: TArrowStyle): TPathD;
Var
   unitVec, basePt: TPointD;
   sDiv40, sDiv50, sDiv60, sDiv120: double;
Begin
   Result := Nil;
   sDiv40 := size * 0.40;
   sDiv50 := size * 0.50;
   sDiv60 := size * 0.60;
   sDiv120 := sDiv60 * 2;
   unitVec := GetUnitVector(ctrlPt, arrowTip);

   Case arrowStyle Of
      asNone:
         Exit;
      asSimple:
         Begin
            setLength(Result, 3);
            basePt := OffsetPoint(arrowTip, -unitVec.X * size, -unitVec.Y * size);
            Result[0] := arrowTip;
            Result[1] := OffsetPoint(basePt, -unitVec.Y * sDiv50, unitVec.X * sDiv50);
            Result[2] := OffsetPoint(basePt, unitVec.Y * sDiv50, -unitVec.X * sDiv50);
         End;
      asFancy:
         Begin
            setLength(Result, 4);
            basePt := OffsetPoint(arrowTip, -unitVec.X * sDiv120, -unitVec.Y * sDiv120);
            Result[0] := OffsetPoint(basePt, -unitVec.Y * sDiv50, unitVec.X * sDiv50);
            Result[1] := OffsetPoint(arrowTip, -unitVec.X * size, -unitVec.Y * size);
            Result[2] := OffsetPoint(basePt, unitVec.Y * sDiv50, -unitVec.X * sDiv50);
            Result[3] := arrowTip;
         End;
      asDiamond:
         Begin
            setLength(Result, 4);
            basePt := OffsetPoint(arrowTip, -unitVec.X * sDiv60, -unitVec.Y * sDiv60);
            Result[0] := arrowTip;
            Result[1] := OffsetPoint(basePt, -unitVec.Y * sDiv50, unitVec.X * sDiv50);
            Result[2] := OffsetPoint(arrowTip, -unitVec.X * sDiv120, -unitVec.Y * sDiv120);
            Result[3] := OffsetPoint(basePt, unitVec.Y * sDiv50, -unitVec.X * sDiv50);
         End;
      asCircle:
         Begin
            basePt := OffsetPoint(arrowTip, -unitVec.X * sDiv50, -unitVec.Y * sDiv50);
            With Point(basePt) Do
               Result := Ellipse(RectD(x - sDiv50, y - sDiv50, x + sDiv50, y + sDiv50));
         End;
      asTail:
         Begin
            setLength(Result, 6);
            basePt := OffsetPoint(arrowTip, -unitVec.X * sDiv60, -unitVec.Y * sDiv60);
            Result[0] := OffsetPoint(arrowTip, -unitVec.X * sDiv50, -unitVec.Y * sDiv50);
            Result[1] := OffsetPoint(arrowTip, -unitVec.Y * sDiv40, unitVec.X * sDiv40);
            Result[2] := OffsetPoint(basePt, -unitVec.Y * sDiv40, unitVec.X * sDiv40);
            Result[3] := OffsetPoint(arrowTip, -unitVec.X * sDiv120, -unitVec.Y * sDiv120);
            Result[4] := OffsetPoint(basePt, unitVec.Y * sDiv40, -unitVec.X * sDiv40);
            Result[5] := OffsetPoint(arrowTip, unitVec.Y * sDiv40, -unitVec.X * sDiv40);
         End;

   End;
End;

Function GetDefaultArrowHeadSize(lineWidth: double): double;
Begin
   Result := lineWidth * 3 + 7;
End;

Function ShortenPath(Const path: TPathD; pathEnd: TPathEnd; amount: double): TPathD;
Var
   len, amount2: double;
   vec: TPointD;
   i, highPath: integer;
Begin
   Result := path;
   highPath := high(path);
   If highPath < 1 Then
      Exit;
   amount2 := amount;

   If pathEnd <> peEnd Then
   Begin

      i := 0;
      While (i < highPath) Do
      Begin
         len := Distance(Result[i], Result[i + 1]);
         If (len >= amount) Then
            Break;
         amount := amount - len;
         Inc(i);
      End;
      If i > 0 Then
      Begin
         Move(path[i], Result[0], (highPath - i + 1) * SizeOf(TPointD));
         Dec(highPath, i);
         SetLength(Result, highPath + 1);
      End;
      If amount > 0 Then
      Begin
         vec := GetUnitVector(Result[0], Result[1]);
         Result[0].X := Result[0].X + vec.X * amount;
         Result[0].Y := Result[0].Y + vec.Y * amount;
      End;
   End;

   If pathEnd <> peStart Then
   Begin

      While (highPath > 1) Do
      Begin
         len := Distance(Result[highPath], Result[highPath - 1]);
         If (len >= amount2) Then
            Break;
         amount2 := amount2 - len;
         Dec(highPath);
      End;
      SetLength(Result, highPath + 1);
      If amount2 > 0 Then
      Begin
         vec := GetUnitVector(Result[highPath], Result[highPath - 1]);
         Result[highPath].X := Result[highPath].X + vec.X * amount2;
         Result[highPath].Y := Result[highPath].Y + vec.Y * amount2;
      End;
   End;
End;

Function GetDashedPath(Const path: TPathD; closed: boolean; Const pattern: TArrayOfInteger; patternOffset: PDouble): TPathsD;
Var
   i, highI, paIdx: integer;
   vecs, path2, dash: TPathD;
   patCnt, patLen: integer;
   dashCapacity, dashCnt, ptsCapacity, ptsCnt: integer;
   segLen, residualPat, patOff: double;
   filling: boolean;
   pt, pt2: TPointD;

   Procedure NewDash;
   Begin
      If ptsCnt = 1 Then
         ptsCnt := 0;
      If ptsCnt = 0 Then
         Exit;
      If dashCnt = dashCapacity Then
      Begin
         Inc(dashCapacity, cnBuffSize);
         setLength(Result, dashCapacity);
      End;
      Result[dashCnt] := Copy(dash, 0, ptsCnt);
      Inc(dashCnt);
      ptsCapacity := cnBuffSize;
      setLength(dash, ptsCapacity);
      ptsCnt := 0;
   End;

   Procedure ExtendDash(Const pt: TPointD);
   Begin
      If ptsCnt = ptsCapacity Then
      Begin
         Inc(ptsCapacity, cnBuffSize);
         setLength(dash, ptsCapacity);
      End;
      dash[ptsCnt] := pt;
      Inc(ptsCnt);
   End;

Begin
   Result := Nil;
   paIdx := 0;
   patCnt := length(pattern);

   path2 := path;
   highI := high(path2);
   If (highI < 1) Or (patCnt = 0) Then
      Exit;

   If closed And ((path2[highI].X <> path2[0].X) Or (path2[highI].Y <> path2[0].Y)) Then
   Begin
      Inc(highI);
      setLength(path2, highI + 2);
      path2[highI] := path2[0];
   End;

   vecs := GetVectors(path2);
   If (vecs[0].X = 0) And (vecs[0].Y = 0) Then
      Exit;

   If Not assigned(patternOffset) Then
      patOff := 0
   Else
      patOff := patternOffset^;

   patLen := 0;
   For i := 0 To patCnt - 1 Do
      Inc(patLen, pattern[i]);

   If patOff < 0 Then
   Begin
      patOff := patLen + patOff;
      While patOff < 0 Do
         patOff := patOff + patLen;
   End
   Else
      While patOff > patLen Do
         patOff := patOff - patLen;

   dashCnt := 0;
   dashCapacity := 0;
   ptsCnt := 0;
   ptsCapacity := 0;

   filling := True;
   While patOff >= pattern[paIdx] Do
   Begin
      filling := Not filling;
      patOff := patOff - pattern[paIdx];
      paIdx := (paIdx + 1) Mod patCnt;
   End;
   residualPat := pattern[paIdx] - patOff;

   pt := path2[0];
   ExtendDash(pt);
   i := 0;
   While (i < highI) Do
   Begin
      segLen := Distance(pt, path2[i + 1]);
      If residualPat > segLen Then
      Begin
         If filling Then
            ExtendDash(path2[i + 1]);
         residualPat := residualPat - segLen;
         pt := path2[i + 1];
         Inc(i);
      End
      Else
      Begin
         pt2.X := pt.X + vecs[i].X * residualPat;
         pt2.Y := pt.Y + vecs[i].Y * residualPat;
         If filling Then
            ExtendDash(pt2);
         filling := Not filling;
         NewDash;
         paIdx := (paIdx + 1) Mod patCnt;
         residualPat := pattern[paIdx];
         pt := pt2;
         ExtendDash(pt);
      End;
   End;
   NewDash;
   SetLength(Result, dashCnt);

   If Not assigned(patternOffset) Then
      Exit;
   patOff := 0;
   For i := 0 To paIdx - 1 Do
      patOff := patOff + pattern[i];
   patternOffset^ := patOff + (pattern[paIdx] - residualPat);
End;

Function GetDashedOutLine(Const path: TPathD; closed: boolean; Const pattern: TArrayOfInteger; patternOffset: PDouble;
   lineWidth: double; joinStyle: TJoinStyle; endStyle: TEndStyle): TPathsD;
Var
   i: integer;
   tmp: TPathsD;
Begin
   Result := Nil;
   tmp := GetDashedPath(path, closed, pattern, patternOffset);
   For i := 0 To high(tmp) Do
      AppendPath(Result, GrowOpenLine(tmp[i], lineWidth, joinStyle, endStyle, 2));
End;

Function gaGetBoundsD(Const paths: TPathsD): TRectD;
Var
   i, j: integer;
   l, t, r, b: double;
   p: PPointD;
Begin
   l := MaxInt;
   t := MaxInt;
   r := -MaxInt;
   b := -MaxInt;
   For i := 0 To high(paths) Do
   Begin
      p := PPointD(paths[i]);
      If Not assigned(p) Then
         Continue;
      For j := 0 To high(paths[i]) Do
      Begin
         If p^.x < l Then
            l := p^.x;
         If p^.x > r Then
            r := p^.x;
         If p^.y < t Then
            t := p^.y;
         If p^.y > b Then
            b := p^.y;
         Inc(p);
      End;
   End;
   If r < l Then
      Result := cnNullRectD
   Else
      Result := RectD(l, t, r, b);
End;

Function gaGetBoundsD(Const path: TPathD): TRectD;
Var
   i, highI: integer;
   l, t, r, b: double;
   p: PPointD;
Begin
   highI := High(path);
   If highI < 0 Then
   Begin
      Result := cnNullRectD;
      Exit;
   End;
   l := path[0].X;
   r := l;
   t := path[0].Y;
   b := t;
   p := PPointD(path);
   For i := 1 To highI Do
   Begin
      Inc(p);
      If p^.x < l Then
         l := p^.x;
      If p^.x > r Then
         r := p^.x;
      If p^.y < t Then
         t := p^.y;
      If p^.y > b Then
         b := p^.y;
   End;
   Result := RectD(l, t, r, b);
End;

Function gaGetBounds(Const path: TPathD): TRect;
Var
   recD: TRectD;
Begin
   recD := gaGetBoundsD(path);
   Result := Rect(recD);
End;

Function gaGetBounds(Const paths: TPathsD): TRect;
Var
   recD: TRectD;
Begin
   recD := gaGetBoundsD(paths);
   Result := Rect(recD);
End;

Function GetPointInQuadBezier(Const a, b, c: TPointD; t: double): TPointD;
Var
   omt: double;
Begin
   If t > 1 Then
      t := 1
   Else If t < 0 Then
      t := 0;
   omt := 1 - t;
   Result.X := a.X * omt * omt + b.X * 2 * omt * t + c.X * t * t;
   Result.Y := a.Y * omt * omt + b.Y * 2 * omt * t + c.Y * t * t;
End;

Function FlattenQBezier(Const firstPt: TPointD; Const pts: TPathD;
   tolerance: double = 0.0): TPathD; Overload;
Begin
   If tolerance <= 0.0 Then
      tolerance := cnBezierTolerance;
   Result := FlattenQBezier(PrePendPoint(firstPt, pts), tolerance);
End;

Function FlattenQBezier(Const pts: TPathsD; tolerance: double = 0.0): TPathsD;
Var
   i, len: integer;
Begin
   len := Length(pts);
   SetLength(Result, len);
   If tolerance <= 0.0 Then
      tolerance := cnBezierTolerance;
   For i := 0 To len - 1 Do
      Result[i] := FlattenQBezier(pts[i], tolerance);
End;

Function FlattenQBezier(Const pts: TPathD; tolerance: double = 0.0): TPathD;
Var
   i, highI: integer;
   p: TPathD;
Begin
   Result := Nil;
   highI := high(pts);
   If highI < 0 Then
      Exit;
   If (highI < 2) Or Odd(highI) Then
      Raise Exception.Create(rsInvalidQBezier);
   If tolerance <= 0.0 Then
      tolerance := cnBezierTolerance;
   setLength(Result, 1);
   Result[0] := pts[0];
   For i := 0 To (highI Div 2) - 1 Do
   Begin
      If PointsEqual(pts[i * 2], pts[i * 2 + 1]) And PointsEqual(pts[i * 2 + 1], pts[i * 2 + 2]) Then
      Begin
         AppendPoint(Result, pts[i * 2]);
         AppendPoint(Result, pts[i * 2 + 2]);
      End
      Else
      Begin
         p := FlattenQBezier(pts[i * 2], pts[i * 2 + 1], pts[i * 2 + 2], tolerance);
         AppendPath(Result, Copy(p, 1, Length(p) - 1));
      End;
   End;
End;

Function FlattenQBezier(Const pt1, pt2, pt3: TPointD; tolerance: double = 0.0): TPathD;
Var
   resultCnt, resultLen: integer;

   Procedure AddPoint(Const pt: TPointD);
   Begin
      If resultCnt = resultLen Then
      Begin
         Inc(resultLen, cnBuffSize);
         setLength(Result, resultLen);
      End;
      Result[resultCnt] := pt;
      Inc(resultCnt);
   End;

   Procedure DoCurve(Const p1, p2, p3: TPointD);
   Var
      p12, p23, p123: TPointD;
   Begin
      If (abs(p1.x + p3.x - 2 * p2.x) + abs(p1.y + p3.y - 2 * p2.y) < tolerance) Then
      Begin
         AddPoint(p3);
      End
      Else
      Begin
         P12.X := (P1.X + P2.X) * 0.5;
         P12.Y := (P1.Y + P2.Y) * 0.5;
         P23.X := (P2.X + P3.X) * 0.5;
         P23.Y := (P2.Y + P3.Y) * 0.5;
         P123.X := (P12.X + P23.X) * 0.5;
         P123.Y := (P12.Y + P23.Y) * 0.5;
         DoCurve(p1, p12, p123);
         DoCurve(p123, p23, p3);
      End;
   End;

Begin
   resultLen := 0;
   resultCnt := 0;
   If tolerance <= 0.0 Then
      tolerance := cnBezierTolerance;

   AddPoint(pt1);
   If ((pt1.X = pt2.X) And (pt1.Y = pt2.Y)) Or ((pt2.X = pt3.X) And (pt2.Y = pt3.Y)) Then
   Begin
      AddPoint(pt3);
   End
   Else
      DoCurve(pt1, pt2, pt3);
   SetLength(Result, resultCnt);
End;

Function GetPointInCubicBezier(Const a, b, c, d: TPointD; t: double): TPointD;
Var
   omt: double;
Begin
   If t > 1 Then
      t := 1
   Else If t < 0 Then
      t := 0;
   omt := 1 - t;
   Result.X := a.X * omt * omt * omt + b.X * 3 * omt * omt * t + c.X * 3 * omt * t * t + d.X * t * t * t;
   Result.Y := a.Y * omt * omt * omt + b.Y * 3 * omt * omt * t + c.Y * 3 * omt * t * t + d.Y * t * t * t;
End;


Function FlattenCBezier(Const firstPt: TPointD; Const pts: TPathD;
   tolerance: double = 0.0): TPathD; Overload;
Begin
   Result := FlattenCBezier(PrePendPoint(firstPt, pts), tolerance);
End;

Function FlattenCBezier(Const pts: TPathsD; tolerance: double = 0.0): TPathsD;
Var
   i, len: integer;
Begin
   len := Length(pts);
   SetLength(Result, len);
   If tolerance <= 0.0 Then
      tolerance := cnBezierTolerance;
   For i := 0 To len - 1 Do
      Result[i] := FlattenCBezier(pts[i], tolerance);
End;

Function FlattenCBezier(Const pts: TPathD; tolerance: double = 0.0): TPathD;
Var
   i, len: integer;
   p: TPathD;
Begin
   Result := Nil;
   len := Length(pts) - 1;
   If len < 0 Then
      Exit;
   If (len < 3) Or (len Mod 3 <> 0) Then
      Raise Exception.Create(rsInvalidCBezier);
   If tolerance <= 0.0 Then
      tolerance := cnBezierTolerance;
   setLength(Result, 1);
   Result[0] := pts[0];
   For i := 0 To (len Div 3) - 1 Do
   Begin
      If PointsEqual(pts[i * 3], pts[i * 3 + 1]) And PointsEqual(pts[i * 3 + 2], pts[i * 3 + 3]) Then
      Begin
         AppendPoint(Result, pts[i * 3]);
         AppendPoint(Result, pts[i * 3 + 3]);
      End
      Else
      Begin
         p := FlattenCBezier(pts[i * 3], pts[i * 3 + 1], pts[i * 3 + 2], pts[i * 3 + 3], tolerance);
         AppendPath(Result, Copy(p, 1, Length(p) - 1));
      End;
   End;
End;

Function FlattenCBezier(Const pt1, pt2, pt3, pt4: TPointD; tolerance: double = 0.0): TPathD;
Var
   resultCnt, resultLen: integer;

   Procedure AddPoint(Const pt: TPointD);
   Begin
      If resultCnt = resultLen Then
      Begin
         Inc(resultLen, cnBuffSize);
         setLength(Result, resultLen);
      End;
      Result[resultCnt] := pt;
      Inc(resultCnt);
   End;

   Procedure DoCurve(Const p1, p2, p3, p4: TPointD);
   Var
      p12, p23, p34, p123, p234, p1234: TPointD;
   Begin
      If (abs(p1.x + p3.x - 2 * p2.x) + abs(p2.x + p4.x - 2 * p3.x) + abs(p1.y + p3.y - 2 * p2.y) + abs(p2.y + p4.y - 2 * p3.y)) < tolerance Then
      Begin
         If resultCnt = length(Result) Then
            setLength(Result, length(Result) + cnBuffSize);
         Result[resultCnt] := p4;
         Inc(resultCnt);
      End
      Else
      Begin
         p12.X := (p1.X + p2.X) / 2;
         p12.Y := (p1.Y + p2.Y) / 2;
         p23.X := (p2.X + p3.X) / 2;
         p23.Y := (p2.Y + p3.Y) / 2;
         p34.X := (p3.X + p4.X) / 2;
         p34.Y := (p3.Y + p4.Y) / 2;
         p123.X := (p12.X + p23.X) / 2;
         p123.Y := (p12.Y + p23.Y) / 2;
         p234.X := (p23.X + p34.X) / 2;
         p234.Y := (p23.Y + p34.Y) / 2;
         p1234.X := (p123.X + p234.X) / 2;
         p1234.Y := (p123.Y + p234.Y) / 2;
         DoCurve(p1, p12, p123, p1234);
         DoCurve(p1234, p234, p34, p4);
      End;
   End;

Begin
   Result := Nil;
   resultLen := 0;
   resultCnt := 0;
   If tolerance <= 0.0 Then
      tolerance := cnBezierTolerance;

   AddPoint(pt1);
   If (pt1.X = pt2.X) And (pt1.Y = pt2.Y) And (pt3.X = pt4.X) And (pt3.Y = pt4.Y) Then
   Begin
      AddPoint(pt4);
   End
   Else
      DoCurve(pt1, pt2, pt3, pt4);
   SetLength(Result, resultCnt);
End;

Function ReflectPoint(Const pt, pivot: TPointD): TPointD;
Begin
   Result.X := pivot.X + (pivot.X - pt.X);
   Result.Y := pivot.Y + (pivot.Y - pt.Y);
End;

Function FlattenCSpline(Const pts: TPathD; tolerance: double = 0.0): TPathD;
Var
   resultCnt, resultLen: integer;

   Procedure AddPoint(Const pt: TPointD);
   Begin
      If resultCnt = resultLen Then
      Begin
         Inc(resultLen, cnBuffSize);
         setLength(Result, resultLen);
      End;
      Result[resultCnt] := pt;
      Inc(resultCnt);
   End;

   Procedure DoCurve(Const p1, p2, p3, p4: TPointD);
   Var
      p12, p23, p34, p123, p234, p1234: TPointD;
   Begin
      If (abs(p1.x + p3.x - 2 * p2.x) + abs(p2.x + p4.x - 2 * p3.x) + abs(p1.y + p3.y - 2 * p2.y) + abs(p2.y + p4.y - 2 * p3.y)) < tolerance Then
      Begin
         If resultCnt = length(Result) Then
            setLength(Result, length(Result) + cnBuffSize);
         Result[resultCnt] := p4;
         Inc(resultCnt);
      End
      Else
      Begin
         p12.X := (p1.X + p2.X) / 2;
         p12.Y := (p1.Y + p2.Y) / 2;
         p23.X := (p2.X + p3.X) / 2;
         p23.Y := (p2.Y + p3.Y) / 2;
         p34.X := (p3.X + p4.X) / 2;
         p34.Y := (p3.Y + p4.Y) / 2;
         p123.X := (p12.X + p23.X) / 2;
         p123.Y := (p12.Y + p23.Y) / 2;
         p234.X := (p23.X + p34.X) / 2;
         p234.Y := (p23.Y + p34.Y) / 2;
         p1234.X := (p123.X + p234.X) / 2;
         p1234.Y := (p123.Y + p234.Y) / 2;
         DoCurve(p1, p12, p123, p1234);
         DoCurve(p1234, p234, p34, p4);
      End;
   End;

Var
   i, len: integer;
   p: PPointD;
   pt1, pt2, pt3, pt4: TPointD;
Begin
   Result := Nil;
   len := Length(pts);
   resultLen := 0;
   resultCnt := 0;
   If (len < 4) Then
      Exit;

   If tolerance <= 0.0 Then
      tolerance := cnBezierTolerance;

   If Odd(len) Then
      Dec(len);
   p := @pts[0];
   AddPoint(p^);
   pt1 := p^;
   Inc(p);
   pt2 := p^;
   Inc(p);
   For i := 0 To (len Shr 1) - 2 Do
   Begin
      pt3 := p^;
      Inc(p);
      pt4 := p^;
      Inc(p);
      DoCurve(pt1, pt2, pt3, pt4);
      pt1 := pt4;
      pt2 := ReflectPoint(pt3, pt1);
   End;
   SetLength(Result, resultCnt);
End;


Function FlattenCSpline(Const priorCtrlPt, startPt: TPointD; Const pts: TPathD; tolerance: double = 0.0): TPathD;
Var
   p: TPathD;
   len: integer;
Begin
   len := Length(pts);
   SetLength(p, len + 2);
   p[0] := startPt;
   p[1] := ReflectPoint(priorCtrlPt, startPt);
   If len > 0 Then
      Move(pts[0], p[2], len * SizeOf(TPointD));
   Result := FlattenCSpline(p, tolerance);
End;

Function FlattenQSpline(Const priorCtrlPt, startPt: TPointD; Const pts: TPathD; tolerance: double = 0.0): TPathD;
Var
   p: TPathD;
   len: integer;
Begin
   len := Length(pts);
   SetLength(p, len + 2);
   p[0] := startPt;
   p[1] := ReflectPoint(priorCtrlPt, startPt);
   If len > 0 Then
      Move(pts[0], p[2], len * SizeOf(TPointD));
   Result := FlattenQSpline(p, tolerance);
End;

Function FlattenQSpline(Const pts: TPathD; tolerance: double = 0.0): TPathD;
Var
   resultCnt, resultLen: integer;

   Procedure AddPoint(Const pt: TPointD);
   Begin
      If resultCnt = resultLen Then
      Begin
         Inc(resultLen, cnBuffSize);
         setLength(Result, resultLen);
      End;
      Result[resultCnt] := pt;
      Inc(resultCnt);
   End;

   Procedure DoCurve(Const p1, p2, p3: TPointD);
   Var
      p12, p23, p123: TPointD;
   Begin
      If (abs(p1.x + p3.x - 2 * p2.x) + abs(p1.y + p3.y - 2 * p2.y) < tolerance) Then
      Begin
         AddPoint(p3);
      End
      Else
      Begin
         P12.X := (P1.X + P2.X) * 0.5;
         P12.Y := (P1.Y + P2.Y) * 0.5;
         P23.X := (P2.X + P3.X) * 0.5;
         P23.Y := (P2.Y + P3.Y) * 0.5;
         P123.X := (P12.X + P23.X) * 0.5;
         P123.Y := (P12.Y + P23.Y) * 0.5;
         DoCurve(p1, p12, p123);
         DoCurve(p123, p23, p3);
      End;
   End;

Var
   i, len: integer;
   p: PPointD;
   pt1, pt2, pt3: TPointD;
Begin
   Result := Nil;
   len := Length(pts);
   If (len < 3) Then
      Exit;
   resultLen := 0;
   resultCnt := 0;
   If tolerance <= 0.0 Then
      tolerance := cnBezierTolerance;

   p := @pts[0];
   AddPoint(p^);
   pt1 := p^;
   Inc(p);
   pt2 := p^;
   Inc(p);
   For i := 0 To len - 3 Do
   Begin
      pt3 := p^;
      Inc(p);
      DoCurve(pt1, pt2, pt3);
      pt1 := pt3;
      pt2 := ReflectPoint(pt2, pt1);
   End;
   SetLength(Result, resultCnt);
End;

Function MakePathI(Const pts: Array Of integer): TPathD; Overload;
Var
   i, j, x, y, len: integer;
Begin
   Result := Nil;
   len := length(pts) Div 2;
   If len < 1 Then
      Exit;

   setlength(Result, len);
   Result[0].X := pts[0];
   Result[0].Y := pts[1];
   j := 0;
   For i := 1 To len - 1 Do
   Begin
      x := pts[i * 2];
      y := pts[i * 2 + 1];
      Inc(j);
      Result[j].X := x;
      Result[j].Y := y;
   End;
   setlength(Result, j + 1);
End;

Function MakePathD(Const pts: Array Of double): TPathD;
Var
   i, j, len: integer;
   x, y: double;
Begin
   Result := Nil;
   len := length(pts) Div 2;
   If len < 1 Then
      Exit;

   setlength(Result, len);
   Result[0].X := pts[0];
   Result[0].Y := pts[1];
   j := 0;
   For i := 1 To len - 1 Do
   Begin
      x := pts[i * 2];
      y := pts[i * 2 + 1];
      Inc(j);
      Result[j].X := x;
      Result[j].Y := y;
   End;
   setlength(Result, j + 1);
End;

Function IsIdentityMatrix(Const matrix: TMatrixD): boolean;
Var
   i, j: integer;
Const
   matVal: Array[boolean] Of double = (0.0, 1.0);
Begin
   Result := False;
   For i := 0 To 2 Do
      For j := 0 To 2 Do
         If matrix[i][j] <> matVal[j = i] Then
            Exit;
   Result := True;
End;

Function IsValidMatrix(Const matrix: TMatrixD): boolean;
Begin
   Result := matrix[2][2] = 1.0;
End;

Function Matrix(Const m00, m01, m02, m10, m11, m12, m20, m21, m22: double): TMatrixD;
Begin
   Result[0, 0] := m00;
   Result[0, 1] := m01;
   Result[0, 2] := m02;
   Result[1, 0] := m10;
   Result[1, 1] := m11;
   Result[1, 2] := m12;
   Result[2, 0] := m20;
   Result[2, 1] := m21;
   Result[2, 2] := m22;
End;

Function Det4(a1, a2, b1, b2: double): double;
Begin
   Result := a1 * b2 - a2 * b1;
End;

Function Det9(a1, a2, a3, b1, b2, b3, c1, c2, c3: double): double;
Begin
   Result := a1 * Det4(b2, b3, c2, c3) - b1 * Det4(a2, a3, c2, c3) + c1 * Det4(a2, a3, b2, b3);
End;

Function MatrixDeterminant(Const matrix: TMatrixD): double;  
//=== ct9999-R160 ========= FPC Internal ERROR 200312205 on ARM32CPU
{$IFDEF CPUARM}
 Begin
   Result := 0;
End;
{$ELSE}
Begin
   Result := Det9(matrix[0, 0], matrix[1, 0], matrix[2, 0], matrix[0, 1], matrix[1, 1], matrix[2, 1], matrix[0, 2], matrix[1, 2], matrix[2, 2]);    
End;
{$ENDIF}
 //==================================================

Function MatrixAdjugate(Const matrix: TMatrixD): TMatrixD;
Begin
   Result[0, 0] := Det4(matrix[1, 1], matrix[1, 2], matrix[2, 1], matrix[2, 2]);
   Result[0, 1] := -Det4(matrix[0, 1], matrix[0, 2], matrix[2, 1], matrix[2, 2]);
   Result[0, 2] := Det4(matrix[0, 1], matrix[0, 2], matrix[1, 1], matrix[1, 2]);

   Result[1, 0] := -Det4(matrix[1, 0], matrix[1, 2], matrix[2, 0], matrix[2, 2]);
   Result[1, 1] := Det4(matrix[0, 0], matrix[0, 2], matrix[2, 0], matrix[2, 2]);
   Result[1, 2] := -Det4(matrix[0, 0], matrix[0, 2], matrix[1, 0], matrix[1, 2]);

   Result[2, 0] := Det4(matrix[1, 0], matrix[1, 1], matrix[2, 0], matrix[2, 1]);
   Result[2, 1] := -Det4(matrix[0, 0], matrix[0, 1], matrix[2, 0], matrix[2, 1]);
   Result[2, 2] := Det4(matrix[0, 0], matrix[0, 1], matrix[1, 0], matrix[1, 1]);
End;

Procedure MatrixApply(Const matrix: TMatrixD; Var x, y: double);
Var
   tmpX: double;
Begin
   tmpX := x;
   x := tmpX * matrix[0, 0] + y * matrix[1, 0] + matrix[2, 0];
   y := tmpX * matrix[0, 1] + y * matrix[1, 1] + matrix[2, 1];
End;

Procedure MatrixApply(Const matrix: TMatrixD; Var pt: TPointD);
Var
   tmpX: double;
Begin
   tmpX := pt.x;
   pt.X := tmpX * matrix[0, 0] + pt.Y * matrix[1, 0] + matrix[2, 0];
   pt.Y := tmpX * matrix[0, 1] + pt.Y * matrix[1, 1] + matrix[2, 1];
End;

Procedure MatrixApply(Const matrix: TMatrixD; Var rec: TRect);
Var
   l, t, b, r, tmpX: double;
Begin
   tmpX := rec.Left;
   l := tmpX * matrix[0, 0] + rec.Top * matrix[1, 0] + matrix[2, 0];
   t := tmpX * matrix[0, 1] + rec.Top * matrix[1, 1] + matrix[2, 1];
   tmpX := rec.Right;
   r := tmpX * matrix[0, 0] + rec.Bottom * matrix[1, 0] + matrix[2, 0];
   b := tmpX * matrix[0, 1] + rec.Bottom * matrix[1, 1] + matrix[2, 1];
   rec := Rect(RectD(l, t, r, b));
End;

Procedure MatrixApply(Const matrix: TMatrixD; Var rec: TRectD);
Var
   path: TPathD;
Begin
   path := Rectangle(rec);
   MatrixApply(matrix, path);
   rec := gaGetBoundsD(path);
End;

Procedure MatrixApply(Const matrix: TMatrixD; Var rec: TRectWH);
Var
   path: TPathD;
   rec2: TRectD;
Begin
   With rec Do
   Begin
      SetLength(path, 2);
      path[0] := PointD(Left, Top);
      path[1] := PointD(Left + Width, Top + Height);
   End;
   MatrixApply(matrix, path);
   rec2 := gaGetBoundsD(path);

   rec.Left := rec2.Left;
   rec.Top := rec2.Top;
   rec.Width := rec2.Right - rec2.Left;
   rec.Height := rec2.Bottom - rec2.Top;
End;

Procedure MatrixApply(Const matrix: TMatrixD; Var path: TPathD);
Var
   i, len: integer;
   tmpX: double;
   pp: PPointD;
Begin
   len := Length(path);
   If (len = 0) Or IsIdentityMatrix(matrix) Then
      Exit;
   pp := @path[0];
   For i := 0 To len - 1 Do
   Begin
      tmpX := pp^.X;
      pp^.X := tmpX * matrix[0, 0] + pp^.Y * matrix[1, 0] + matrix[2, 0];
      pp^.Y := tmpX * matrix[0, 1] + pp^.Y * matrix[1, 1] + matrix[2, 1];
      Inc(pp);
   End;
End;

Procedure MatrixApply(Const matrix: TMatrixD; Var paths: TPathsD);
Var
   i, j, len: integer;
   tmpX: double;
   pp: PPointD;
Begin
   If Not Assigned(paths) Or IsIdentityMatrix(matrix) Then
      Exit;

   For i := 0 To High(paths) Do
   Begin
      len := Length(paths[i]);
      If len = 0 Then
         Continue;
      pp := @paths[i][0];
      For j := 0 To High(paths[i]) Do
      Begin
         tmpX := pp^.X;
         pp^.X := tmpX * matrix[0, 0] + pp^.Y * matrix[1, 0] + matrix[2, 0];
         pp^.Y := tmpX * matrix[0, 1] + pp^.Y * matrix[1, 1] + matrix[2, 1];
         Inc(pp);
      End;
   End;
End;

Function MatrixMultiply(Const modifier, matrix: TMatrixD): TMatrixD;
Var
   i, j: integer;
Begin
   For i := 0 To 2 Do
      For j := 0 To 2 Do
         Result[i, j] :=
            (modifier[0, j] * matrix[i, 0]) + (modifier[1, j] * matrix[i, 1]) + (modifier[2, j] * matrix[i, 2]);
End;

Procedure MatrixScale(Var matrix: TMatrixD; scaleX, scaleY: double);
Var
   m: TMatrixD;
Begin
   m := cnIdentityMatrix;
   If (scaleX = 0) Or (scaleY = 0) Then
      Raise Exception(rsInvalidScale);

   If ValueAlmostOne(scaleX) And ValueAlmostOne(scaleY) Then
      Exit;
   m[0, 0] := scaleX;
   m[1, 1] := scaleY;
   matrix := MatrixMultiply(m, matrix);
End;

Procedure MatrixScale(Var matrix: TMatrixD; scale: double);
Begin
   If (scale = 0) Or (scale = 1) Then
      Exit;
   MatrixScale(matrix, scale, scale);
End;

Procedure MatrixRotate(Var matrix: TMatrixD; Const center: TPointD; angRad: double);
Var
   m: TMatrixD;
   sinA, cosA: double;
   origOffset: boolean;
Begin
   NormalizeAngle(angRad);
   If angRad = 0 Then
      Exit;
   If VarClockwiseRotationIsAnglePositive Then
      angRad := -angRad;
   m := cnIdentityMatrix;
   origOffset := (center.X <> 0) Or (center.Y <> 0);
   If origOffset Then
      MatrixTranslate(matrix, -center.X, -center.Y);
   GetSinCos(angRad, sinA, cosA);
   m := cnIdentityMatrix;
   m[0, 0] := cosA;
   m[1, 0] := sinA;
   m[0, 1] := -sinA;
   m[1, 1] := cosA;
   matrix := MatrixMultiply(m, matrix);
   If origOffset Then
      MatrixTranslate(matrix, center.X, center.Y);
End;

Procedure MatrixTranslate(Var matrix: TMatrixD; dx, dy: double);
Var
   m: TMatrixD;
Begin
   If ValueAlmostZero(dx) And ValueAlmostZero(dy) Then
      Exit;
   m := cnIdentityMatrix;
   m[2, 0] := dx;
   m[2, 1] := dy;
   matrix := MatrixMultiply(m, matrix);
End;

Procedure ScaleInternal(Var matrix: TMatrixD; s: double);
Var
   i, j: integer;
Begin
   For i := 0 To 2 Do
      For j := 0 To 2 Do
         matrix[i, j] := matrix[i, j] * s;
End;

Function MatrixInvert(Var matrix: TMatrixD): boolean;
Var
   d: double;
Const
   tolerance = 1.0E-5;
Begin
   d := MatrixDeterminant(matrix);
   Result := abs(d) > tolerance;
   If Result Then
   Begin
      matrix := MatrixAdjugate(matrix);
      ScaleInternal(matrix, 1 / d);
   End;
End;

Procedure MatrixSkew(Var matrix: TMatrixD; angleX, angleY: double);
Var
   m: TMatrixD;
Begin
   If ValueAlmostZero(angleX) And ValueAlmostZero(angleY) Then
      Exit;
   m := cnIdentityMatrix;
   m[1, 0] := tan(angleX);
   m[0, 1] := tan(angleY);
   matrix := MatrixMultiply(m, matrix);
End;

Procedure MatrixMulCoord(Const matrix: TMatrixD; Var x, y, z: double);
Var
   xx, yy: double;
Begin
   xx := x;
   yy := y;
   x := matrix[0, 0] * xx + matrix[0, 1] * yy + matrix[0, 2] * z;
   y := matrix[1, 0] * xx + matrix[1, 1] * yy + matrix[1, 2] * z;
   z := matrix[2, 0] * xx + matrix[2, 1] * yy + matrix[2, 2] * z;
End;

Function BasisToPoints(x1, y1, x2, y2, x3, y3, x4, y4: double): TMatrixD;
//=== ct9999-R160 ========= FPC Internal ERROR 200312205 on ARM32CPU
{$IFDEF CPUARM}
Begin
   Result := cnIdentityMatrix;
End;
 {$ELSE}              
Var
   m, m2: TMatrixD;
   z4: double;
Begin
   m := Matrix(x1, x2, x3, y1, y2, y3, 1, 1, 1);
   m2 := MatrixAdjugate(m);
   z4 := 1;
   MatrixMulCoord(m2, x4, y4, z4);
   m2 := Matrix(x4, 0, 0, 0, y4, 0, 0, 0, z4);
   Result := MatrixMultiply(m2, m);
End;           
{$ENDIF}
//================================================

Procedure GetSrcCoords256(Const matrix: TMatrixD; Var x, y: integer);
Var
   xx, yy, zz: double;
Const
   Q: integer = MaxInt Div 256;
Begin
   xx := x;
   yy := y;
   zz := 1;
   MatrixMulCoord(matrix, xx, yy, zz);

   If zz = 0 Then
   Begin
      If xx >= 0 Then
         x := Q
      Else
         x := -MaxInt;
      If yy >= 0 Then
         y := Q
      Else
         y := -MaxInt;
   End
   Else
   Begin
      xx := xx / zz;
      If xx > Q Then
         x := MaxInt
      Else If xx < -Q Then
         x := -MaxInt
      Else
         x := Round(xx * 256);

      yy := yy / zz;
      If yy > Q Then
         y := MaxInt
      Else If yy < -Q Then
         y := -MaxInt
      Else
         y := Round(yy * 256);
   End;
End;

Function GetProjectionMatrix(Const srcPts, dstPts: TPathD): TMatrixD;
Var
   srcMat, dstMat: TMatrixD;
Begin
   If (length(srcPts) <> 4) Or (length(dstPts) <> 4) Then
   Begin
      Result := cnIdentityMatrix;
      Exit;
   End;
   srcMat := BasisToPoints(srcPts[0].X, srcPts[0].Y, srcPts[1].X, srcPts[1].Y, srcPts[2].X, srcPts[2].Y, srcPts[3].X, srcPts[3].Y);
   dstMat := BasisToPoints(dstPts[0].X, dstPts[0].Y, dstPts[1].X, dstPts[1].Y, dstPts[2].X, dstPts[2].Y, dstPts[3].X, dstPts[3].Y);
   Result := MatrixMultiply(MatrixAdjugate(dstMat), srcMat);
End;

Function ReColor(color, newColor: TColor32): TColor32;
Begin
   Result := Alpha(color) Or newColor;
End;

Function InterpolateSegX(Const pt1, pt2: TPointD): TPathD;
Var
   i, x1, x2: integer;
   xo, dydx: double;
Begin
   Result := Nil;
   If pt2.X > pt1.X Then
   Begin
      x1 := Ceil(pt1.X);
      x2 := Ceil(pt2.X);
      If x1 = x2 Then
         Exit;
      dydx := (pt2.Y - pt1.Y) / (pt2.X - pt1.X);
      xo := x1 - pt1.X;
      SetLength(Result, x2 - x1);
      For i := 0 To x2 - x1 - 1 Do
      Begin
         Result[i].X := x1 + i;
         Result[i].Y := pt1.Y + dydx * (xo + i);
      End;
   End
   Else
   Begin
      x1 := Floor(pt1.X);
      x2 := Floor(pt2.X);
      If x1 = x2 Then
         Exit;
      dydx := (pt2.Y - pt1.Y) / (pt2.X - pt1.X);
      xo := x1 - pt1.X;
      SetLength(Result, x1 - x2);
      For i := 0 To x1 - x2 - 1 Do
      Begin
         Result[i].X := x1 - i;
         Result[i].Y := pt1.Y + dydx * (xo - i);
      End;
   End;
End;

Function InterpolateSegY(Const pt1, pt2: TPointD): TPathD;
Var
   i, y1, y2: integer;
   yo, dxdy: double;
Begin
   Result := Nil;
   If pt2.Y > pt1.Y Then
   Begin
      y1 := Ceil(pt1.Y);
      y2 := Ceil(pt2.Y);
      If y1 = y2 Then
         Exit;
      dxdy := (pt2.X - pt1.X) / (pt2.Y - pt1.Y);
      yo := y1 - pt1.Y;
      SetLength(Result, y2 - y1);
      For i := 0 To y2 - y1 - 1 Do
      Begin
         Result[i].Y := y1 + i;
         Result[i].X := pt1.X + dxdy * (yo + i);
      End;
   End
   Else
   Begin
      y1 := Floor(pt1.Y);
      y2 := Floor(pt2.Y);
      If y1 = y2 Then
         Exit;
      dxdy := (pt2.X - pt1.X) / (pt2.Y - pt1.Y);
      yo := y1 - pt1.Y;
      SetLength(Result, y1 - y2);
      For i := 0 To y1 - y2 - 1 Do
      Begin
         Result[i].Y := y1 - i;
         Result[i].X := pt1.X + dxdy * (yo - i);
      End;
   End;
End;

Function InterpolatePathForX(Const path: TPathD): TPathD;
Var
   i, len: integer;
   tmp: TPathD;
Begin
   Result := Nil;
   len := length(path);
   If len < 2 Then
      Exit;
   For i := 1 To len - 1 Do
   Begin
      tmp := InterpolateSegX(path[i - 1], path[i]);
      AppendPath(Result, tmp);
   End;
End;

Function InterpolatePathForY(Const path: TPathD): TPathD;
Var
   i, len: integer;
   tmp: TPathD;
Begin
   Result := Nil;
   len := length(path);
   If len < 2 Then
      Exit;
   For i := 1 To len - 1 Do
   Begin
      tmp := InterpolateSegY(path[i - 1], path[i]);
      AppendPath(Result, tmp);
   End;
End;

Procedure ExtractAllFromMatrix(Const mat: TMatrixD; Out angle: double; Out scale, skew, trans: TPointD);
Var
   a, b, c, d, e, f: double;
   delta, r, s: double;
Begin
   a := mat[0][0];
   b := mat[1][0];
   c := mat[0][1];
   d := mat[1][1];
   e := mat[2][0];
   f := mat[2][1];

   delta := a * d - b * c;
   trans := PointD(e, f);
   angle := 0;
   scale := PointD(1, 1);
   skew := cnNullPointD;

   If (a <> 0) Or (b <> 0) Then
   Begin
      r := Sqrt(a * a + b * b);
      angle := ArcCos(a / r);
      If b < 0 Then
         angle := -angle;
      scale.X := r;
      scale.Y := delta / r;
      skew.X := ArcTan((a * c + b * d) / (r * r));
   End
   Else If (c <> 0) Or (d <> 0) Then
   Begin
      s := Sqrt(c * c + d * d);
      If d > 0 Then
         angle := cnAngle90 - ArcCos(-c / s)
      Else
         angle := cnAngle90 + ArcCos(c / s);
      scale.X := delta / s;
      scale.Y := s;
      skew.Y := ArcTan((a * c + b * d) / (s * s));
   End;
   angle := -angle;
   NormalizeAngle(angle);
End;

Function ExtractAngleFromMatrix(Const mat: TMatrixD): double;
Var
   a, b, c, d: double;
   r, s: double;
Begin
   a := mat[0][0];
   b := mat[1][0];
   c := mat[0][1];
   d := mat[1][1];

   If (a <> 0) Or (b <> 0) Then
   Begin
      r := Sqrt(a * a + b * b);
      Result := ArcCos(a / r);
      If b < 0 Then
         Result := -Result;
   End
   Else If (c <> 0) Or (d <> 0) Then
   Begin
      s := Sqrt(c * c + d * d);
      If d > 0 Then
         Result := cnAngle90 - ArcCos(-c / s)
      Else
         Result := cnAngle90 + ArcCos(c / s);
   End
   Else
   Begin
      Result := cnInvalidD;
      Exit;
   End;
   Result := -Result;
   NormalizeAngle(Result);
End;

Function ExtractScaleFromMatrix(Const mat: TMatrixD): TSizeD;
Var
   a, b, c, d: double;
   delta, q: double;
Begin
   a := mat[0][0];
   b := mat[1][0];
   c := mat[0][1];
   d := mat[1][1];

   delta := a * d - b * c;
   If (a <> 0) Or (b <> 0) Then
   Begin
      q := Sqrt(a * a + b * b);
      Result.sx := q;
      Result.sy := delta / q;
   End
   Else If (c <> 0) Or (d <> 0) Then
   Begin
      q := Sqrt(c * c + d * d);
      Result.sx := delta / q;
      Result.sy := q;
   End
   Else
      Result := SizeD(0.0, 0.0);
End;

Function ExtractAvgScaleFromMatrix(Const mat: TMatrixD): double;
Var
   scale: TSizeD;
Begin
   scale := ExtractScaleFromMatrix(mat);
   Result := Average(Abs(scale.sx), Abs(scale.sy));
End;

Type

   TFragment = Record
      botX, topX, dy, dydx: double;
   End;
   TFragmentArray = Array[0..(Maxint Div SizeOf(TFragment)) - 1] Of TFragment;
   PFragments = ^TFragmentArray;
   PFragment = ^TFragment;

   TScanLine = Record
      Y: integer;
      minX, maxX: integer;
      fragCnt: integer;
      fragments: PFragments;
   End;
   PScanline = ^TScanline;
   TArrayOfScanline = Array Of TScanline;

Type
   PArgbs = ^TArgbs;
   TArgbs = Array[0..(Maxint Div SizeOf(TARGB)) - 1] Of TARGB;

Function GetPixel(current: PARGB; delta: integer): PARGB;
Begin
   Result := current;
   Inc(Result, delta);
End;

Function ReverseColors(Const colors: TArrayOfGradientColor): TArrayOfGradientColor;
Var
   i, highI: integer;
Begin
   highI := High(colors);
   SetLength(Result, highI + 1);
   For i := 0 To highI Do
   Begin
      Result[i].color := colors[highI - i].color;
      Result[i].offset := 1 - colors[highI - i].offset;
   End;
End;

Procedure SwapColors(Var color1, color2: TColor32);
Var
   c: TColor32;
Begin
   c := color1;
   color1 := color2;
   color2 := c;
End;

Procedure SwapPoints(Var point1, point2: TPoint); Overload;
Var
   pt: TPoint;
Begin
   pt := point1;
   point1 := point2;
   point2 := pt;
End;

Procedure SwapPoints(Var point1, point2: TPointD); Overload;
Var
   pt: TPointD;
Begin
   pt := point1;
   point1 := point2;
   point2 := pt;
End;

Function ClampQ(q, endQ: integer): integer;
Begin
   If q < 0 Then
      Result := 0
   Else If q >= endQ Then
      Result := endQ - 1
   Else
      Result := q;
End;

Function ClampD(d: double; colorCnt: integer): integer;
Begin
   Dec(colorCnt);
   If d < 0 Then
      Result := 0
   Else If d >= 1 Then
      Result := colorCnt
   Else
      Result := Round(d * colorCnt);
End;

Function MirrorQ(q, endQ: integer): integer;
Begin
   Result := q Mod endQ;
   If (Result < 0) Then
      Result := -Result;
   If Odd(q Div endQ) Then
      Result := (endQ - 1) - Result;
End;

Function MirrorD(d: double; colorCnt: integer): integer;
Begin
   Dec(colorCnt);
   If Odd(Round(d)) Then
      Result := Round((1 - frac(d)) * colorCnt)
   Else
      Result := Round(frac(d) * colorCnt);
End;

Function RepeatQ(q, endQ: integer): integer;
Begin
   If (q < 0) Or (q >= endQ) Then
   Begin
      endQ := Abs(endQ);
      Result := q Mod endQ;
      If Result < 0 Then
         Inc(Result, endQ);
   End
   Else
      Result := q;
End;

Function SoftRptQ(q, endQ: integer): integer;
Begin
   If (q < 0) Then
      Result := endQ + (q Mod endQ)
   Else
      Result := (q Mod endQ);
   If Result = 0 Then
      Result := endQ Div 2;
End;

Function RepeatD(d: double; colorCnt: integer): integer;
Begin
   Dec(colorCnt);
   If (d < 0) Then
      Result := Round((1 + frac(d)) * colorCnt)
   Else
      Result := Round(frac(d) * colorCnt);
End;

Function BlendColorUsingMask(bgColor, fgColor: TColor32; mask: byte): TColor32;
Var
   bg: TARGB Absolute bgColor;
   fg: TARGB Absolute fgColor;
   res: TARGB Absolute Result;
   R, invR: PByteArray;
Begin
   If fg.A = 0 Then
   Begin
      Result := bgColor;
      res.A := MulBytes(res.A, Not mask);
   End
   Else If bg.A = 0 Then
   Begin
      Result := fgColor;
      res.A := MulBytes(res.A, mask);
   End
   Else If (mask = 0) Then
      Result := bgColor
   Else If (mask = 255) Then
      Result := fgColor
   Else
   Begin
      R := PByteArray(@VarMulTable[mask]);
      InvR := PByteArray(@VarMulTable[Not mask]);
      res.A := R^[fg.A] + InvR^[bg.A];
      res.R := R^[fg.R] + InvR^[bg.R];
      res.G := R^[fg.G] + InvR^[bg.G];
      res.B := R^[fg.B] + InvR^[bg.B];
   End;
End;

Function MakeColorGradient(Const gradColors: TArrayOfGradientColor; len: integer): TArrayOfColor32;
Var
   i, j, lenC: integer;
   dist, offset1, offset2, step, pos: double;
   color1, color2: TColor32;
Begin
   lenC := length(gradColors);
   If (len = 0) Or (lenC < 2) Then
      Exit;
   SetLength(Result, len);

   color2 := gradColors[0].color;
   Result[0] := color2;
   If len = 1 Then
      Exit;

   step := 1 / (len - 1);
   pos := step;
   offset2 := 0;
   i := 1;
   j := 1;
   Repeat
      offset1 := offset2;
      offset2 := gradColors[i].offset;
      dist := offset2 - offset1;
      color1 := color2;
      color2 := gradColors[i].color;
      While (pos <= dist) And (j < len) Do
      Begin
         Result[j] := BlendColorUsingMask(color1, color2, Round(pos / dist * 255));
         Inc(j);
         pos := pos + step;
      End;
      pos := pos - dist;
      Inc(i);
   Until i = lenC;
   If j < len Then
      Result[j] := Result[j - 1];
End;

Procedure AllocateScanlines(Const polygons: TPathsD; Var scanlines: TArrayOfScanline; clipBottom, clipRight: integer);
Var
   i, j, highI, highJ: integer;
   y1, y2: integer;
   psl: PScanline;
Begin

   For i := 0 To high(polygons) Do
   Begin
      highJ := high(polygons[i]);
      If highJ < 2 Then
         continue;
      y1 := Round(polygons[i][highJ].Y);
      For j := 0 To highJ Do
      Begin
         y2 := Round(polygons[i][j].Y);
         If y1 < y2 Then
         Begin

            If (y2 >= 0) And (y1 <= clipBottom) Then
            Begin
               If (y1 > 0) And (y1 <= clipBottom) Then
                  Dec(scanlines[y1 - 1].fragCnt);
               If y2 >= clipBottom Then
                  Inc(scanlines[clipBottom].fragCnt)
               Else
                  Inc(scanlines[y2].fragCnt);
            End;
         End
         Else
         Begin

            If (y1 >= 0) And (y2 <= clipBottom) Then
            Begin
               If (y2 > 0) Then
                  Dec(scanlines[y2 - 1].fragCnt);
               If y1 >= clipBottom Then
                  Inc(scanlines[clipBottom].fragCnt)
               Else
                  Inc(scanlines[y1].fragCnt);
            End;
         End;
         y1 := y2;
      End;
   End;

   j := 0;
   highI := high(scanlines);
   psl := @scanlines[highI];

   For i := highI Downto 0 Do
   Begin
      Inc(j, psl^.fragCnt);
      If j > 0 Then
         GetMem(psl^.fragments, j * SizeOf(TFragment));

      psl^.fragCnt := 0;
      psl^.minX := clipRight;
      psl^.maxX := 0;
      psl^.Y := i;
      Dec(psl);
   End;
End;

Procedure SplitEdgeIntoFragments(Const pt1, pt2: TPointD; Const scanlines: TArrayOfScanline; Const clipRec: TRect);
Var
   x, y, dx, dy, absDx, dydx, dxdy: double;
   i, scanlineY, maxY, maxX: integer;
   psl: PScanLine;
   pFrag: PFragment;
   bot, top: TPointD;
Begin
   dy := pt1.Y - pt2.Y;
   dx := pt2.X - pt1.X;
   maxY := RectHeight(clipRec);
   maxX := RectWidth(clipRec);
   absDx := abs(dx);

   If dy > 0 Then
   Begin

      If dy < 0.0001 Then
         Exit;
      bot := pt1;
      top := pt2;

      If (top.Y >= maxY) Or (bot.Y <= 0) Then
         Exit;
   End
   Else
   Begin

      If dy > -0.0001 Then
         Exit;
      bot := pt2;
      top := pt1;

      If (top.Y >= maxY) Or (bot.Y <= 0) Then
         Exit;
   End;

   If absDx < 0.000001 Then
   Begin

      top.X := bot.X;

      If bot.X > maxX Then
      Begin
         For i := Min(maxY, Round(bot.Y)) Downto Max(0, Round(top.Y)) Do
            scanlines[i].maxX := maxX;
         Exit;
      End;

      dxdy := 0;
      If dy > 0 Then
         dydx := 1
      Else
         dydx := -1;
   End
   Else
   Begin
      dxdy := dx / dy;
      dydx := dy / absDx;
   End;

   If bot.X >= maxX Then
   Begin
      If top.X >= maxX Then
      Begin
         For i := Min(maxY, Round(bot.Y)) Downto Max(0, Round(top.Y)) Do
            scanlines[i].maxX := maxX;
         Exit;
      End;

      y := bot.Y - (bot.X - maxX) * Abs(dydx);
      For i := Min(maxY, Round(bot.Y)) Downto Max(0, Round(y)) Do
         scanlines[i].maxX := maxX;
      bot.Y := y;
      If bot.Y <= 0 Then
         Exit;
      bot.X := maxX;
   End
   Else If top.X > maxX Then
   Begin

      y := top.Y + (top.X - maxX) * Abs(dydx);
      For i := Min(maxY, Round(y)) Downto Max(0, Round(top.Y)) Do
         scanlines[i].maxX := maxX;
      top.Y := y;
      If top.Y >= maxY Then
         Exit;
      top.X := maxX;
   End;
   If bot.Y > maxY Then
   Begin
      bot.X := bot.X + dxdy * (bot.Y - maxY);
      If (bot.X > maxX) Then
         Exit;
      bot.Y := maxY;
   End;
   If top.Y < 0 Then
   Begin
      top.X := top.X + (dxdy * top.Y);
      If (top.X > maxX) Then
         Exit;
      top.Y := 0;
   End;

   scanlineY := Round(bot.Y);
   If bot.Y = scanlineY Then
      Dec(scanlineY);

   If scanlineY < 0 Then
      Exit;

   psl := @scanlines[scanlineY];
   If Not assigned(psl^.fragments) Then
      Exit;

   pFrag := @psl^.fragments^[psl^.fragCnt];
   Inc(psl^.fragCnt);

   pFrag^.botX := bot.X;
   If scanlineY <= top.Y Then
   Begin
      pFrag^.topX := top.X;
      pFrag^.dy   := bot.Y - top.Y;
      pFrag^.dydx := dydx;
      Exit;
   End;

   x := bot.X + (bot.Y - scanlineY) * dxdy;
   pFrag^.topX := x;
   pFrag^.dy   := bot.Y - scanlineY;
   pFrag^.dydx := dydx;

   Dec(psl);
   While psl^.Y > top.Y Do
   Begin
      pFrag := @psl^.fragments^[psl^.fragCnt];
      Inc(psl^.fragCnt);
      pFrag^.botX := x;
      x := x + dxdy;
      pFrag^.topX := x;
      pFrag^.dy := 1;
      pFrag^.dydx := dydx;
      Dec(psl);
   End;
   pFrag := @psl^.fragments^[psl^.fragCnt];
   Inc(psl^.fragCnt);
   pFrag^.botX := x;
   pFrag^.topX := top.X;
   pFrag^.dy   := psl^.Y + 1 - top.Y;
   pFrag^.dydx := dydx;
End;

Procedure InitializeScanlines(Var polygons: TPathsD; Const scanlines: TArrayOfScanline; Const clipRec: TRect);
Var
   i, j, highJ: integer;
   pt1, pt2: PPointD;
Begin
   For i := 0 To high(polygons) Do
   Begin
      highJ := high(polygons[i]);
      If highJ < 2 Then
         continue;
      pt1 := @polygons[i][highJ];
      pt2 := @polygons[i][0];
      For j := 0 To highJ Do
      Begin
         SplitEdgeIntoFragments(pt1^, pt2^, scanlines, clipRec);
         pt1 := pt2;
         Inc(pt2);
      End;
   End;
End;

Procedure ProcessScanlineFragments(Var scanline: TScanLine; Var buffer: TArrayOfDouble);
Var
   i, j, leftXi, rightXi: integer;
   fracX, yy, q, windDir: double;
   pd: PDouble;
   frag: PFragment;
Begin
   frag := @scanline.fragments[0];
   For i := 1 To scanline.fragCnt Do
   Begin
      If frag^.botX > frag^.topX Then
      Begin

         q := frag^.botX;
         frag^.botX := frag^.topX;
         frag^.topX := q;
      End;

      leftXi  := Max(0, Round(frag^.botX));
      rightXi := Max(0, Round(frag^.topX));

      If (leftXi = rightXi) Then
      Begin
         If frag^.dydx < 0 Then
            windDir := -1.0
         Else
            windDir := 1.0;

         If leftXi < scanline.minX Then
            scanline.minX := leftXi;
         If rightXi > scanline.maxX Then
            scanline.maxX := rightXi;
         pd := @buffer[leftXi];
         If (frag^.botX <= 0) Then
         Begin
            pd^ := pd^ + frag^.dy * windDir;
         End
         Else
         Begin
            q := (frag^.botX + frag^.topX) * 0.5 - leftXi;
            pd^ := pd^ + (1 - q) * frag^.dy * windDir;
            Inc(pd);
            pd^ := pd^ + q * frag^.dy * windDir;
         End;
      End
      Else
      Begin
         If leftXi < scanline.minX Then
            scanline.minX := leftXi;
         If rightXi > scanline.maxX Then
            scanline.maxX := rightXi;
         pd := @buffer[leftXi];

         fracX := leftXi + 1 - frag^.botX;
         yy := frag^.dydx * fracX;
         q := fracX * yy * 0.5;
         pd^ := pd^ + q;
         q := yy - q;
         Inc(pd);

         For j := leftXi + 1 To rightXi - 1 Do
         Begin
            pd^ := pd^ + q + frag^.dydx * 0.5;
            q := frag^.dydx * 0.5;
            Inc(pd);
         End;

         fracX := frag^.topX - rightXi;
         yy := fracX * frag^.dydx;
         pd^ := pd^ + q + (1 - fracX * 0.5) * yy;
         Inc(pd);

         pd^ := pd^ + fracX * 0.5 * yy;
      End;
      Inc(frag);
   End;
End;

Type
   TRoundingMode = TFPURoundingMode;

Procedure Rasterize(Const paths: TPathsD; Const clipRec: TRect; fillRule: TFillRule; renderer: TgaRendererCustom);
Var
   i, j, xli, xri, maxW, maxH, aa: integer;
   clipRec2: TRect;
   paths2: TPathsD;
   accum: double;
   windingAccum: TArrayOfDouble;
   byteBuffer: TArrayOfByte;
   scanlines: TArrayOfScanline;
   scanline: PScanline;
   savedRoundMode: TRoundingMode;
Begin

   If Not assigned(renderer) Then
      Exit;
   Types.IntersectRect(clipRec2, clipRec, gaGetBounds(paths));
   If IsEmptyRect(clipRec2) Then
      Exit;

   paths2 := OffsetPath(paths, -clipRec2.Left, -clipRec2.Top);

   savedRoundMode := SetRoundMode(rmDown);

   maxW := RectWidth(clipRec2);
   maxH := RectHeight(clipRec2);
   SetLength(scanlines, maxH + 1);
   SetLength(windingAccum, maxW + 2);
   AllocateScanlines(paths2, scanlines, maxH, maxW - 1);
   InitializeScanlines(paths2, scanlines, clipRec2);
   SetLength(byteBuffer, RectWidth(clipRec2));
   If byteBuffer = Nil Then
      Exit;

   scanline := @scanlines[0];
   For i := 0 To high(scanlines) Do
   Begin
      If scanline^.fragCnt = 0 Then
      Begin
         FreeMem(scanline^.fragments);
         Inc(scanline);
         Continue;
      End;

      ProcessScanlineFragments(scanline^, windingAccum);

      xli := scanline^.minX;
      xri := Min(maxW - 1, scanline^.maxX + 1);
      FillChar(byteBuffer[xli], xri - xli + 1, 0);

      accum := 0;
      For j := xli To xri Do
      Begin
         accum := accum + windingAccum[j];
         Case fillRule Of
            frEvenOdd:
               Begin
                  aa := Round(Abs(accum) * 1275) Mod 2550;
                  If aa > 1275 Then
                     byteBuffer[j] := Min(255, (2550 - aa) Shr 2)
                  Else
                     byteBuffer[j] := Min(255, aa Shr 2);
               End;
            frNonZero:
               Begin
                  byteBuffer[j] := Min(255, Round(Abs(accum) * 318));
               End;
            frPositive:
               Begin
                  If accum > 0.002 Then
                     byteBuffer[j] := Min(255, Round(accum * 318));
               End;
            frNegative:
               Begin
                  If accum < -0.002 Then
                     byteBuffer[j] := Min(255, Round(-accum * 318));
               End;
         End;
      End;
      renderer.RenderProc(clipRec2.Left + xli, clipRec2.Left + xri,
         clipRec2.Top + i, @byteBuffer[xli]);

      FillChar(windingAccum[xli], (xri - xli + 1) * sizeOf(double), 0);
      FreeMem(scanline^.fragments);
      Inc(scanline);
   End;
   SetRoundMode(savedRoundMode);
End;

Function TgaRendererCustom.Initialize(imgBase: Pointer; imgWidth, imgHeight, pixelSize: integer): boolean;
Begin
   fImgBase := imgBase;
   fImgWidth := ImgWidth;
   fImgHeight := ImgHeight;
   fPixelSize := pixelSize;

   fCurrLinePtr := fImgBase;
   fCurrY := 0;
   Result := True;
End;

Procedure TgaRendererCustom.NotifyChange;
Begin
   If assigned(fChangeProc) Then
      fChangeProc;
End;

Type
   THackedImage32 = Class(TgaGraphMem);

Function TgaRendererCustom.Initialize(targetImage: TgaGraphMem): boolean;
Begin
   fChangeProc := @THackedImage32(targetImage).Changed;
   With targetImage Do
      Result := Initialize(PixelBase, Width, Height, 4);
End;

Function TgaRendererCustom.GetDstPixel(x, y: integer): Pointer;
Begin
   If (y <> fCurrY) Then
   Begin
      fCurrY := y;
      fCurrLinePtr := fImgBase;
      Inc(pbyte(fCurrLinePtr), fCurrY * fImgWidth * fPixelSize);
   End;
   Result := fCurrLinePtr;
   Inc(pbyte(Result), x * fPixelSize);
End;

Constructor TgaRendererColor.Create(color: TColor32 = clNone32);
Begin
   If color <> clNone32 Then
      SetColor(color);
End;

Function TgaRendererColor.Initialize(targetImage: TgaGraphMem): boolean;
Begin
   Result := Inherited Initialize(targetImage) And (fAlpha > 0);
End;

Procedure TgaRendererColor.SetColor(Value: TColor32);
Begin
   fColor := Value And $FFFFFF;
   fAlpha := Value Shr 24;
End;

Procedure TgaRendererColor.RenderProc(x1, x2, y: integer; alpha: pbyte);
Var
   i: integer;
   dst: PColor32;
Begin
   dst := GetDstPixel(x1, y);
   For i := x1 To x2 Do
   Begin
      dst^ := BlendToAlpha(dst^, ((Ord(alpha^) * fAlpha) Shr 8) Shl 24 Or fColor);
      Inc(dst);
      Inc(alpha);
   End;
End;

Constructor TgaRendererImage.Create(tileFillStyle: TTileFillStyle; brushImage: TgaGraphMem);
Begin
   fImage := TgaGraphMem.Create(brushImage);
   SetTileFillStyle(tileFillStyle);
End;

Destructor TgaRendererImage.Destroy;
Begin
   fImage.Free;
   Inherited;
End;

Procedure TgaRendererImage.SetTileFillStyle(Value: TTileFillStyle);
Begin
   Case Value Of
      tfsRepeat: fBoundsProc     := @RepeatQ;
      tfsMirrorHorz: fBoundsProc := @MirrorQ;
      tfsMirrorVert: fBoundsProc := @RepeatQ;
      tfsRotate180: fBoundsProc  := @MirrorQ;
   End;
   fMirrorY := Value In [tfsMirrorVert, tfsRotate180];
End;

Function TgaRendererImage.Initialize(targetImage: TgaGraphMem): boolean;
Begin
   Result := Inherited Initialize(targetImage) And (Not fImage.IsEmpty);
   If Not Result Then
      Exit;
   fLastYY := 0;
   fBrushPixel := PARGB(fImage.PixelBase);
End;

Procedure TgaRendererImage.RenderProc(x1, x2, y: integer; alpha: pbyte);
Var
   i: integer;
   pDst: PColor32;
   pBrush: PARGB;
Begin
   pDst := GetDstPixel(x1, y);
   Dec(x1, fOffset.X);
   Dec(x2, fOffset.X);
   Dec(y, fOffset.Y);
   pBrush := GetFirstBrushPixel(x1, y);
   For i := x1 To x2 Do
   Begin
      pDst^ := BlendToAlpha(pDst^, MulBytes(pBrush^.A, Ord(alpha^)) Shl 24 Or (pBrush^.Color And $FFFFFF));
      Inc(pDst);
      Inc(alpha);
      pBrush := GetPixel(fBrushPixel, fBoundsProc(i, fImage.Width));
   End;
End;

Function TgaRendererImage.GetFirstBrushPixel(x, y: integer): PARGB;
Begin
   If fMirrorY Then
      y := MirrorQ(y, fImage.Height)
   Else
      y := RepeatQ(y, fImage.Height);
   If y <> fLastYY Then
   Begin
      fBrushPixel := PARGB(fImage.PixelRow[y]);
      fLastYY := y;
   End;
   x := fBoundsProc(x, fImage.Width);
   Result := GetPixel(fBrushPixel, x);
End;

Constructor TgaRendererGradientCustom.Create;
Begin
   fBoundsProc := @ClampQ;
End;

Procedure TgaRendererGradientCustom.Clear;
Begin
   fGradientColors := Nil;
   fColors := Nil;
End;

Procedure TgaRendererGradientCustom.SetGradientFillStyle(Value: TGradientFillStyle);
Begin
   Case Value Of
      gfsClamp: fBoundsProc  := @ClampQ;
      gfsMirror: fBoundsProc := @MirrorQ;
      Else
         fBoundsProc := @RepeatQ;
   End;
End;

Procedure TgaRendererGradientCustom.SetParameters(startColor, endColor: TColor32; gradFillStyle: TGradientFillStyle = gfsClamp);
Begin
   SetGradientFillStyle(gradFillStyle);

   SetLength(fGradientColors, 2);
   fGradientColors[0].offset := 0;
   fGradientColors[0].color := startColor;
   fGradientColors[1].offset := 1;
   fGradientColors[1].color := endColor;
End;

Procedure TgaRendererGradientCustom.InsertColorStop(offsetFrac: double; color: TColor32);
Var
   i, len: integer;
   gradColor: TGradientColor;
Begin
   len := Length(fGradientColors);

   If len = 0 Then
      Exit;

   If offsetFrac < 0 Then
      offsetFrac := 0
   Else If offsetFrac > 1 Then
      offsetFrac := 1;

   If offsetFrac = 0 Then
   Begin
      fGradientColors[0].color := color;
      Exit;
   End
   Else If offsetFrac = 1 Then
   Begin
      fGradientColors[len - 1].color := color;
      Exit;
   End;
   gradColor.offset := offsetFrac;
   gradColor.color := color;

   i := 1;
   While (i < len - 1) And (fGradientColors[i].offset <= offsetFrac) Do
      Inc(i);
   SetLength(fGradientColors, len + 1);
   Move(fGradientColors[i],
      fGradientColors[i + 1], (len - i) * SizeOf(TGradientColor));
   fGradientColors[i] := gradColor;
End;

Procedure TgaRendererGradientLinear.SetParameters(Const startPt, endPt: TPointD; startColor, endColor: TColor32; gradFillStyle: TGradientFillStyle);
Begin
   Inherited SetParameters(startColor, endColor, gradFillStyle);
   fStartPt := startPt;
   fEndPt := endPt;
End;

Function TgaRendererGradientLinear.Initialize(targetImage: TgaGraphMem): boolean;
Var
   i: integer;
   dx, dy, dxdy, dydx: double;
Begin
   Result := Inherited Initialize(targetImage) And assigned(fGradientColors);
   If Not Result Then
      Exit;

   If abs(fEndPt.Y - fStartPt.Y) > abs(fEndPt.X - fStartPt.X) Then
   Begin
      If (fEndPt.Y < fStartPt.Y) Then
      Begin
         fGradientColors := ReverseColors(fGradientColors);
         SwapPoints(fStartPt, fEndPt);
      End;
      fIsVert := True;
      dx := (fEndPt.X - fStartPt.X);
      dy := (fEndPt.Y - fStartPt.Y);
      dxdy := dx / dy;

      fColorsCnt := Ceil(dy + dxdy * (fEndPt.X - fStartPt.X));
      fColors := MakeColorGradient(fGradientColors, fColorsCnt);

      SetLength(fPerpendicOffsets, ImgWidth);

      For i := 0 To ImgWidth - 1 Do
         fPerpendicOffsets[i] := Round(dxdy * (fStartPt.X - i) + fStartPt.Y);
   End
   Else
   Begin
      If (fEndPt.X = fStartPt.X) Then
      Begin
         Result := False;
         Exit;
      End;
      If (fEndPt.X < fStartPt.X) Then
      Begin
         fGradientColors := ReverseColors(fGradientColors);
         SwapPoints(fStartPt, fEndPt);
      End;
      fIsVert := False;
      dx := (fEndPt.X - fStartPt.X);
      dy := (fEndPt.Y - fStartPt.Y);
      dydx := dy / dx;

      fColorsCnt := Ceil(dx + dydx * (fEndPt.Y - fStartPt.Y));
      fColors := MakeColorGradient(fGradientColors, fColorsCnt);
      SetLength(fPerpendicOffsets, ImgHeight);

      For i := 0 To ImgHeight - 1 Do
         fPerpendicOffsets[i] := Round(dydx * (fStartPt.Y - i) + fStartPt.X);
   End;
End;

Procedure TgaRendererGradientLinear.RenderProc(x1, x2, y: integer; alpha: pbyte);
Var
   i, off: integer;
   pDst: PColor32;
   color: TARGB;
Begin
   pDst := GetDstPixel(x1, y);
   For i := x1 To x2 Do
   Begin
      If fIsVert Then
      Begin

         off := fPerpendicOffsets[i];
         color.Color := fColors[fBoundsProc(y - off, fColorsCnt)];
      End
      Else
      Begin

         off := fPerpendicOffsets[y];
         color.Color := fColors[fBoundsProc(i - off, fColorsCnt)];
      End;
      pDst^ := BlendToAlpha(pDst^, MulBytes(color.A, Ord(alpha^)) Shl 24 Or (color.Color And $FFFFFF));
      Inc(pDst);
      Inc(alpha);
   End;
End;

Function TgaRendererGradientRadial.Initialize(targetImage: TgaGraphMem): boolean;
Begin
   Result := Inherited Initialize(targetImage) And (fColorsCnt > 1);
   If Result Then
      fColors := MakeColorGradient(fGradientColors, fColorsCnt);
End;

Procedure TgaRendererGradientRadial.SetParameters(Const focalRect: TRect; innerColor, outerColor: TColor32; gradientFillStyle: TGradientFillStyle);
Var
   radX, radY: double;
Begin
   Inherited SetParameters(innerColor, outerColor, gradientFillStyle);
   fColorsCnt := 0;
   If IsEmptyRect(focalRect) Then
      Exit;

   fCenterPt.X := (focalRect.Left + focalRect.Right) * 0.5;
   fCenterPt.Y := (focalRect.Top + focalRect.Bottom) * 0.5;
   radX := RectWidth(focalRect) * 0.5;
   radY := RectHeight(focalRect) * 0.5;
   If radX >= radY Then
   Begin
      fScaleX := 1;
      fScaleY := radX / radY;
      fColorsCnt := Ceil(radX) + 1;
   End
   Else
   Begin
      fScaleX := radY / radX;
      fScaleY := 1;
      fColorsCnt := Ceil(radY) + 1;
   End;
End;

Procedure TgaRendererGradientRadial.RenderProc(x1, x2, y: integer; alpha: pbyte);
Var
   i: integer;
   dist: double;
   color: TARGB;
   pDst: PColor32;
Begin
   pDst := GetDstPixel(x1, y);
   For i := x1 To x2 Do
   Begin
      dist := Hypot((y - fCenterPt.Y) * fScaleY, (i - fCenterPt.X) * fScaleX);
      color.Color := fColors[fBoundsProc(Round(dist), fColorsCnt)];
      pDst^ := BlendToAlpha(pDst^, MulBytes(color.A, Ord(alpha^)) Shl 24 Or (color.Color And $FFFFFF));
      Inc(pDst);
      Inc(alpha);
   End;
End;

Function TgaRendererGradientSvgRadial.Initialize(targetImage: TgaGraphMem): boolean;
Begin
   Result := Inherited Initialize(targetImage) And (fColorsCnt > 1);
   If Result Then
      fColors := MakeColorGradient(fGradientColors, fColorsCnt);
End;

Procedure TgaRendererGradientSvgRadial.SetParameters(Const ellipseRect: TRect; Const focus: TPoint;
   innerColor, outerColor: TColor32; gradientFillStyle: TGradientFillStyle = gfsClamp);
Var
   radii: TPointD;
Begin
   Inherited SetParameters(innerColor, outerColor);
   Case gradientFillStyle Of
      gfsMirror: fBoundsProcD := @MirrorD;
      gfsRepeat: fBoundsProcD := @RepeatD;
      Else
         fBoundsProcD := @ClampD;
   End;

   fColorsCnt := 0;
   If IsEmptyRect(ellipseRect) Then
      Exit;

   fCenterPt := RectD(ellipseRect).MidPoint;
   radii.X := RectWidth(ellipseRect) * 0.5;
   radii.Y := RectHeight(ellipseRect) * 0.5;

   fA := RectWidth(ellipseRect) * 0.5;
   fB := RectHeight(ellipseRect) * 0.5;
   fFocusPt.X := focus.X - fCenterPt.X;
   fFocusPt.Y := focus.Y - fCenterPt.Y;
   fColorsCnt := Ceil(Hypot(fA * 2, fB * 2)) + 1;
   fAA := fA * fA;
   fBB := fB * fB;
End;

Procedure TgaRendererGradientSvgRadial.RenderProc(x1, x2, y: integer; alpha: pbyte);
Var
   i: integer;
   q, m, c, qa, qb, qc, qs: double;
   dist, dist2: double;
   color: TARGB;
   pDst: PColor32;
   pt, ellipsePt: TPointD;
Begin

   pDst := GetDstPixel(x1, y);
   pt.X := x1 - fCenterPt.X;
   pt.Y := y - fCenterPt.Y;
   For i := x1 To x2 Do
   Begin

      If (pt.X = fFocusPt.X) Then
      Begin

         q := Sqrt(fBB * (1 - Sqr(pt.X) / fAA));
         ellipsePt.X := pt.X;
         If pt.Y >= fFocusPt.Y Then
            ellipsePt.Y := q
         Else
            ellipsePt.Y := -q;
         dist := abs(pt.Y - fFocusPt.Y);
         dist2 := abs(ellipsePt.Y - fFocusPt.Y);
         If dist2 = 0 Then
            q := 1
         Else
            q := dist / dist2;
      End
      Else
      Begin

         m := (pt.Y - fFocusPt.Y) / (pt.X - fFocusPt.X);
         c := pt.Y - m * pt.X;

         qa := (fBB / fAA + (m * m));
         qb := 2 * m * c;
         qc := (c * c) - fBB;
         qs := (qb * qb) - 4 * qa * qc;
         If qs >= 0 Then
         Begin
            qs := Sqrt(qs);
            If pt.X <= fFocusPt.X Then
               ellipsePt.X := (-qb - qs) / (2 * qa)
            Else
               ellipsePt.X := (-qb + qs) / (2 * qa);
            ellipsePt.Y := m * ellipsePt.X + c;
            dist := Hypot(pt.X - fFocusPt.X, pt.Y - fFocusPt.Y);
            dist2 := Hypot(ellipsePt.X - fFocusPt.X, ellipsePt.Y - fFocusPt.Y);
            If dist2 = 0 Then
               q := 1
            Else
               q := dist / dist2;
         End
         Else
            q := 1;
      End;
      color.Color := fColors[fBoundsProcD(Abs(q), fColorsCnt)];
      pDst^ := BlendToAlpha(pDst^, MulBytes(color.A, Ord(alpha^)) Shl 24 Or (color.Color And $FFFFFF));
      Inc(pDst);
      pt.X := pt.X + 1;
      Inc(alpha);
   End;
End;

Procedure TgaRendererErase.RenderProc(x1, x2, y: integer; alpha: pbyte);
Var
   i: integer;
   dst: PARGB;
Begin
   dst := PARGB(GetDstPixel(x1, y));
   For i := x1 To x2 Do
   Begin
      dst^.A := MulBytes(dst^.A, Not alpha^);
      Inc(dst);
      Inc(alpha);
   End;
End;

Procedure TgaRendererInverse.RenderProc(x1, x2, y: integer; alpha: pbyte);
Var
   i: integer;
   dst: PARGB;
   c: TARGB;
Begin
   dst := PARGB(GetDstPixel(x1, y));
   For i := x1 To x2 Do
   Begin
      c.Color := Not dst^.Color;
      c.A := MulBytes(dst^.A, Ord(alpha^));
      dst^.Color := BlendToAlpha(dst^.Color, c.Color);
      Inc(dst);
      Inc(alpha);
   End;
End;

Procedure TgaRendererBarycentric.SetParameters(Const aa, bb, cc: TPointD; cc1, cc2, cc3: TColor32);
Begin

   self.a := aa;
   self.c1.Color := cc1;
   self.c2.Color := cc2;
   self.c3.Color := cc3;

   v0.X := bb.X - aa.X;
   v0.Y := bb.Y - aa.Y;
   v1.X := cc.X - aa.X;
   v1.Y := cc.Y - aa.Y;
   d00 := (v0.X * v0.X + v0.Y * v0.Y);
   d01 := (v0.X * v1.X + v0.Y * v1.Y);
   d11 := (v1.X * v1.X + v1.Y * v1.Y);
   invDenom := 1 / (d00 * d11 - d01 * d01);
End;

Function TgaRendererBarycentric.GetColor(Const pt: TPointD): TColor32;
Var
   v2: TPointD;
   d20, d21, v, w, u: double;
   res: TARGB Absolute Result;
Begin
   Result := 0;
   v2.X := pt.X - a.X;
   v2.Y := pt.Y - a.Y;
   d20 := (v2.X * v0.X + v2.Y * v0.Y);
   d21 := (v2.X * v1.X + v2.Y * v1.Y);

   v := (d11 * d20 - d01 * d21) * invDenom;
   w := (d00 * d21 - d01 * d20) * invDenom;
   u := 1.0 - v - w;

   Res.A := ClampByte(c1.A * u + c2.A * v + c3.A * w);
   Res.R := ClampByte(c1.R * u + c2.R * v + c3.R * w);
   Res.G := ClampByte(c1.G * u + c2.G * v + c3.G * w);
   Res.B := ClampByte(c1.B * u + c2.B * v + c3.B * w);
End;

Procedure TgaRendererBarycentric.RenderProc(x1, x2, y: integer; alpha: pbyte);
Var
   x: integer;
   p: PARGB;
   c: TARGB;
Begin
   p := PARGB(fImgBase);
   Inc(p, y * ImgWidth + x1);
   For x := x1 To x2 Do
   Begin
      c.Color := GetColor(PointD(x, y));
      c.A := c.A * Ord(alpha^) Shr 8;
      p^.Color := BlendToAlpha(p^.Color, c.Color);
      Inc(p);
      Inc(alpha);
   End;
End;

Const
   FloodFillDefaultRGBTolerance: byte = 20;
   FloodFillDefaultHueTolerance: byte = 1;
   MaxBlur = 100;

Type
   PColor32Array = ^TColor32Array;
   TColor32Array = Array[0..maxint Div SizeOf(TColor32) - 1] Of TColor32;

   PgaWeightedColorArray = ^TgaWeightedColorArray;
   TgaWeightedColorArray = Array[0..$FFFFFF] Of TgaWeightedColor;

Function ColorDifference(color1, color2: TColor32): cardinal;

Var
   c1: TARGB Absolute color1;
   c2: TARGB Absolute color2;
Begin
   Result := Abs(c1.R - c2.R) + Abs(c1.G - c2.G) + Abs(c1.B - c2.B);
End;

Function RainbowColor(fraction: double): TColor32;
Begin
   If (fraction >= 1) Or (fraction <= 0) Then
      Result := clRed32
   Else
   Begin
      fraction := fraction * 6;
      Case trunc(fraction) Of
         0: Result := GradientColor(clRed32, clYellow32, frac(fraction));
         1: Result := GradientColor(clYellow32, clLime32, frac(fraction));
         2: Result := GradientColor(clLime32, clAqua32, frac(fraction));
         3: Result := GradientColor(clAqua32, clBlue32, frac(fraction));
         4: Result := GradientColor(clBlue32, clFuchsia32, frac(fraction));
         Else
            Result := GradientColor(clFuchsia32, clRed32, frac(fraction));
      End;
   End;
End;

Function GradientColor(color1, color2: TColor32; frac: single): TColor32;
Var
   c1: TARGB Absolute color1;
   c2: TARGB Absolute color2;
   r: TARGB Absolute Result;
Begin
   If frac >= 1 Then
      Result := color2
   Else If frac <= 0 Then
      Result := color1
   Else
   Begin
      r.B := trunc(c1.B * (1 - frac) + c2.B * frac);
      r.G := trunc(c1.G * (1 - frac) + c2.G * frac);
      r.R := trunc(c1.R * (1 - frac) + c2.R * frac);
      r.A := trunc(c1.A * (1 - frac) + c2.A * frac);
   End;
End;

Function MakeDarker(color: TColor32; percent: cardinal): TColor32;
Begin
   Result := GradientColor(color, $FF000000, percent / 100);
End;

Function MakeLighter(color: TColor32; percent: cardinal): TColor32;
Begin
   Result := GradientColor(color, $FFFFFFFF, percent / 100);
End;

Function AlphaAverage(color1, color2: TColor32): cardinal;

Var
   c1: TARGB Absolute color1;
   c2: TARGB Absolute color2;
Begin
   Result := (c1.A + c2.A) Shr 1;
End;

Function BlendAverage(bgColor, fgColor: TColor32): TColor32;
Var
   res: TARGB Absolute Result;
   bg: TARGB Absolute bgColor;
   fg: TARGB Absolute fgColor;
Begin
   res.A := (fg.A + bg.A) Shr 1;
   res.R := (fg.R + bg.R) Shr 1;
   res.G := (fg.G + bg.G) Shr 1;
   res.B := (fg.B + bg.B) Shr 1;
End;

Function BlendLinearBurn(bgColor, fgColor: TColor32): TColor32;
Var
   res: TARGB Absolute Result;
   bg: TARGB Absolute bgColor;
   fg: TARGB Absolute fgColor;
Begin
   res.A := 255;
   res.R := Max(0, bg.R + fg.R - 255);
   res.G := Max(0, bg.G + fg.G - 255);
   res.B := Max(0, bg.B + fg.B - 255);
End;

Function BlendColorDodge(bgColor, fgColor: TColor32): TColor32;
Var
   res: TARGB Absolute Result;
   bg: TARGB Absolute bgColor;
   fg: TARGB Absolute fgColor;
Begin
   res.A := 255;
   res.R := VarDivTable[bg.R, Not fg.R];
   res.G := VarDivTable[bg.G, Not fg.G];
   res.B := VarDivTable[bg.B, Not fg.B];
End;

Type
   PFloodFillRec = ^TFloodFillRec;

   TFloodFillRec = Record
      xLeft: integer;
      xRight: integer;
      y: integer;
      dirY: integer;
      Next: PFloodFillRec;
   End;

   TFloodFillStack = Class
      First: PFloodFillRec;
      maxY: integer;
      Constructor Create(aMaxY: integer);
      Destructor Destroy; Override;
      Procedure Push(xLeft, xRight, y, direction: integer);
      Procedure Pop(Out xLeft, xRight, y, direction: integer);
      Function IsEmpty: boolean;
   End;

   TFloodFillMask = Object
      mask: TArrayOfByte;
      Width: integer;
      Height: integer;
      tolerance: integer;
      initialColor: TColor32;
      colorsBase: PColor32Array;
      colorsRow: PColor32Array;
      maskRow: PByteArray;
      compareFunc: TCompareFunction;
      Procedure Reset(w, h, x, y: integer; pixelBase: PColor32; compFunc: TCompareFunction; aTolerance: integer = 0);
      Procedure SetCurrentY(y: integer);
      Function IsMatch(x: integer): boolean;
   End;

Constructor TFloodFillStack.Create(AMaxY: integer);
Begin
   self.maxY := aMaxY;
End;

Destructor TFloodFillStack.Destroy;
Var
   ffr: PFloodFillRec;
Begin
   While assigned(First) Do
   Begin
      ffr := First;
      First := First^.Next;
      dispose(ffr);
   End;
End;

Procedure TFloodFillStack.Push(xLeft, xRight, y, direction: integer);
Var
   ffr: PFloodFillRec;
Begin
   If ((y = 0) And (direction = -1)) Or ((y = maxY) And (direction = 1)) Then
      Exit;
   new(ffr);
   ffr^.xLeft := xLeft;
   ffr^.xRight := xRight;
   ffr^.y := y;
   ffr^.dirY := direction;
   ffr^.Next := First;
   First := ffr;
End;

Procedure TFloodFillStack.Pop(Out xLeft, xRight, y, direction: integer);
Var
   ffr: PFloodFillRec;
Begin
   xLeft := First^.xLeft;
   xRight := First^.xRight;
   direction := First^.dirY;
   y := First^.y + direction;
   ffr := First;
   First := First^.Next;
   dispose(ffr);
End;

Function TFloodFillStack.IsEmpty: boolean;
Begin
   Result := Not assigned(First);
End;

Procedure TFloodFillMask.Reset(w, h, x, y: integer; pixelBase: PColor32; compFunc: TCompareFunction; aTolerance: integer);
Begin
   mask := Nil;

   setLength(mask, w * h);
   Self.Width := w;
   Self.Height := h;
   colorsBase := PColor32Array(pixelBase);
   Self.initialColor := colorsBase^[x + y * w];
   Self.compareFunc := compFunc;
   Self.tolerance := aTolerance;

End;

Procedure TFloodFillMask.SetCurrentY(y: integer);
Begin
   colorsRow := @colorsBase[y * Width];
   maskRow := @mask[y * Width];
End;

Function TFloodFillMask.IsMatch(x: integer): boolean;
Begin
   Result := (maskRow^[x] = 0) And compareFunc(initialColor, colorsRow^[x], tolerance);
   If Result Then
      maskRow^[x] := 1;
End;

Function IncPWeightColor(pwc: PgaWeightedColor; cnt: integer): PgaWeightedColor;
Begin
   Result := PgaWeightedColor(pbyte(pwc) + cnt * SizeOf(TgaWeightedColor));
End;

Function Intensity(color: TColor32): byte;
Var
   c: TARGB Absolute color;
Begin
   Result := (c.R * 61 + c.G * 174 + c.B * 21) Shr 8;
End;

Function Gray(color: TColor32): TColor32;
Var
   c: TARGB Absolute color;
   res: TARGB Absolute Result;
Begin
   res.A := c.A;
   res.R := Intensity(color);
   res.G := res.R;
   res.B := res.R;
End;

Type
   TPt2Container = Class;

   TPt2 = Class
      pt: TPointD;
      owner: TPt2Container;
      isStart: boolean;
      isHole: boolean;
      nextInPath: TPt2;
      prevInPath: TPt2;
      nextInRow: TPt2;
      prevInRow: TPt2;
      Destructor Destroy; Override;
      Procedure Update(x, y: double);
      Function GetCount: integer;
      Function GetPoints: TPathD;
      Property IsAscending: boolean Read isStart;
   End;

   TPt2Container = Class
      prevRight: integer;
      leftMostPt, rightMost: TPt2;
      solution: TPathsD;
      Procedure AddToSolution(Const path: TPathD);
      Function StartNewPath(insertBefore: TPt2; xLeft, xRight, y: integer; isHole: boolean): TPt2;
      Procedure AddRange(Var current: TPt2; xLeft, xRight, y: integer);
      Function JoinAscDesc(path1, path2: TPt2): TPt2;
      Function JoinDescAsc(path1, path2: TPt2): TPt2;
      Procedure CheckRowEnds(pt2Left, pt2Right: TPt2);
   End;

Destructor TPt2.Destroy;
Var
   startPt, endPt, xpt: TPt2;
Begin
   If Not isStart Then
      Exit;
   startPt := self;
   endPt := startPt.prevInPath;

   If endPt = owner.rightMost Then
      owner.rightMost := endPt.prevInRow
   Else If assigned(endPt.nextInRow) Then
      endPt.nextInRow.prevInRow := endPt.prevInRow;
   If endPt = owner.leftMostPt Then
      owner.leftMostPt := endPt.nextInRow
   Else If assigned(endPt.prevInRow) Then
      endPt.prevInRow.nextInRow := endPt.nextInRow;

   If startPt = owner.leftMostPt Then
      owner.leftMostPt := startPt.nextInRow
   Else If assigned(startPt.prevInRow) Then
      startPt.prevInRow.nextInRow := startPt.nextInRow;
   If assigned(startPt.nextInRow) Then
      startPt.nextInRow.prevInRow := startPt.prevInRow;

   owner.AddToSolution(GetPoints);

   xpt := startPt.nextInPath;
   While xpt <> startPt Do
   Begin
      endPt := xpt;
      xpt := xpt.nextInPath;
      endPt.Free;
   End;
End;

Function IsColinear(Const pt1, pt2, pt3: TPoint): boolean; Overload;
Begin
   Result := (pt1.X - pt2.X) * (pt2.Y - pt3.Y) = (pt2.X - pt3.X) * (pt1.Y - pt2.Y);
End;

Function IsColinear(Const pt1, pt2, pt3, pt4: TPoint): boolean; Overload;
Begin
   Result := (pt1.X - pt2.X) * (pt3.Y - pt4.Y) = (pt3.X - pt4.X) * (pt1.Y - pt2.Y);
End;

Function CreatePt2After(pt: TPt2; Const p: TPointD): TPt2;
Begin
   Result := TPt2.Create;
   Result.pt := p;
   Result.nextInPath := pt.nextInPath;
   Result.prevInPath := pt;
   pt.nextInPath.prevInPath := Result;
   pt.nextInPath := Result;
End;

Procedure TPt2.Update(x, y: double);
Var
   newPt2: TPt2;
Begin

   If isStart Then
   Begin
      If (x = pt.X) And (pt.X = nextInPath.pt.X) Then
      Begin
         pt := PointD(x, y);
         Exit;
      End;

      CreatePt2After(self, pt);
      If (x <> pt.X) Or (x <> nextInPath.pt.X) Then
      Begin
         If IsAscending Then
            CreatePt2After(self, PointD(pt.X, y))
         Else
            CreatePt2After(self, PointD(x, pt.Y));
      End;
      pt := PointD(x, y);

   End
   Else
   Begin
      If (x = pt.X) And (pt.X = prevInPath.pt.X) Then
      Begin
         pt := PointD(x, y);
         Exit;
      End;

      newPt2 := CreatePt2After(prevInPath, pt);
      If (x <> pt.X) Or (x <> prevInPath.pt.X) Then
      Begin
         If IsAscending Then
            CreatePt2After(newPt2, PointD(x, pt.Y))
         Else
            CreatePt2After(newPt2, PointD(pt.X, y));
      End;
      pt := PointD(x, y);
   End;
End;

Function TPt2.GetCount: integer;
Var
   pt2: TPt2;
Begin
   Result := 1;
   pt2 := nextInPath;
   While pt2 <> self Do
   Begin
      Inc(Result);
      pt2 := pt2.nextInPath;
   End;
End;

Function TPt2.GetPoints: TPathD;
Var
   i, Count: integer;
   pt2: TPt2;
Begin
   Update(pt.X, pt.Y + 1);
   With prevInPath Do
      Update(pt.X, pt.Y + 1);

   Count := GetCount;
   SetLength(Result, Count);
   pt2 := self;
   For i := 0 To Count - 1 Do
   Begin
      Result[i] := pt2.pt;
      pt2 := pt2.nextInPath;
   End;
End;

Procedure TPt2Container.AddToSolution(Const path: TPathD);
Var
   len: integer;
Begin
   If Length(path) < 2 Then
      Exit;
   len := Length(solution);
   SetLength(solution, len + 1);
   solution[len] := path;
End;

Function TPt2Container.StartNewPath(insertBefore: TPt2; xLeft, xRight, y: integer; isHole: boolean): TPt2;
Var
   pt2Left, pt2Right: TPt2;
Begin
   Inc(xRight);
   pt2Left := TPt2.Create;
   pt2Left.owner := self;
   pt2Left.isStart := Not isHole;
   pt2Left.isHole := isHole;
   pt2Left.pt := PointD(xLeft, y);

   pt2Right := TPt2.Create;
   pt2Right.owner := self;
   pt2Right.isStart := isHole;
   pt2Right.isHole := isHole;
   pt2Right.pt := PointD(xRight, y);

   pt2Left.nextInPath := pt2Right;
   pt2Left.prevInPath := pt2Right;
   pt2Right.nextInPath := pt2Left;
   pt2Right.prevInPath := pt2Left;

   pt2Left.nextInRow := pt2Right;
   pt2Right.prevInRow := pt2Left;
   If Not Assigned(insertBefore) Then
   Begin
      pt2Left.prevInRow := rightMost;
      If Assigned(rightMost) Then
         rightMost.nextInRow := pt2Left;
      pt2Right.nextInRow := Nil;
      rightMost := pt2Right;
      If Not Assigned(leftMostPt) Then
         leftMostPt := pt2Left;
   End
   Else
   Begin
      pt2Right.nextInRow := insertBefore;
      If leftMostPt = insertBefore Then
      Begin
         leftMostPt := pt2Left;
         pt2Left.prevInRow := Nil;
      End
      Else
      Begin
         pt2Left.prevInRow := insertBefore.prevInRow;
         insertBefore.prevInRow.nextInRow := pt2Left;
      End;
      insertBefore.prevInRow := pt2Right;
   End;
   Result := pt2Right.nextInRow;
End;

Procedure TPt2Container.CheckRowEnds(pt2Left, pt2Right: TPt2);
Begin
   If pt2Left = leftMostPt Then
      leftMostPt := pt2Right.nextInRow;
   If pt2Right = rightMost Then
      rightMost := pt2Left.prevInRow;
End;

Function TPt2Container.JoinAscDesc(path1, path2: TPt2): TPt2;
Begin
   Result := path2.nextInRow;
   CheckRowEnds(path1, path2);
   If path2 = path1.prevInPath Then
   Begin
      path1.Free;
      Exit;
   End;

   With path1 Do
      Update(pt.X, pt.Y + 1);
   With path2 Do
      Update(pt.X, pt.Y + 1);
   path1.isStart := False;
   If assigned(path1.nextInRow) Then
      path1.nextInRow.prevInRow := path1.prevInRow;
   If assigned(path1.prevInRow) Then
      path1.prevInRow.nextInRow := path1.nextInRow;

   If assigned(path2.nextInRow) Then
      path2.nextInRow.prevInRow := path2.prevInRow;
   If assigned(path2.prevInRow) Then
      path2.prevInRow.nextInRow := path2.nextInRow;

   path1.prevInPath.nextInPath := path2.nextInPath;
   path2.nextInPath.prevInPath := path1.prevInPath;
   path2.nextInPath := path1;
   path1.prevInPath := path2;
End;

Function TPt2Container.JoinDescAsc(path1, path2: TPt2): TPt2;
Begin
   Result := path2.nextInRow;
   CheckRowEnds(path1, path2);
   If path1 = path2.prevInPath Then
   Begin
      path2.Free;
      Exit;
   End;

   With path1 Do
      Update(pt.X, pt.Y + 1);
   With path2 Do
      Update(pt.X, pt.Y + 1);
   path2.isStart := False;
   If assigned(path1.nextInRow) Then
      path1.nextInRow.prevInRow := path1.prevInRow;
   If assigned(path1.prevInRow) Then
      path1.prevInRow.nextInRow := path1.nextInRow;
   If assigned(path2.nextInRow) Then
      path2.nextInRow.prevInRow := path2.prevInRow;
   If assigned(path2.prevInRow) Then
      path2.prevInRow.nextInRow := path2.nextInRow;

   path1.nextInPath.prevInPath := path2.prevInPath;
   path2.prevInPath.nextInPath := path1.nextInPath;
   path1.nextInPath := path2;
   path2.prevInPath := path1;
End;

Function IsHeadingLeft(current: TPt2; r: integer): boolean;

Begin
   Result := r <= current.pt.X;
End;

Procedure TPt2Container.AddRange(Var current: TPt2; xLeft, xRight, y: integer);
Begin
   If (prevRight > 0) Then
   Begin
      If xLeft < current.pt.X Then
      Begin
         current := StartNewPath(current, prevRight, xLeft - 1, y, True);
         prevRight := xRight;
         Exit;
      End;

      While assigned(current) And assigned(current.nextInRow) And (prevRight > current.nextInRow.pt.X) Do
      Begin
         current := JoinDescAsc(current, current.nextInRow);
      End;

      If (xLeft < current.pt.X) Then
      Begin
         current := StartNewPath(current, prevRight, xLeft - 1, y, True);
         prevRight := xRight;
         Exit;
      End;

      current.Update(prevRight, y);
      current := current.nextInRow;
      prevRight := 0;
   End;

   While assigned(current) And assigned(current.nextInRow) And (xLeft > current.nextInRow.pt.X) Do
      current := JoinAscDesc(current, current.nextInRow);

   If Not assigned(current) Or (xRight < current.pt.X) Then
   Begin
      StartNewPath(current, xLeft, xRight - 1, y, False);
   End
   Else
   Begin
      If IsHeadingLeft(current, xRight) Then
      Begin
         If current.isHole Then
         Begin
            current.Update(xLeft, y);
            current := current.nextInRow;
         End;

         current.Update(xRight, y);
         current.Update(xLeft, y);
         If current.IsAscending Then
            prevRight := xRight
         Else
            prevRight := 0;
         current := current.nextInRow;
      End
      Else
      Begin
         current.Update(xLeft, y);
         current := current.nextInRow;
         prevRight := xRight;
      End;
   End;
End;

Function VectorizeMask(Const mask: TArrayOfByte; maskWidth: integer): TPathsD;
Var
   i, j, len, Height, blockStart: integer;
   current: TPt2;
   ba: PByteArray;
   pt2Container: TPt2Container;
Begin
   Result := Nil;
   len := Length(mask);
   If (len = 0) Or (maskWidth = 0) Or (len Mod maskWidth <> 0) Then
      Exit;
   Height := len Div maskWidth;

   pt2Container := TPt2Container.Create;
   Try
      For i := 0 To Height - 1 Do
      Begin
         ba := @mask[maskWidth * i];
         blockStart := -2;
         current := pt2Container.leftMostPt;
         For j := 0 To maskWidth - 1 Do
         Begin
            If (ba^[j] > 0) = (blockStart >= 0) Then
               Continue;
            If blockStart >= 0 Then
            Begin
               pt2Container.AddRange(current, blockStart, j, i);
               blockStart := -1;
            End
            Else
               blockStart := j;
         End;

         If blockStart >= 0 Then
            pt2Container.AddRange(current, blockStart, maskWidth, i);

         If (pt2Container.prevRight > 0) Then
         Begin
            While Assigned(current.nextInRow) And (pt2Container.prevRight >= current.nextInRow.pt.X) Do
            Begin
               If current.isStart Then
                  current := pt2Container.JoinAscDesc(current, current.nextInRow)
               Else
                  current := pt2Container.JoinDescAsc(current, current.nextInRow);
            End;
            current.Update(pt2Container.prevRight, i);
            current := current.nextInRow;
            pt2Container.prevRight := 0;
         End;

         While assigned(current) Do
         Begin
            If current.isStart Then
               current := pt2Container.JoinAscDesc(current, current.nextInRow)
            Else
               current := pt2Container.JoinDescAsc(current, current.nextInRow);
         End;
      End;

      With pt2Container Do
         While Assigned(leftMostPt) Do
            If leftMostPt.isStart Then
               JoinAscDesc(leftMostPt, leftMostPt.nextInRow)
            Else
               JoinDescAsc(leftMostPt, leftMostPt.nextInRow);

      Result := pt2Container.solution;
   Finally
      pt2Container.Free;
   End;
End;

Function Tidy(Const poly: TPathD; tolerance: integer): TPathD;
Var
   i, j, highI: integer;
   prev: TPointD;
   tolSqrd: double;
Begin
   highI := High(poly);
   While (HighI >= 0) And PointsEqual(poly[highI], poly[0]) Do
      Dec(highI);
   If highI < 1 Then
   Begin
      Result := Nil;
      Exit;
   End;
   tolSqrd := Sqr(Max(2.02, Min(16.1, tolerance + 0.01)));
   SetLength(Result, highI + 1);
   prev := poly[highI];
   Result[0] := prev;
   Result[1] := poly[0];
   j := 1;
   For i := 1 To highI - 1 Do
   Begin
      If ((DistanceSqrd(prev, Result[j]) > tolSqrd) And (DistanceSqrd(Result[j], poly[i]) > tolSqrd)) Or
         (TurnsRight(prev, Result[j], poly[i]) Or TurnsLeft(Result[j], poly[i], poly[i + 1])) Then
      Begin
         prev := Result[j];
         Inc(j);
      End;
      Result[j] := poly[i];
   End;
   If ((DistanceSqrd(prev, Result[j]) > tolSqrd) And (DistanceSqrd(Result[j], Result[0]) > tolSqrd)) Or
      TurnsRight(prev, Result[j], Result[0]) Or TurnsLeft(Result[j], Result[0], Result[1]) Then
      SetLength(Result, j + 1)
   Else
      SetLength(Result, j);
End;

Procedure RDP(Const path: TPathD; startIdx, endIdx: integer; epsilonSqrd: double; Const flags: TArrayOfInteger);
Var
   i, idx: integer;
   d, maxD: double;
Begin
   idx := 0;
   maxD := 0;
   For i := startIdx + 1 To endIdx - 1 Do
   Begin
      d := PerpendicularDistSqrd(path[i], path[startIdx], path[endIdx]);
      If d <= maxD Then
         Continue;
      maxD := d;
      idx := i;
   End;
   If maxD < epsilonSqrd Then
      Exit;
   flags[idx] := 1;
   If idx > startIdx + 1 Then
      RDP(path, startIdx, idx, epsilonSqrd, flags);
   If endIdx > idx + 1 Then
      RDP(path, idx, endIdx, epsilonSqrd, flags);
End;

Function RamerDouglasPeucker(Const path: TPathD; epsilon: double): TPathD;
Var
   i, j, len: integer;
   buffer: TArrayOfInteger;
Begin
   len := length(path);
   If len < 5 Then
   Begin
      Result := Copy(path, 0, len);
      Exit;
   End;
   SetLength(buffer, len);

   buffer[0] := 1;
   buffer[len - 1] := 1;
   RDP(path, 0, len - 1, Sqr(epsilon), buffer);
   j := 0;
   SetLength(Result, len);
   For i := 0 To len - 1 Do
      If buffer[i] = 1 Then
      Begin
         Result[j] := path[i];
         Inc(j);
      End;
   SetLength(Result, j);
End;

Function RamerDouglasPeucker(Const paths: TPathsD; epsilon: double): TPathsD;
Var
   i, j, len: integer;
Begin
   j := 0;
   len := length(paths);
   setLength(Result, len);
   For i := 0 To len - 1 Do
   Begin
      Result[i] := RamerDouglasPeucker(paths[i], epsilon);
      If Result[i] <> Nil Then
         Inc(j);
   End;
   setLength(Result, j);
End;

Function Scale(Const vec: TPointD; newLen: double): TPointD;
Begin
   Result.X := vec.X * newLen;
   Result.Y := vec.Y * newLen;
End;

Function Mul(Const vec: TPointD; val: double): TPointD;
Begin
   Result.X := vec.X * val;
   Result.Y := vec.Y * val;
End;

Function AddVecs(Const vec1, vec2: TPointD): TPointD;
Begin
   Result.X := vec1.X + vec2.X;
   Result.Y := vec1.Y + vec2.Y;
End;

Function SubVecs(Const vec1, vec2: TPointD): TPointD;
Begin
   Result.X := vec1.X - vec2.X;
   Result.Y := vec1.Y - vec2.Y;
End;

Function DotProdVecs(Const vec1, vec2: TPointD): double;
Begin
   Result := (vec1.X * vec2.X + vec1.Y * vec2.Y);
End;

Function NormalizeVec(Const vec: TPointD): TPointD;
Var
   len: double;
Begin
   len := Sqrt(vec.X * vec.X + vec.Y * vec.Y);
   If len <> 0 Then
   Begin
      Result.X := vec.X / len;
      Result.Y := vec.Y / len;
   End
   Else
      Result := vec;
End;

Function NormalizeTPt(Const pt: PPt): TPointD;
Begin
   With pt^ Do
      If len <> 0 Then
      Begin
         Result.X := vec.X / len;
         Result.Y := vec.Y / len;
      End
      Else
         Result := vec;
End;

Function NegateVec(vec: TPointD): TPointD;
Begin
   Result.X := -vec.X;
   Result.Y := -vec.Y;
End;

Function B0(u: double): double;
Var
   tmp: double;
Begin
   tmp := 1.0 - u;
   Result := tmp * tmp * tmp;
End;

Function B1(u: double): double;
Var
   tmp: double;
Begin
   tmp := 1.0 - u;
   Result := 3 * u * tmp * tmp;
End;

Function B2(u: double): double;
Begin
   Result := 3 * u * u * (1.0 - u);
End;

Function B3(u: double): double;
Begin
   Result := u * u * u;
End;

Function HardBreakCheck(ppt: PPt; compareLen: double): boolean;
Var
   q: double;
Const
   longLen = 15;
Begin

   If (ppt^.prev^.len * 4 < ppt^.len) Or (ppt^.len * 4 < ppt^.prev^.len) Then
   Begin

      Result := True;
   End
   Else If ((ppt^.prev^.len > longLen) And (ppt^.len > longLen)) Then
   Begin

      q := (Sqr(ppt^.prev^.len) + Sqr(ppt^.len) - DistanceSqrd(ppt^.prev^.pt, ppt^.Next^.pt)) / (2 * ppt^.prev^.len * ppt^.len);
      Result := (1 - abs(q)) > 0.3;
   End
   Else If ((TurnsLeft(ppt^.prev^.prev^.pt, ppt^.prev^.pt, ppt^.pt) = TurnsRight(ppt^.prev^.pt, ppt^.pt, ppt^.Next^.pt)) And
      (ppt^.prev^.len > compareLen) And (ppt^.len > compareLen)) Then
   Begin
      Result := True;
   End
   Else
   Begin
      q := (Sqr(ppt^.prev^.len) + Sqr(ppt^.len) - DistanceSqrd(ppt^.prev^.pt, ppt^.Next^.pt)) / (2 * ppt^.prev^.len * ppt^.len);
      Result := (q > -0.2);
   End;
End;

Function BoxesForGauss(stdDev, boxCnt, maxVal: integer): TArrayOfInteger;
Var
   i, wl, wu, m: integer;
   wIdeal, mIdeal: double;
Begin
   wIdeal := Sqrt((12 * stdDev * stdDev / boxCnt) + 1);
   wl := Floor(wIdeal);
   If (wl Mod 2 = 0) Then
      Dec(wl);
   wu := wl + 2;

   mIdeal := (12 * stdDev * stdDev - boxCnt * wl * wl - 4 * boxCnt * wl - 3 * boxCnt) / (-4 * wl - 4);
   m := Round(mIdeal);
   SetLength(Result, boxCnt);
   wl := Min(maxVal, wl);
   wu := Min(maxVal, wu);

   For i := 0 To boxCnt - 1 Do
      If i < m Then
         Result[i] := (wl - 1) Div 2
      Else
         Result[i] := (wu - 1) Div 2;
End;

Procedure BoxBlurH(Var src, dst: TArrayOfColor32; w, h, stdDev: integer);
Var
   i, j, ti, li, ri: integer;
   fv, lv, val: TgaWeightedColor;
Begin
   For i := 0 To h - 1 Do
   Begin
      ti := i * w;
      li := ti;
      ri := ti + stdDev;

      fv.Reset;
      lv.Reset;
      val.Reset;

      fv.Add(src[ti], 1);
      lv.Add(src[ti + w - 1], 1);
      val.Add(fv.Color, stdDev + 1);

      For j := 0 To stdDev - 1 Do
         val.Add(src[ti + j]);
      For j := 0 To stdDev Do
      Begin
         val.Add(src[ri]);
         Inc(ri);
         val.Subtract(fv);
         dst[ti] := val.Color;
         Inc(ti);
      End;
      For j := stdDev + 1 To w - stdDev - 1 Do
      Begin
         val.Add(src[ri]);
         Inc(ri);
         val.Subtract(src[li]);
         Inc(li);
         dst[ti] := val.Color;
         Inc(ti);
      End;
      For j := w - stdDev To w - 1 Do
      Begin
         val.Add(lv);
         val.Subtract(src[li]);
         Inc(li);
         dst[ti] := val.Color;
         Inc(ti);
      End;
   End;
End;

Procedure BoxBlurT(Var src, dst: TArrayOfColor32; w, h, stdDev: integer);
Var
   i, j, ti, li, ri: integer;
   fv, lv, val: TgaWeightedColor;
Begin
   For i := 0 To w - 1 Do
   Begin
      ti := i;
      li := ti;
      ri := ti + stdDev * w;

      fv.Reset;
      lv.Reset;
      val.Reset;

      fv.Add(src[ti]);
      lv.Add(src[ti + w * (h - 1)], 1);
      val.Add(fv.Color, stdDev + 1);

      For j := 0 To stdDev - 1 Do
         val.Add(src[ti + j * w]);
      For j := 0 To stdDev Do
      Begin
         val.Add(src[ri]);
         Inc(ri, w);
         val.Subtract(fv);
         dst[ti] := val.Color;
         Inc(ti, w);

      End;
      For j := stdDev + 1 To h - stdDev - 1 Do
      Begin
         val.Add(src[ri]);
         Inc(ri, w);
         val.Subtract(src[li]);
         Inc(li, w);
         dst[ti] := val.Color;
         Inc(ti, w);
      End;
      For j := h - stdDev To h - 1 Do
      Begin
         val.Add(lv);
         val.Subtract(src[li]);
         Inc(li, w);
         dst[ti] := val.Color;
         Inc(ti, w);
      End;
   End;
End;

Function GeTgaWeightedColor(Const srcBits: TArrayOfColor32; x256, y256, xx256, yy256, maxX: integer): TColor32;
Var
   i, j, xi, yi, xxi, yyi, weight: integer;
   xf, yf, xxf, yyf: cardinal;
   color: TgaWeightedColor;
Begin
   color.Reset;

   xi := x256 Shr 8;
   xf := x256 And $FF;
   yi := y256 Shr 8;
   yf := y256 And $FF;
   xxi := xx256 Shr 8;
   xxf := xx256 And $FF;
   yyi := yy256 Shr 8;
   yyf := yy256 And $FF;

   weight := (($100 - xf) * ($100 - yf)) Shr 8;
   color.Add(srcBits[xi + yi * maxX], weight);
   weight := (xxf * ($100 - yf)) Shr 8;
   If (weight <> 0) Then
      color.Add(srcBits[xxi + yi * maxX], weight);
   weight := (($100 - xf) * yyf) Shr 8;
   If (weight <> 0) Then
      color.Add(srcBits[xi + yyi * maxX], weight);
   weight := (xxf * yyf) Shr 8;
   If (weight <> 0) Then
      color.Add(srcBits[xxi + yyi * maxX], weight);

   If (yi + 1 < yyi) Then
   Begin
      xf := $100 - xf;
      For i := yi + 1 To yyi - 1 Do
         color.Add(srcBits[xi + i * maxX], xf);
      If (xxf <> 0) Then
         For i := yi + 1 To yyi - 1 Do
            color.Add(srcBits[xxi + i * maxX], xxf);
   End;
   If (xi + 1 < xxi) Then
   Begin
      yf := $100 - yf;
      For i := xi + 1 To xxi - 1 Do
         color.Add(srcBits[i + yi * maxX], yf);
      If (yyf <> 0) Then
         For i := xi + 1 To xxi - 1 Do
            color.Add(srcBits[i + yyi * maxX], yyf);
   End;

   For i := xi + 1 To xxi - 1 Do
      For j := yi + 1 To yyi - 1 Do
         color.Add(srcBits[i + j * maxX], $100);

   If color.AddCount = 0 Then
      Result := srcBits[xi + yi * maxX]
   Else
      Result := color.Color;
End;

Function NearestResampler(img: TgaGraphMem; x256, y256: integer): TColor32;
Begin
   If (x256 < -$7F) Then
   Begin
      Result := clNone32;
      Exit;
   End;

   If (y256 < -$7F) Then
   Begin
      Result := clNone32;
      Exit;
   End;

   If (x256 And $FF > $7F) Then
      Inc(x256, $100);
   x256 := x256 Shr 8;
   If y256 And $FF > $7F Then
      Inc(y256, $100);
   y256 := y256 Shr 8;

   If (x256 < 0) Or (x256 >= img.Width) Or (y256 < 0) Or (y256 >= img.Height) Then
      Result := clNone32
   Else
      Result := img.Pixels[y256 * img.Width + x256];
End;

Function BilinearResampler(img: TgaGraphMem; x256, y256: integer): TColor32;
Var
   xi, yi, weight: integer;
   iw, ih: integer;
   pixels: TArrayOfColor32;
   color: TgaWeightedColor;
   xf, yf: cardinal;
Begin
   iw := img.Width;
   ih := img.Height;
   pixels := img.Pixels;

   If (x256 <= -$100) Or (x256 >= iw * $100) Or (y256 <= -$100) Or (y256 >= ih * $100) Then
   Begin
      Result := clNone32;
      Exit;
   End;

   If x256 < 0 Then
      xi := -1
   Else
      xi := x256 Shr 8;

   If y256 < 0 Then
      yi := -1
   Else
      yi := y256 Shr 8;

   xf := x256 And $FF;
   yf := y256 And $FF;

   color.Reset;

   weight := (($100 - xf) * ($100 - yf)) Shr 8;
   If (xi < 0) Or (yi < 0) Then
      color.AddWeight(weight)
   Else
      color.Add(pixels[xi + yi * iw], weight);

   weight := (xf * ($100 - yf)) Shr 8;
   If ((xi + 1) >= iw) Or (yi < 0) Then
      color.AddWeight(weight)
   Else
      color.Add(pixels[(xi + 1) + yi * iw], weight);

   weight := (($100 - xf) * yf) Shr 8;
   If (xi < 0) Or ((yi + 1) >= ih) Then
      color.AddWeight(weight)
   Else
      color.Add(pixels[(xi) + (yi + 1) * iw], weight);

   weight := (xf * yf) Shr 8;
   If (xi + 1 >= iw) Or (yi + 1 >= ih) Then
      color.AddWeight(weight)
   Else
      color.Add(pixels[(xi + 1) + (yi + 1) * iw], weight);

   Result := color.Color;
End;

Type
   TBiCubicEdgeAdjust = (eaNone, eaOne, eaTwo, eaThree, eaFour);

Var
   byteFrac: Array[0..255] Of double;
   byteFracSq: Array[0..255] Of double;
   byteFracCubed: Array[0..255] Of double;

Function CubicHermite(aclr: PColor32; t: byte; bce: TBiCubicEdgeAdjust): TColor32;
Var
   a, b, c, d: PARGB;
   q: TARGB;
   aa, bb, cc: integer;
   t1, t2, t3: double;
   res: TARGB Absolute Result;
Const
   clTrans: TColor32 = clNone32;
Begin
   Case bce Of
      eaOne:
         Begin
            a := @clTrans;
            b := @clTrans;
            c := PARGB(aclr);
            Inc(aclr);
            d := PARGB(aclr);
         End;
      eaTwo:
         Begin
            a := PARGB(aclr);
            b := a;
            Inc(aclr);
            c := PARGB(aclr);
            Inc(aclr);
            d := PARGB(aclr);
         End;
      eaThree:
         Begin
            a := PARGB(aclr);
            Inc(aclr);
            b := PARGB(aclr);
            Inc(aclr);
            c := PARGB(aclr);
            d := c;
         End;
      eaFour:
         Begin
            a := PARGB(aclr);
            Inc(aclr);
            b := PARGB(aclr);
            c := @clTrans;
            d := @clTrans;
         End;
      Else
         Begin
            a := PARGB(aclr);
            Inc(aclr);
            b := PARGB(aclr);
            Inc(aclr);
            c := PARGB(aclr);
            Inc(aclr);
            d := PARGB(aclr);
         End;
   End;

   If (b^.A = 0) And (c^.A = 0) Then
   Begin
      Result := clNone32;
      Exit;
   End
   Else If b^.A = 0 Then
   Begin
      q := c^;
      q.A := 0;
      b := @q;
   End;
   If c^.A = 0 Then
   Begin
      q := b^;
      q.A := 0;
      c := @q;
   End;

   t1 := byteFrac[t];
   t2 := byteFracSq[t];
   t3 := byteFracCubed[t];

   aa := integer(-a^.A + 3 * b^.A - 3 * c^.A + d^.A) Div 2;
   bb := integer(2 * a^.A - 5 * b^.A + 4 * c^.A - d^.A) Div 2;
   cc := integer(-a^.A + c^.A) Div 2;
   Res.A := ClampByte(aa * t3 + bb * t2 + cc * t1 + b^.A);

   aa := integer(-a^.R + 3 * b^.R - 3 * c^.R + d^.R) Div 2;
   bb := integer(2 * a^.R - 5 * b^.R + 4 * c^.R - d^.R) Div 2;
   cc := integer(-a^.R + c^.R) Div 2;
   Res.R := ClampByte(aa * t3 + bb * t2 + cc * t1 + b^.R);

   aa := integer(-a^.G + 3 * b^.G - 3 * c^.G + d^.G) Div 2;
   bb := integer(2 * a^.G - 5 * b^.G + 4 * c^.G - d^.G) Div 2;
   cc := integer(-a^.G + c^.G) Div 2;
   Res.G := ClampByte(aa * t3 + bb * t2 + cc * t1 + b^.G);

   aa := integer(-a^.B + 3 * b^.B - 3 * c^.B + d^.B) Div 2;
   bb := integer(2 * a^.B - 5 * b^.B + 4 * c^.B - d^.B) Div 2;
   cc := integer(-a^.B + c^.B) Div 2;
   Res.B := ClampByte(aa * t3 + bb * t2 + cc * t1 + b^.B);
End;

Function BicubicResampler(img: TgaGraphMem; x256, y256: integer): TColor32;
Var
   i, dx, dy, pi, iw, w, h: integer;
   c: Array[0..3] Of TColor32;
   x, y: integer;
   bceX, bceY: TBiCubicEdgeAdjust;
Begin
   Result := clNone32;

   iw := img.Width;
   w := iw - 1;
   h := img.Height - 1;

   x := Abs(x256) Shr 8;
   y := Abs(y256) Shr 8;

   If (x256 < -$FF) Or (x > w) Or (y256 < -$FF) Or (y > h) Then
      Exit;

   If (x256 < 0) Then
      bceX := eaOne
   Else If (x = 0) Then
      bceX := eaTwo
   Else If (x256 > w Shl 8) Then
      bceX := eaFour
   Else If (x256 > (w - 1) Shl 8) Then
      bceX := eaThree
   Else
      bceX := eaNone;

   If (bceX = eaOne) Or (bceX = eaTwo) Then
      dx := 1
   Else
      dx := 0;

   If (y256 < 0) Then
      bceY := eaOne
   Else If y = 0 Then
      bceY := eaTwo
   Else If y = h - 1 Then
      bceY := eaThree
   Else If y = h Then
      bceY := eaFour
   Else
      bceY := eaNone;

   If (bceY = eaOne) Or (bceY = eaTwo) Then
      dy := 1
   Else
      dy := 0;

   pi := (y - 1 + dy) * iw + (x - 1 + dx);

   If bceY = eaFour Then
      dx := 2
   Else If bceY = eaThree Then
      dx := 1
   Else
      dx := 0;

   For i := dy To 3 - dx Do
   Begin
      c[i] := CubicHermite(@img.Pixels[pi], x256 And $FF, bceX);
      Inc(pi, iw);
   End;
   Result := CubicHermite(@c[dy], y256 And $FF, bceY);
End;

Function SmoothToBezier(Const path: TPathD; closed: boolean; tolerance: double; minSegLength: double): TPathD;
Var
   paths, solution: TPathsD;
Begin
   SetLength(paths, 1);
   paths[0] := path;
   solution := SmoothToBezier(paths, closed, tolerance, minSegLength);
   If solution <> Nil Then
      Result := solution[0];
End;

Function SmoothToBezier(Const paths: TPathsD; closed: boolean; tolerance: double; minSegLength: double): TPathsD;
Var
   i, j, len: integer;
Begin
   j := 0;
   len := Length(paths);
   SetLength(Result, len);
   With TgaFitCurveContainer.Create Do
      Try
         For i := 0 To len - 1 Do
            If (paths[i] <> Nil) And (Abs(Area(paths[i])) > Sqr(tolerance)) Then
            Begin
               Result[j] := FitCurve(paths[i], closed, tolerance, minSegLength);
               Inc(j);
            End;
      Finally
         Free;
      End;
   SetLength(Result, j);
End;

Function TgaFitCurveContainer.AddPt(Const pt: TPointD): PPt;
Begin
   new(Result);
   Result^.pt := pt;
   If Not assigned(ppts) Then
   Begin
      Result^.prev := Result;
      Result^.Next := Result;
      ppts := Result;
   End
   Else
   Begin
      Result^.prev := ppts^.prev;
      ppts^.prev^.Next := Result;
      ppts^.prev := Result;
      Result^.Next := ppts;
   End;
End;

Procedure TgaFitCurveContainer.Clear;
Var
   p: PPt;
Begin
   solution := Nil;
   ppts^.prev^.Next := Nil;
   While assigned(ppts) Do
   Begin
      p := ppts;
      ppts := ppts^.Next;
      Dispose(p);
   End;
End;

Function TgaFitCurveContainer.Count(First, last: PPt): integer;
Begin
   If First = last Then
      Result := 0
   Else
      Result := 1;
   Repeat
      Inc(Result);
      First := First^.Next;
   Until (First = last);
End;

Function TgaFitCurveContainer.ComputeLeftTangent(p: PPt): TPointD;
Begin
   Result := NormalizeTPt(p);
End;

Function TgaFitCurveContainer.ComputeRightTangent(p: PPt): TPointD;
Begin
   Result := NegateVec(NormalizeTPt(p^.prev));
End;

Function TgaFitCurveContainer.ComputeCenterTangent(p: PPt): TPointD;
Var
   v1, v2: TPointD;
Begin
   v1 := SubVecs(p^.pt, p^.prev^.pt);
   v2 := SubVecs(p^.Next^.pt, p^.pt);
   Result := AddVecs(v1, v2);
   Result := NormalizeVec(Result);
End;

Function TgaFitCurveContainer.ChordLengthParameterize(First: PPt; cnt: integer): TArrayOfDouble;
Var
   d: double;
   i: integer;
Begin
   SetLength(Result, cnt);
   Result[0] := 0;
   d := 0;
   For i := 1 To cnt - 1 Do
   Begin
      d := d + First^.len;
      Result[i] := d;
      First := First^.Next;
   End;
   For i := 1 To cnt - 1 Do
      Result[i] := Result[i] / d;
End;

Function TgaFitCurveContainer.GenerateBezier(First, last: PPt; cnt: integer; Const u: TArrayOfDouble; Const firstTan, lastTan: TPointD): TPathD;
Var
   i: integer;
   p: PPt;
   dist, epsilon: double;
   v1, v2, tmp: TPointD;
   a0, a1: TPathD;
   c: Array[0..1, 0..1] Of double;
   x: Array[0..1] Of double;
   det_c0_c1, det_c0_x, det_x_c1, alphaL, alphaR: double;
Begin
   SetLength(a0, cnt);
   SetLength(a1, cnt);
   dist := Distance(First^.pt, last^.pt);

   For i := 0 To cnt - 1 Do
   Begin
      v1 := Scale(firstTan, B1(u[i]));
      v2 := Scale(lastTan, B2(u[i]));
      a0[i] := v1;
      a1[i] := v2;
   End;

   FillChar(c[0][0], 4 * SizeOf(double), 0);
   FillChar(x[0], 2 * SizeOf(double), 0);

   p := First;
   For i := 0 To cnt - 1 Do
   Begin
      c[0][0] := c[0][0] + DotProdVecs(a0[i], (a0[i]));
      c[0][1] := c[0][1] + DotProdVecs(a0[i], (a1[i]));
      c[1][0] := c[0][1];
      c[1][1] := c[1][1] + DotProdVecs(a1[i], (a1[i]));

      tmp := SubVecs(p^.pt, AddVecs(Mul(First^.pt, B0(u[i])), AddVecs(Mul(First^.pt, B1(u[i])),
         AddVecs(Mul(last^.pt, B2(u[i])), Mul(last^.pt, B3(u[i]))))));

      x[0] := x[0] + DotProdVecs(a0[i], tmp);
      x[1] := x[1] + DotProdVecs(a1[i], tmp);
      p := p^.Next;
   End;

   det_c0_c1 := c[0][0] * c[1][1] - c[1][0] * c[0][1];
   det_c0_x := c[0][0] * x[1] - c[1][0] * x[0];
   det_x_c1 := x[0] * c[1][1] - x[1] * c[0][1];

   If det_c0_c1 = 0 Then
      alphaL := 0
   Else
      alphaL := det_x_c1 / det_c0_c1;

   If det_c0_c1 = 0 Then
      alphaR := 0
   Else
      alphaR := det_c0_x / det_c0_c1;

   If (alphaL > dist * 2) Then
      alphaL := 0
   Else If (alphaR > dist * 2) Then
      alphaR := 0;
   epsilon := 1.0E-6 * dist;

   SetLength(Result, 4);
   Result[0] := First^.pt;
   Result[3] := last^.pt;
   If (alphaL < epsilon) Or (alphaR < epsilon) Then
   Begin
      dist := dist / 3;
      Result[1] := AddVecs(Result[0], Scale(firstTan, dist));
      Result[2] := AddVecs(Result[3], Scale(lastTan, dist));
   End
   Else
   Begin
      Result[1] := AddVecs(Result[0], Scale(firstTan, alphaL));
      Result[2] := AddVecs(Result[3], Scale(lastTan, alphaR));
   End;
End;

Function TgaFitCurveContainer.Reparameterize(First: PPt; cnt: integer; Const u: TArrayOfDouble; Const bezier: TPathD): TArrayOfDouble;
Var
   i: integer;
Begin
   SetLength(Result, cnt);
   For i := 0 To cnt - 1 Do
   Begin
      Result[i] := NewtonRaphsonRootFind(bezier, First^.pt, u[i]);
      First := First^.Next;
   End;
End;

Function BezierII(degree: integer; Const v: Array Of TPointD; t: double): TPointD;
Var
   i, j: integer;
   tmp: Array[0..3] Of TPointD;
Begin
   Move(v[0], tmp[0], (degree + 1) * sizeOf(TPointD));
   For i := 1 To degree Do
      For j := 0 To degree - i Do
      Begin
         tmp[j].x := (1.0 - t) * tmp[j].x + t * tmp[j + 1].x;
         tmp[j].y := (1.0 - t) * tmp[j].y + t * tmp[j + 1].y;
      End;
   Result := tmp[0];
End;

Function TgaFitCurveContainer.ComputeMaxErrorSqrd(First, last: PPt; Const bezier: TPathD; Const u: TArrayOfDouble; Out SplitPoint: PPt): double;
Var
   i: integer;
   distSqrd: double;
   pt: TPointD;
   p: PPt;
Begin
   Result := 0;
   i := 1;
   SplitPoint := First^.Next;
   p := First^.Next;
   While p <> last Do
   Begin
      pt := BezierII(3, bezier, u[i]);
      distSqrd := DistanceSqrd(pt, p^.pt);
      If (distSqrd >= Result) Then
      Begin
         Result := distSqrd;
         SplitPoint := p;
      End;
      Inc(i);
      p := p^.Next;
   End;
End;

Function TgaFitCurveContainer.NewtonRaphsonRootFind(Const q: TPathD; Const pt: TPointD; u: double): double;
Var
   numerator, denominator: double;
   qu, q1u, q2u: TPointD;
   q1: Array[0..2] Of TPointD;
   q2: Array[0..1] Of TPointD;
Begin

   q1[0].x := (q[1].x - q[0].x) * 3.0;
   q1[0].y := (q[1].y - q[0].y) * 3.0;
   q1[1].x := (q[2].x - q[1].x) * 3.0;
   q1[1].y := (q[2].y - q[1].y) * 3.0;
   q1[2].x := (q[3].x - q[2].x) * 3.0;
   q1[2].y := (q[3].y - q[2].y) * 3.0;

   q2[0].x := (q1[1].x - q1[0].x) * 2.0;
   q2[0].y := (q1[1].y - q1[0].y) * 2.0;
   q2[1].x := (q1[2].x - q1[1].x) * 2.0;
   q2[1].y := (q1[2].y - q1[1].y) * 2.0;

   qu := BezierII(3, q, u);
   q1u := BezierII(2, q1, u);
   q2u := BezierII(1, q2, u);

   numerator := (qu.x - pt.x) * (q1u.x) + (qu.y - pt.y) * (q1u.y);
   denominator := (q1u.x) * (q1u.x) + (q1u.y) * (q1u.y) + (qu.x - pt.x) * (q2u.x) + (qu.y - pt.y) * (q2u.y);

   If (denominator = 0) Then
      Result := u
   Else
      Result := u - (numerator / denominator);
End;

Function TgaFitCurveContainer.FitCubic(First, last: PPt; firstTan, lastTan: TPointD): boolean;
Var
   i, cnt: integer;
   splitPoint: PPt;
   centerTan: TPointD;
   bezier: TPathD;
   clps, uPrime: TArrayOfDouble;
   maxErrorSqrd: double;
Const
   maxRetries = 4;
Begin
   Result := True;
   cnt := Count(First, last);
   If cnt = 2 Then
   Begin
      SetLength(bezier, 4);
      bezier[0] := First^.pt;
      bezier[3] := last^.pt;
      bezier[1] := bezier[0];
      bezier[2] := bezier[3];
      AppendSolution(bezier);
      Exit;
   End
   Else If cnt = 3 Then
   Begin
      If TurnsLeft(First^.prev^.pt, First^.pt, First^.Next^.pt) = TurnsLeft(First^.pt, First^.Next^.pt, last^.pt) Then
         firstTan := ComputeCenterTangent(First);
      If TurnsLeft(last^.prev^.pt, last^.pt, last^.Next^.pt) = TurnsLeft(First^.pt, First^.Next^.pt, last^.pt) Then
         lastTan := NegateVec(ComputeCenterTangent(last));
   End;

   clps := ChordLengthParameterize(First, cnt);
   bezier := GenerateBezier(First, last, cnt, clps, firstTan, lastTan);
   maxErrorSqrd := ComputeMaxErrorSqrd(First, last, bezier, clps, splitPoint);
   If (maxErrorSqrd < tolSqrd) Then
   Begin
      AppendSolution(bezier);
      Exit;
   End;

   If (maxErrorSqrd < tolSqrd * 4) Then
   Begin
      For i := 1 To maxRetries Do
      Begin
         uPrime := Reparameterize(First, cnt, clps, bezier);
         bezier := GenerateBezier(First, last, cnt, uPrime, firstTan, lastTan);
         maxErrorSqrd :=
            ComputeMaxErrorSqrd(First, last, bezier, uPrime, splitPoint);
         If (maxErrorSqrd < tolSqrd) Then
         Begin
            AppendSolution(bezier);
            Exit;
         End;
         clps := uPrime;
      End;
   End;

   If TurnsLeft(splitPoint^.prev^.prev^.pt, splitPoint^.prev^.pt, splitPoint^.pt) <> TurnsLeft(splitPoint^.prev^.pt,
      splitPoint^.pt, splitPoint^.Next^.pt) Then
   Begin
      centerTan := ComputeRightTangent(splitPoint);
      FitCubic(First, splitPoint, firstTan, centerTan);
      centerTan := ComputeLeftTangent(splitPoint);
      FitCubic(splitPoint, last, centerTan, lastTan);
   End
   Else
   Begin
      centerTan := ComputeCenterTangent(splitPoint);
      FitCubic(First, splitPoint, firstTan, NegateVec(centerTan));
      FitCubic(splitPoint, last, centerTan, lastTan);
   End;
End;

Function TgaFitCurveContainer.FitCurve(Const path: TPathD; closed: boolean; tolerance: double; minSegLength: double): TPathD;
Var
   i, highI: integer;
   d: double;
   p, p2, pEnd: PPt;
Begin
   tolSqrd := Sqr(Max(1, Min(10, tolerance)));

   highI := High(path);
   If closed Then
      While (highI > 0) And (Distance(path[highI], path[0]) < minSegLength) Do
         Dec(highI);

   p := AddPt(path[0]);
   For i := 1 To highI Do
   Begin
      d := Distance(p^.pt, path[i]);

      If d < minSegLength Then
         Continue;
      p := AddPt(path[i]);
      p^.prev^.len := d;
      p^.prev^.vec := SubVecs(p^.pt, p^.prev^.pt);
   End;
   p^.len := Distance(ppts^.pt, p^.pt);
   p^.vec := SubVecs(p^.Next^.pt, p^.pt);
   p := ppts;

   If (p^.Next = p) Or (closed And (p^.Next = p^.prev)) Then
   Begin
      Clear;
      Result := Nil;
      Exit;
   End;

   If closed Then
   Begin
      Repeat
         If HardBreakCheck(p, tolerance) Then
            break;
         p := p^.Next;
      Until p = ppts;
      pEnd := p;
   End
   Else
      pEnd := ppts^.prev;

   p2 := p^.Next;
   Repeat
      If HardBreakCheck(p2, tolerance) Then
      Begin
         FitCubic(p, p2, ComputeLeftTangent(p), ComputeRightTangent(p2));
         p := p2;
      End;
      p2 := p2^.Next;
   Until (p2 = pEnd);
   FitCubic(p, p2, ComputeLeftTangent(p), ComputeRightTangent(p2));

   Result := solution;
   Clear;
End;

Procedure TgaFitCurveContainer.AppendSolution(Const bezier: TPathD);
Var
   i, len: integer;
Begin
   len := Length(solution);
   If len > 0 Then
   Begin
      SetLength(solution, len + 3);
      For i := 0 To 2 Do
         solution[len + i] := bezier[i + 1];
   End
   Else
      solution := bezier;
End;

Procedure TgaWeightedColor.Reset;
Begin
   fAddCount := 0;
   fAlphaTot := 0;
   fColorTotR := 0;
   fColorTotG := 0;
   fColorTotB := 0;
End;

Procedure TgaWeightedColor.AddWeight(w: integer);
Begin
   Inc(fAddCount, w);
End;

Procedure TgaWeightedColor.Add(c: TColor32; w: integer);
Var
   a: integer;
   argb: TARGB Absolute c;
Begin
   Inc(fAddCount, w);
   a := w * argb.A;
   If a = 0 Then
      Exit;
   Inc(fAlphaTot, a);
   Inc(fColorTotB, (a * argb.B));
   Inc(fColorTotG, (a * argb.G));
   Inc(fColorTotR, (a * argb.R));
End;

Procedure TgaWeightedColor.Add(Const other: TgaWeightedColor);
Begin
   Inc(fAddCount, other.fAddCount);
   Inc(fAlphaTot, other.fAlphaTot);
   Inc(fColorTotR, other.fColorTotR);
   Inc(fColorTotG, other.fColorTotG);
   Inc(fColorTotB, other.fColorTotB);
End;

Procedure TgaWeightedColor.Subtract(c: TColor32; w: integer);
Var
   a: integer;
   argb: TARGB Absolute c;
Begin
   Dec(fAddCount, w);
   a := w * argb.A;
   If a = 0 Then
      Exit;
   Dec(fAlphaTot, a);
   Dec(fColorTotB, (a * argb.B));
   Dec(fColorTotG, (a * argb.G));
   Dec(fColorTotR, (a * argb.R));
End;

Procedure TgaWeightedColor.Subtract(Const other: TgaWeightedColor);
Begin
   Dec(fAddCount, other.fAddCount);
   Dec(fAlphaTot, other.fAlphaTot);
   Dec(fColorTotR, other.fColorTotR);
   Dec(fColorTotG, other.fColorTotG);
   Dec(fColorTotB, other.fColorTotB);
End;

Function TgaWeightedColor.GetColor: TColor32;
Var
   invAlpha: double;
   res: TARGB Absolute Result;
Begin
   If (fAlphaTot <= 0) Or (fAddCount <= 0) Then
   Begin
      Result := clNone32;
      Exit;
   End;
   res.A := Min(255, (fAlphaTot + (fAddCount Shr 1)) Div fAddCount);
   invAlpha := 1 / fAlphaTot;
   res.R := ClampByte(fColorTotR * invAlpha);
   res.G := ClampByte(fColorTotG * invAlpha);
   res.B := ClampByte(fColorTotB * invAlpha);
End;

Function TSizeD.average: double;
Begin
   Result := (sx + sy) * 0.5;
End;

Function TRectWH.IsEmpty: boolean;
Begin
   Result := (Width <= 0) Or (Height <= 0);
End;

Function TRectWH.Right: double;
Begin
   Result := Left + Width;
End;

Function TRectWH.Bottom: double;
Begin
   Result := Top + Height;
End;

Function TRectWH.Contains(Const Pt: TPoint): boolean;
Begin
   Result := (pt.X >= Left) And (pt.X <= Left + Width) And (pt.Y >= Top) And (pt.Y <= Top + Height);
End;

Function TRectWH.Contains(Const Pt: TPointD): boolean;
Begin
   Result := (pt.X >= Left) And (pt.X <= Left + Width) And (pt.Y >= Top) And (pt.Y <= Top + Height);
End;

Function TRectWH.MidPoint: TPointD;
Begin
   Result := PointD(left + Width * 0.5, top + Height * 0.5);
End;

Function TRectWH.RectD: TRectD;
Begin
   Result := gaEngine.RectD(left, top, left + Width, top + Height);
End;

Function TRectD.IsEmpty: boolean;
Begin
   Result := (right <= left) Or (bottom <= top);
End;

Function TRectD.Width: double;
Begin
   Result := Max(0, right - left);
End;

Function TRectD.Height: double;
Begin
   Result := Max(0, bottom - top);
End;

Function TRectD.TopLeft: TPointD;
Begin
   Result := PointD(Left, Top);
End;

Function TRectD.BottomRight: TPointD;
Begin
   Result := PointD(Right, Bottom);
End;

Function TRectD.MidPoint: TPointD;
Begin
   Result.X := (Right + Left) / 2;
   Result.Y := (Bottom + Top) / 2;
End;

Function TRectD.Normalize: boolean;
Var
   d: double;
Begin
   Result := False;
   If Left > Right Then
   Begin
      d := Left;
      Left := Right;
      Right := d;
      Result := True;
   End;
   If Top > Bottom Then
   Begin
      d := Top;
      Top := Bottom;
      Bottom := d;
      Result := True;
   End;
End;

Function TRectD.Contains(Const Pt: TPoint): boolean;
Begin
   Result := (pt.X >= Left) And (pt.X < Right) And (pt.Y >= Top) And (pt.Y < Bottom);
End;

Function TRectD.Contains(Const Pt: TPointD): boolean;
Begin
   Result := (pt.X >= Left) And (pt.X < Right) And (pt.Y >= Top) And (pt.Y < Bottom);
End;

Constructor TgaGraphMemList.Create;
Begin
   fList := TList.Create;
   fIsImageOwner := True;
End;

Destructor TgaGraphMemList.Destroy;
Begin
   Clear;
   fList.Free;
   Inherited;
End;

Function TgaGraphMemList.Count: integer;
Begin
   Result := fList.Count;
End;

Procedure TgaGraphMemList.Clear;
Var
   i: integer;
Begin
   If IsImageOwner Then
      For i := 0 To fList.Count - 1 Do
         TgaGraphMem(fList[i]).Free;
   fList.Clear;
End;

Function TgaGraphMemList.GetImage(index: integer): TgaGraphMem;
Begin
   Result := TgaGraphMem(fList[index]);
End;

Procedure TgaGraphMemList.SetImage(index: integer; img: TgaGraphMem);
Begin
   If fIsImageOwner Then
      TgaGraphMem(fList[index]).Free;
   fList[index] := img;
End;

Function TgaGraphMemList.GetLast: TgaGraphMem;
Begin
   If Count = 0 Then
      Result := Nil
   Else
      Result := TgaGraphMem(fList[Count - 1]);
End;

Procedure TgaGraphMemList.Add(image: TgaGraphMem);
Begin
   fList.Add(image);
End;

Function TgaGraphMemList.Add(Width, Height: integer): TgaGraphMem;
Begin
   Result := TgaGraphMem.Create(Width, Height);
   fList.Add(Result);
End;

Procedure TgaGraphMemList.Insert(index: integer; image: TgaGraphMem);
Begin
   fList.Insert(index, image);
End;

Procedure TgaGraphMemList.Move(currentIndex, newIndex: integer);
Begin
   fList.Move(currentIndex, newIndex);
End;

Procedure TgaGraphMemList.Delete(index: integer);
Begin
   If fIsImageOwner Then
      TgaGraphMem(fList[index]).Free;
   fList.Delete(index);
End;

Function TgaImageFormat.LoadFromFile(Const filename: String; gaBitMap: TgaGraphMem): boolean;
Var
   fs: TFileStream;
Begin
   Result := FileExists(filename);
   If Not Result Then
      Exit;
   fs := TFileStream.Create(filename, fmOpenRead Or fmShareDenyWrite);
   Try
      Result := LoadFromStream(fs, gaBitMap);
   Finally
      fs.Free;
   End;
End;

Function TgaImageFormat.SaveToFile(Const filename: String; gaBitMap: TgaGraphMem): boolean;
Var
   fs: TFileStream;
Begin
   Result := (pos('.', filename) = 1) Or DirectoryExists(ExtractFilePath(filename));
   If Not Result Then
      Exit;

   fs := TFileStream.Create(filename, fmCreate);
   Try
      SaveToStream(fs, gaBitMap);
   Finally
      fs.Free;
   End;
End;

Class Function TgaImageFormat.CanCopyToClipboard: boolean;
Begin
   Result := False;
End;

Constructor TgaGraphMem.Create(Const aWidth: integer = 0; aHeight: integer = 0);
Begin
   fResampler := VarDefaultResampler;
   fwidth  := Max(0, aWidth);
   fheight := Max(0, aHeight);
   SetLength(fPixels, fwidth * fheight);
End;

Constructor TgaGraphMem.Create(aSrc: TgaGraphMem);
Begin
   Assign(aSrc);
End;

Constructor TgaGraphMem.Create(aSrc: TgaGraphMem; Const srcRec: TRect);
Begin
   fResampler := aSrc.fResampler;
   fUpdateCnt := 1;
   SetSize(RectWidth(srcRec), RectHeight(srcRec));
   If Not IsEmptyRect(srcRec) Then
      fPixels := aSrc.CopyPixels(srcRec);
   fUpdateCnt := 0;
End;

Destructor TgaGraphMem.Destroy;
Begin
   fPixels := Nil;
   Inherited;
End;

Class Function TgaGraphMem.IsRegisteredFormat(Const ext: String): boolean;
Begin
   Result := Assigned(TgaGraphMem.GeTgaImageFormatClass(ext));
End;

Class Procedure TgaGraphMem.RegisterImageFormatClass(ext: String; bm32ExClass: TgaImageFormatClass; clipPriority: TClipboardPriority);
Var
   i: integer;
   imgFmtRec: PImgFmtRec;
Begin
   If Not Assigned(ImageFormatClassList) Then
      CreateImageFormatList;

   If (ext = '') Or (ext = '.') Then
      Exit;
   If (ext[1] = '.') Then
      Delete(ext, 1, 1);
   If Not IsAlphaChar(ext[1]) Then
      Exit;

   For i := 0 To imageFormatClassList.Count - 1 Do
   Begin
      imgFmtRec := PImgFmtRec(imageFormatClassList[i]);
      If SameText(imgFmtRec^.Fmt, ext) Then
         Exit;
   End;

   new(imgFmtRec);
   imgFmtRec^.Fmt := ext;
   imgFmtRec^.SortOrder := clipPriority;
   imgFmtRec^.Obj := bm32ExClass;
   ImageFormatClassList.Add(imgFmtRec);

   ImageFormatClassList.Sort(@ImageFormatClassListSort);
End;

Class Function TgaGraphMem.GeTgaImageFormatClass(Const ext: String): TgaImageFormatClass;
Var
   i: integer;
   pattern: String;
   imgFmtRec: PImgFmtRec;
Begin
   Result := Nil;
   pattern := ext;
   If (pattern = '') Or (pattern = '.') Then
      Exit;
   If pattern[1] = '.' Then
      Delete(pattern, 1, 1);

   For i := imageFormatClassList.Count - 1 Downto 0 Do
   Begin
      imgFmtRec := PImgFmtRec(imageFormatClassList[i]);
      If Not SameText(imgFmtRec^.Fmt, pattern) Then
         Continue;
      Result := imgFmtRec^.Obj;
      break;
   End;
End;

Procedure TgaGraphMem.Assign(src: TgaGraphMem);
Begin
   If assigned(src) Then
      src.AssignTo(self);
End;

Procedure TgaGraphMem.AssignTo(dst: TgaGraphMem);
Begin
   If dst = self Then
      Exit;
   dst.BeginUpdate;
   Try
      dst.fResampler := fResampler;
      dst.fIsPremultiplied := fIsPremultiplied;
      dst.fColorCount := 0;
      Try
         dst.SetSize(Width, Height);
         If (Width > 0) And (Height > 0) Then
            move(fPixels[0], dst.fPixels[0], Width * Height * SizeOf(TColor32));
      Except
         dst.SetSize(0, 0);
      End;
   Finally
      dst.EndUpdate;
   End;
End;

Procedure TgaGraphMem.Changed;
Begin
   If fUpdateCnt <> 0 Then
      Exit;
   fColorCount := 0;
   If Assigned(fOnChange) Then
      fOnChange(Self);
End;

Procedure TgaGraphMem.Resized;
Begin
   If Assigned(fOnResize) Then
      fOnResize(Self);
   Changed;
End;

Procedure TgaGraphMem.BeginUpdate;
Begin
   Inc(fUpdateCnt);
End;

Procedure TgaGraphMem.EndUpdate;
Begin
   Dec(fUpdateCnt);
   Changed;
End;

Procedure TgaGraphMem.BlockUpdate;
Begin
   Inc(fUpdateCnt);
End;

Procedure TgaGraphMem.UnblockUpdate;
Begin
   Dec(fUpdateCnt);
End;

Procedure TgaGraphMem.SetBackgroundColor(bgColor: TColor32);
Var
   i: integer;
   pc: PColor32;
Begin
   pc := Pixelbase;
   For i := 0 To high(fPixels) Do
   Begin
      pc^ := BlendToOpaque(bgColor, pc^);
      Inc(pc);
   End;
   Changed;
End;

Procedure TgaGraphMem.Clear(color: TColor32);
Var
   i: integer;
   pc: PColor32;
Begin
   fIsPremultiplied := False;
   If IsEmpty Then
      Exit;
   If color = clNone32 Then
      FillChar(fPixels[0], Width * Height * SizeOf(TColor32), 0)
   Else
   Begin
      pc := PixelBase;
      For i := 0 To Width * Height - 1 Do
      Begin
         pc^ := color;
         Inc(pc);
      End;
   End;
   Changed;
End;

Procedure TgaGraphMem.Clear(Const rec: TRect; color: TColor32 = 0);
Begin
   FillRect(rec, color);
End;

Procedure TgaGraphMem.FillRect(rec: TRect; color: TColor32);
Var
   i, j, rw: integer;
   c: PColor32;
Begin
   Types.IntersectRect(rec, rec, bounds);
   If IsEmptyRect(rec) Then
      Exit;
   rw := RectWidth(rec);
   c := @Pixels[rec.Top * Width + rec.Left];
   For i := rec.Top To rec.Bottom - 1 Do
   Begin
      For j := 1 To rw Do
      Begin
         c^ := color;
         Inc(c);
      End;
      Inc(c, Width - rw);
   End;
   Changed;
End;

Procedure CheckBlendFill(pc: PColor32; color: TColor32);

Begin
   If Not assigned(pc) Then
      Exit;
   pc^ := BlendToAlpha(pc^, color);
End;

Function TgaGraphMem.CopyPixels(rec: TRect): TArrayOfColor32;
Var
   i, clipW, w, h: integer;
   pSrc, pDst, pDst2: PColor32;
   recClipped: TRect;
Begin
   w := RectWidth(rec);
   h := RectHeight(rec);
   setLength(Result, w * h);

   If w * h = 0 Then
      Exit;
   Types.IntersectRect(recClipped, rec, Bounds);

   If IsEmptyRect(recClipped) Then
   Begin
      FillChar(Result[0], w * h * SizeOf(TColor32), 0);
      Changed;
      Exit;
   End;

   If RectsEqual(recClipped, rec) Then
   Begin
      pDst := @Result[0];
      pSrc := @fPixels[recClipped.Top * Width + rec.Left];
      For i := recClipped.Top To recClipped.Bottom - 1 Do
      Begin
         Move(pSrc^, pDst^, w * SizeOf(TColor32));
         Inc(pSrc, Width);
         Inc(pDst, w);
      End;
      Changed;
      Exit;
   End;

   pDst := @Result[0];
   For i := rec.Top To -1 Do
   Begin
      FillChar(pDst^, w * SizeOf(TColor32), 0);
      Inc(pDst, w);
   End;

   pSrc := @fPixels[recClipped.Top * Width + Max(0, rec.Left)];
   If (rec.Left < 0) Or (rec.Right > Width) Then
   Begin
      clipW := RectWidth(recClipped);
      pDst2 := IncPColor32(pDst, -Min(0, rec.Left));
      For i := recClipped.Top To recClipped.Bottom - 1 Do
      Begin
         FillChar(pDst^, w * SizeOf(TColor32), 0);
         Move(pSrc^, pDst2^, clipW * SizeOf(TColor32));
         Inc(pDst, w);
         Inc(pDst2, w);
         Inc(pSrc, Width);
      End;
   End
   Else
   Begin
      For i := recClipped.Top To recClipped.Bottom - 1 Do
      Begin
         Move(pSrc^, pDst^, w * SizeOf(TColor32));
         Inc(pSrc, Width);
         Inc(pDst, w);
      End;
   End;
   For i := Height To rec.Bottom - 1 Do
   Begin
      FillChar(pDst^, w * SizeOf(TColor32), 0);
      Inc(pDst, w);
   End;
   Changed;
End;

Procedure TgaGraphMem.Crop(Const rec: TRect);
Var
   newPixels: TArrayOfColor32;
Begin
   newPixels := CopyPixels(rec);
   BeginUpdate;
   Try
      SetSize(RectWidth(rec), RectHeight(rec));
      If Not IsEmptyRect(rec) Then
         fPixels := newPixels;
   Finally
      EndUpdate;
   End;
End;

Function TgaGraphMem.GetBounds: TRect;
Begin
   Result := Types.Rect(0, 0, Width, Height);
End;

Function TgaGraphMem.GetMidPoint: TPointD;
Begin
   Result := PointD(fWidth * 0.5, fHeight * 0.5);
End;

Procedure TgaGraphMem.SetSize(newWidth, newHeight: integer; color: TColor32);
Begin
   fwidth := Max(0, newWidth);
   fheight := Max(0, newHeight);
   fPixels := Nil;
   SetLength(fPixels, fwidth * fheight);
   fIsPremultiplied := False;
   BeginUpdate;
   Try
      If color > 0 Then
         Clear(color);
      Resized;
   Finally
      EndUpdate;
   End;
End;

Procedure TgaGraphMem.Resize(newWidth, newHeight: integer; stretchImage: boolean);
Var
   tmp: TgaGraphMem;
   rec: TRect;
Begin
   If (newWidth <= 0) Or (newHeight <= 0) Then
   Begin
      SetSize(0, 0);
      Exit;

   End;
   If (newWidth = fwidth) And (newHeight = fheight) Then
      Exit;

   If IsEmpty Then
   Begin
      SetSize(newWidth, newHeight);
      Exit;
   End;

   BeginUpdate;
   Try
      If stretchImage Then
      Begin
         If fResampler = 0 Then
            NearestNeighborResize(newWidth, newHeight)
         Else
            ResamplerResize(newWidth, newHeight);
      End
      Else
      Begin
         tmp := TgaGraphMem.Create(self);
         Try
            rec := Bounds;
            SetSize(newWidth, newHeight, clNone32);
            Copy(tmp, rec, rec);
         Finally
            tmp.Free;
         End;
      End;
   Finally
      Resized;
      EndUpdate;
   End;
End;

Procedure TgaGraphMem.NearestNeighborResize(newWidth, newHeight: integer);
Var
   x, y, srcY: integer;
   scaledXi, scaledYi: TArrayOfInteger;
   tmp: TArrayOfColor32;
   pc: PColor32;
Begin

   If (newWidth = fWidth) And (newHeight = fHeight) Then
      Exit;
   SetLength(tmp, newWidth * newHeight * SizeOf(TColor32));

   SetLength(scaledXi, newWidth);
   For x := 0 To newWidth - 1 Do
      scaledXi[x] := Floor(x * fWidth / newWidth);
   SetLength(scaledYi, newHeight);
   For y := 0 To newHeight - 1 Do
      scaledYi[y] := Floor(y * fHeight / newHeight);

   pc := @tmp[0];
   For y := 0 To newHeight - 1 Do
   Begin
      srcY := scaledYi[y];
      If (srcY < 0) Or (srcY >= fHeight) Then
         Continue;
      For x := 0 To newWidth - 1 Do
      Begin
         pc^ := fPixels[scaledXi[x] + srcY * fWidth];
         Inc(pc);
      End;
   End;

   fPixels := tmp;
   fwidth := newWidth;
   fheight := newHeight;
End;

Procedure TgaGraphMem.ResamplerResize(newWidth, newHeight: integer);
Var
   mat: TMatrixD;
Begin
   mat := cnIdentityMatrix;
   MatrixScale(mat, newWidth / fWidth, newHeight / fHeight);
   self.AffineTransformImage(mat);
End;

Procedure TgaGraphMem.Scale(s: double);
Begin
   Scale(s, s);
End;

Procedure TgaGraphMem.Scale(sx, sy: double);
Begin
   If (sx > 0) And (sy > 0) Then
      ReSize(Round(Width * sx), Round(Height * sy));
End;

Procedure TgaGraphMem.ScaleToFit(Width, Height: integer);
Var
   sx, sy: double;
Begin
   If IsEmpty Or (Width <= 0) Or (Height <= 0) Then
      Exit;
   sx := Width / self.Width;
   sy := Height / self.Height;
   If sx <= sy Then
      Scale(sx)
   Else
      Scale(sy);
End;

Procedure TgaGraphMem.ScaleToFitCentered(Width, Height: integer);
Var
   sx, sy: double;
   tmp: TgaGraphMem;
   rec2: TRect;
Begin
   If IsEmpty Or (Width <= 0) Or (Height <= 0) Then
      Exit;
   sx := Width / self.Width;
   sy := Height / self.Height;
   BeginUpdate;
   Try
      If sx <= sy Then
      Begin
         Scale(sx);
         If Height = self.Height Then
            Exit;
         rec2 := Bounds;
         Types.OffsetRect(rec2, 0, (Height - self.Height) Div 2);
         tmp := TgaGraphMem.Create(self);
         Try
            SetSize(Width, Height);
            CopyInternal(tmp, tmp.Bounds, rec2, Nil);
         Finally
            tmp.Free;
         End;
      End
      Else
      Begin
         Scale(sy);
         If Width = self.Width Then
            Exit;
         rec2 := Bounds;
         Types.OffsetRect(rec2, (Width - self.Width) Div 2, 0);
         tmp := TgaGraphMem.Create(self);
         Try
            SetSize(Width, Height);
            CopyInternal(tmp, tmp.Bounds, rec2, Nil);
         Finally
            tmp.Free;
         End;
      End;
   Finally
      EndUpdate;
   End;
End;

Procedure TgaGraphMem.RotateLeft90;
Var
   x, y, xx: integer;
   src, dst: PColor32;
   tmp: TgaGraphMem;
Begin
   If IsEmpty Then
      Exit;

   BeginUpdate;
   tmp := TgaGraphMem.Create(Self);
   Try
      SetSize(Height, Width);
      xx := (Width - 1) * Height;
      dst := PixelBase;
      For y := 0 To Height - 1 Do
      Begin
         src := @tmp.Pixels[xx + y];
         For x := 0 To Width - 1 Do
         Begin
            dst^ := src^;
            Inc(dst);
            Dec(src, Height);
         End;
      End;
   Finally
      tmp.Free;
      EndUpdate;
   End;
End;

Procedure TgaGraphMem.RotateRight90;
Var
   x, y: integer;
   src, dst: PColor32;
   tmp: TgaGraphMem;
Begin
   If IsEmpty Then
      Exit;

   BeginUpdate;
   tmp := TgaGraphMem.Create(Self);
   Try
      SetSize(Height, Width);
      dst := PixelBase;
      For y := 0 To Height - 1 Do
      Begin
         src := @tmp.Pixels[Height - 1 - y];
         For x := 0 To Width - 1 Do
         Begin
            dst^ := src^;
            Inc(dst);
            Inc(src, Height);
         End;
      End;
   Finally
      tmp.Free;
      EndUpdate;
   End;
End;

Procedure TgaGraphMem.Rotate180;
Var
   x, y: integer;
   src, dst: PColor32;
   tmp: TgaGraphMem;
Begin
   If IsEmpty Then
      Exit;
   tmp := TgaGraphMem.Create(Self);
   Try
      dst := PixelBase;
      src := @tmp.Pixels[Width * Height - 1];
      For y := 0 To Height - 1 Do
      Begin
         For x := 0 To Width - 1 Do
         Begin
            dst^ := src^;
            Inc(dst);
            Dec(src);
         End;
      End;
   Finally
      tmp.Free;
   End;
   Changed;
End;

Function TgaGraphMem.GetColorCount: integer;
Var
   allColors: PByteArray;
   i: integer;
   c: PColor32;
Const
   cube256 = 256 * 256 * 256;
Begin
   Result := 0;
   If IsEmpty Then
      Exit;
   If fColorCount > 0 Then
   Begin
      Result := fColorCount;
      Exit;
   End;
   allColors := AllocMem(cube256);
   Try
      c := PixelBase;
      For i := 0 To Width * Height - 1 Do
      Begin
         If (c^ Shr 24) > $80 Then
            allColors^[c^ And $FFFFFF] := 1;
         Inc(c);
      End;
      For i := 0 To cube256 - 1 Do
         If allColors^[i] = 1 Then
            Inc(Result);
   Finally
      FreeMem(allColors);
   End;
   fColorCount := Result;
End;

Function TgaGraphMem.GetHasTransparency: boolean;
Var
   i: integer;
   pc: PARGB;
Begin
   Result := True;
   If IsEmpty Then
      Exit;
   pc := PARGB(PixelBase);
   For i := 0 To Width * Height - 1 Do
   Begin
      If pc^.A < 255 Then
         Exit;
      Inc(pc);
   End;
   Result := False;
End;

Function TgaGraphMem.SaveToFile(filename: String): boolean;
Var
   fileFormatClass: TgaImageFormatClass;
Begin
   Result := False;
   If IsEmpty Or (length(filename) < 5) Then
      Exit;
   If ExtractFilePath(filename) = '' Then
      filename := GetCurrentDir + '\' + filename;
   fileFormatClass := GeTgaImageFormatClass(ExtractFileExt(filename));
   If assigned(fileFormatClass) Then
      With fileFormatClass.Create Do
         Try
            Result := SaveToFile(filename, self);
         Finally
            Free;
         End;
End;

Function TgaGraphMem.SaveToStream(stream: TStream; Const FmtExt: String): boolean;
Var
   fileFormatClass: TgaImageFormatClass;
Begin
   Result := False;
   fileFormatClass := GeTgaImageFormatClass(FmtExt);
   If assigned(fileFormatClass) Then
      With fileFormatClass.Create Do
         Try
            SaveToStream(stream, self);
            Result := True;
         Finally
            Free;
         End;
End;

Function TgaGraphMem.LoadFromFile(Const filename: String): boolean;
Var
   image32FileFmtClass: TgaImageFormatClass;
Begin
   image32FileFmtClass := GeTgaImageFormatClass(ExtractFileExt(filename));
   Result := assigned(image32FileFmtClass);
   If Not Result Then
      Exit;

   With image32FileFmtClass.Create Do
      Try
         Result := LoadFromFile(filename, self);
      Finally
         Free;
      End;
End;

Function TgaGraphMem.LoadFromStream(stream: TStream): boolean;
Var
   i: integer;
Begin
   Result := False;
   For i := 0 To imageFormatClassList.Count - 1 Do
      With PImgFmtRec(imageFormatClassList[i])^ Do
         If Obj.IsValidImageStream(stream) Then
         Begin
            With obj.Create Do
               Try
                  Result := LoadFromStream(stream, self);
               Finally
                  Free;
               End;
            break;
         End;
End;

Function TgaGraphMem.GetPixel(x, y: integer): TColor32;
Begin
   If (x < 0) Or (x >= Width) Or (y < 0) Or (y >= Height) Then
      Result := clNone32
   Else
      Result := fPixels[y * Width + x];
End;

Procedure TgaGraphMem.SetPixel(x, y: integer; color: TColor32);
Begin
   If (x < 0) Or (x >= Width) Or (y < 0) Or (y >= Height) Then
      Exit;
   fPixels[y * Width + x] := color;
End;

Function TgaGraphMem.GetIsBlank: boolean;
Var
   i: integer;
   pc: PARGB;
Begin
   Result := IsEmpty;
   If Result Then
      Exit;
   pc := PARGB(PixelBase);
   For i := 0 To Width * Height - 1 Do
   Begin
      If pc^.A > 0 Then
         Exit;
      Inc(pc);
   End;
   Result := True;
End;

Function TgaGraphMem.GetIsEmpty: boolean;
Begin
   Result := fPixels = Nil;
End;

Function TgaGraphMem.GetPixelBase: PColor32;
Begin
   If IsEmpty Then
      Result := Nil
   Else
      Result := @fPixels[0];
End;

Function TgaGraphMem.GetPixelRow(row: integer): PColor32;
Begin
   If IsEmpty Then
      Result := Nil
   Else
      Result := @fPixels[row * Width];
End;

Procedure TgaGraphMem.CopyInternal(src: TgaGraphMem; Const srcRec, dstRec: TRect; blendFunc: TBlendFunction);
Var
   i, j, srcRecWidth: integer;
   s, d: PColor32;
Begin
   srcRecWidth := srcRec.Right - srcRec.Left;
   s := @src.Pixels[srcRec.Top * src.Width + srcRec.Left];
   d := @Pixels[dstRec.top * Width + dstRec.Left];

   If assigned(blendFunc) Then
      For i := srcRec.Top To srcRec.Bottom - 1 Do
      Begin
         For j := 1 To srcRecWidth Do
         Begin
            d^ := blendFunc(d^, s^);
            Inc(s);
            Inc(d);
         End;
         Inc(s, src.Width - srcRecWidth);
         Inc(d, Width - srcRecWidth);
      End
   Else
      For i := srcRec.Top To srcRec.Bottom - 1 Do
      Begin
         move(s^, d^, srcRecWidth * SizeOf(TColor32));
         Inc(s, src.Width);
         Inc(d, Width);
      End;
End;

Function TgaGraphMem.Copy(src: TgaGraphMem; srcRec, dstRec: TRect): boolean;
Begin
   Result := CopyBlend(src, srcRec, dstRec, Nil);
End;

Function TgaGraphMem.RectHasTransparency(rec: TRect): Boolean;
Var
   i, j, rw: Integer;
   c: PARGB;
Begin
   Result := True;
   Types.IntersectRect(rec, rec, bounds);
   If IsEmptyRect(rec) Then
      Exit;
   rw := RectWidth(rec);
   c := @Pixels[rec.Top * Width + rec.Left];
   For i := rec.Top To rec.Bottom - 1 Do
   Begin
      For j := 1 To rw Do
      Begin
         If c^.A < 254 Then
            Exit;
         inc(c);
      End;
      inc(c, Width - rw);
   End;
   Result := False;
End;

Function TgaGraphMem.CopyBlend(src: TgaGraphMem; srcRec, dstRec: TRect; blendFunc: TBlendFunction): boolean;
Var
   tmp: TgaGraphMem;
   srcRecClipped, dstRecClipped, r: TRect;
   xscale, scaleSrc, scaleDst: TPointD;
Begin
   Result := False;
   If IsEmptyRect(srcRec) Or IsEmptyRect(dstRec) Then
      Exit;
   Types.IntersectRect(srcRecClipped, srcRec, src.Bounds);

   xscale.X := RectWidth(dstRec) / RectWidth(srcRec);
   xscale.Y := RectHeight(dstRec) / RectHeight(srcRec);

   If Not RectsEqual(srcRecClipped, srcRec) Then
   Begin
      If IsEmptyRect(srcRecClipped) Then
         Exit;
      scaleSrc.X := RectWidth(srcRecClipped) / RectWidth(srcRec);
      scaleSrc.Y := RectHeight(srcRecClipped) / RectHeight(srcRec);
      ScaleRect(dstRec, scaleSrc);
      Types.OffsetRect(dstRec,
         srcRecClipped.Left - srcRec.Left,
         srcRecClipped.Top - srcRec.Top);
   End;

   If (xscale.X <> 1.0) Or (xscale.Y <> 1.0) Then
   Begin
      tmp := TgaGraphMem.Create(src, srcRecClipped);
      Try
         tmp.Scale(xscale.X, xscale.Y);
         Result := CopyBlend(tmp, tmp.Bounds, dstRec, blendFunc);
      Finally
         tmp.Free;
      End;
      Exit;
   End;

   Types.IntersectRect(dstRecClipped, dstRec, Bounds);
   If IsEmptyRect(dstRecClipped) Then
      Exit;

   If Not RectsEqual(dstRecClipped, dstRec) Then
   Begin
      scaleDst.X := RectWidth(dstRecClipped) / RectWidth(dstRec);
      scaleDst.Y := RectHeight(dstRecClipped) / RectHeight(dstRec);
      ScaleRect(srcRecClipped, scaleDst);
      Types.OffsetRect(srcRecClipped,
         dstRecClipped.Left - dstRec.Left,
         dstRecClipped.Top - dstRec.Top);
   End;

   If (src = self) And Types.IntersectRect(r, srcRecClipped, dstRecClipped) Then
   Begin
      tmp := TgaGraphMem.Create(self, srcRecClipped);
      Try
         Result := src.CopyBlend(tmp, tmp.Bounds, dstRecClipped, blendFunc);
      Finally
         tmp.Free;
      End;
      Exit;
   End;

   CopyInternal(src, srcRecClipped, dstRecClipped, blendFunc);
   Result := True;
   Changed;
End;

Function TgaGraphMem.LoadFromResource(Const resName: String; resType: PChar): boolean;
Var
   resStream: TResourceStream;
Begin
   resStream := CreateResourceStream(resName, resType);
   Try
      Result := assigned(resStream) And LoadFromStream(resStream);
   Finally
      resStream.Free;
   End;
End;


Function TgaGraphMem.CopyToGR32(aGR32Bits: PGRColor32Array): Boolean;
Begin
   Result := False;
   If aGR32Bits = Nil Then
      Exit;

   System.Move(Self.Pixels[0],
      aGR32Bits^[0],
      Self.Width * Self.Height * (SizeOf(TColor32)));

   Result := True;
End;

Function TgaGraphMem.CopyFromGR32(aGR32Bits: PGRColor32Array; Const ScrWidth, ScrHeight: Integer): Boolean;
Begin
   Result := False;
   If aGR32Bits = Nil Then
      Exit;
   If ScrWidth < 1 Then
      Exit;
   If ScrHeight < 1 Then
      Exit;

   Self.SetSize(ScrWidth, ScrHeight);

   System.Move(aGR32Bits^[0],
      Self.Pixels[0],
      Self.Width * Self.Height * (SizeOf(TColor32)));
   Result := True;
End;

Function TgaGraphMem.CopyToClipBoard: boolean;
Var
   i: integer;
   formatClass: TgaImageFormatClass;
Begin
   Result := Not IsEmpty;
   If Not Result Then
      Exit;
   Result := False;

   For i := ImageFormatClassList.Count - 1 Downto 0 Do
   Begin
      formatClass := PImgFmtRec(ImageFormatClassList[i])^.Obj;
      If Not formatClass.CanCopyToClipboard Then
         Continue;
      With formatClass.Create Do
         Try
            If CopyToClipboard(self) Then
               Result := True;
         Finally
            Free;
         End;
   End;
End;

Class Function TgaGraphMem.CanPasteFromClipBoard: boolean;
Var
   i: integer;
   formatClass: TgaImageFormatClass;
Begin
   Result := False;
   For i := ImageFormatClassList.Count - 1 Downto 0 Do
   Begin
      formatClass := PImgFmtRec(ImageFormatClassList[i])^.Obj;
      If formatClass.CanPasteFromClipboard Then
      Begin
         Result := True;
         Exit;
      End;
   End;
End;

Function TgaGraphMem.PasteFromClipBoard: boolean;
Var
   i: integer;
   formatClass: TgaImageFormatClass;
Begin
   Result := False;
   For i := ImageFormatClassList.Count - 1 Downto 0 Do
   Begin
      formatClass := PImgFmtRec(ImageFormatClassList[i])^.Obj;
      If Not formatClass.CanPasteFromClipboard Then
         Continue;

      With formatClass.Create Do
         Try
            Result := PasteFromClipboard(self);
            If Not Result Then
               Continue;
         Finally
            Free;
         End;
      Changed;
      Break;
   End;
End;

Procedure TgaGraphMem.ConvertToBoolMask(reference: TColor32; tolerance: integer; colorFunc: TCompareFunction; maskBg: TColor32; maskFg: TColor32);
Var
   i: integer;
   mask: TArrayOfByte;
   c: PColor32;
   b: pbyte;
Begin
   If IsEmpty Then
      Exit;
   mask := self.GetBoolMask(reference, colorFunc, tolerance);
   c := PixelBase;
   b := @mask[0];
   For i := 0 To Width * Height - 1 Do
   Begin
      If b^ = 0 Then
         c^ := maskBg
      Else
         c^ := maskFg;

      Inc(c);
      Inc(b);
   End;
   Changed;
End;

Procedure TgaGraphMem.ConvertToAlphaMask(reference: TColor32; colorFunc: TCompareFunctionEx);
Var
   i: integer;
   mask: TArrayOfByte;
   c: PColor32;
   b: pbyte;
Begin
   If IsEmpty Then
      Exit;
   mask := self.GetByteMask(reference, colorFunc);
   c := PixelBase;
   b := @mask[0];
   For i := 0 To Width * Height - 1 Do
   Begin
      c^ := b^ Shl 24;
      Inc(c);
      Inc(b);
   End;
   Changed;
End;

Procedure TgaGraphMem.FlipVertical;
Var
   i: integer;
   a: TArrayOfColor32;
   src, dst: PColor32;
Begin
   If IsEmpty Then
      Exit;
   SetLength(a, fWidth * fHeight);
   src := @fPixels[(Height - 1) * Width];
   dst := @a[0];
   For i := 0 To fHeight - 1 Do
   Begin
      move(src^, dst^, fWidth * SizeOf(TColor32));
      Dec(src, fWidth);
      Inc(dst, fWidth);
   End;
   fPixels := a;
   Changed;
End;

Procedure TgaGraphMem.FlipHorizontal;
Var
   i, j, widthLess1: integer;
   a: TArrayOfColor32;
   row: PColor32;
Begin
   If IsEmpty Then
      Exit;
   SetLength(a, fWidth);
   widthLess1 := fWidth - 1;
   row := @fPixels[(Height - 1) * Width];
   For i := 0 To fHeight - 1 Do
   Begin
      move(row^, a[0], fWidth * SizeOf(TColor32));
      For j := 0 To widthLess1 Do
      Begin
         row^ := a[widthLess1 - j];
         Inc(row);
      End;
      Dec(row, fWidth * 2);
   End;
   Changed;
End;

Procedure TgaGraphMem.PreMultiply;
Var
   i: integer;
   c: PARGB;
Begin
   If IsEmpty Or fIsPremultiplied Then
      Exit;
   fIsPremultiplied := True;
   c := PARGB(PixelBase);
   For i := 0 To Width * Height - 1 Do
   Begin
      If c^.A > 0 Then
      Begin
         c^.R := VarMulTable[c^.R, c^.A];
         c^.G := VarMulTable[c^.G, c^.A];
         c^.B := VarMulTable[c^.B, c^.A];
      End
      Else
         c^.Color := 0;
      Inc(c);
   End;
End;

Procedure TgaGraphMem.SetRGB(rgbColor: TColor32);
Var
   rgb: TARGB Absolute rgbColor;
   r, g, b: byte;
   i: integer;
   pc: PARGB;
Begin
   If IsEmpty Then
      Exit;
   r := rgb.R;
   g := rgb.G;
   b := rgb.B;
   pc := PARGB(PixelBase);
   For i := 0 To Width * Height - 1 Do
   Begin
      pc^.R := r;
      pc^.G := g;
      pc^.B := b;
      Inc(pc);
   End;
   Changed;
End;

Procedure TgaGraphMem.SetRGB(rgbColor: TColor32; rec: TRect);
Var
   rgb: TARGB Absolute rgbColor;
   r, g, b: byte;
   i, j, dx: integer;
   pc: PARGB;
Begin
   Types.IntersectRect(rec, rec, bounds);
   If IsEmptyRect(rec) Then
      Exit;
   r := rgb.R;
   g := rgb.G;
   b := rgb.B;
   pc := PARGB(PixelBase);
   Inc(pc, rec.Left);
   dx := Width - RectWidth(rec);
   For i := rec.Top To rec.Bottom - 1 Do
   Begin
      For j := rec.Left To rec.Right - 1 Do
      Begin
         pc^.R := r;
         pc^.G := g;
         pc^.B := b;
         Inc(pc);
      End;
      Inc(pc, dx);
   End;
   Changed;
End;

Procedure TgaGraphMem.SetAlpha(alpha: byte);
Var
   i: integer;
   c: PARGB;
Begin
   If IsEmpty Then
      Exit;
   c := PARGB(PixelBase);
   For i := 0 To Width * Height - 1 Do
   Begin
      c^.A := alpha;
      Inc(c);
   End;
   Changed;
End;

Procedure TgaGraphMem.ReduceOpacity(opacity: byte);
Var
   i: integer;
   c: PARGB;
Begin
   If opacity = 255 Then
      Exit;
   c := PARGB(PixelBase);
   For i := 0 To Width * Height - 1 Do
   Begin
      c^.A := VarMulTable[c^.A, opacity];
      Inc(c);
   End;
   Changed;
End;

Procedure TgaGraphMem.ReduceOpacity(opacity: byte; rec: TRect);
Var
   i, j, rw: integer;
   c: PARGB;
Begin
   Types.IntersectRect(rec, rec, bounds);
   If IsEmptyRect(rec) Then
      Exit;
   rw := RectWidth(rec);
   c := @Pixels[rec.Top * Width + rec.Left];
   For i := rec.Top To rec.Bottom - 1 Do
   Begin
      For j := 1 To rw Do
      Begin
         c^.A := VarMulTable[c^.A, opacity];
         Inc(c);
      End;
      Inc(c, Width - rw);
   End;
   Changed;
End;

Procedure TgaGraphMem.Grayscale;
Begin
   AdjustSaturation(-100);
End;

Procedure TgaGraphMem.InvertColors;
Var
   pc: PARGB;
   i: integer;
Begin
   pc := PARGB(PixelBase);
   For i := 0 To Width * Height - 1 Do
   Begin
      pc^.R := 255 - pc^.R;
      pc^.G := 255 - pc^.G;
      pc^.B := 255 - pc^.B;
      Inc(pc);
   End;
   Changed;
End;

Procedure TgaGraphMem.InvertAlphas;
Var
   pc: PARGB;
   i: integer;
Begin
   pc := PARGB(PixelBase);
   For i := 0 To Width * Height - 1 Do
   Begin
      pc^.A := 255 - pc^.A;
      Inc(pc);
   End;
   Changed;
End;

Procedure TgaGraphMem.Tile(newWidth, newHeight: integer; style: TTileFillStyle = tfsRepeat);
Var
   i, j, w, h, w2, h2: integer;
   pc1, pc2: PColor32;
   tmp: TgaGraphMem;
Begin
   w := Min(Width, newWidth);
   h := Min(Height, newHeight);
   w2 := Min(w * 2, newWidth);
   h2 := Min(h * 2, newHeight);

   tmp := TgaGraphMem.Create(self, Types.Rect(0, 0, w, h));
   Try
      SetSize(newWidth, newHeight);
      For i := 0 To h - 1 Do
      Begin
         pc1 := tmp.PixelRow[i];
         pc2 := PixelRow[i];
         Move(pc1^, pc2^, w * SizeOf(TColor32));
      End;

      Case style Of
         tfsMirrorHorz: tmp.FlipHorizontal;
         tfsMirrorVert: tmp.FlipVertical;
         tfsRotate180: tmp.Rotate180;
      End;

      If w2 > w Then
      Begin
         For i := 0 To h - 1 Do
         Begin
            pc1 := tmp.PixelRow[i];
            pc2 := @Pixels[i * newWidth + w];
            Move(pc1^, pc2^, (w2 - w) * SizeOf(TColor32));
         End;
      End;

      If h2 > h Then
      Begin
         For i := 0 To w2 - 1 Do
         Begin
            pc1 := @Pixels[i];
            pc2 := @Pixels[h * newWidth + i];
            For j := h To h2 - 1 Do
            Begin
               pc2^ := pc1^;
               Inc(pc1, newWidth);
               Inc(pc2, newWidth);
            End;
         End;
      End;
   Finally
      tmp.Free;
   End;

   w := w2;
   h := h2;
   While (w2 < newWidth) Do
   Begin
      j := Min(w2 + w, newWidth);
      For i := 0 To h - 1 Do
      Begin
         pc1 := PixelRow[i];
         pc2 := IncPColor32(pc1, w2);
         Move(pc1^, pc2^, (j - w2) * SizeOf(TColor32));
      End;
      Inc(w2, w);
   End;

   While (h2 < newHeight) Do
   Begin
      j := Min(h2 + h, newHeight);
      For i := h2 To j - 1 Do
      Begin
         pc1 := PixelRow[i - h2];
         pc2 := PixelRow[i];
         Move(pc1^, pc2^, newWidth * SizeOf(TColor32));
      End;
      Inc(h2, h);
   End;
   Changed;
End;

Procedure TgaGraphMem.AdjustHue(percent: integer);
Var
   i: integer;
   tmpImage: TArrayofHSL;
   lut: Array[byte] Of byte;
Begin
   percent := percent Mod 100;
   If percent < 0 Then
      Inc(percent, 100);
   percent := Round(percent * 255 / 100);
   If (percent = 0) Or IsEmpty Then
      Exit;
   For i := 0 To 255 Do
      lut[i] := (i + percent) Mod 255;
   tmpImage := ArrayOfColor32ToArrayHSL(fPixels);
   For i := 0 To high(tmpImage) Do
      tmpImage[i].hue := lut[tmpImage[i].hue];
   fPixels := ArrayOfHSLToArrayColor32(tmpImage);
   Changed;
End;

Procedure TgaGraphMem.AdjustLuminance(percent: integer);
Var
   i: integer;
   tmpImage: TArrayofHSL;
   pc: double;
   lut: Array[byte] Of byte;
Begin
   If (percent = 0) Or IsEmpty Then
      Exit;
   percent := percent Mod 101;
   pc := percent / 100;
   If pc > 0 Then
      For i := 0 To 255 Do
         lut[i] := Round(i + (255 - i) * pc)
   Else
      For i := 0 To 255 Do
         lut[i] := Round(i + (i * pc));

   tmpImage := ArrayOfColor32ToArrayHSL(fPixels);
   For i := 0 To high(tmpImage) Do
      tmpImage[i].lum := lut[tmpImage[i].lum];
   fPixels := ArrayOfHSLToArrayColor32(tmpImage);
   Changed;
End;

Procedure TgaGraphMem.AdjustSaturation(percent: integer);
Var
   i: integer;
   tmpImage: TArrayofHSL;
   lut: Array[byte] Of byte;
   pc: double;
Begin
   If (percent = 0) Or IsEmpty Then
      Exit;
   percent := percent Mod 101;
   pc := percent / 100;
   If pc > 0 Then
      For i := 0 To 255 Do
         lut[i] := Round(i + (255 - i) * pc)
   Else
      For i := 0 To 255 Do
         lut[i] := Round(i + (i * pc));

   tmpImage := ArrayOfColor32ToArrayHSL(fPixels);
   For i := 0 To high(tmpImage) Do
      tmpImage[i].sat := lut[tmpImage[i].sat];
   fPixels := ArrayOfHSLToArrayColor32(tmpImage);
   Changed;
End;

Function TgaGraphMem.CropTransparentPixels: TRect;
Var
   x, y, x1, x2, y1, y2: integer;
   found: boolean;
Begin
   y1 := 0;
   y2 := 0;
   found := False;
   For y := 0 To Height - 1 Do
   Begin
      For x := 0 To Width - 1 Do
         If TARGB(fPixels[y * Width + x]).A > 0 Then
         Begin
            y1 := y;
            found := True;
            break;
         End;
      If found Then
         break;
   End;

   If Not found Then
   Begin
      SetSize(0, 0);
      Exit;
   End;

   found := False;
   For y := Height - 1 Downto 0 Do
   Begin
      For x := 0 To Width - 1 Do
         If TARGB(fPixels[y * Width + x]).A > 0 Then
         Begin
            y2 := y;
            found := True;
            break;
         End;
      If found Then
         break;
   End;

   x1 := Width;
   x2 := 0;
   For y := y1 To y2 Do
      For x := 0 To Width - 1 Do
         If TARGB(fPixels[y * Width + x]).A > 0 Then
         Begin
            If x < x1 Then
               x1 := x;
            If x > x2 Then
               x2 := x;
         End;

   Result := Types.Rect(x1, y1, x2 + 1, y2 + 1);
   Crop(Result);
End;

Procedure TgaGraphMem.Rotate(angleRads: double);
Var
   rec: TRectD;
   mat: TMatrixD;
Begin
   If Not VarClockwiseRotationIsAnglePositive Then
      angleRads := -angleRads;

   NormalizeAngle(angleRads);
   If IsEmpty Or (angleRads = 0) Then
      Exit;

   If angleRads = cnAngle180 Then
   Begin
      Rotate180;
   End
   Else If angleRads = cnAngle90 Then
   Begin
      RotateRight90;
   End
   Else If angleRads = -cnAngle90 Then
   Begin
      RotateLeft90;
   End
   Else
   Begin
      mat := cnIdentityMatrix;
      MatrixTranslate(mat, Width / 2, Height / 2);
      rec := RectD(Bounds);
      rec := GetRotatedRectBounds(rec, angleRads);
      MatrixRotate(mat, cnNullPointD, angleRads);
      MatrixTranslate(mat, rec.Width / 2, rec.Height / 2);
      self.AffineTransformImage(mat);
   End;
End;

Procedure TgaGraphMem.RotateRect(Const rec: TRect; angleRads: double; eraseColor: TColor32 = 0);
Var
   tmp: TgaGraphMem;
   rec2: TRect;
   recWidth, recHeight: integer;
Begin
   recWidth := rec.Right - rec.Left;
   recHeight := rec.Bottom - rec.Top;
   tmp := TgaGraphMem.Create(self, rec);
   Try
      tmp.Rotate(angleRads);
      rec2.Left := (tmp.Width - recWidth) Div 2;
      rec2.Top := (tmp.Height - recHeight) Div 2;
      rec2.Right := rec2.Left + recWidth;
      rec2.Bottom := rec2.Top + recHeight;
      FillRect(rec, eraseColor);
      CopyBlend(tmp, rec2, rec);
   Finally
      tmp.Free;
   End;
End;

Procedure TgaGraphMem.Skew(dx, dy: double);
Var
   mat: TMatrixD;
Begin
   If IsEmpty Or ((dx = 0) And (dy = 0)) Then
      Exit;
   dx := ClampRange(dx, -2.0, 2.0);
   dy := ClampRange(dy, -2.0, 2.0);
   mat := cnIdentityMatrix;
   MatrixSkew(mat, dx, dy);
   self.AffineTransformImage(mat);
End;

Procedure TgaGraphMem.ScaleAlpha(aScale: double);
Var
   i: integer;
   pb: PARGB;
Begin
   pb := PARGB(PixelBase);
   For i := 0 To Width * Height - 1 Do
   Begin
      pb^.A := ClampByte(Round(pb^.A * aScale));
      Inc(pb);
   End;
   Changed;
End;

Function TgaGraphMem.GetTransformBounds(Const matrix: TMatrixD): TRect;
Var
   pts: TPathD;
Begin
   pts := Rectangle(self.Bounds);
   MatrixApply(matrix, pts);
   Result := gaGetBounds(pts);
End;

Function TgaGraphMem.AffineTransformImage(matrix: TMatrixD): TPoint;
Var
   i, j, srcWidth, srcHeight: integer;
   newWidth, newHeight: integer;
   x, y: double;
   pc: PColor32;
   tmp: TArrayOfColor32;
   dstRec: TRect;
   xresampler: TResamplerFunction;
Begin
   Result := cnNullPoint;
   srcWidth := self.Width;
   srcHeight := self.Height;

   If self.Resampler = 0 Then
      xresampler := Nil
   Else
      xresampler := GetResampler(self.Resampler);

   If Not Assigned(xresampler) Or (srcWidth * srcHeight = 0) Or IsIdentityMatrix(matrix) Then
      Exit;

   dstRec := GetTransformBounds(matrix);
   newWidth := RectWidth(dstRec);
   newHeight := RectHeight((dstRec));

   Result := dstRec.TopLeft;

   If Not MatrixInvert(matrix) Then
      Exit;

   SetLength(tmp, RectWidth(dstRec) * RectHeight(dstRec));
   pc := @tmp[0];

   For i := dstRec.Top To +dstRec.Bottom - 1 Do
      For j := dstRec.Left To dstRec.Right - 1 Do
      Begin

         x := j;
         y := i;
         MatrixApply(matrix, x, y);

         pc^ := xresampler(self, Round(x * 256), Round(y * 256));
         Inc(pc);
      End;
   self.BeginUpdate;
   Try
      self.SetSize(newWidth, newHeight);
      Move(tmp[0], self.Pixels[0], newWidth * newHeight * sizeOf(TColor32));
   Finally
      self.EndUpdate;
   End;
End;

Function TgaGraphMem.ProjectiveTransform(Const srcPts, dstPts: TPathD; Const margins: TRect): boolean;
Var
   w, h, i, j: integer;
   x, y: integer;
   rec: TRect;
   dstPts2: TPathD;
   mat: TMatrixD;
   tmp: TArrayOfColor32;
   pc: PColor32;
   xresampler: TResamplerFunction;
Begin

   If self.Resampler = 0 Then
      xresampler := Nil
   Else
      xresampler := GetResampler(self.Resampler);

   Result := Assigned(xresampler) And Not self.IsEmpty And (Length(dstPts) = 4) And IsPathConvex(dstPts);
   If Not Result Then
      Exit;

   rec := gaGetBounds(dstPts);
   Dec(rec.Left, margins.Left);
   Dec(rec.Top, margins.Top);
   Inc(rec.Right, margins.Right);
   Inc(rec.Bottom, margins.Bottom);
   dstPts2 := OffsetPath(dstPts, -rec.Left, -rec.Top);

   mat := GetProjectionMatrix(srcPts, dstPts2);
   w := RectWidth(rec);
   h := RectHeight(rec);
   SetLength(tmp, w * h);
   pc := @tmp[0];
   For i := 0 To h - 1 Do
      For j := 0 To w - 1 Do
      Begin
         x := j;
         y := i;
         GetSrcCoords256(mat, x, y);
         pc^ := xresampler(self, x, y);
         Inc(pc);
      End;
   self.SetSize(w, h);
   Move(tmp[0], self.PixelBase^, w * h * sizeOf(TColor32));
End;

Function TgaGraphMem.SplineVertTransform(Const topSpline: TPathD; splineType: TSplineType; backColor: TColor32;
   reverseFill: boolean; Out offset: TPoint): boolean;
Var
   i, j, w, h, len: integer;
   y, q: double;
   distances: TArrayOfDouble;
   pc: PColor32;
   rec: TRect;
   tmp: TArrayOfColor32;
   topPath: TPathD;
   prevX: double;
   xresampler: TResamplerFunction;
   backColoring, allowBackColoring: boolean;
Begin
   offset := cnNullPoint;
   If self.Resampler = 0 Then
      xresampler := Nil
   Else
      xresampler := GetResampler(self.Resampler);

   If splineType = stQuadratic Then
      topPath := FlattenQSpline(topSpline)
   Else
      topPath := FlattenCSpline(topSpline);

   rec := gaGetBounds(topPath);
   Result := Assigned(xresampler) And Not IsEmptyRect(rec);
   If Not Result Then
      Exit;

   offset := rec.TopLeft;
   topPath := InterpolatePathForX(topPath);
   len := Length(topPath);
   Inc(rec.Bottom, self.Height);
   w := RectWidth(rec);
   h := RectHeight(rec);
   SetLength(tmp, (w + 1) * h);

   prevX := topPath[0].X;
   allowBackColoring := (backColor Shr 24) > 2;
   backColor := backColor And $00FFFFFF;

   distances := GetCumulativeDistances(topPath);
   q := self.Width * 256 / distances[High(distances)];
   ;
   For i := 0 To len - 1 Do
   Begin
      pc := @tmp[Round(topPath[i].X) - rec.Left];
      backColoring := allowBackColoring And (prevX >= topPath[i].X);
      prevX := topPath[i].X;
      y := topPath[i].Y;
      For j := rec.top To rec.bottom - 1 Do
      Begin
         If (j > y - 1.0) And (j < y + self.Height) Then
            If backColoring Then
               pc^ := BlendToAlpha(pc^, ReColor(xresampler(self, Round(Distances[i] * q), Round((j - y) * 256)), backColor))
            Else
               pc^ := BlendToAlpha(pc^, xresampler(self, Round(Distances[i] * q), Round((j - y) * 256)));
         Inc(pc, w);
      End;
   End;

   self.BeginUpdate;
   self.SetSize(w, h);
   Move(tmp[0], self.Pixels[0], self.Width * self.Height * SizeOf(TColor32));
   self.EndUpdate;
End;

Function TgaGraphMem.SplineHorzTransform(Const leftSpline: TPathD; splineType: TSplineType; backColor: TColor32;
   reverseFill: boolean; Out offset: TPoint): boolean;
Var
   i, j, len, w, h: integer;
   x, q, prevY: double;
   leftPath: TPathD;
   distances: TArrayOfDouble;
   rec: TRect;
   pc: PColor32;
   tmp: TArrayOfColor32;
   backColoring, allowBackColoring: boolean;
   xresampler: TResamplerFunction;
Begin
   offset := cnNullPoint;

   If self.Resampler = 0 Then
      xresampler := Nil
   Else
      xresampler := GetResampler(self.Resampler);

   If splineType = stQuadratic Then
      leftPath := FlattenQSpline(leftSpline)
   Else
      leftPath := FlattenCSpline(leftSpline);
   rec := gaGetBounds(leftPath);

   Result := Assigned(xresampler) And Not IsEmptyRect(rec);
   If Not Result Then
      Exit;

   offset := rec.TopLeft;
   leftPath := InterpolatePathForY(leftPath);
   len := Length(leftPath);
   Inc(rec.Right, self.Width);
   w := RectWidth(rec);
   h := RectHeight(rec);
   SetLength(tmp, w * (h + 1));

   prevY := leftPath[0].Y;
   allowBackColoring := (backColor Shr 24) > 2;
   backColor := backColor And $00FFFFFF;

   distances := GetCumulativeDistances(leftPath);
   q := self.Height * 256 / distances[High(distances)];
   ;
   For i := 0 To len - 1 Do
   Begin
      pc := @tmp[Round(leftPath[i].Y - rec.Top) * w];
      backColoring := allowBackColoring And (prevY >= leftPath[i].Y);
      prevY := leftPath[i].Y;
      x := leftPath[i].X;
      For j := rec.left To rec.right - 1 Do
      Begin
         If (j > x - 1.0) And (j < x + self.Width) Then
            If backColoring Then
               pc^ := BlendToAlpha(pc^, ReColor(xresampler(self, Round((j - x) * 256), Round(Distances[i] * q)), backColor))
            Else
               pc^ := BlendToAlpha(pc^, xresampler(self, Round((j - x) * 256), Round(Distances[i] * q)));
         Inc(pc);
      End;
   End;

   self.BeginUpdate;
   self.SetSize(w, h);
   Move(tmp[0], self.Pixels[0], self.Width * self.Height * SizeOf(TColor32));
   self.EndUpdate;
End;

Procedure TgaGraphMem.DrawPoint(Const pt: TPointD; radius: double; color: TColor32);
Var
   path: TPathD;
Begin
   If radius <= 1 Then
      path := Rectangle(pt.X - radius, pt.Y - radius, pt.X + radius, pt.Y + radius)
   Else
      path := Ellipse(RectD(pt.X - radius, pt.Y - radius, pt.X + radius, pt.Y + radius));
   DrawPolygon(path, frEvenOdd, color);
End;

Procedure TgaGraphMem.DrawPoint(Const pt: TPointD; radius: double; renderer: TgaRendererCustom);
Var
   path: TPathD;
Begin
   path := Ellipse(RectD(pt.X - radius, pt.Y - radius, pt.X + radius, pt.Y + radius));
   DrawPolygon(path, frEvenOdd, renderer);
End;

Procedure TgaGraphMem.DrawPoint(Const points: TPathD; radius: double; color: TColor32);
Var
   i: integer;
Begin
   For i := 0 To high(points) Do
      DrawPoint(points[i], radius, color);
End;

Procedure TgaGraphMem.DrawPoint(Const paths: TPathsD; radius: double; color: TColor32);
Var
   i: integer;
Begin
   For i := 0 To high(paths) Do
      DrawPoint(paths[i], radius, color);
End;

Procedure TgaGraphMem.DrawLine(Const pt1, pt2: TPointD; lineWidth: double; color: TColor32);
Var
   lines: TPathsD;
Begin
   setLength(lines, 1);
   setLength(lines[0], 2);
   lines[0][0] := pt1;
   lines[0][1] := pt2;
   DrawLine(lines, lineWidth, color, esRound);
End;

Procedure TgaGraphMem.DrawLine(Const line: TPathD; lineWidth: double; color: TColor32; endStyle: TEndStyle;
   joinStyle: TJoinStyle; miterLimit: double);
Var
   Lines: TPathsD;
Begin
   setLength(Lines, 1);
   Lines[0] := line;
   DrawLine(Lines, lineWidth, color, endStyle, joinStyle, miterLimit);
End;

Procedure TgaGraphMem.DrawLine(Const line: TPathD; lineWidth: double; renderer: TgaRendererCustom; endStyle: TEndStyle;
   joinStyle: TJoinStyle; miterLimit: double);
Var
   Lines: TPathsD;
Begin
   setLength(Lines, 1);
   Lines[0] := line;
   DrawLine(Lines, lineWidth, renderer, endStyle, joinStyle, miterLimit);
End;

Procedure TgaGraphMem.DrawInvertedLine(Const line: TPathD; lineWidth: double; endStyle: TEndStyle; joinStyle: TJoinStyle = jsAuto);
Var
   Lines: TPathsD;
Begin
   setLength(Lines, 1);
   Lines[0] := line;
   DrawInvertedLine(Lines, lineWidth, endStyle, joinStyle);
End;

Procedure TgaGraphMem.DrawLine(Const Lines: TPathsD; lineWidth: double; color: TColor32; endStyle: TEndStyle;
   joinStyle: TJoinStyle; miterLimit: double);
Var
   lines2: TPathsD;
   cr: TgaRendererColor;
Begin
   If Not assigned(Lines) Then
      exit;
   If (lineWidth < VarMinStrokeWidth) Then
      lineWidth := VarMinStrokeWidth;
   lines2 := Outline(Lines, lineWidth, joinStyle, endStyle, miterLimit);

   cr := TgaRendererColor.Create(color);
   Try
      If cr.Initialize(self) Then
      Begin
         Rasterize(lines2, self.bounds, frNonZero, cr);
         cr.NotifyChange;
      End;
   Finally
      cr.Free;
   End;
End;

Procedure TgaGraphMem.DrawLine(Const Lines: TPathsD; lineWidth: double; renderer: TgaRendererCustom; endStyle: TEndStyle;
   joinStyle: TJoinStyle; miterLimit: double);
Var
   lines2: TPathsD;
Begin
   If (Not assigned(Lines)) Or (Not assigned(renderer)) Then
      exit;
   If (lineWidth < VarMinStrokeWidth) Then
      lineWidth := VarMinStrokeWidth;
   lines2 := Outline(Lines, lineWidth, joinStyle, endStyle, miterLimit);
   If renderer.Initialize(self) Then
   Begin
      Rasterize(lines2, self.bounds, frNonZero, renderer);
      renderer.NotifyChange;
   End;
End;

Procedure TgaGraphMem.DrawInvertedLine(Const Lines: TPathsD; lineWidth: double; endStyle: TEndStyle; joinStyle: TJoinStyle = jsAuto);
Var
   lines2: TPathsD;
   ir: TgaRendererInverse;
Begin
   If Not assigned(Lines) Then
      exit;
   If (lineWidth < VarMinStrokeWidth) Then
      lineWidth := VarMinStrokeWidth;
   lines2 := Outline(Lines, lineWidth, joinStyle, endStyle, 2);
   ir := TgaRendererInverse.Create;
   Try
      If ir.Initialize(self) Then
      Begin
         Rasterize(lines2, self.bounds, frNonZero, ir);
         ir.NotifyChange;
      End;
   Finally
      ir.Free;
   End;
End;

Procedure TgaGraphMem.DrawDashedLine(Const line: TPathD; dashPattern: TArrayOfInteger; patternOffset: PDouble;
   lineWidth: double; color: TColor32; endStyle: TEndStyle; joinStyle: TJoinStyle);
Var
   Lines: TPathsD;
   cr: TgaRendererColor;
Begin
   If (lineWidth < VarMinStrokeWidth) Then
      lineWidth := VarMinStrokeWidth;
   If Not assigned(line) Then
      exit;
   Lines := GetDashedPath(line, endStyle = esPolygon, dashPattern, patternOffset);
   If Length(Lines) = 0 Then
      Exit;
   Case joinStyle Of
      jsSquare, jsMiter:
         endStyle := esSquare;
      jsRound:
         endStyle := esRound;
      Else
         endStyle := esButt;
   End;
   Lines := Outline(Lines, lineWidth, joinStyle, endStyle);
   cr := TgaRendererColor.Create(color);
   Try
      If cr.Initialize(self) Then
      Begin
         Rasterize(Lines, self.bounds, frNonZero, cr);
         cr.NotifyChange;
      End;
   Finally
      cr.Free;
   End;
End;

Procedure TgaGraphMem.DrawDashedLine(Const Lines: TPathsD; dashPattern: TArrayOfInteger; patternOffset: PDouble;
   lineWidth: double; color: TColor32; endStyle: TEndStyle; joinStyle: TJoinStyle);
Var
   i: integer;
Begin
   If Not assigned(Lines) Then
      exit;
   For i := 0 To high(Lines) Do
      DrawDashedLine(Lines[i],
         dashPattern, patternOffset, lineWidth, color, endStyle, joinStyle);
End;

Procedure TgaGraphMem.DrawDashedLine(Const line: TPathD; dashPattern: TArrayOfInteger; patternOffset: PDouble;
   lineWidth: double; renderer: TgaRendererCustom; endStyle: TEndStyle; joinStyle: TJoinStyle);
Var
   Lines: TPathsD;
Begin
   If (Not assigned(line)) Or (Not assigned(renderer)) Then
      exit;
   If (lineWidth < VarMinStrokeWidth) Then
      lineWidth := VarMinStrokeWidth;
   Lines := GetDashedPath(line, endStyle = esPolygon, dashPattern, patternOffset);
   If Length(Lines) = 0 Then
      Exit;
   Lines := Outline(Lines, lineWidth, joinStyle, endStyle);
   If renderer.Initialize(self) Then
   Begin
      Rasterize(Lines, self.bounds, frNonZero, renderer);
      renderer.NotifyChange;
   End;
End;

Procedure TgaGraphMem.DrawDashedLine(Const Lines: TPathsD; dashPattern: TArrayOfInteger; patternOffset: PDouble;
   lineWidth: double; renderer: TgaRendererCustom; endStyle: TEndStyle; joinStyle: TJoinStyle);
Var
   i: integer;
Begin
   If Not assigned(Lines) Then
      exit;
   For i := 0 To high(Lines) Do
      DrawDashedLine(Lines[i],
         dashPattern, patternOffset, lineWidth, renderer, endStyle, joinStyle);
End;

Procedure TgaGraphMem.DrawInvertedDashedLine(Const line: TPathD; dashPattern: TArrayOfInteger; patternOffset: PDouble;
   lineWidth: double; endStyle: TEndStyle; joinStyle: TJoinStyle = jsAuto);
Var
   Lines: TPathsD;
   renderer: TgaRendererInverse;
Begin
   If Not assigned(line) Then
      exit;
   If (lineWidth < VarMinStrokeWidth) Then
      lineWidth := VarMinStrokeWidth;
   Lines := GetDashedPath(line, endStyle = esPolygon, dashPattern, patternOffset);
   If Length(Lines) = 0 Then
      Exit;
   Lines := Outline(Lines, lineWidth, joinStyle, endStyle);
   renderer := TgaRendererInverse.Create;
   Try
      If renderer.Initialize(self) Then
      Begin
         Rasterize(Lines, self.bounds, frNonZero, renderer);
         renderer.NotifyChange;
      End;
   Finally
      renderer.Free;
   End;
End;

Procedure TgaGraphMem.DrawInvertedDashedLine(Const Lines: TPathsD; dashPattern: TArrayOfInteger; patternOffset: PDouble;
   lineWidth: double; endStyle: TEndStyle; joinStyle: TJoinStyle = jsAuto);
Var
   i: integer;
Begin
   If Not assigned(Lines) Then
      exit;
   For i := 0 To high(Lines) Do
      DrawInvertedDashedLine(Lines[i],
         dashPattern, patternOffset, lineWidth, endStyle, joinStyle);
End;

Procedure TgaGraphMem.DrawPolygon(Const polygon: TPathD; fillRule: TFillRule; color: TColor32);
Var
   polygons: TPathsD;
Begin
   If Not assigned(polygon) Then
      exit;
   setLength(polygons, 1);
   polygons[0] := polygon;
   DrawPolygon(polygons, fillRule, color);
End;

Procedure TgaGraphMem.DrawPolygon(Const polygon: TPathD; fillRule: TFillRule; renderer: TgaRendererCustom);
Var
   polygons: TPathsD;
Begin
   If (Not assigned(polygon)) Or (Not assigned(renderer)) Then
      exit;
   setLength(polygons, 1);
   polygons[0] := polygon;
   If renderer.Initialize(self) Then
   Begin
      Rasterize(polygons, self.Bounds, fillRule, renderer);
      renderer.NotifyChange;
   End;
End;

Procedure TgaGraphMem.DrawPolygon(Const polygons: TPathsD; fillRule: TFillRule; color: TColor32);
Var
   cr: TgaRendererColor;
Begin
   If Not assigned(polygons) Then
      exit;
   cr := TgaRendererColor.Create(color);
   Try
      If cr.Initialize(self) Then
      Begin
         Rasterize(polygons, self.bounds, fillRule, cr);
         cr.NotifyChange;
      End;
   Finally
      cr.Free;
   End;
End;

Procedure TgaGraphMem.DrawPolygon(Const polygons: TPathsD; fillRule: TFillRule; renderer: TgaRendererCustom);
Begin
   If (Not assigned(polygons)) Or (Not assigned(renderer)) Then
      exit;
   If renderer.Initialize(self) Then
   Begin
      Rasterize(polygons, self.bounds, fillRule, renderer);
      renderer.NotifyChange;
   End;
End;

Procedure TgaGraphMem.DrawPolygon_ClearType(Const polygons: TPathsD; fillRule: TFillRule; color: TColor32; backColor: TColor32);
Var
   tmpImg: TgaGraphMem;
   rec: TRect;
   tmpPolygons: TPathsD;
   cr: TgaRendererColor;
Begin
   If Not assigned(polygons) Then
      exit;

   rec := gaGetBounds(polygons);
   tmpImg := TgaGraphMem.Create(RectWidth(rec) * 3, RectHeight(rec));
   Try
      tmpPolygons := OffsetPath(polygons, -rec.Left, -rec.Top);
      tmpPolygons := ScalePath(tmpPolygons, 3, 1);
      cr := TgaRendererColor.Create(clBlack32);
      Try
         If cr.Initialize(tmpImg) Then
         Begin
            Rasterize(tmpPolygons, tmpImg.bounds, fillRule, cr);
            cr.NotifyChange;
         End;
      Finally
         cr.Free;
      End;
      tmpImg.ApplyClearType(color, backColor);
      self.CopyBlend(tmpImg, tmpImg.Bounds, rec, @BlendToAlpha);
   Finally
      tmpImg.Free;
   End;
End;

Procedure TgaGraphMem.ErasePolygon(Const polygon: TPathD; fillRule: TFillRule);
Var
   polygons: TPathsD;
Begin
   If Not assigned(polygon) Then
      exit;
   setLength(polygons, 1);
   polygons[0] := polygon;
   ErasePolygon(polygons, fillRule);
End;

Procedure TgaGraphMem.ErasePolygon(Const polygons: TPathsD; fillRule: TFillRule);
Var
   er: TgaRendererErase;
Begin
   er := TgaRendererErase.Create;
   Try
      If er.Initialize(self) Then
      Begin
         Rasterize(polygons, self.bounds, fillRule, er);
         er.NotifyChange;
      End;
   Finally
      er.Free;
   End;
End;

Procedure TgaGraphMem.DrawBoolMask(Const mask: TArrayOfByte; color: TColor32);
Var
   i, len: integer;
   pc: PColor32;
   pb: pbyte;
Begin
   len := Length(mask);
   If (len = 0) Or (len <> self.Width * self.Height) Then
      Exit;
   pc := self.PixelBase;
   pb := @mask[0];
   For i := 0 To len - 1 Do
   Begin
      If pb^ > 0 Then
         pc^ := color
      Else
         pc^ := clNone32;
      Inc(pc);
      Inc(pb);
   End;
End;

Procedure TgaGraphMem.DrawAlphaMask(Const mask: TArrayOfByte; color: TColor32);
Var
   i, len: integer;
   pc: PColor32;
   pb: pbyte;
Begin
   len := Length(mask);
   If (len = 0) Or (len <> self.Width * self.Height) Then
      Exit;
   color := color And $FFFFFF;
   pc := self.PixelBase;
   pb := @mask[0];
   For i := 0 To len - 1 Do
   Begin
      If pb^ > 0 Then
         pc^ := color Or pb^ Shl 24
      Else
         pc^ := clNone32;
      Inc(pc);
      Inc(pb);
   End;
End;

Procedure TgaGraphMem.SymmetricCropTransparent;
Var
   rec: TRect;
Begin
   rec := self.GetSymmetricCropTransparentRect;
   If (rec.Top > 0) Or (rec.Left > 0) Then
      self.Crop(rec);
End;

Procedure TgaGraphMem.DrawShadow(Const polygon: TPathD; fillRule: TFillRule; depth: double; angleRads: double;
   color: TColor32; cutoutInsideShadow: boolean);
Var
   polygons: TPathsD;
Begin
   setlength(polygons, 1);
   polygons[0] := polygon;
   DrawShadow(polygons, fillRule, depth,
      angleRads, color, cutoutInsideShadow);
End;

Procedure TgaGraphMem.DrawShadow(Const polygons: TPathsD; fillRule: TFillRule; depth: double; angleRads: double;
   color: TColor32; cutoutInsideShadow: boolean);
Var
   x, y: double;
   blurSize: integer;
   rec: TRect;
   polys, shadowPolys: TPathsD;
   shadowImg: TgaGraphMem;
Begin
   rec := gaGetBounds(polygons);
   If IsEmptyRect(rec) Or (depth < 1) Then
      Exit;
   If Not VarClockwiseRotationIsAnglePositive Then
      angleRads := -angleRads;
   GetSinCos(angleRads, y, x);
   x := depth * x;
   y := depth * y;
   blurSize := Max(1, Round(depth / 4));
   gaEngine.InflateRect(rec, Ceil(depth * 2), Ceil(depth * 2));
   polys := OffsetPath(polygons, -rec.Left, -rec.Top);
   shadowPolys := OffsetPath(polys, x, y);
   shadowImg := TgaGraphMem.Create(RectWidth(rec), RectHeight(rec));
   Try
      shadowImg.DrawPolygon(shadowPolys, fillRule, color);
      shadowImg.FastGaussianBlur(shadowImg.Bounds, blurSize, 1);
      If cutoutInsideShadow Then
         shadowImg.EraseInsidePaths(polys, fillRule);
      self.CopyBlend(shadowImg, shadowImg.Bounds, rec, @BlendToAlpha);
   Finally
      shadowImg.Free;
   End;
End;

Procedure TgaGraphMem.DrawGlow(Const polygon: TPathD; fillRule: TFillRule; color: TColor32; blurRadius: integer);
Var
   polygons: TPathsD;
Begin
   setlength(polygons, 1);
   polygons[0] := polygon;
   DrawGlow(polygons, fillRule, color, blurRadius);
End;

Procedure TgaGraphMem.DrawGlow(Const polygons: TPathsD; fillRule: TFillRule; color: TColor32; blurRadius: integer);
Var
   rec: TRect;
   glowPolys: TPathsD;
   glowImg: TgaGraphMem;
Begin
   rec := gaGetBounds(polygons);
   glowPolys := OffsetPath(polygons, blurRadius - rec.Left + 1, blurRadius - rec.Top + 1);
   gaEngine.InflateRect(rec, blurRadius + 1, blurRadius + 1);
   glowImg := TgaGraphMem.Create(RectWidth(rec), RectHeight(rec));
   Try
      glowImg.DrawPolygon(glowPolys, fillRule, color);
      glowImg.FastGaussianBlur(glowImg.Bounds, blurRadius);
      glowImg.ScaleAlpha(4);
      self.CopyBlend(glowImg, glowImg.Bounds, rec, @BlendToAlpha);
   Finally
      glowImg.Free;
   End;
End;

Procedure TgaGraphMem.Sharpen(radius: integer; amount: integer);
Var
   i: integer;
   amt: double;
   weightAmount: Array[-255..255] Of integer;
   bmpBlur: TgaGraphMem;
   pColor, pBlur: PARGB;
Begin
   If radius = 0 Then
      Exit;
   amt := ClampRange(amount / 10, 0.1, 5);
   radius := ClampRange(radius, 1, 10);
   For i := -255 To 255 Do
      weightAmount[i] := Round(amt * i);

   bmpBlur := TgaGraphMem.Create(self);
   Try
      pColor := PARGB(self.pixelBase);
      bmpBlur.FastGaussianBlur(bmpBlur.Bounds, radius);
      pBlur := PARGB(bmpBlur.pixelBase);
      For i := 1 To self.Width * self.Height Do
      Begin
         If (pColor^.A > 0) Then
         Begin
            pColor^.R := ClampByte(pColor^.R + weightAmount[pColor^.R - pBlur^.R]);
            pColor^.G := ClampByte(pColor^.G + weightAmount[pColor^.G - pBlur^.G]);
            pColor^.B := ClampByte(pColor^.B + weightAmount[pColor^.B - pBlur^.B]);
         End;
         Inc(pColor);
         Inc(pBlur);
      End;
   Finally
      bmpBlur.Free;
   End;
End;

Procedure TgaGraphMem.HatchBackground(color1: TColor32; color2: TColor32; hatchSize: integer);
Var
   i, j: integer;
   pc: PColor32;
   colors: Array[boolean] Of TColor32;
   hatch: boolean;
Begin
   colors[False] := color1;
   colors[True] := color2;
   self.BeginUpdate;
   Try
      pc := self.Pixelbase;
      For i := 0 To self.Height - 1 Do
      Begin
         hatch := Odd(i Div hatchSize);
         For j := 0 To self.Width - 1 Do
         Begin
            If (j + 1) Mod hatchSize = 0 Then
               hatch := Not hatch;
            pc^ := BlendToOpaque(colors[hatch], pc^);
            Inc(pc);
         End;
      End;
   Finally
      self.EndUpdate;
   End;
End;

Procedure TgaGraphMem.GridBackground(majorInterval, minorInterval: integer; fillColor: TColor32; majColor: TColor32; minColor: TColor32);
Var
   i, x, y, w, h: integer;
   path: TPathD;
Begin
   self.Clear(fillColor);
   w := self.Width;
   h := self.Height;
   SetLength(path, 2);
   If minorInterval > 0 Then
   Begin
      x := minorInterval;
      path[0] := PointD(x, 0);
      path[1] := PointD(x, h);
      ;
      For i := 1 To (w Div minorInterval) Do
      Begin
         self.DrawLine(path, 1, minColor, esSquare);
         path[0].X := path[0].X + minorInterval;
         path[1].X := path[1].X + minorInterval;
      End;
      y := minorInterval;
      path[0] := PointD(0, y);
      path[1] := PointD(w, y);
      For i := 1 To (h Div minorInterval) Do
      Begin
         self.DrawLine(path, 1, minColor, esSquare);
         path[0].Y := path[0].Y + minorInterval;
         path[1].Y := path[1].Y + minorInterval;
      End;
   End;
   If majorInterval > minorInterval Then
   Begin
      x := majorInterval;
      path[0] := PointD(x, 0);
      path[1] := PointD(x, h);
      ;
      For i := 1 To (w Div majorInterval) Do
      Begin
         self.DrawLine(path, 1, majColor, esSquare);
         path[0].X := path[0].X + majorInterval;
         path[1].X := path[1].X + majorInterval;
      End;
      y := majorInterval;
      path[0] := PointD(0, y);
      path[1] := PointD(w, y);
      For i := 1 To (h Div majorInterval) Do
      Begin
         self.DrawLine(path, 1, majColor, esSquare);
         path[0].Y := path[0].Y + majorInterval;
         path[1].Y := path[1].Y + majorInterval;
      End;
   End;
End;

Procedure TgaGraphMem.ReplaceColor(oldColor, newColor: TColor32);
Var
   color: PColor32;
   i: integer;
Begin
   color := self.PixelBase;
   For i := 0 To self.Width * self.Height - 1 Do
   Begin
      If color^ = oldColor Then
         color^ := newColor;
      Inc(color);
   End;
End;

Procedure TgaGraphMem.EraseColor(color: TColor32);
Var
   fg: TARGB Absolute color;
   bg: PARGB;
   i: integer;
   Q: byte;
Begin
   If fg.A = 0 Then
      Exit;
   bg := PARGB(self.PixelBase);

   For i := 0 To self.Width * self.Height - 1 Do
   Begin
      If bg^.A > 0 Then
      Begin
         If (bg^.R > fg.R) Then
            Q := VarDivTable[bg^.R - fg.R, Not fg.R]
         Else If (bg^.R < fg.R) Then
            Q := VarDivTable[fg.R - bg^.R, fg.R]
         Else
            Q := 0;
         If (bg^.G > fg.G) Then
            Q := Max(Q, VarDivTable[bg^.G - fg.G, Not fg.G])
         Else If (bg^.G < fg.G) Then
            Q := Max(Q, VarDivTable[fg.G - bg^.G, fg.G]);
         If (bg^.B > fg.B) Then
            Q := Max(Q, VarDivTable[bg^.B - fg.B, Not fg.B])
         Else If (bg^.B < fg.B) Then
            Q := Max(Q, VarDivTable[fg.B - bg^.B, fg.B]);
         If (Q > 0) Then
         Begin
            bg^.A := VarMulTable[bg^.A, Q];
            bg^.R := VarDivTable[bg^.R - VarMulTable[fg.R, Not Q], Q];
            bg^.G := VarDivTable[bg^.G - VarMulTable[fg.G, Not Q], Q];
            bg^.B := VarDivTable[bg^.B - VarMulTable[fg.B, Not Q], Q];
         End
         Else
            bg^.Color := clNone32;
      End;
      Inc(bg);
   End;
End;

Procedure TgaGraphMem.RedEyeRemove(Const rect: TRect);
Var
   k: integer;
   cutout, mask: TgaGraphMem;
   path: TPathD;
   cutoutRec, rect3: TRect;
   radGrad: TgaRendererGradientRadial;
Begin
   k := RectWidth(rect) * RectHeight(rect);
   If k < 120 Then
      k := 2
   Else If k < 230 Then
      k := 3
   Else
      k := 4;
   cutoutRec := rect;
   gaEngine.InflateRect(cutoutRec, k, k);

   cutout := TgaGraphMem.Create(self, cutoutRec);
   mask := TgaGraphMem.Create(cutout.Width, cutout.Height);
   radGrad := TgaRendererGradientRadial.Create;
   Try
      rect3 := cutout.Bounds;
      gaEngine.InflateRect(rect3, -k, -k);
      path := Ellipse(rect3);
      Mask.DrawPolygon(path, frNonZero, clBlack32);
      Mask.GaussianBlur(Mask.Bounds, k);
      self.CopyBlend(mask, Mask.Bounds, cutoutRec, @BlendToOpaque);

      path := Ellipse(cutoutRec);
      radGrad.SetParameters(rect3, clBlack32, clNone32);
      Mask.DrawPolygon(path, frNonZero, radGrad);
      cutout.CopyBlend(mask, Mask.Bounds, cutout.Bounds, @BlendMask);

      cutout.EraseColor(clRed32);

      self.CopyBlend(cutout, cutout.Bounds, cutoutRec, @BlendToOpaque);
   Finally
      Mask.Free;
      cutout.Free;
      radGrad.Free;
   End;
End;

Procedure TgaGraphMem.EraseInsidePath(Const path: TPathD; fillRule: TFillRule);
Begin
   If assigned(path) Then
      self.ErasePolygon(path, fillRule);
End;

Procedure TgaGraphMem.EraseInsidePaths(Const paths: TPathsD; fillRule: TFillRule);
Begin
   If assigned(paths) Then
      self.ErasePolygon(paths, fillRule);
End;

Procedure TgaGraphMem.EraseOutsidePath(Const path: TPathD; fillRule: TFillRule; Const outsideBounds: TRect);
Var
   mask: TgaGraphMem;
   p: TPathD;
Begin
   If Not assigned(path) Then
      Exit;
   mask := TgaGraphMem.Create(RectWidth(outsideBounds), RectHeight(outsideBounds));
   Try
      p := OffsetPath(path, -outsideBounds.Left, -outsideBounds.top);
      Mask.DrawPolygon(p, fillRule, clBlack32);
      self.CopyBlend(mask, Mask.Bounds, outsideBounds, @BlendMask);
   Finally
      Mask.Free;
   End;
End;

Procedure TgaGraphMem.EraseOutsidePaths(Const paths: TPathsD; fillRule: TFillRule; Const outsideBounds: TRect);
Var
   mask: TgaGraphMem;
   pp: TPathsD;
Begin
   If Not assigned(paths) Then
      Exit;
   mask := TgaGraphMem.Create(RectWidth(outsideBounds), RectHeight(outsideBounds));
   Try
      pp := OffsetPath(paths, -outsideBounds.Left, -outsideBounds.top);
      Mask.DrawPolygon(pp, fillRule, clBlack32);
      self.CopyBlend(mask, Mask.Bounds, outsideBounds, @BlendMask);
   Finally
      Mask.Free;
   End;
End;

Procedure TgaGraphMem.Draw3D(Const polygon: TPathD; fillRule: TFillRule; Height, blurRadius: double; colorLt: TColor32;
   colorDk: TColor32; angleRads: double);
Var
   polygons: TPathsD;
Begin
   setLength(polygons, 1);
   polygons[0] := polygon;
   self.Draw3D(polygons, fillRule, Height, blurRadius, colorLt, colorDk, angleRads);
End;

Procedure TgaGraphMem.Draw3D(Const polygons: TPathsD; fillRule: TFillRule; Height, blurRadius: double; colorLt: TColor32;
   colorDk: TColor32; angleRads: double);
Var
   tmp: TgaGraphMem;
   recI: TRect;
   recD: TRectD;
   paths, paths2: TPathsD;
   x, y: double;
Begin
   If Not VarClockwiseRotationIsAnglePositive Then
      angleRads := -angleRads;
   GetSinCos(angleRads, y, x);
   recD := gaGetBoundsD(polygons);
   If recD.IsEmpty Then
      Exit;
   recI := Rect(recD);
   paths := OffsetPath(polygons, -recI.Left, -recI.Top);
   tmp := TgaGraphMem.Create(rectWidth(recI), rectHeight(recI));
   Try
      If colorLt Shr 24 > 0 Then
      Begin
         tmp.Clear(colorLt);
         paths2 := OffsetPath(paths, Height * x, Height * y);
         tmp.EraseInsidePaths(paths2, fillRule);
         tmp.FastGaussianBlur(tmp.Bounds, Round(blurRadius), 0);
         tmp.EraseOutsidePaths(paths, fillRule, tmp.Bounds);
         self.CopyBlend(tmp, tmp.Bounds, recI, @BlendToAlpha);
      End;

      If colorDk Shr 24 > 0 Then
      Begin
         tmp.Clear(colorDk);
         paths2 := OffsetPath(paths, -Height * x, -Height * y);
         tmp.EraseInsidePaths(paths2, fillRule);
         tmp.FastGaussianBlur(tmp.Bounds, Round(blurRadius), 0);
         tmp.EraseOutsidePaths(paths, fillRule, tmp.Bounds);
         self.CopyBlend(tmp, tmp.Bounds, recI, @BlendToAlpha);
      End;
   Finally
      tmp.Free;
   End;
End;

Function TgaGraphMem.DrawButton(Const pt: TPointD; size: double; color: TColor32; buttonShape: TgaButtonShape;
   buttonAttributes: TgaButtonAttributes): TPathD;
Var
   i: integer;
   radius: double;
   rec: TRectD;
   shadowSize, shadowAngle: double;
Begin
   self.Clear;
   If (size < 5) Then
      Exit;
   radius := size * 0.5;
   shadowSize := radius * 0.25;

   rec := RectD(pt.X - radius, pt.Y - radius, pt.X + radius, pt.Y + radius);

   Case buttonShape Of
      bsDiamond:
         Begin
            SetLength(Result, 4);
            For i := 0 To 3 Do
               Result[i] := pt;
            Result[0].X := Result[0].X - radius;
            Result[1].Y := Result[1].Y - radius;
            Result[2].X := Result[2].X + radius;
            Result[3].Y := Result[3].Y + radius;
         End;
      bsSquare:
         Begin
            gaEngine.InflateRect(rec, -1, -1);
            Result := Rectangle(rec);
         End;
      Else
         Result := Ellipse(rec);
   End;
   shadowAngle := cnAngle45;

   self.BeginUpdate;
   Try

      If baShadow In buttonAttributes Then
         DrawShadow(Result, frNonZero, shadowSize,
            shadowAngle, $AA000000, color Shr 24 < 254);

      If color Shr 24 > 2 Then
         DrawPolygon(Result, frNonZero, color);

      If ba3D In buttonAttributes Then
         Draw3D(Result, frNonZero, shadowSize * 2,
            Ceil(shadowSize), $CCFFFFFF, $AA000000, shadowAngle);
      DrawLine(Result, VarDpiAwareI, clBlack32, esPolygon);
   Finally
      EndUpdate;
   End;

End;

Procedure TgaGraphMem.PencilEffect(intensity: integer);
Var
   w, h: integer;
   img2: TgaGraphMem;
Begin
   w := self.Width;
   h := self.Height;
   If w * h = 0 Then
      Exit;

   intensity := max(1, min(10, intensity));
   self.Grayscale;
   img2 := TgaGraphMem.Create(self);
   Try
      img2.InvertColors;
      img2.FastGaussianBlur(img2.Bounds, intensity);
      self.CopyBlend(img2, img2.Bounds, self.Bounds, @BlendColorDodge);
   Finally
      img2.Free;
   End;
End;

Procedure TgaGraphMem.TraceContours(intensity: integer);
Var
   i, j, w, h: integer;
   tmp, tmp2: TArrayOfColor32;
   s: PColor32;
   d: PARGB;
Begin
   w := self.Width;
   h := self.Height;
   If w * h = 0 Then
      Exit;
   SetLength(tmp, w * h);
   SetLength(tmp2, w * h);
   s := self.PixelRow[0];
   d := @tmp[0];
   For j := 0 To h - 1 Do
   Begin
      For i := 0 To w - 2 Do
      Begin
         d^.A := Min($FF, ColorDifference(s^, IncPColor32(s, 1)^));
         Inc(s);
         Inc(d);
      End;
      Inc(s);
      Inc(d);
   End;

   For j := 0 To w - 1 Do
   Begin
      s := @tmp[j];
      d := @tmp2[j];
      For i := 0 To h - 2 Do
      Begin
         d^.A := Min($FF, AlphaAverage(s^, IncPColor32(s, w)^));
         Inc(s, w);
         Inc(d, w);
      End;
   End;
   Move(tmp2[0], self.PixelBase^, w * h * sizeOf(TColor32));

   If intensity < 1 Then
      Exit;
   If intensity > 10 Then
      intensity := 10;
   self.ScaleAlpha(intensity);
End;

Function TgaGraphMem.GetFloodFillMask(x, y: integer; compareFunc: TCompareFunction; tolerance: integer): TArrayOfByte;
Var
   xl, xr, xr2, dirY: integer;
   maxX, maxY: integer;
   ffs: TFloodFillStack;
   ffm: TFloodFillMask;
Begin
   Result := Nil;
   If (x < 0) Or (x >= self.Width) Or (y < 0) Or (y >= self.Height) Then
      Exit;
   maxX := self.Width - 1;
   maxY := self.Height - 1;

   If Not Assigned(compareFunc) Then
      compareFunc := @CompareRGB;

   ffs := TFloodFillStack.Create(maxY);
   Try
      xl := x;
      xr := x;
      ffm.Reset(self.Width, self.Height, x, y,
         self.PixelBase, compareFunc, tolerance);
      ffm.SetCurrentY(y);
      ffm.IsMatch(x);

      While (xl > 0) And ffm.IsMatch(xl - 1) Do
         Dec(xl);
      While (xr < maxX) And ffm.IsMatch(xr + 1) Do
         Inc(xr);
      ffs.Push(xl, xr, y, -1);
      ffs.Push(xl, xr, y, 1);
      While Not ffs.IsEmpty Do
      Begin
         ffs.Pop(xl, xr, y, dirY);
         ffm.SetCurrentY(y);
         xr2 := xl;
         If ffm.IsMatch(xl) Then
         Begin
            While (xl > 0) And ffm.IsMatch(xl - 1) Do
               Dec(xl);
            If xl <= xr2 - 2 Then
               ffs.Push(xl, xr2 - 2, y, -dirY);
            While (xr2 < maxX) And ffm.IsMatch(xr2 + 1) Do
               Inc(xr2);
            ffs.Push(xl, xr2, y, dirY);
            If xr2 >= xr + 2 Then
               ffs.Push(xr + 2, xr2, y, -dirY);
            xl := xr2 + 2;
         End;
         While (xl <= xr) And Not ffm.IsMatch(xl) Do
            Inc(xl);
         While (xl <= xr) Do
         Begin
            xr2 := xl;
            While (xr2 < maxX) And ffm.IsMatch(xr2 + 1) Do
               Inc(xr2);
            ffs.Push(xl, xr2, y, dirY);
            If xr2 >= xr + 2 Then
            Begin
               ffs.Push(xr + 2, xr2, y, -dirY);
               break;
            End;
            Inc(xl, 2);
            While (xl <= xr) And Not ffm.IsMatch(xl) Do
               Inc(xl);
         End;
      End;
      Result := ffm.mask;
   Finally
      ffs.Free;
   End;
End;

Procedure TgaGraphMem.FloodFill(x, y: integer; newColor: TColor32; compareFunc: TCompareFunction; tolerance: integer);
Var
   i: integer;
   ba: TArrayOfByte;
   pb: pbyte;
   pc: PColor32;
Begin
   If Not assigned(compareFunc) Then
   Begin
      compareFunc := @CompareRGB;
      tolerance := FloodFillDefaultRGBTolerance;
   End;

   If (tolerance < 0) Then
   Begin
      If compareFunc = @CompareRGB Then
         tolerance := FloodFillDefaultRGBTolerance
      Else If compareFunc = @CompareHue Then
         tolerance := FloodFillDefaultHueTolerance;
   End;

   ba := self.GetFloodFillMask(x, y, compareFunc, tolerance);
   If ba = Nil Then
      Exit;
   pb := @ba[0];
   pc := self.PixelBase;
   For i := 0 To High(ba) Do
   Begin
      If Ord(pb^) > 0 Then
         pc^ := newColor;
      Inc(pb);
      Inc(pc);
   End;
End;

Procedure TgaGraphMem.Emboss(radius: integer; depth: integer; luminance: integer; preserveColor: boolean);
Var
   yy, xx, x, y, w, h: integer;
   b: byte;
   kernel: Array[0..MaxBlur, 0..MaxBlur] Of integer;
   wca: TArrayOfWeightedColor;
   pc0, pcf, pcb: PColor32;
   pw0, pw: PgaWeightedColor;
   customGray: TColor32;
   pc: PColor32;
Const
   maxDepth = 50;
Begin
   luminance := ClampRange(luminance, 0, 100);
   b := luminance * 255 Div 100;
   customGray := $FF000000 + b Shl 16 + b Shl 8 + b;

   ClampRange(radius, 1, 5);
   Inc(depth);
   ClampRange(depth, 2, maxDepth);

   kernel[0][0] := 1;
   For y := 1 To radius Do
      For x := 1 To radius Do
         kernel[y][x] := depth;

   w := self.Width;
   h := self.Height;

   SetLength(wca, w * h);

   pc0 := IncPColor32(self.PixelBase, radius * w);
   pw0 := @wca[radius * w];
   For y := radius To h - 1 - radius Do
   Begin
      For x := radius To w - 1 - radius Do
      Begin
         pw := IncPWeightColor(pw0, x);
         pcb := IncPColor32(pc0, x - 1);
         If preserveColor Then
         Begin
            pcf := IncPColor32(pc0, x);
            pw^.Add(pcf^, kernel[0, 0]);
            Inc(pcf);
         End
         Else
         Begin
            pw^.Add(customGray, kernel[0, 0]);
            pcf := IncPColor32(pc0, x + 1);
         End;

         For yy := 1 To radius Do
         Begin
            For xx := 1 To radius Do
            Begin
               pw^.Subtract(Gray(pcf^), kernel[yy, xx]);
               pw^.Add(Gray(pcb^), kernel[yy, xx]);
               Dec(pcb);
               Inc(pcf);
            End;
            Dec(pcb, self.Width - radius);
            Inc(pcf, self.Width - radius);
         End;
      End;
      Inc(pc0, self.Width);
      Inc(pw0, self.Width);
   End;

   pc := @self.Pixels[0];
   pw := @wca[0];
   For x := 0 To self.Width * self.Height - 1 Do
   Begin
      pc^ := pw^.Color Or $FF000000;
      Inc(pc);
      Inc(pw);
   End;
End;

Function TgaGraphMem.Vectorize(compareColor: TColor32; compareFunc: TCompareFunction; colorTolerance: integer; roundingTolerance: integer): TPathsD;
Var
   i: integer;
   mask: TArrayOfByte;
Begin
   mask := self.GetBoolMask(compareColor, compareFunc, colorTolerance);
   Result := VectorizeMask(mask, self.Width);
   For i := 0 To high(Result) Do
      Result[i] := Tidy(Result[i], roundingTolerance);
End;

Procedure TgaGraphMem.GaussianBlur(rec: TRect; radius: integer);
Var
   i, x, y, yy, z: integer;
   gaussTable: Array[-MaxBlur..MaxBlur] Of cardinal;
   wc: TgaWeightedColor;
   wca: TArrayOfWeightedColor;
   row: PColor32Array;
   wcRow: PgaWeightedColorArray;
Begin
   Types.IntersectRect(rec, rec, self.Bounds);
   If IsEmptyRect(rec) Or (radius < 1) Then
      Exit
   Else If radius > MaxBlur Then
      radius := MaxBlur;

   For i := 0 To radius Do
   Begin
      gaussTable[i] := Sqr(Radius - i + 1);
      gaussTable[-i] := gaussTable[i];
   End;

   setLength(wca, RectWidth(rec) * RectHeight(rec));

   For y := 0 To RectHeight(rec) - 1 Do
   Begin
      row := PColor32Array(@self.Pixels[(y + rec.Top) * self.Width + rec.Left]);
      wcRow := PgaWeightedColorArray(@wca[y * RectWidth(rec)]);
      For x := 0 To RectWidth(rec) - 1 Do
         For z := max(0, x - radius) To min(self.Width - 1, x + radius) Do
            wcRow^[x].Add(row^[z], gaussTable[x - z]);
   End;

   For x := 0 To RectWidth(rec) - 1 Do
   Begin
      For y := 0 To RectHeight(rec) - 1 Do
      Begin
         wc.Reset;
         yy := max(0, y - radius) * RectWidth(rec);
         For z := max(0, y - radius) To min(RectHeight(rec) - 1, y + radius) Do
         Begin
            wc.Add(wca[x + yy].Color, gaussTable[y - z]);
            Inc(yy, RectWidth(rec));
         End;
         self.Pixels[x + rec.Left + (y + rec.Top) * self.Width] := wc.Color;
      End;
   End;
End;

Procedure TgaGraphMem.FastGaussianBlur(Const rec: TRect; stdDev: integer; repeats: integer);
Var
   i, j, len, w, h, maxStdDev: integer;
   rec2: TRect;
   boxes: TArrayOfInteger;
   src, dst: TArrayOfColor32;
   blurFullImage: boolean;
   p: PColor32;
Begin
   If Not Assigned(self) Then
      Exit;
   Types.IntersectRect(rec2, rec, self.Bounds);
   If IsEmptyRect(rec2) Then
      Exit;
   blurFullImage := RectsEqual(rec2, self.Bounds);

   w := RectWidth(rec2);
   h := RectHeight(rec2);
   len := w * h;
   SetLength(src, len);
   SetLength(dst, len);

   maxStdDev := Min(w, h) Div 2 - 1;
   If stdDev > maxStdDev Then
      stdDev := maxStdDev;
   If stdDev < 1 Then
      Exit;

   If blurFullImage Then
   Begin
      Move(self.PixelBase^, dst[0], len * SizeOf(TColor32));
   End
   Else
   Begin
      p := @dst[0];
      For i := rec2.Top To rec2.Bottom - 1 Do
      Begin
         Move(self.Pixels[i * self.Width + rec2.Left],
            p^, w * SizeOf(TColor32));
         Inc(p, w);
      End;
   End;

   Inc(repeats);
   boxes := BoxesForGauss(stdDev, repeats, maxStdDev);
   For j := 0 To repeats - 1 Do
   Begin
      BoxBlurH(dst, src, w, h, boxes[j]);
      BoxBlurT(src, dst, w, h, boxes[j]);
   End;
   self.BeginUpdate;
   Try
      If blurFullImage Then
      Begin
         Move(dst[0], self.PixelBase^, len * SizeOf(TColor32));
      End
      Else
      Begin
         p := @dst[0];
         For i := rec2.Top To rec2.Bottom - 1 Do
         Begin
            Move(p^, self.Pixels[i * self.Width + rec2.Left], w * SizeOf(TColor32));
            Inc(p, w);
         End;
      End;
   Finally
      self.EndUpdate;
   End;
End;

Procedure TgaGraphMem.BoxDownSampling(newWidth, newHeight: integer);
Var
   x, y, x256, y256, xx256, yy256: integer;
   sx, sy: double;
   tmp: TArrayOfColor32;
   pc: PColor32;
   scaledX: Array Of integer;
Begin
   sx := self.Width / newWidth * 256;
   sy := self.Height / newHeight * 256;
   SetLength(tmp, newWidth * newHeight);

   SetLength(scaledX, newWidth + 1);
   For x := 0 To newWidth - 1 Do
      scaledX[x] := Round((x + 1) * sx);

   y256 := 0;
   pc := @tmp[0];
   For y := 0 To newHeight - 1 Do
   Begin
      x256 := 0;
      yy256 := Round((y + 1) * sy);
      For x := 0 To newWidth - 1 Do
      Begin
         xx256 := scaledX[x];
         pc^ := GeTgaWeightedColor(self.Pixels, x256, y256, xx256, yy256, self.Width);
         x256 := xx256;
         Inc(pc);
      End;
      y256 := yy256;
   End;

   self.BeginUpdate;
   self.SetSize(newWidth, newHeight);
   Move(tmp[0], self.Pixels[0], newWidth * newHeight * SizeOf(TColor32));
   self.EndUpdate;
End;

Function TgaGraphMem.GetBoolMask(reference: TColor32; compareFunc: TCompareFunction; tolerance: integer): TArrayOfByte;
Var
   i: integer;
   pa: pbyte;
   pc: PColor32;
Begin
   Result := Nil;
   If self.IsEmpty Then
      Exit;
   If Not Assigned(compareFunc) Then
      compareFunc := @CompareRGB;
   SetLength(Result, self.Width * self.Height);
   pa := @Result[0];
   pc := self.PixelBase;
   For i := 0 To self.Width * self.Height - 1 Do
   Begin
      If compareFunc(reference, pc^, tolerance) Then
         pa^ := 1
      Else
         pa^ := 0;

      Inc(pc);
      Inc(pa);
   End;
End;

Function TgaGraphMem.GetByteMask(reference: TColor32; compareFunc: TCompareFunctionEx): TArrayOfByte;
Var
   i: integer;
   pa: pbyte;
   pc: PColor32;
Begin
   Result := Nil;
   If self.IsEmpty Then
      Exit;
   If Not Assigned(compareFunc) Then
      compareFunc := @GetAlpha;
   SetLength(Result, self.Width * self.Height);
   pa := @Result[0];
   pc := self.PixelBase;
   For i := 0 To self.Width * self.Height - 1 Do
   Begin
      pa^ := compareFunc(reference, pc^);
      Inc(pc);
      Inc(pa);
   End;
End;

Procedure TgaGraphMem.ApplyClearType(textColor: TColor32 = clBlack32; bkColor: TColor32 = clWhite32);
Const
   centerWeighting = 5;
Var
   h, w: integer;
   src, dst: PARGB;
   srcArr: PArgbs;
   fgColor: TARGB Absolute textColor;
   bgColor: TARGB Absolute bkColor;
   diff_R, diff_G, diff_B: integer;
   bg8_R, bg8_G, bg8_B: integer;
   rowBuffer: TArrayOfARGB;
   primeTbl, nearTbl, FarTbl: PByteArray;
Begin
   primeTbl := PByteArray(@VarMulTable[85 + centerWeighting * 2]);
   nearTbl := PByteArray(@VarMulTable[57]);
   farTbl := PByteArray(@VarMulTable[28 - centerWeighting]);

   SetLength(rowBuffer, self.Width + 4);
   FillChar(rowBuffer[0], Length(rowBuffer) * SizeOf(TColor32), 0);

   For h := 0 To self.Height - 1 Do
   Begin
      dst := PARGB(@self.Pixels[h * self.Width]);

      src := PARGB(@rowBuffer[2]);
      Move(dst^, src^, self.Width * SizeOf(TColor32));
      srcArr := PArgbs(rowBuffer);

      w := 2;
      While w < self.Width Do
      Begin
         dst^.R := primeTbl^[srcArr^[w].A] + nearTbl^[srcArr^[w - 1].A] + farTbl^[srcArr^[w - 2].A] + nearTbl^[srcArr^[w + 1].A] + farTbl^[srcArr^[w + 2].A];
         Inc(w);
         dst^.G := primeTbl^[srcArr^[w].A] + nearTbl^[srcArr^[w - 1].A] + farTbl^[srcArr^[w - 2].A] + nearTbl^[srcArr^[w + 1].A] + farTbl^[srcArr^[w + 2].A];
         Inc(w);
         dst^.B := primeTbl^[srcArr^[w].A] + nearTbl^[srcArr^[w - 1].A] + farTbl^[srcArr^[w - 2].A] + nearTbl^[srcArr^[w + 1].A] + farTbl^[srcArr^[w + 2].A];
         Inc(w);
         dst^.A := 255;
         Inc(dst);
      End;
   End;

   self.Crop(Types.Rect(0, 0, self.Width Div 3, self.Height));

   diff_R := fgColor.R - bgColor.R;
   diff_G := fgColor.G - bgColor.G;
   diff_B := fgColor.B - bgColor.B;
   bg8_R := bgColor.R Shl 8;
   bg8_G := bgColor.G Shl 8;
   bg8_B := bgColor.B Shl 8;
   dst := PARGB(self.PixelBase);
   For h := 0 To self.Width * self.Height - 1 Do
   Begin
      If dst^.R > 0 Then
      Begin
         dst^.R := (bg8_R + diff_R * dst^.R) Shr 8;
         dst^.G := (bg8_G + diff_G * dst^.G) Shr 8;
         dst^.B := (bg8_B + diff_B * dst^.B) Shr 8;
      End
      Else
         dst^.Color := 0;
      Inc(dst);
   End;
End;

Function TgaGraphMem.GetSymmetricCropTransparentRect: TRect;
Var
   w, h, x, y, x1, y1: integer;
   p1, p2: PARGB;
   opaquePxlFound: boolean;
Begin
   Result := self.Bounds;
   w := self.Width;
   y1 := 0;
   opaquePxlFound := False;
   For y := 0 To (self.Height Div 2) - 1 Do
   Begin
      p1 := PARGB(self.PixelRow[y]);
      p2 := PARGB(self.PixelRow[self.Height - y - 1]);
      For x := 0 To w - 1 Do
      Begin
         If (p1^.A > 0) Or (p2^.A > 0) Then
         Begin
            y1 := y;
            opaquePxlFound := True;
            break;
         End;
         Inc(p1);
         Inc(p2);
      End;
      If opaquePxlFound Then
         break;
   End;

   If Not opaquePxlFound Then
      Exit;

   If y1 > 0 Then
   Begin
      Inc(Result.Top, y1);
      Dec(Result.Bottom, y1);
   End;

   x1 := 0;
   h := RectHeight(Result);
   opaquePxlFound := False;
   For x := 0 To (w Div 2) - 1 Do
   Begin
      p1 := PARGB(@self.Pixels[Result.Top * w + x]);
      p2 := PARGB(@self.Pixels[Result.Top * w + (w - 1) - x]);
      For y := 0 To h - 1 Do
      Begin
         If (p1^.A > 0) Or (p2^.A > 0) Then
         Begin
            x1 := x;
            opaquePxlFound := True;
            break;
         End;
         Inc(p1, w);
         Inc(p2, w);
      End;
      If opaquePxlFound Then
         break;
   End;

   If Not opaquePxlFound Then
      Exit;
   Inc(Result.Left, x1);
   Dec(Result.Right, x1);
End;

Function TgaGraphMem.GetFloodFillBounds(x, y: integer; Const byteArray: TArrayOfByte): TRect;

   Function RowHasFill(i: integer): boolean;
   Var
      pb, pEnd: pbyte;
   Begin
      Result := True;
      pb := @byteArray[i * self.Width];
      pEnd := pb + self.Width;
      While (pb < pEnd) Do
         If Ord(pb^) = 1 Then
            Exit
         Else
            Inc(pb);
      Result := False;
   End;

   Function ColHasFill(i: integer): boolean;
   Var
      pb, pEnd: pbyte;
   Begin
      Result := True;
      pb := @byteArray[i];
      pEnd := @byteArray[length(byteArray) - 1];
      While (pb < pEnd) Do
         If Ord(pb^) > 0 Then
            Exit
         Else
            Inc(pb, self.Width);
      Result := False;
   End;

Begin
   Result := Types.Rect(x, y, x, y);
   While (Result.Top > 0) And RowHasFill(Result.Top - 1) Do
      Dec(Result.Top);
   While (Result.Bottom < self.Height - 1) And RowHasFill(Result.Bottom) Do
      Inc(Result.Bottom);
   While (Result.Left > 0) And ColHasFill(Result.Left - 1) Do
      Dec(Result.Left);
   While (Result.Right < self.Width - 1) And ColHasFill(Result.Right) Do
      Inc(Result.Right);
End;

Procedure InitByteExponents;
Var
   i: integer;
Const
   inv255: double = 1 / 255;
   inv255sqrd: double = 1 / (255 * 255);
   inv255cubed: double = 1 / (255 * 255 * 255);
Begin
   For i := 0 To 255 Do
   Begin
      byteFrac[i] := i * inv255;
      byteFracSq[i] := i * i * inv255sqrd;
      byteFracCubed[i] := i * i * i * inv255cubed;
   End;
End;


Function PrePendPoint(Const pt: TPointD; Const p: TPathD): TPathD;
Var
   len: integer;
Begin
   len := Length(p);
   SetLength(Result, len + 1);
   Result[0] := pt;
   If len > 0 Then
      Move(p[0], Result[1], len * SizeOf(TPointD));

End;

Function PrePendPoints(Const pt1, pt2: TPointD; Const p: TPathD): TPathD;
Var
   len: integer;
Begin
   len := Length(p);
   SetLength(Result, len + 2);
   Result[0] := pt1;
   Result[1] := pt2;
   If len > 0 Then
      Move(p[0], Result[2], len * SizeOf(TPointD));

End;


Function GetPtOnRotatedEllipseFromAngle(Const ellipseRect: TRectD;
   ellipseRotAngle, angle: double): TPointD;
Begin
   Result := GetPtOnEllipseFromAngle(ellipseRect, angle);
   If ellipseRotAngle <> 0 Then
      gaEngine.RotatePoint(Result, ellipseRect.MidPoint, ellipseRotAngle);
End;

Function GetClosestPtOnRotatedEllipse(Const ellipseRect: TRectD;
   ellipseRotation: double; Const pt: TPointD): TPointD;
Var
   pt2: TPointD;
   angle: double;
Begin
   pt2 := pt;
   gaEngine.RotatePoint(pt2, ellipseRect.MidPoint, -ellipseRotation);
   angle := GetEllipticalAngleFromPoint(ellipseRect, pt2);
   Result := GetPtOnEllipseFromAngle(ellipseRect, angle);
   gaEngine.RotatePoint(Result, ellipseRect.MidPoint, ellipseRotation);
End;

Function GetPtOnEllipseFromAngle(Const ellipseRect: TRectD;
   angle: double): TPointD;
Var
   sn, co: double;
Begin
   NormalizeAngle(angle);
   GetSinCos(angle, sn, co);
   Result.X := ellipseRect.MidPoint.X + ellipseRect.Width / 2 * co;
   Result.Y := ellipseRect.MidPoint.Y + ellipseRect.Height / 2 * sn;
End;

Function GetEllipticalAngleFromPoint(Const ellipseRect: TRectD;
   Const pt: TPointD): double;
Begin
   With ellipseRect Do
      Result := ArcTan2(Width / Height * (pt.Y - MidPoint.Y), (pt.X - MidPoint.X));
End;

Function GetRotatedEllipticalAngleFromPoint(Const ellipseRect: TRectD;
   ellipseRotAngle: double; pt: TPointD): double;
Begin
   Result := 0;
   If ellipseRect.IsEmpty Then
      Exit;
   RotatePoint(pt, ellipseRect.MidPoint, -ellipseRotAngle);
   Result := GetEllipticalAngleFromPoint(ellipseRect, pt);
End;


Function MakePath(Const pts: Array Of integer): TPathD;
Var
   i, j, x, y, len: Integer;
Begin
   Result := Nil;
   len := length(pts) Div 2;
   If len < 1 Then
      Exit;
   setlength(Result, len);
   Result[0].X := pts[0];
   Result[0].Y := pts[1];
   j := 0;
   For i := 1 To len - 1 Do
   Begin
      x := pts[i * 2];
      y := pts[i * 2 + 1];
      inc(j);
      Result[j].X := x;
      Result[j].Y := y;
   End;
   setlength(Result, j + 1);
End;

Function MakePath(Const pts: Array Of double): TPathD;
Var
   i, j, len: Integer;
   x, y: double;
Begin
   Result := Nil;
   len := length(pts) Div 2;
   If len = 0 Then
      Exit;
   setlength(Result, len);
   Result[0].X := pts[0];
   Result[0].Y := pts[1];
   j := 0;
   For i := 1 To len - 1 Do
   Begin
      x := pts[i * 2];
      y := pts[i * 2 + 1];
      inc(j);
      Result[j].X := x;
      Result[j].Y := y;
   End;
   setlength(Result, j + 1);
End;



//========================================================================  
//========================================================================
//========================================================================

Initialization
   CreateImageFormatList;
   MakeBlendTables;
   InitByteExponents;
   VarNearestResampler := RegisterResampler(@NearestResampler,  'NearestNeighbor');
   VarBilinearResampler:= RegisterResampler(@BilinearResampler, 'Bilinear');
   VarBicubicResampler := RegisterResampler(@BicubicResampler,  'HermiteBicubic');
   VarDefaultResampler := VarBilinearResampler;

Finalization
   CleanUpImageFormatClassList;
   CleanUpResamplerClassList;

End.
